/*****************************************************************************/
/*****************************************************************************/
/********************* TEST FILE FOR CodesOverZ4.m ***************************/
/*****************************************************************************/
/*****************************************************************************/
/*
	Put the file CodesOverZ4.m in the same directory than this test file.
	Execute the function test()
*/
Z  := Integers();
Z2 := Integers(2);
Z4 := Integers(4);


/*****************************************************************************/
GammaRM:=function(s,r,m)
	sum := r div 2;
	aux := 0;
	for i:=0 to sum do
		aux := aux + Binomial(m-2*s-1,r-2*i)*Binomial(s,i);
	end for;
	return aux;
end function;
/*****************************************************************************/
DeltaRM:=function(s,r,m)
	aux := 0;
	for i:=0 to r do
		aux := aux + Binomial(m,i);
	end for;
	aux := aux - GammaRM(s,r,m);
	aux := aux div 2;
	return aux;
end function;
/*****************************************************************************/
TestDualKronecker:=function()

	//equivalent to ExtendedPerfectCodeZ4(2,4)
	ExtendedPerfectCodeZ4 := LinearCode<Z4, 8 |
								[1,0,0,1,0,0,1,3],
								[0,1,0,1,0,0,2,2],
								[0,0,1,1,0,0,1,1],
								[0,0,0,2,0,0,0,2],
								[0,0,0,0,1,0,3,2],
								[0,0,0,0,0,1,2,3]>;
								
	//equivalent to HadamardCodeZ4(2,4)
	HadamardCodeZ4 := LinearCode<Z4, 8 |
						[1,0,3,2,1,0,3,2],
						[0,1,2,3,0,1,2,3],
						[0,0,0,0,2,2,2,2]>;
						
	return 			   DualKroneckerZ4(HadamardCodeZ4) eq ExtendedPerfectCodeZ4 
			and DualKroneckerZ4(ExtendedPerfectCodeZ4) eq HadamardCodeZ4;
			
end function;

/*****************************************************************************/
/******************************************************************************
** TestDimensionOfSpanZ2 function
******************************************************************************/
//RM(2,1,5)=(0,3,4,7)-RM(2,2,5)=(2,7,11,21)-RM(2,3,5)=(0,13,14,29)-RM(2,4,5)=(1,15,31,31)-

TestDimensionOfSpanZ2:=function()

	C := ReedMullerCodeRMZ4(2,2,5);
	D := ReedMullerCodeRMZ4(2,4,5);
	rC := DimensionOfSpanZ2(C);
	rD := DimensionOfSpanZ2(D);
	return rC eq 21 and rD eq 31;
	
end function;
/*****************************************************************************/
/******************************************************************************
** TestDimensionOfSpanZ2ForAllFamily function
******************************************************************************/

// TestDimensionOfSpanZ2ForAllFamily:=function(s,m)
// 
// 	SuposedKer := [];
// 	DKerList   := [DimensionOfKernelZ2(ReedMullerCodeRMZ4(0,r,m)) : r in {0..m}];
// 	GammaDelta := [GammaRM(s,r,m) + DeltaRM(s,r,m) : r in {0..m}];
// 	if s eq 0 then
// 		SuposedKer := [i + m : i in GammaDelta]; 
// 	end if;
// 	if s eq 1 then
// 		SuposedKer := [i + 2 : i in GammaDelta]; 
// 	end if;
// 	if s gt 1 then
// 		SuposedKer := [i + 1 : i in GammaDelta]; 
// 	end if;
// 	if s eq 2 and m eq 5 then
// 		SuposedKer[3] := SuposedKer[3] + 1;
// 	end if;
// 	
// 	SuposedKer[m+1] := SuposedKer[m+1]
// 	
// 	return DKerList eq SuposedKer; 
// 	
// end function;

/*****************************************************************************/
/******************************************************************************
** TestDimensionOfKernelZ2 function
******************************************************************************/
//RM(2,1,5)=(0,3,4,7)-RM(2,2,5)=(2,7,11,21)-RM(2,3,5)=(0,13,14,29)-RM(2,4,5)=(1,15,31,31)-

TestDimensionOfKernelZ2:=function()

	C := ReedMullerCodeRMZ4(2,2,5);
	D := ReedMullerCodeRMZ4(2,4,5);
	kC := DimensionOfKernelZ2(KernelZ2CodeZ4(C));
	kD := DimensionOfKernelZ2(KernelZ2CodeZ4(D));
	return kC eq 11 and kD eq 31;
	
end function;

/*****************************************************************************/
/******************************************************************************
** TestReedMullerCodesRMZ4 function
******************************************************************************/
//TODO: test more properties 
TestReedMullerCodesRMZ4:=function()
	F := ReedMullerCodesRMZ4(1,5);
	n := #F;
	
	itsok := true;
	for i:=1 to n-1 do
		itsok := itsok and F[i] subset F[i+1]; 
	end for;
	
	return itsok;
end function;

/*****************************************************************************/
/******************************************************************************
** TestCodesOverZ4 function
******************************************************************************/

TestCodesOverZ4:=function()
	allok:=true;
	printf "\n";
	printf "Testing package CodesOverZ4.m\n\n";
	
	if(TestDualKronecker()) then
		printf "DualKronecker							[OK]\n";
	else
		allok = false;
		printf "DualKronecker							[FAIL]\n";
	end if;
	
	if(TestReedMullerCodesRMZ4()) then
		printf "ReedMullerCodesRMZ4						[OK]\n";
	else
		allok = false;
		printf "ReedMullerCodesRMZ4						[FAIL]\n";
	end if;
	
	if(TestDimensionOfSpanZ2()) then
		printf "DimensionOfSpanZ2		    				[OK]\n";
	else
		allok = false;
		printf "DimensionOfSpanZ2							[FAIL]\n";
	end if;
	
	if(TestDimensionOfKernelZ2()) then
		printf "DimensionOfKernelZ2			   			[OK]\n";
	else
		allok = false;
		printf "DimensionOfKernelZ2						[FAIL]\n";
	end if;
	
	
	return allok;
end function;

/*****************************************************************************/

TestCodesOverZ4();
