freeze;

import "cuspidal.m":
    set_cuspidal_space;
import "setlevel.m":
    SetLevel;
import "sharbly.m":
    SetSharblies;
import "cohomology.m":
    HomologyData;
import "setup.m":
    Preset;
import "findperfectform.m":
    FindPerfectForm;
import "defs.m":
    ModFrmDataRec,
    LevelRec;
import "findcells.m":
    FindCells;

/**********************************************************************
*                                                                     *
*                   BIANCHI MODULAR FORMS                             *
*                                                                     *
*     Definition of new attributes on ModFrmBianchi           	      *
*                                                                     *
*                    Dan Yasaki May 2009                              *
*                                                                     *
***********************************************************************/


/**********************************************************************
  ModFrmBianchi attributes  ...  
  Let M be a ModFrmBianchi over the FldNum F    
**********************************************************************/

declare type ModFrmBianchiElt; // Not used yet?

declare type ModFrmBianchi [ModFrmBianchiElt]: ModFrmHil;

declare attributes ModFrmBianchi:
    Full_Dimension,//Dimension of H_0(S_*(Gamma0))
    //Hecke matrices for cusp space used to compute T_p for
    //non-principal primes in class number 3. Should not get used often
    Hecke_ap,//T_a,a,T_p from Cremona, restricted to cuspidal space
    //EHecke_ap,//T_a,a,T_p from Cremona, on the homology
    //needs p to be a square in the class group
    EHecke, //Hecke matrices for full space (including Eisenstein series)
	//Should not get used often
    ModFrmData, //Should eventually get organized.
    LevelData, //Should eventually get organized.
    is_Bianchi,//set true for Bianchi modular forms
    homology_coefficients,//Coefficients for homology computations.
	//Should be a field.  Don't know the interpretation when this is not.
    Hecke_pp //Hecke matrices for Tp,p action
;

/*****************   Constructors for Bianchi spaces   *********************/

function BMF(F, N ,G,V, Coefficients)
    M := HackobjCreateRaw(ModFrmBianchi);
    M`is_Bianchi := true;
    M`Field := F;
    M`Level := N;
    M`NewLevel := 1*Integers(F);
    M`DirichletCharacter := 1;
    M`Weight := [2];
    M`CentralCharacter := 0;
    M`is_cuspidal := true; // always true, currently
    M`homology_coefficients:=Coefficients;
    M`Hecke := [* *];
    M`Hecke_pp := [* *];
    M`EHecke := [* *];
    M`Hecke_ap := [* *];
    M`ModFrmData:=V;
    M`LevelData:=rec<LevelRec|>;
    Preset(M);
    M`ModFrmData`initial_form:=FindPerfectForm(M);
    FindCells(M);
    SetSharblies(M);
    return M;
end function;

intrinsic BianchiCuspForms(F::FldNum, N::RngOrdIdl:
    Coefficients:=RationalField(),G:="GL",VorData:=rec<ModFrmDataRec|G:=G,ideal_reps:=[]>)
    -> ModFrmBianchi 
    {The space of Bianchi modular forms over the imaginary quadratic field F
    with level N and weight 2.  Here N should be an ideal in the 
    maximal order of F.}
    
    require BaseField(F) eq Rationals() and Degree(F) eq 2 and Discriminant(F) lt 0 :
	"The base field F must be an imaginary quadratic extension of the rationals";
    require NumberField(Order(N)) eq F :
	"The level N must be an ideal in the base field";
    return BMF(F, N,G,VorData,Coefficients);
end intrinsic;

function DimensionBianchi(M)
    vprint Bianchi,1:"Computing dimension of the cuspidal space.";
     if not assigned M`Dimension then 
	set_cuspidal_space(M);
	M`Dimension:=NumberOfRows(M`basis_matrix);
	
	//M`Dimension:=FullDimension(M)-NumberOfCusps(M) +1;
    end if;
    return M`Dimension; 
end function;

intrinsic FullDimension(M::ModFrmBianchi) -> RngIntElt
    {The dimension of the space M of H_0(S_*(Gamma0))}
    if not assigned M`Full_Dimension then 
	HomologyData(M);
	M`Full_Dimension:=Dimension(M`LevelData`homology`H);
    end if;
    return M`Full_Dimension; 
end intrinsic;

intrinsic VoronoiData(M::ModFrmBianchi) -> Rec
    {Returns the Voronoi data for BaseField(M)}
    return M`ModFrmData;
end intrinsic;

intrinsic HomologyData(M::ModFrmBianchi) -> Rec
    {Returns the Homology data for BaseField(M)}
    d:=Dimension(M);
    return M`LevelData`homology;
end intrinsic;

