freeze;

/////////////////////////////////////////////////////
//                                                 //
//     INVARIANT THEORY FOR GENUS ONE CURVES       //
//            "hessepolynomials.m"                 //
//                                                 //
//                T. A. Fisher                     //
//                                                 //
//           Version: 13th April 2006              //
//                                                 //
/////////////////////////////////////////////////////

// This file contains intrinsics :-  
//
//   HessePolynomials(n,r,invariants)
//   RubinSilverbergPolynomials(n,J,t)


intrinsic HessePolynomials(n::RngIntElt,r::RngIntElt,invariants::SeqEnum[RngElt] : 
                           Variables := 0)
      -> RngElt,RngElt,RngElt,FldRatElt,FldRatElt
{The Hesse polynomials DD,cc4,cc6. These polynomials give the invariants for 
certain families of genus one curves.}

  if Variables cmpeq 0 then 
     _<x,y> := PolynomialRing(Universe(invariants),2);
  else
     x,y := Explode(Variables);
  end if;

  c4,c6 := Explode(invariants);
  require GCD(r,n) eq 1 : r,"and",n,"are not coprime.";
  r := r mod n;
  eps := 1;
  alpha := 1;
  case [n,r]:
    when [3,2]: eps:=2;alpha:=1/3;
    when [4,3]: eps:=12;
    when [5,2]: eps:=2*3^3;alpha:=3^3;
    when [5,3]: eps:=2^7*3;alpha:=2^6;
    when [5,4]: eps:=12^4;alpha:=12^3;
  end case;

case [n,r]:

  when [2,1]:

DD := x^3 - 3*c4*x*y^2 - 2*c6*y^3;
cc4 := c4*x^2 + 2*c6*x*y + c4^2*y^2;
cc6 := c6*x^3 + 3*c4^2*x^2*y + 3*c4*c6*x*y^2 + (-c4^3 + 2*c6^2)*y^3;

  when [3,1]:

DD := x^4 - 6*c4*x^2*y^2 - 8*c6*x*y^3 - 3*c4^2*y^4;
cc4 := c4*x^4 + 4*c6*x^3*y + 6*c4^2*x^2*y^2 + 4*c4*c6*x*y^3 + (-3*c4^3 + 4*c6^2)*y^4;
cc6 := c6*x^6 + 6*c4^2*x^5*y + 15*c4*c6*x^4*y^2 + 20*c6^2*x^3*y^3 + 15*c4^2*c6*x^2*y^4 + (18*c4^4 - 12*c4*c6^2)*x*y^5 + (9*c4^3*c6 - 8*c6^3)*y^6;

  when [3,2]:

DD := 3*c4*x^4 + 4*c6*x^3*y - 6*c4^2*x^2*y^2 - 12*c4*c6*x*y^3 + (-c4^3 - 4*c6^2)*y^4;
cc4 := (3*c4^3 + c6^2)*x^4 + 16*c4^2*c6*x^3*y + (6*c4^4 + 18*c4*c6^2)*x^2*y^2 + (8*c4^3*c6 + 8*c6^3)*x*y^3 + (-c4^5 + 5*c4^2*c6^2)*y^4;

cc6 := (9*c4^3*c6 - c6^3)*x^6
    +6*(3*c4^5 + 5*c4^2*c6^2)*x^5*y
    +15*(5*c4^4*c6 + 3*c4*c6^3)*x^4*y^2
    +20*(7*c4^3*c6^2 + c6^4)*x^3*y^3
    +15*(c4^5*c6 + 7*c4^2*c6^3)*x^2*y^4
    +6*(c4^7 + c4^4*c6^2 + 6*c4*c6^4)*x*y^5
    +(5*c4^6*c6 - 5*c4^3*c6^3 + 8*c6^5)*y^6;

  when [4,1]:

DD := x^6 - 15*c4*x^4*y^2 - 40*c6*x^3*y^3 - 45*c4^2*x^2*y^4 - 24*c4*c6*x*y^5 + (27*c4^3 - 32*c6^2)*y^6;

cc4 := c4*x^8
    +8*c6*x^7*y
    +28*c4^2*x^6*y^2
    +56*c4*c6*x^5*y^3
    +14*(-3*c4^3 + 8*c6^2)*x^4*y^4
    +56*c4^2*c6*x^3*y^5
    +28*(9*c4^4 - 8*c4*c6^2)*x^2*y^6
    +8*(33*c4^3*c6 - 32*c6^3)*x*y^7
    +(81*c4^5 - 80*c4^2*c6^2)*y^8;

cc6 := c6*x^12
    +12*c4^2*x^11*y
    +66*c4*c6*x^10*y^2
    +44*(c4^3 + 4*c6^2)*x^9*y^3
    +495*c4^2*c6*x^8*y^4
    +792*c4^4*x^7*y^5
    +924*c4^3*c6*x^6*y^6
    +792*(-3*c4^5 + 4*c4^2*c6^2)*x^5*y^7
    +99*(-59*c4^4*c6 + 64*c4*c6^3)*x^4*y^8
    +44*(-27*c4^6 - 96*c4^3*c6^2 + 128*c6^4)*x^3*y^9
    +66*(-63*c4^5*c6 + 64*c4^2*c6^3)*x^2*y^10
    +12*(-243*c4^7 + 372*c4^4*c6^2 - 128*c4*c6^4)*x*y^11
    +(-1215*c4^6*c6 + 2240*c4^3*c6^3 - 1024*c6^5)*y^12;

  when [4,3]:

DD := (27*c4^3 + c6^2)*x^6 + 96*c4^2*c6*x^5*y 
  + (-45*c4^4 + 105*c4*c6^2)*x^4*y^2 
  + (-200*c4^3*c6 + 40*c6^3)*x^3*y^3 + (-15*c4^5 - 285*c4^2*c6^2)*x^2*y^4 
  + (-24*c4^4*c6 - 168*c4*c6^3)*x*y^5 + (c4^6 - 13*c4^3*c6^2 - 32*c6^4)*y^6;

cc4 := (81*c4^7 + 70*c4^4*c6^2 - 7*c4*c6^4)*x^8
    +8*(111*c4^6*c6 + 34*c4^3*c6^3 - c6^5)*x^7*y
    +28*(9*c4^8 + 118*c4^5*c6^2 + 17*c4^2*c6^4)*x^6*y^2
    +56*(23*c4^7*c6 + 114*c4^4*c6^3 + 7*c4*c6^5)*x^5*y^3
    +14*(-3*c4^9 + 214*c4^6*c6^2 + 501*c4^3*c6^4 + 8*c6^6)*x^4*y^4
    +56*(-c4^8*c6 + 66*c4^5*c6^3 + 79*c4^2*c6^5)*x^3*y^5
    +28*(c4^10 - 2*c4^7*c6^2 + 89*c4^4*c6^4 + 56*c4*c6^6)*x^2*y^6
    +8*(7*c4^9*c6 - 14*c4^6*c6^3 + 119*c4^3*c6^5 + 32*c6^7)*x*y^7
    +(c4^11 + 22*c4^8*c6^2 - 55*c4^5*c6^4 + 176*c4^2*c6^6)*y^8;

cc6 := (1701*c4^9*c6 - 91*c4^6*c6^3 + 119*c4^3*c6^5 - c6^7)*x^12
    +36*(81*c4^11 + 473*c4^8*c6^2 + 11*c4^5*c6^4 + 11*c4^2*c6^6)*x^11*y
    +66*(477*c4^10*c6 + 1213*c4^7*c6^3 + 31*c4^4*c6^5 + 7*c4*c6^7)*x^10*y^2
    +44*(27*c4^12 + 3615*c4^9*c6^2 + 4933*c4^6*c6^4 + 61*c4^3*c6^6 + 4*c6^8)*x^9*y^3
    +99*(169*c4^11*c6 + 4713*c4^8*c6^3 + 3747*c4^5*c6^5 + 11*c4^2*c6^7)*x^8*y^4
    +792*(3*c4^13 + 91*c4^10*c6^2 + 1105*c4^7*c6^4 + 529*c4^4*c6^6)*x^7*y^5
    +924*(13*c4^12*c6 + 173*c4^9*c6^3 + 1199*c4^6*c6^5 + 343*c4^3*c6^7)*x^6*y^6
    +792*(-c4^14 + 35*c4^11*c6^2 + 273*c4^8*c6^4 + 1225*c4^5*c6^6 + 196*c4^2*c6^8)*x^5*y^7
    +99*(-39*c4^13*c6 + 409*c4^10*c6^3 + 1907*c4^7*c6^5 + 5915*c4^4*c6^7 + 448*c4*c6^9)*x^4*y^8
    +44*(-c4^15 - 169*c4^12*c6^2 + 965*c4^9*c6^4 + 2309*c4^6*c6^6 + 5408*c4^3*c6^8 + 128*c6^10)*x^3*y^9
    +198*(-c4^14*c6 - 33*c4^11*c6^3 + 149*c4^8*c6^5 + 141*c4^5*c6^7 + 320*c4^2*c6^9)*x^2*y^10
    +12*(-c4^16 - 13*c4^13*c6^2 - 223*c4^10*c6^4 + 953*c4^7*c6^6 + 116*c4^4*c6^8 + 896*c4*c6^10)*x*y^11
    +(-11*c4^15*c6 - 11*c4^12*c6^3 - 473*c4^9*c6^5 + 1903*c4^6*c6^7 - 704*c4^3*c6^9 + 1024*c6^11)*y^12;

  when [5,1]:

DD := x^12
    -66*c4*x^10*y^2
    -440*c6*x^9*y^3
    -1485*c4^2*x^8*y^4
    -3168*c4*c6*x^7*y^5
    +660*(9*c4^3 - 16*c6^2)*x^6*y^6
    -4752*c4^2*c6*x^5*y^7
    +495*(-135*c4^4 + 128*c4*c6^2)*x^4*y^8
    +1760*(-81*c4^3*c6 + 80*c6^3)*x^3*y^9
    +594*(-225*c4^5 + 224*c4^2*c6^2)*x^2*y^10
    +120*(-513*c4^4*c6 + 512*c4*c6^3)*x*y^11
    +(91125*c4^6 - 193536*c4^3*c6^2 + 102400*c6^4)*y^12;

cc4 := c4*x^20
    +20*c6*x^19*y
    +190*c4^2*x^18*y^2
    +1140*c4*c6*x^17*y^3
    +285*(-3*c4^3 + 20*c6^2)*x^16*y^4
    +15504*c4^2*c6*x^15*y^5
    +2280*(45*c4^4 - 28*c4*c6^2)*x^14*y^6
    +4560*(57*c4^3*c6 - 40*c6^3)*x^13*y^7
    +7410*(-135*c4^5 + 152*c4^2*c6^2)*x^12*y^8
    +9880*(-783*c4^4*c6 + 800*c4*c6^3)*x^11*y^9
    +988*(-10125*c4^6 - 5688*c4^3*c6^2 + 16000*c6^4)*x^10*y^10
    +9880*(-5535*c4^5*c6 + 5552*c4^2*c6^3)*x^9*y^11
    +7410*(-6075*c4^7 + 972*c4^4*c6^2 + 5120*c4*c6^4)*x^8*y^12
    +4560*(-7695*c4^6*c6 - 5088*c4^3*c6^3 + 12800*c6^5)*x^7*y^13
    +2280*(91125*c4^8 - 217188*c4^5*c6^2 + 126080*c4^2*c6^4)*x^6*y^14
    +912*(455625*c4^7*c6 - 967608*c4^4*c6^3 + 512000*c4*c6^5)*x^5*y^15
    +285*(-273375*c4^9 + 2507760*c4^6*c6^2 - 4282368*c4^3*c6^4 + 2048000*c6^6)*x^4*y^16
    +1140*(309825*c4^8*c6 - 627264*c4^5*c6^3 + 317440*c4^2*c6^5)*x^3*y^17
    +190*(4100625*c4^10 - 11489040*c4^7*c6^2 + 10665216*c4^4*c6^4 - 3276800*c4*c6^6)*x^2*y^18
    +20*(34718625*c4^9*c6 - 102220704*c4^6*c6^3 + 100270080*c4^3*c6^5 - 32768000*c6^7)*x*y^19
    +(184528125*c4^11 - 549301500*c4^8*c6^2 + 544997376*c4^5*c6^4 - 180224000*c4^2*c6^6)*y^20;

cc6 := c6*x^30
    +30*c4^2*x^29*y
    +435*c4*c6*x^28*y^2
    +580*(2*c4^3 + 5*c6^2)*x^27*y^3
    +27405*c4^2*c6*x^26*y^4
    +1566*(61*c4^4 + 30*c4*c6^2)*x^25*y^5
    +6525*(67*c4^3*c6 + 24*c6^3)*x^24*y^6
    +156600*(-12*c4^5 + 25*c4^2*c6^2)*x^23*y^7
    +450225*(-19*c4^4*c6 + 32*c4*c6^3)*x^22*y^8
    +100050*(243*c4^6 - 420*c4^3*c6^2 + 320*c6^4)*x^21*y^9
    +330165*(747*c4^5*c6 - 656*c4^2*c6^3)*x^20*y^10
    +600300*(270*c4^7 + 2253*c4^4*c6^2 - 2432*c4*c6^4)*x^19*y^11
    +950475*(4347*c4^6*c6 - 1696*c4^3*c6^3 - 2560*c6^5)*x^18*y^12
    +17108550*(405*c4^8 + 306*c4^5*c6^2 - 704*c4^2*c6^4)*x^17*y^13
    +145422675*(297*c4^7*c6 - 296*c4^4*c6^3)*x^16*y^14
    +155117520*(513*c4^6*c6^2 - 512*c4^3*c6^4)*x^15*y^15
    +145422675*(-1215*c4^8*c6 + 3264*c4^5*c6^3 - 2048*c4^2*c6^5)*x^14*y^16
    +17108550*(-18225*c4^10 + 4536*c4^7*c6^2 + 54656*c4^4*c6^4 - 40960*c4*c6^6)*x^13*y^17
    +950475*(-1782405*c4^9*c6 + 2921184*c4^6*c6^3 - 319488*c4^3*c6^5 - 819200*c6^7)*x^12*y^18
    +600300*(-546750*c4^11 - 5800167*c4^8*c6^2 + 13494528*c4^5*c6^4 - 7147520*c4^2*c6^6)*x^11*y^19
    +30015*(-95772375*c4^10*c6 + 80789184*c4^7*c6^3 + 129672192*c4^4*c6^5 - 114688000*c4*c6^7)*x^10*y^20
    +100050*(-22143375*c4^12 + 30756510*c4^9*c6^2 - 27077760*c4^6*c6^4 + 51232768*c4^3*c6^6 - 32768000*c6^8)*x^9*y^21
    +450225*(-3116475*c4^11*c6 - 3830328*c4^8*c6^3 + 17104896*c4^5*c6^5 - 10158080*c4^2*c6^7)*x^8*y^22
    +156600*(49207500*c4^13 - 213374655*c4^10*c6^2 + 318574080*c4^7*c6^4 - 193728512*c4^4*c6^6 + 39321600*c4*c6^8)*x^7*y^23
    +6525*(779938875*c4^12*c6 - 3476968416*c4^9*c6^3 + 5401700352*c4^6*c6^5 - 3491102720*c4^3*c6^7 + 786432000*c6^9)*x^6*y^24
    +1566*(-11256215625*c4^14 + 44762422500*c4^11*c6^2 - 67387360320*c4^8*c6^4 + 45513793536*c4^5*c6^6 - 11632640000*c4^2*c6^8)*x^5*y^25
    +3915*(-10468895625*c4^13*c6 + 41806400400*c4^10*c6^3 - 62691729408*c4^7*c6^5 + 41839984640*c4^4*c6^7 - 10485760000*c4*c6^9)*x^4*y^26
    +580*(-16607531250*c4^15 + 13972469625*c4^12*c6^2 + 110010648960*c4^9*c6^4 - 247937200128*c4^6*c6^6 + 192990412800*c4^3*c6^8 - 52428800000*c6^10)*x^3*y^27
    +435*(-49269009375*c4^14*c6 + 197055698400*c4^11*c6^3 - 295587408384*c4^8*c6^5 + 197083791360*c4^5*c6^7 - 49283072000*c4^2*c6^9)*x^2*y^28
    +30*(-373669453125*c4^16 + 1662450783750*c4^13*c6^2 - 2913163081920*c4^10*c6^4 + 2501423824896*c4^7*c6^6 - 1044814233600*c4^4*c6^8 + 167772160000*c4*c6^10)*x*y^29
    +(-4110363984375*c4^15*c6 + 19796866155000*c4^12*c6^3 - 38083944628224*c4^9*c6^5 + 36574189977600*c4^6*c6^7 - 17532190720000*c4^3*c6^9 + 3355443200000*c6^11)*y^30;

  when [5,2]:

DD := (-125*c4^3 - 64*c6^2)*x^12
    -1620*c4^2*c6*x^11*y
    +66*(-25*c4^4 - 56*c4*c6^2)*x^10*y^2
    +220*(-11*c4^3*c6 - 16*c6^3)*x^9*y^3
    +1485*(5*c4^5 + 4*c4^2*c6^2)*x^8*y^4
    +792*(53*c4^4*c6 + 28*c4*c6^3)*x^7*y^5
    +660*(9*c4^6 + 164*c4^3*c6^2 + 16*c6^4)*x^6*y^6
    +2376*(19*c4^5*c6 + 44*c4^2*c6^3)*x^5*y^7
    +495*(27*c4^7 + 104*c4^4*c6^2 + 112*c4*c6^4)*x^4*y^8
    +220*(81*c4^6*c6 + 136*c4^3*c6^3 + 80*c6^5)*x^3*y^9
    +594*(-9*c4^8 + 32*c4^5*c6^2 + 16*c4^2*c6^4)*x^2*y^10
    +60*(-135*c4^7*c6 + 328*c4^4*c6^3 - 112*c4*c6^5)*x*y^11
    +(-729*c4^9 - 108*c4^6*c6^2 + 2896*c4^3*c6^4 - 1600*c6^6)*y^12;

cc4 := (-3125*c4^7 + 9625*c4^4*c6^2 - 3584*c4*c6^4)*x^20
    +80*(250*c4^6*c6 + 607*c4^3*c6^3 - 128*c6^5)*x^19*y
    +190*(625*c4^8 + 1315*c4^5*c6^2 + 976*c4^2*c6^4)*x^18*y^2
    +2280*(575*c4^7*c6 + 659*c4^4*c6^3 + 224*c4*c6^5)*x^17*y^3
    +285*(375*c4^9 + 30365*c4^6*c6^2 + 17552*c4^3*c6^4 + 1280*c6^6)*x^16*y^4
    +3648*(1250*c4^8*c6 + 9434*c4^5*c6^3 + 1709*c4^2*c6^5)*x^15*y^5
    +2280*(1125*c4^10 + 15383*c4^7*c6^2 + 31496*c4^4*c6^4 + 1568*c4*c6^6)*x^14*y^6
    +9120*(3255*c4^9*c6 + 10633*c4^6*c6^3 + 10738*c4^3*c6^5 + 160*c6^7)*x^13*y^7+7410*(675*c4^11 + 13793*c4^8*c6^2 + 22256*c4^5*c6^4 + 12848*c4^2*c6^6)*x^12*y^8
    +39520*(270*c4^10*c6 + 5909*c4^7*c6^3 + 4814*c4^4*c6^5 + 1400*c4*c6^7)*x^11*y^9
    +988*(-10125*c4^12 + 33273*c4^9*c6^2 + 398896*c4^6*c6^4 + 107248*c4^3*c6^6 + 16000*c6^8)*x^10*y^10
    +19760*(-1593*c4^11*c6 + 6271*c4^8*c6^3 + 20236*c4^5*c6^5 - 128*c4^2*c6^7)*x^9*y^11
    +7410*(1215*c4^13 - 2187*c4^10*c6^2 + 20576*c4^7*c6^4 + 34448*c4^4*c6^6 - 4480*c4*c6^8)*x^8*y^12
    +18240*(3726*c4^12*c6 - 2352*c4^9*c6^3 + 3683*c4^6*c6^5 + 7736*c4^3*c6^7 - 400*c6^9)*x^7*y^13
    +2280*(3645*c4^14 + 65583*c4^11*c6^2 - 82264*c4^8*c6^4 + 13328*c4^5*c6^6 + 49280*c4^2*c6^8)*x^6*y^14
    +1824*(16767*c4^13*c6 + 80901*c4^10*c6^3 - 169874*c4^7*c6^5 + 68992*c4^4*c6^7 + 28000*c4*c6^9)*x^5*y^15
    +285*(2187*c4^15 + 187353*c4^12*c6^2 + 67872*c4^9*c6^4 - 628192*c4^6*c6^6 + 388352*c4^3*c6^8 + 32000*c6^10)*x^4*y^16
    +4560*(1458*c4^14*c6 + 4779*c4^11*c6^3 - 4436*c4^8*c6^5 - 10352*c4^5*c6^7 + 9280*c4^2*c6^9)*x^3*y^17
    +190*(6561*c4^16 + 8019*c4^13*c6^2 + 18576*c4^10*c6^4 - 40736*c4^7*c6^6 - 34304*c4^4*c6^8 + 44800*c4*c6^10)*x^2*y^18
    +40*(28431*c4^15*c6 - 54837*c4^12*c6^3 + 64728*c4^9*c6^5 - 45184*c4^6*c6^7 - 23680*c4^3*c6^9 + 32000*c6^11)*x*y^19
    +(-59049*c4^17 + 656829*c4^14*c6^2 - 1843344*c4^11*c6^4 + 2497504*c4^8*c6^6 - 1793024*c4^5*c6^8 + 544000*c4^2*c6^10)*y^20;

cc6 := (546875*c4^9*c6 - 931875*c4^6*c6^3 + 575232*c4^3*c6^5 - 32768*c6^7)*x^30
    +30*(78125*c4^11 + 69375*c4^8*c6^2 - 121620*c4^5*c6^4 + 131584*c4^2*c6^6)*x^29*y
    +435*(59375*c4^10*c6 + 60825*c4^7*c6^3 + 8592*c4^4*c6^5 + 28672*c4*c6^7)*x^28*y^2
    +580*(-31250*c4^12 + 535375*c4^9*c6^2 + 497835*c4^6*c6^4 + 79808*c4^3*c6^6 + 20480*c6^8)*x^27*y^3
    +3915*(18125*c4^11*c6 + 711075*c4^8*c6^3 + 375864*c4^5*c6^5 - 2816*c4^2*c6^7)*x^26*y^4
    +1566*(190625*c4^13 + 2255625*c4^10*c6^2 + 9735630*c4^7*c6^4 + 2254864*c4^4*c6^6 - 107520*c4*c6^8)*x^25*y^5
    +6525*(942125*c4^12*c6 + 4763787*c4^9*c6^3 + 7734480*c4^6*c6^5 + 901120*c4^3*c6^7 - 12288*c6^9)*x^24*y^6
    +156600*(7500*c4^14 + 361205*c4^11*c6^2 + 871275*c4^8*c6^4 + 754572*c4^5*c6^6 + 52480*c4^2*c6^8)*x^23*y^7
    +450225*(48175*c4^13*c6 + 639961*c4^10*c6^3 + 892800*c4^7*c6^5 + 451760*c4^4*c6^7 + 14336*c4*c6^9)*x^22*y^8
    +100050*(30375*c4^15 + 1499865*c4^12*c6^2 + 10244200*c4^9*c6^4 + 8374272*c4^6*c6^6 + 2348160*c4^3*c6^8 + 20480*c6^10)*x^21*y^9
    +210105*(122625*c4^14*c6 + 3484471*c4^11*c6^3 + 12576480*c4^8*c6^5 + 5497680*c4^5*c6^7 + 836096*c4^2*c6^9)*x^20*y^10
    +600300*(-6750*c4^16 + 422547*c4^13*c6^2 + 4143371*c4^10*c6^4 + 7923288*c4^7*c6^6 + 1710576*c4^4*c6^8 + 136192*c4*c6^10)*x^19*y^11
    +950475*(64665*c4^15*c6 + 1446727*c4^12*c6^3 + 5701240*c4^9*c6^5 + 6508080*c4^6*c6^7 + 588032*c4^3*c6^9 + 20480*c6^11)*x^18*y^12
    +17108550*(2025*c4^17 + 38709*c4^14*c6^2 + 218042*c4^11*c6^4 + 474096*c4^8*c6^6 + 360672*c4^5*c6^8 + 8704*c4^2*c6^10)*x^17*y^13
    +145422675*(2295*c4^16*c6 + 14897*c4^13*c6^3 + 43680*c4^10*c6^5 + 62160*c4^7*c6^7 + 34432*c4^4*c6^9)*x^16*y^14
    +155117520*(6777*c4^15*c6^2 + 26891*c4^12*c6^4 + 50628*c4^9*c6^6 + 51216*c4^6*c6^8 + 21952*c4^3*c6^10)*x^15*y^15
    +145422675*(-1701*c4^17*c6 + 13917*c4^14*c6^3 + 42176*c4^11*c6^5 + 52896*c4^8*c6^7 + 37632*c4^5*c6^9 + 12544*c4^2*c6^11)*x^14*y^16
    +17108550*(-3645*c4^19 - 83511*c4^16*c6^2 + 207772*c4^13*c6^4 + 472576*c4^10*c6^6 + 335232*c4^7*c6^8 + 137984*c4^4*c6^10 + 35840*c4*c6^12)*x^13*y^17
    +950475*(-422091*c4^18*c6 - 4277853*c4^15*c6^3 + 7680336*c4^12*c6^5 + 9708896*c4^9*c6^7 + 1742592*c4^6*c6^9 - 205056*c4^3*c6^11 + 102400*c6^13)*x^12*y^18
    +600300*(21870*c4^20 - 2437533*c4^17*c6^2 - 11402649*c4^14*c6^4 + 23836208*c4^11*c6^6 + 10530336*c4^8*c6^8 - 4850688*c4^5*c6^10 - 1368320*c4^2*c6^12)*x^11*y^19
    +30015*(-166941*c4^19*c6 - 116476083*c4^16*c6^3 - 197567544*c4^13*c6^5 + 662181856*c4^10*c6^7 - 47601408*c4^7*c6^9 - 130204416*c4^4*c6^11 - 12544000*c4*c6^13)*x^10*y^20
    +100050*(-177147*c4^21 - 2918187*c4^18*c6^2 - 49513410*c4^15*c6^4 - 2973040*c4^12*c6^6 + 163674560*c4^9*c6^8 - 64265472*c4^6*c6^10 - 20797952*c4^3*c6^12 - 512000*c6^14)*x^9*y^21
    +450225*(-347733*c4^20*c6 - 1517859*c4^17*c6^3 - 7666992*c4^14*c6^5 + 6922208*c4^11*c6^7 + 18195456*c4^8*c6^9 - 12350208*c4^5*c6^11 - 1187840*c4^2*c6^13)*x^8*y^22
    +156600*(-78732*c4^22 - 3151467*c4^19*c6^2 + 309555*c4^16*c6^4 - 10571620*c4^13*c6^6 + 15548640*c4^10*c6^8 + 18227328*c4^7*c6^10 - 17699072*c4^4*c6^12 - 537600*c4*c6^14)*x^7*y^23
    +6525*(-12538071*c4^21*c6 - 80989713*c4^18*c6^3 + 170565696*c4^15*c6^5 - 246218704*c4^12*c6^7 + 209071872*c4^9*c6^9 + 144155904*c4^6*c6^11 - 168181760*c4^3*c6^13 - 1536000*c6^15)*x^6*y^24
    +1566*(-3601989*c4^23 - 84630339*c4^20*c6^2 - 106084080*c4^17*c6^4 + 706103040*c4^14*c6^6 - 1120238080*c4^11*c6^8 + 789217536*c4^8*c6^10 + 53723136*c4^5*c6^12 - 220160000*c4^2*c6^14)*x^5*y^25
    +3915*(-3090231*c4^22*c6 - 26728785*c4^19*c6^3 + 40746240*c4^16*c6^5 + 73614640*c4^13*c6^7 - 216577280*c4^10*c6^9 + 185382144*c4^7*c6^11 - 34324480*c4^4*c6^13 - 17920000*c4*c6^15)*x^4*y^26
    +580*(1062882*c4^24 - 31709313*c4^21*c6^2 - 15647985*c4^18*c6^4 + 160370280*c4^15*c6^6 - 44570160*c4^12*c6^8 - 320455168*c4^9*c6^10 + 383584512*c4^6*c6^12 - 118732800*c4^3*c6^14 - 12800000*c6^16)*x^3*y^27
    +435*(177147*c4^23*c6 - 19912635*c4^20*c6^3 + 30822120*c4^17*c6^5 + 24660720*c4^14*c6^7 - 50506240*c4^11*c6^9 - 31849728*c4^8*c6^11 + 81070080*c4^5*c6^13 - 34304000*c4^2*c6^15)*x^2*y^28
    +30*(-4782969*c4^25 + 31827411*c4^22*c6^2 - 162690930*c4^19*c6^4 + 339506640*c4^16*c6^6 - 264350880*c4^13*c6^8 + 12083456*c4^10*c6^10 + 31259136*c4^7*c6^12 + 53145600*c4^4*c6^14 - 35840000*c4*c6^16)*x*y^29
    +(-81310473*c4^24*c6 + 614011185*c4^21*c6^3 - 2272380480*c4^18*c6^5 + 4720025520*c4^15*c6^7 - 5743802240*c4^12*c6^9 + 4195985152*c4^9*c6^11 - 1895731200*c4^6*c6^13 + 565760000*c4^3*c6^15 - 102400000*c6^17)*y^30;

  when [5,3]:

DD := (-729*c4^6 - 999*c4^3*c6^2 + 64*c6^4)*x^12
    +60*(-189*c4^5*c6 - 67*c4^2*c6^3)*x^11*y
    +66*(-81*c4^7 - 711*c4^4*c6^2 - 104*c4*c6^4)*x^10*y^2
    +220*(-81*c4^6*c6 - 415*c4^3*c6^3 - 16*c6^5)*x^9*y^3
    +495*(27*c4^8 + 17*c4^5*c6^2 - 172*c4^2*c6^4)*x^8*y^4
    +792*(123*c4^7*c6 + 185*c4^4*c6^3 - 52*c4*c6^5)*x^7*y^5
    +660*(9*c4^9 + 491*c4^6*c6^2 + 412*c4^3*c6^4 - 16*c6^6)*x^6*y^6
    +792*(67*c4^8*c6 + 665*c4^5*c6^3 + 292*c4^2*c6^5)*x^5*y^7
    +495*(15*c4^10 + 217*c4^7*c6^2 + 968*c4^4*c6^4 + 208*c4*c6^6)*x^4*y^8
    +220*(71*c4^9*c6 + 465*c4^6*c6^3 + 1176*c4^3*c6^5 + 80*c6^7)*x^3*y^9
    +66*(-25*c4^11 + 249*c4^8*c6^2 + 816*c4^5*c6^4 + 1136*c4^2*c6^6)*x^2*y^10
    +60*(-53*c4^10*c6 + 237*c4^7*c6^3 + 120*c4^4*c6^5 + 208*c4*c6^7)*x*y^11
    +(-125*c4^12 - 719*c4^9*c6^2 + 3804*c4^6*c6^4 - 1616*c4^3*c6^6 + 1600*c6^8)*y^12;

cc4 := (-59049*c4^13 + 293787*c4^10*c6^2 - 147771*c4^7*c6^4 + 53833*c4^4*c6^6 + 6656*c4*c6^8)*x^20
    +80*(2187*c4^12*c6 + 39447*c4^9*c6^3 - 10575*c4^6*c6^5 + 5677*c4^3*c6^7 + 128*c6^9)*x^19*y
    +190*(6561*c4^14 + 38637*c4^11*c6^2 + 103059*c4^8*c6^4 - 6257*c4^5*c6^6 + 5456*c4^2*c6^8)*x^18*y^2
    +2280*(8019*c4^13*c6 + 29079*c4^10*c6^3 + 36985*c4^7*c6^5 - 771*c4^4*c6^7 + 416*c4*c6^9)*x^17*y^3
    +285*(2187*c4^15 + 518319*c4^12*c6^2 + 1201233*c4^9*c6^4 + 804901*c4^6*c6^6 - 21168*c4^3*c6^8 + 1280*c6^10)*x^16*y^4
    +3648*(8019*c4^14*c6 + 213444*c4^11*c6^3 + 294970*c4^8*c6^5 + 112884*c4^5*c6^7 - 2629*c4^2*c6^9)*x^15*y^5
    +2280*(3645*c4^16 + 134433*c4^13*c6^2 + 1162415*c4^10*c6^4 + 979691*c4^7*c6^6 + 229480*c4^4*c6^8 - 2912*c4*c6^10)*x^14*y^6
    +9120*(13203*c4^15*c6 + 155553*c4^12*c6^3 + 675835*c4^9*c6^5 + 356923*c4^6*c6^7 + 52022*c4^3*c6^9 - 160*c6^11)*x^13*y^7
    +7410*(1215*c4^17 + 84483*c4^14*c6^2 + 534125*c4^11*c6^4 + 1393361*c4^8*c6^6 + 454480*c4^5*c6^8 + 39088*c4^2*c6^10)*x^12*y^8
    +39520*(999*c4^16*c6 + 48889*c4^13*c6^3 + 191675*c4^10*c6^5 + 321579*c4^7*c6^7 + 60946*c4^4*c6^9 + 2600*c4*c6^11)*x^11*y^9
    +988*(-10125*c4^18 + 106983*c4^15*c6^2 + 4252745*c4^12*c6^4 + 10387741*c4^9*c6^6 + 11679120*c4^6*c6^8 + 1141808*c4^3*c6^10 + 16000*c6^12)*x^10*y^10
    +19760*(-2565*c4^17*c6 + 16667*c4^14*c6^3 + 332413*c4^11*c6^5 + 495945*c4^8*c6^7 + 394148*c4^5*c6^9 + 16768*c4^2*c6^11)*x^9*y^11
    +7410*(675*c4^19 - 10297*c4^16*c6^2 + 96105*c4^13*c6^4 + 988381*c4^10*c6^6 + 879680*c4^7*c6^8 + 543888*c4^4*c6^10 + 8320*c4*c6^12)*x^8*y^12
    +18240*(2535*c4^18*c6 - 2314*c4^15*c6^3 + 46644*c4^12*c6^5 + 324610*c4^9*c6^7 + 165269*c4^6*c6^9 + 89544*c4^3*c6^11 + 400*c6^13)*x^7*y^13
    +2280*(1125*c4^20 + 64793*c4^17*c6^2 - 38745*c4^14*c6^4 + 240931*c4^11*c6^6 + 1639400*c4^8*c6^8 + 383568*c4^5*c6^10 + 215680*c4^2*c6^12)*x^6*y^14
    +1824*(6875*c4^19*c6 + 130517*c4^16*c6^3 - 176657*c4^13*c6^5 + 181815*c4^10*c6^7 + 1054538*c4^7*c6^9 + 4288*c4^4*c6^11 + 52000*c4*c6^13)*x^5*y^15
    +285*(375*c4^21 + 97915*c4^18*c6^2 + 675397*c4^15*c6^4 - 1517239*c4^12*c6^6 + 1178208*c4^9*c6^8 + 2550304*c4^6*c6^10 - 510208*c4^3*c6^12 + 32000*c6^14)*x^4*y^16
    +4560*(275*c4^20*c6 + 5723*c4^17*c6^3 + 16837*c4^14*c6^5 - 60855*c4^11*c6^7 + 56852*c4^8*c6^9 + 33712*c4^5*c6^11 - 15680*c4^2*c6^13)*x^3*y^17
    +190*(625*c4^22 + 6365*c4^19*c6^2 + 60291*c4^16*c6^4 + 67583*c4^13*c6^6 - 493616*c4^10*c6^8 + 569952*c4^7*c6^10 + 19456*c4^4*c6^12 - 83200*c4*c6^14)*x^2*y^18
    +40*(3875*c4^21*c6 + 3391*c4^18*c6^3 + 68769*c4^15*c6^5 - 2315*c4^12*c6^7 - 420856*c4^9*c6^9 + 595584*c4^6*c6^11 - 142720*c4^3*c6^13 - 32000*c6^15)*x*y^19
    +(-3125*c4^23 + 83375*c4^20*c6^2 - 251919*c4^17*c6^4 + 827333*c4^14*c6^6 - 974096*c4^11*c6^8 - 576288*c4^8*c6^10 + 1778176*c4^5*c6^12 - 736000*c4^2*c6^14)*y^20;

cc6 := (62178597*c4^18*c6 - 134080596*c4^15*c6^3 + 167458590*c4^12*c6^5 - 37170900*c4^9*c6^7 + 1441765*c4^6*c6^9 - 3171584*c4^3*c6^11 - 32768*c6^13)*x^30
    +270*(531441*c4^20 + 2939328*c4^17*c6^2 - 4303530*c4^14*c6^4 + 9226260*c4^11*c6^6 - 1795295*c4^8*c6^8 - 248892*c4^5*c6^10 - 57856*c4^2*c6^12)*x^29*y
    +435*(5491557*c4^19*c6 + 19709244*c4^16*c6^3 + 663390*c4^13*c6^5 + 40437900*c4^10*c6^7 - 8727835*c4^7*c6^9 - 897904*c4^4*c6^11 - 53248*c4*c6^13)*x^28*y^2
    +580*(-1062882*c4^21 + 50841189*c4^18*c6^2 + 157464000*c4^15*c6^4 + 93168630*c4^12*c6^6 + 124506090*c4^9*c6^8 - 27214691*c4^6*c6^10 - 1320128*c4^3*c6^12 - 20480*c6^14)*x^27*y^3
    +3915*(-137781*c4^20*c6 + 78857388*c4^17*c6^3 + 181450530*c4^14*c6^5 + 93192100*c4^11*c6^7 + 52598955*c4^8*c6^9 - 9422568*c4^5*c6^11 - 176896*c4^2*c6^13)*x^26*y^4
    +1566*(3601989*c4^22 + 90948582*c4^19*c6^2 + 1520861670*c4^16*c6^4 + 2377114440*c4^13*c6^6 + 894208445*c4^10*c6^8 + 300389882*c4^7*c6^10 - 34222864*c4^4*c6^12 - 199680*c4*c6^14)*x^25*y^5
    +6525*(22300839*c4^21*c6 + 305144820*c4^18*c6^3 + 1979347050*c4^15*c6^5 + 2133346436*c4^12*c6^7 + 585609895*c4^9*c6^9 + 134834128*c4^6*c6^11 - 7868416*c4^3*c6^13 - 12288*c6^15)*x^24*y^6
    +156600*(78732*c4^23 + 11116521*c4^20*c6^2 + 90144360*c4^17*c6^4 + 326815510*c4^14*c6^6 + 250328320*c4^11*c6^8 + 49657161*c4^8*c6^10 + 8165828*c4^5*c6^12 - 206080*c4^2*c6^14)*x^23*y^7
    +450225*(636417*c4^22*c6 + 27577260*c4^19*c6^3 + 146992950*c4^16*c6^5 + 342876828*c4^13*c6^7 + 188964385*c4^10*c6^9 + 26032304*c4^7*c6^11 + 3046832*c4^4*c6^13 - 26624*c4*c6^15)*x^22*y^8
    +100050*(177147*c4^24 + 27559116*c4^21*c6^2 + 613143810*c4^18*c6^4 + 2265685540*c4^15*c6^6 + 3619536475*c4^12*c6^8 + 1432562296*c4^9*c6^10 + 128087488*c4^6*c6^12 + 10372480*c4^3*c6^14 - 20480*c6^16)*x^21*y^9
    +90045*(2473983*c4^23*c6 + 191099700*c4^20*c6^3 + 2511557450*c4^17*c6^5 + 6561229732*c4^14*c6^7 + 7420389215*c4^11*c6^9 + 2091428656*c4^8*c6^11 + 109310288*c4^5*c6^13 + 5753344*c4^2*c6^15)*x^20*y^10
    +600300*(-21870*c4^25 + 3720573*c4^22*c6^2 + 130195440*c4^19*c6^4 + 1065597766*c4^16*c6^6 + 1991081998*c4^13*c6^8 + 1629911173*c4^10*c6^10 + 323621656*c4^7*c6^12 + 8342800*c4^4*c6^14 + 252928*c4*c6^16)*x^19*y^11
    +950475*(168399*c4^24*c6 + 16496460*c4^21*c6^3 + 275240490*c4^18*c6^5 + 1482291956*c4^15*c6^7 + 2002622735*c4^12*c6^9 + 1207196168*c4^9*c6^11 + 167079344*c4^6*c6^13 + 1586432*c4^3*c6^15 + 20480*c6^17)*x^18*y^12
    +17108550*(3645*c4^26 + 177066*c4^23*c6^2 + 4037030*c4^20*c6^4 + 38219232*c4^17*c6^6 + 143279781*c4^14*c6^8 + 141074526*c4^11*c6^10 + 63490032*c4^8*c6^12 + 6068640*c4^5*c6^14 + 11776*c4^2*c6^16)*x^17*y^13
    +145422675*(5589*c4^25*c6 + 120540*c4^22*c6^3 + 1398670*c4^19*c6^5 + 8590156*c4^16*c6^7 + 23516725*c4^13*c6^9 + 16889008*c4^10*c6^11 + 5730064*c4^7*c6^13 + 372352*c4^4*c6^15)*x^16*y^14
    +155117520*(26433*c4^24*c6^2 + 370840*c4^21*c6^4 + 2775686*c4^18*c6^6 + 12055268*c4^15*c6^8 + 24991793*c4^12*c6^10 + 12942236*c4^9*c6^12 + 3320240*c4^6*c6^14 + 140608*c4^3*c6^16)*x^15*y^15
    +145422675*(-1755*c4^26*c6 + 80012*c4^23*c6^3 + 886334*c4^20*c6^5 + 4763628*c4^17*c6^7 + 15388965*c4^14*c6^9 + 24771936*c4^11*c6^11 + 8960160*c4^8*c6^13 + 1730560*c4^5*c6^15 + 43264*c4^2*c6^17)*x^14*y^16
    +17108550*(-2025*c4^28 - 132264*c4^25*c6^2 + 1280250*c4^22*c6^4 + 12824732*c4^19*c6^6 + 51473671*c4^16*c6^8 + 126410356*c4^13*c6^10 + 160828672*c4^10*c6^12 + 38097280*c4^7*c6^14 + 5514496*c4^4*c6^16 + 66560*c4*c6^18)*x^13*y^17
    +950475*(-328185*c4^27*c6 - 10114988*c4^24*c6^3 + 34297354*c4^21*c6^5 + 322233028*c4^18*c6^7 + 944237255*c4^15*c6^9 + 1766478896*c4^12*c6^11 + 1818664160*c4^9*c6^13 + 249541120*c4^6*c6^15 + 27591424*c4^3*c6^17 + 102400*c6^19)*x^12*y^18
    +1800900*(2250*c4^29 - 802829*c4^26*c6^2 - 14104480*c4^23*c6^4 + 27887002*c4^20*c6^6 + 200676606*c4^17*c6^8 + 397155291*c4^14*c6^10 + 571910352*c4^11*c6^12 + 501310880*c4^8*c6^14 + 30635776*c4^5*c6^16 + 2896640*c4^2*c6^18)*x^11*y^19
    +30015*(217125*c4^28*c6 - 146905868*c4^25*c6^3 - 1491780386*c4^22*c6^5 + 2856594268*c4^19*c6^7 + 11499863045*c4^16*c6^9 + 13794026696*c4^13*c6^11 + 16488946400*c4^10*c6^13 + 13515020800*c4^7*c6^15 + 140449024*c4^4*c6^17 + 23296000*c4*c6^19)*x^10*y^20
    +100050*(-30375*c4^30 - 1003290*c4^27*c6^2 - 91334050*c4^24*c6^4 - 527278312*c4^21*c6^6 + 1257360401*c4^18*c6^8 + 2421063370*c4^15*c6^10 + 1467622960*c4^12*c6^12 + 1944669760*c4^9*c6^14 + 1646902016*c4^6*c6^16 - 21380608*c4^3*c6^18 + 512000*c6^20)*x^9*y^21
    +450225*(-72575*c4^29*c6 - 1060884*c4^26*c6^3 - 27830298*c4^23*c6^5 - 90750116*c4^20*c6^7 + 299516865*c4^17*c6^9 + 235334448*c4^14*c6^11 + 24284000*c4^11*c6^13 + 161629440*c4^8*c6^15 + 137056512*c4^5*c6^17 - 2007040*c4^2*c6^19)*x^8*y^22
    +156600*(-7500*c4^31 - 879055*c4^28*c6^2 - 5046520*c4^25*c6^4 - 73121722*c4^22*c6^6 - 127587224*c4^19*c6^8 + 655069105*c4^16*c6^10 + 75516820*c4^13*c6^12 - 109414240*c4^10*c6^14 + 192327296*c4^7*c6^16 + 130241792*c4^4*c6^18 - 998400*c4*c6^20)*x^7*y^23
    +6525*(-1461625*c4^30*c6 - 41280428*c4^27*c6^3 - 74072486*c4^24*c6^5 - 1137513852*c4^21*c6^7 - 800525785*c4^18*c6^9 + 8639711216*c4^15*c6^11 - 3012757200*c4^12*c6^13 - 1159055360*c4^9*c6^15 + 1904114944*c4^6*c6^17 + 837079040*c4^3*c6^19 - 1536000*c6^21)*x^6*y^24
    +1566*(-190625*c4^32 - 15327500*c4^29*c6^2 - 178237830*c4^26*c6^4 + 103969636*c4^23*c6^6 - 2575650593*c4^20*c6^8 + 406774320*c4^17*c6^10 + 14424885440*c4^14*c6^12 - 10536666880*c4^11*c6^14 + 247640832*c4^8*c6^16 + 2594545664*c4^5*c6^18 + 680960000*c4^2*c6^20)*x^5*y^25
    +3915*(-238125*c4^31*c6 - 7742300*c4^28*c6^3 - 37133086*c4^25*c6^5 + 99101012*c4^22*c6^7 - 530089485*c4^19*c6^9 + 430092400*c4^16*c6^11 + 1592784880*c4^13*c6^13 - 1813896192*c4^10*c6^15 + 409243904*c4^7*c6^17 + 220958720*c4^4*c6^19 + 33280000*c4*c6^21)*x^4*y^26
    +580*(31250*c4^33 - 2680375*c4^30*c6^2 - 32821360*c4^27*c6^4 - 48806418*c4^24*c6^6 + 364619454*c4^21*c6^8 - 1504137135*c4^18*c6^10 + 1832133480*c4^15*c6^12 + 1587643440*c4^12*c6^14 - 3181852416*c4^9*c6^16 + 1228683008*c4^6*c6^18 + 140748800*c4^3*c6^20 + 12800000*c6^22)*x^3*y^27
    +3915*(3125*c4^32*c6 - 300700*c4^29*c6^3 - 1403762*c4^26*c6^5 + 1033724*c4^23*c6^7 + 12651445*c4^20*c6^9 - 60646600*c4^17*c6^11 + 92523760*c4^14*c6^13 - 8068864*c4^11*c6^15 - 66497792*c4^8*c6^17 + 38533120*c4^5*c6^19 - 1536000*c4^2*c6^21)*x^2*y^28
    +30*(-78125*c4^34 + 808750*c4^31*c6^2 - 17121830*c4^28*c6^4 - 4833744*c4^25*c6^6 + 56832267*c4^22*c6^8 + 146362170*c4^19*c6^10 - 1129864560*c4^16*c6^12 + 2092991520*c4^13*c6^14 - 1095030528*c4^10*c6^16 - 427721216*c4^7*c6^18 + 500838400*c4^4*c6^20 - 66560000*c4*c6^22)*x*y^29
    +(-1796875*c4^33*c6 + 19377500*c4^30*c6^3 - 156270418*c4^27*c6^5 + 385955916*c4^24*c6^7 - 540905835*c4^21*c6^9 + 994455600*c4^18*c6^11 - 2916950160*c4^15*c6^13 + 5153887104*c4^12*c6^15 - 3906870528*c4^9*c6^17 + 715980800*c4^6*c6^19 + 412160000*c4^3*c6^21 - 102400000*c6^23)*y^30;

  when [5,4]:

DD := (91125*c4^9 + 7749*c4^6*c6^2 - 377*c4^3*c6^4 - c6^6)*x^12
    +240*(3699*c4^8*c6 + 194*c4^5*c6^3 - 5*c4^2*c6^5)*x^11*y
    +66*(-2025*c4^10 + 57231*c4^7*c6^2 + 1837*c4^4*c6^4 - 19*c4*c6^6)*x^10*y^2
    +440*(-3321*c4^9*c6 + 21073*c4^6*c6^3 + 393*c4^3*c6^5 - c6^7)*x^9*y^3
    +495*(-135*c4^11 - 13927*c4^8*c6^2 + 29331*c4^5*c6^4 + 283*c4^2*c6^6)*x^8*y^4
    +3168*(-156*c4^10*c6 - 5917*c4^7*c6^3 + 4758*c4^4*c6^5 + 19*c4*c6^7)*x^7*y^5
    +660*(9*c4^12 - 2479*c4^9*c6^2 - 49597*c4^6*c6^4 + 15763*c4^3*c6^6 + 16*c6^8)*x^6*y^6
    +1584*(17*c4^11*c6 - 1981*c4^8*c6^3 - 24281*c4^5*c6^5 + 2917*c4^2*c6^7)*x^5*y^7
    +495*(-3*c4^13 + 109*c4^10*c6^2 - 7649*c4^7*c6^4 - 62281*c4^4*c6^6 + 2432*c4*c6^8)*x^4*y^8
    +880*(-7*c4^12*c6 + 76*c4^9*c6^3 - 3339*c4^6*c6^5 - 18922*c4^3*c6^7 + 160*c6^9)*x^3*y^9
    +66*(-c4^14 - 137*c4^11*c6^2 + 837*c4^8*c6^4 - 21787*c4^5*c6^6 - 87776*c4^2*c6^8)*x^2*y^10
    +120*(-c4^13*c6 - 47*c4^10*c6^3 + 225*c4^7*c6^5 - 3409*c4^4*c6^7 - 9728*c4*c6^9)*x*y^11
    +(c4^15 - 63*c4^12*c6^2 - 1237*c4^9*c6^4 + 5587*c4^6*c6^6 - 52224*c4^3*c6^8 - 102400*c6^10)*y^12;

cc4 := (184528125*c4^19 + 273557250*c4^16*c6^2 - 37558269*c4^13*c6^4 + 9328924*c4^10*c6^6 + 121011*c4^7*c6^8 + 4674*c4^4*c6^10 - 19*c4*c6^12)*x^20
    +20*(227721375*c4^18*c6 + 218503494*c4^15*c6^3 - 21278799*c4^12*c6^5 + 4986388*c4^9*c6^7 + 48273*c4^6*c6^9 + 966*c4^3*c6^11 - c6^13)*x^19*y
    +190*(4100625*c4^20 + 259589610*c4^17*c6^2 + 175757391*c4^14*c6^4 - 11995316*c4^11*c6^6 + 2512671*c4^8*c6^8 + 16554*c4^5*c6^10 + 161*c4^2*c6^12)*x^18*y^2
    +1140*(11718675*c4^19*c6 + 282684654*c4^16*c6^3 + 141190301*c4^13*c6^5 - 6798652*c4^10*c6^7 + 1181853*c4^7*c6^9 + 4846*c4^4*c6^11 + 19*c4*c6^13)*x^17*y^3
    +285*(-273375*c4^21 + 384758910*c4^18*c6^2 + 5070881127*c4^15*c6^4 + 1911860728*c4^12*c6^6 - 66275393*c4^9*c6^8 + 8716758*c4^6*c6^10 + 20057*c4^3*c6^12 + 20*c6^14)*x^16*y^4
    +912*(-455625*c4^20*c6 + 623568438*c4^17*c6^3 + 5220302617*c4^14*c6^5 + 1501142836*c4^11*c6^7 - 38306439*c4^8*c6^9 + 3433142*c4^5*c6^11 + 3863*c4^2*c6^13)*x^15*y^5
    +2280*(91125*c4^22 + 522342*c4^19*c6^2 + 911511923*c4^16*c6^4 + 5262122536*c4^13*c6^6 + 1156205675*c4^10*c6^8 - 21964994*c4^7*c6^10 + 1199693*c4^4*c6^12 + 532*c4*c6^14)*x^14*y^6
    +4560*(566595*c4^21*c6 + 4056438*c4^18*c6^3 + 1242606077*c4^15*c6^5 + 5204670556*c4^12*c6^7 + 869612621*c4^9*c6^9 - 12182618*c4^6*c6^11 + 359123*c4^3*c6^13 + 40*c6^15)*x^13*y^7
    +7410*(-6075*c4^23 + 2012202*c4^20*c6^2 + 11393963*c4^17*c6^4 + 1609521316*c4^14*c6^6 + 5057516555*c4^11*c6^8 + 635501626*c4^8*c6^10 - 6337627*c4^5*c6^12 + 86872*c4^2*c6^14)*x^12*y^8
    +9880*(-59265*c4^22*c6 + 5399398*c4^19*c6^3 + 23259537*c4^16*c6^5 + 2004326196*c4^13*c6^7 + 4831377201*c4^10*c6^9 + 448355622*c4^7*c6^11 - 2985057*c4^4*c6^13 + 15200*c4*c6^15)*x^11*y^9
    +988*(-10125*c4^24 - 3437658*c4^21*c6^2 + 135098813*c4^18*c6^4 + 435078396*c4^15*c6^6 + 26609728125*c4^12*c6^8 + 49914045846*c4^9*c6^10 + 3329459763*c4^6*c6^12 - 13402008*c4^3*c6^14 + 16000*c6^16)*x^10*y^10
    +9880*(-10557*c4^23*c6 - 1175138*c4^20*c6^3 + 25240077*c4^17*c6^5 + 58879284*c4^14*c6^7 + 2846150829*c4^11*c6^9 + 4187980542*c4^8*c6^11 + 193028067*c4^5*c6^13 - 404272*c4^2*c6^15)*x^9*y^11
    +7410*(-135*c4^25 - 63898*c4^22*c6^2 - 3519937*c4^19*c6^4 + 48707616*c4^16*c6^6 + 77571975*c4^13*c6^8 + 3278827086*c4^10*c6^10 + 3794172513*c4^7*c6^12 + 114090892*c4^4*c6^14 - 97280*c4*c6^16)*x^8*y^12
    +4560*(-1497*c4^24*c6 - 273418*c4^21*c6^3 - 8906263*c4^18*c6^5 + 90221844*c4^15*c6^7 + 87870729*c4^12*c6^9 + 3712172982*c4^9*c6^11 + 3367951047*c4^6*c6^13 + 60666208*c4^3*c6^15 - 12800*c6^17)*x^7*y^13
    +2280*(45*c4^26 - 9298*c4^23*c6^2 - 930517*c4^20*c6^4 - 19717584*c4^17*c6^6 + 162007635*c4^14*c6^8 + 75099606*c4^11*c6^10 + 4145030133*c4^8*c6^12 + 2920613932*c4^5*c6^14 + 27594880*c4^2*c6^16)*x^6*y^14
    +912*(643*c4^25*c6 - 43258*c4^22*c6^3 - 2679203*c4^19*c6^5 - 39054516*c4^16*c6^7 + 282600909*c4^13*c6^9 + 13952022*c4^10*c6^11 + 4583110067*c4^7*c6^13 + 2462074168*c4^4*c6^15 + 9728000*c4*c6^17)*x^5*y^15
    +285*(-3*c4^27 + 4978*c4^24*c6^2 - 169885*c4^21*c6^4 - 6762068*c4^18*c6^6 - 70194957*c4^15*c6^8 + 478021842*c4^12*c6^10 - 135629235*c4^9*c6^12 + 5043176048*c4^6*c6^14 + 1999194112*c4^3*c6^16 + 2048000*c6^18)*x^4*y^16
    +1140*(-c4^26*c6 + 1606*c4^23*c6^3 - 35119*c4^20*c6^5 - 897708*c4^17*c6^7 - 6794703*c4^14*c6^9 + 45994566*c4^11*c6^11 - 25419809*c4^8*c6^13 + 326928704*c4^5*c6^15 + 90204160*c4^2*c6^17)*x^3*y^17
    +190*(c4^28 - 6*c4^25*c6^2 + 6975*c4^22*c6^4 - 113444*c4^19*c6^6 - 1826961*c4^16*c6^8 - 10297254*c4^13*c6^10 + 72602225*c4^10*c6^12 - 56164176*c4^7*c6^14 + 363515136*c4^4*c6^16 + 62259200*c4*c6^18)*x^2*y^18
    +20*(19*c4^27*c6 - 114*c4^24*c6^3 + 25821*c4^21*c6^5 - 341468*c4^18*c6^7 - 3341283*c4^15*c6^9 - 14464434*c4^12*c6^11 + 110542931*c4^9*c6^13 - 106118496*c4^6*c6^15 + 410910720*c4^3*c6^17 + 32768000*c6^19)*x*y^19
    +(c4^29 + 174*c4^26*c6^2 - 1305*c4^23*c6^4 + 86536*c4^20*c6^6 - 969441*c4^17*c6^8 - 5361114*c4^14*c6^10 - 19134055*c4^11*c6^12 + 162677124*c4^8*c6^14 - 182452224*c4^5*c6^16 + 475136000*c4^2*c6^18)*y^20;

cc6 := (7099719609375*c4^27*c6 + 722009345625*c4^24*c6^3 + 1403913480876*c4^21*c6^5 - 310180334364*c4^18*c6^7 + 1937555666*c4^15*c6^9 - 1278908050*c4^12*c6^11 - 20419300*c4^9*c6^13 + 120916*c4^6*c6^15 - 2489*c4^3*c6^17 + c6^19)*x^30
    +270*(41518828125*c4^29 + 760005736875*c4^26*c6^2 + 108606531780*c4^23*c6^4 +102242844396*c4^20*c6^6 - 21665064394*c4^17*c6^8 + 31038410*c4^14*c6^10 - 
 61384300*c4^11*c6^12 - 705860*c4^8*c6^14 + 2581*c4^5*c6^16 - 29*c4^2*c6^18)*x^29*y
    +435*(764500021875*c4^28*c6 + 6630097335525*c4^25*c6^3 + 1042742211516*c4^22*c6^5 + 599406407316*c4^19*c6^7 - 120267663254*c4^16*c6^9 - 151687850*c4^13*c6^11 - 224350100*c4^10*c6^13 - 1830844*c4^7*c6^15 + 4091*c4^4*c6^17 - 19*c4*c6^19)*x^28*y^2
    +580*(16607531250*c4^30 + 8264549707875*c4^27*c6^2 + 44931149867115*c4^24*c6^4 + 6990936234588*c4^21*c6^6 + 2718570122568*c4^18*c6^8 - 507261990550*c4^15*c6^10 - 1248589150*c4^12*c6^12 - 596387380*c4^9*c6^14 - 3362882*c4^6*c6^16 + 4363*c4^3*c6^18 - 5*c6^20)*x^27*y^3
    +3915*(78473660625*c4^29*c6 + 11338919198775*c4^26*c6^3 + 43255926806292*c4^23*c6^5 + 6283843169372*c4^20*c6^7 + 1753897960782*c4^17*c6^9 - 297340567950*c4^14*c6^11 - 806071900*c4^11*c6^13 - 210231828*c4^8*c6^15 - 786903*c4^5*c6^17 + 527*c4^2*c6^19)*x^26*y^4
    +1566*(11256215625*c4^31 + 2824465393125*c4^28*c6^2 + 188752989992670*c4^25*c6^4 + 537010126138164*c4^22*c6^6 + 70308831130054*c4^19*c6^8 + 14659087109110*c4^16*c6^10 - 2195391861200*c4^13*c6^12 - 5335768540*c4^10*c6^14 - 885482071*c4^7*c6^16 - 2076211*c4^4*c6^18 + 570*c4*c6^20)*x^25*y^5
    +6525*(60050372625*c4^30*c6 + 6017256272799*c4^27*c6^3 + 230528201413356*c4^24*c6^5 + 508028485468924*c4^21*c6^7 + 58545230118062*c4^18*c6^9 + 9391692158882*c4^15*c6^11 - 1203207286668*c4^12*c6^13 - 2307153908*c4^9*c6^15 - 260221751*c4^6*c6^17 - 351049*c4^3*c6^19 + 24*c6^21)*x^24*y^6
    +156600*(-49207500*c4^32 + 26981926605*c4^29*c6^2 + 1561882312215*c4^26*c6^4 + 38815713245252*c4^23*c6^6 + 67965363224372*c4^20*c6^8 + 6774705544470*c4^17*c6^10 + 855343041890*c4^14*c6^12 - 90501110380*c4^11*c6^14 - 123410688*c4^8*c6^16 - 9732283*c4^5*c6^18 - 6625*c4^2*c6^20)*x^23*y^7
    +450225*(-352489725*c4^31*c6 + 65388931317*c4^28*c6^3 + 2548488204828*c4^25*c6^5 + 44471599972692*c4^22*c6^7 + 62976533820746*c4^19*c6^9 + 5352321070006*c4^16*c6^11 + 540912120556*c4^13*c6^13 - 45543647484*c4^10*c6^15 - 39942133*c4^7*c6^17 - 2212867*c4^4*c6^19 - 608*c4*c6^21)*x^22*y^8
    +100050*(22143375*c4^33 - 15614389035*c4^30*c6^2 + 1487745679680*c4^27*c6^4 + 42316700625812*c4^24*c6^6 + 547452732935042*c4^21*c6^8 + 634884661017590*c4^18*c6^10 + 45401549571060*c4^15*c6^12 + 3714178094340*c4^12*c6^14 - 239486018553*c4^9*c6^16 - 120930083*c4^6*c6^18 - 4628300*c4^3*c6^20 - 320*c6^22)*x^21*y^9
    +90045*(524849625*c4^32*c6 - 108601015617*c4^29*c6^3 + 6454263997732*c4^26*c6^5 + 140483676509948*c4^23*c6^7 + 1401110628880894*c4^20*c6^9 + 1342318842221794*c4^17*c6^11 + 79692754227044*c4^14*c6^13 + 5305293951244*c4^11*c6^15 - 251801822927*c4^8*c6^17 - 63949993*c4^5*c6^19 - 1614992*c4^2*c6^21)*x^20*y^10
    +600300*(546750*c4^34 + 809245647*c4^31*c6^2 - 72887726781*c4^28*c6^4 + 3031211906060*c4^25*c6^6 + 52018668688784*c4^22*c6^8 + 411821183394962*c4^19*c6^10 + 327808754995842*c4^16*c6^12 + 15924468625212*c4^13*c6^14 + 861816707274*c4^10*c6^16 - 28882474025*c4^7*c6^18 - 3072221*c4^4*c6^20 - 46208*c4*c6^22)*x^19*y^11
    +950475*(9043245*c4^33*c6 + 3333776859*c4^30*c6^3 - 157118979804*c4^27*c6^5 + 4925225988844*c4^24*c6^7 + 67972145509462*c4^21*c6^9 + 436901557383082*c4^18*c6^11 + 289899747378132*c4^15*c6^13 + 11336491103292*c4^12*c6^15 + 494940620709*c4^9*c6^17 - 11190653869*c4^6*c6^19 - 376096*c4^3*c6^21 - 2560*c6^23)*x^18*y^12
    +17108550*(18225*c4^35 + 5486049*c4^32*c6^2 + 854104378*c4^29*c6^4 - 23708622300*c4^26*c6^6 + 587964700518*c4^23*c6^8 + 6615394338494*c4^20*c6^10 + 35148001663704*c4^17*c6^12 + 19462354954644*c4^14*c6^14 + 601274261633*c4^11*c6^16 + 20864089065*c4^8*c6^18 - 301854762*c4^5*c6^20 - 1856*c4^2*c6^22)*x^17*y^13
    +145422675*(35235*c4^34*c6 + 4123869*c4^31*c6^3 + 346872596*c4^28*c6^5 - 6213405196*c4^25*c6^7 + 125559417962*c4^22*c6^9 + 1162889504742*c4^19*c6^11 + 5182404676092*c4^16*c6^13 + 2392442631972*c4^13*c6^15 + 57140056459*c4^10*c6^17 + 1539880741*c4^7*c6^19 - 13346216*c4^4*c6^21)*x^16*y^14
    +155117520*(253017*c4^33*c6^2 + 16610639*c4^30*c6^4 + 870176500*c4^27*c6^6 - 10880303876*c4^24*c6^8 + 181930292222*c4^21*c6^10 + 1396222825602*c4^18*c6^12 + 5283395043972*c4^15*c6^14 + 2027311203724*c4^12*c6^16 + 36498770225*c4^9*c6^18 + 739088039*c4^6*c6^20 - 3511808*c4^3*c6^22)*x^15*y^15
    +145422675*(-513*c4^35*c6 + 1285561*c4^32*c6^3 + 55117804*c4^29*c6^5 + 1970012772*c4^26*c6^7 - 18462473710*c4^23*c6^9 + 256821071534*c4^20*c6^11 + 1639719654684*c4^17*c6^13 + 5324264372180*c4^14*c6^15 + 1689151228967*c4^11*c6^17 + 22257485409*c4^8*c6^19 + 323432896*c4^5*c6^21 - 739328*c4^2*c6^23)*x^14*y^16
    +17108550*(-405*c4^37 - 54171*c4^34*c6^2 + 36397188*c4^31*c6^4 + 1102871380*c4^28*c6^6 + 28631572058*c4^25*c6^8 - 214669470826*c4^22*c6^10 + 2483465864764*c4^19*c6^12 + 13208640780804*c4^16*c6^14 + 37147072433043*c4^13*c6^16 + 9667489675565*c4^10*c6^18 + 90053251048*c4^7*c6^20 + 880595584*c4^4*c6^22 - 778240*c4*c6^24)*x^13*y^17
    +950475*(-93663*c4^36*c6 - 5560249*c4^33*c6^3 + 1624819124*c4^30*c6^5 + 36351115292*c4^27*c6^7 + 717312713390*c4^24*c6^9 - 4562967037486*c4^21*c6^11 + 43824927801764*c4^18*c6^13 + 193859129964460*c4^15*c6^15 + 476285387696377*c4^12*c6^17 + 100564173873439*c4^9*c6^19 + 635393753056*c4^6*c6^21 + 3812564992*c4^3*c6^23 - 819200*c6^25)*x^12*y^18
    +1800900*(-90*c4^38 - 292771*c4^35*c6^2 - 10143587*c4^32*c6^4 + 1644065060*c4^29*c6^6 + 27792670888*c4^26*c6^8 + 431692995734*c4^23*c6^10 - 2452519987346*c4^20*c6^12 + 19561032389044*c4^17*c6^14 + 71612943251258*c4^14*c6^16 + 155009484234805*c4^11*c6^18 + 26156064327533*c4^8*c6^20 + 106582161664*c4^5*c6^22 + 341258240*c4^2*c6^24)*x^11*y^19
    +30015*(-68193*c4^37*c6 - 63305959*c4^34*c6^3 - 1418731156*c4^31*c6^5 + 149575353572*c4^28*c6^7 + 1928471456210*c4^25*c6^9 + 24274188707054*c4^22*c6^11 - 128613599652196*c4^19*c6^13 + 855711089906260*c4^16*c6^15 + 2570367798459847*c4^13*c6^17 + 4946121872645569*c4^10*c6^19 + 653465347335616*c4^7*c6^21 + 1611107525632*c4^4*c6^23 + 2179072000*c4*c6^25)*x^10*y^20
    +100050*(-243*c4^39 - 108543*c4^36*c6^2 - 46398098*c4^33*c6^4 - 703905500*c4^30*c6^6 + 54509601790*c4^27*c6^8 + 535424107582*c4^24*c6^10 + 5610956183872*c4^21*c6^12 - 28722810027308*c4^18*c6^14 + 160934881319885*c4^15*c6^16 + 391506441167945*c4^12*c6^18 + 676902610617722*c4^9*c6^20 + 68087842289792*c4^6*c6^22 + 93226483712*c4^3*c6^24 + 32768000*c6^26)*x^9*y^21
    +450225*(-409*c4^38*c6 - 73527*c4^35*c6^3 - 18020268*c4^32*c6^5 - 187716444*c4^29*c6^7 + 11847634530*c4^26*c6^9 + 87768942222*c4^23*c6^11 + 788093862572*c4^20*c6^13 - 4007620094700*c4^17*c6^15 + 19177003806831*c4^14*c6^17 + 37102875773937*c4^11*c6^19 + 58441765339848*c4^8*c6^21 + 4304889839616*c4^5*c6^23 + 2886533120*c4^2*c6^25)*x^8*y^22
    +156600*(12*c4^40 - 4163*c4^37*c6^2 - 417233*c4^34*c6^4 - 66624380*c4^31*c6^6 - 470862620*c4^28*c6^8 + 26775806902*c4^25*c6^10 + 146676718162*c4^22*c6^12 + 1169286078932*c4^19*c6^14 - 6029881626760*c4^16*c6^16 + 25071228087605*c4^13*c6^18 + 37635483016727*c4^10*c6^20 + 55078694676992*c4^7*c6^22 + 2810833866752*c4^4*c6^24 + 747110400*c4*c6^26)*x^7*y^23
    +6525*(1717*c4^39*c6 - 216109*c4^36*c6^3 - 13504636*c4^33*c6^5 - 1531961908*c4^30*c6^7 - 7016485690*c4^27*c6^9 + 402612501434*c4^24*c6^11 + 1577858042484*c4^21*c6^13 + 11740428016540*c4^18*c6^15 - 62218973284883*c4^15*c6^17 + 229062682715659*c4^12*c6^19 + 257809347814816*c4^9*c6^21 + 361234532411392*c4^6*c6^23 + 11764428308480*c4^3*c6^25 + 786432000*c6^27)*x^6*y^24
    +1566*(-61*c4^41 + 19049*c4^38*c6^2 - 1302936*c4^35*c6^4 - 52904220*c4^32*c6^6 - 4557768390*c4^29*c6^8 - 12092175666*c4^26*c6^10 + 824646196244*c4^23*c6^12 + 2185361480084*c4^20*c6^14 + 16425096458235*c4^17*c6^16 - 89885842810095*c4^14*c6^18 + 298551401031564*c4^11*c6^20 + 239597451791424*c4^8*c6^22 + 337518267736064*c4^5*c6^24 + 6165463040000*c4^2*c6^26)*x^5*y^25
    +3915*(-123*c4^40*c6 + 11987*c4^37*c6^3 - 521420*c4^34*c6^5 - 13806260*c4^31*c6^7 - 958871450*c4^28*c6^9 - 1060213958*c4^25*c6^11 + 124082583892*c4^22*c6^13 + 199498445340*c4^19*c6^15 + 1744885271165*c4^16*c6^17 - 9803757357685*c4^13*c6^19 + 29881659306512*c4^10*c6^21 + 15751991140352*c4^7*c6^23 + 24317147709440*c4^4*c6^25 + 199229440000*c4*c6^27)*x^4*y^26
    +580*(-2*c4^42 - 1637*c4^39*c6^2 + 81943*c4^36*c6^4 - 2439940*c4^33*c6^6 - 40850080*c4^30*c6^8 - 2440689862*c4^27*c6^10 + 429228378*c4^24*c6^12 + 231837782508*c4^21*c6^14 + 171648722970*c4^18*c6^16 + 2415795735235*c4^15*c6^18 - 13677800127977*c4^12*c6^20 + 38796757161088*c4^9*c6^22 + 11408947539968*c4^6*c6^24 + 23015142195200*c4^3*c6^26 + 52428800000*c6^28)*x^3*y^27
    +3915*(-c4^41*c6 - 231*c4^38*c6^3 + 7820*c4^35*c6^5 - 168220*c4^32*c6^7 - 1648750*c4^29*c6^9 - 92816146*c4^26*c6^11 + 117303004*c4^23*c6^13 + 6559725140*c4^20*c6^15 + 21740455*c4^17*c6^17 + 54135520705*c4^14*c6^19 - 299123236256*c4^11*c6^21 + 797563201024*c4^8*c6^23 + 78128087040*c4^5*c6^25 + 353370112000*c4^2*c6^27)*x^2*y^28
    +30*(-c4^43 - 121*c4^40*c6^2 - 13906*c4^37*c6^4 + 377980*c4^34*c6^6 - 6231590*c4^31*c6^8 - 30273806*c4^28*c6^10 - 1906223976*c4^25*c6^12 + 4193117964*c4^22*c6^14 + 100093350735*c4^19*c6^16 - 64488375745*c4^16*c6^18 + 710988880274*c4^13*c6^20 - 3679483160896*c4^10*c6^22 + 9274149535744*c4^7*c6^24 - 615081574400*c4^4*c6^26 + 3187671040000*c4*c6^28)*x*y^29
    +(-29*c4^42*c6 - 899*c4^39*c6^3 - 75980*c4^36*c6^5 + 1826420*c4^33*c6^7 - 24544150*c4^30*c6^9 - 35187034*c4^27*c6^11 - 4203625284*c4^24*c6^13 + 12590561700*c4^21*c6^15 + 160378782795*c4^18*c6^17 - 199721488955*c4^15*c6^19 + 1076947257976*c4^12*c6^21 - 5013938683904*c4^9*c6^23 + 11961358745600*c4^6*c6^25 - 2432696320000*c4^3*c6^27 + 3355443200000*c6^29)*y^30;

  end case;
  return DD,cc4,cc6,eps,alpha;
end intrinsic;

intrinsic RubinSilverbergPolynomials(n::RngIntElt,J::RngElt : 
                                     Parameter:=PolynomialRing(Parent(J)).1) 
       -> RngElt,RngElt
{Polynomials alpha(t) and beta(t) that determine a family of elliptic curves with fixed $n$-torsion.
Let n = 2,3,4 or 5 and let E : y^2 = x^3 + A x + B be an elliptic curve over Q 
with j-invariant 1728J.  Then every non-singular member of the family 
EE(t) : y^2 = x^3 + A alpha(t) x + B beta(t) has n-torsion isomorphic to that of E 
(as a Galois module equipped with the Weil pairing). 
If n = 3,4,5 with A and B nonzero, then all such n-congruent curves arise in this way.} 

  t := Parameter;

  require n in [2..5] : "The integer n must be in the range [2..5]";
  
  case n :

  when 2 :

    a := (J - 1)*t^2 + 1;
    b := (-J + 1)*t^3 + (-3*J + 3)*t^2 + 3*t + 1;

  when 3 :

    a := (-3*J^2 + 6*J - 3)*t^4 + (-8*J^2 + 16*J - 8)*t^3 + (6*J - 6)*t^2 + 1;
    b := (-8*J^4 + 23*J^3 - 21*J^2 + 5*J + 1)*t^6 + (12*J^3 - 18*J^2 + 
      6)*t^5 + (15*J^2 - 30*J + 15)*t^4 + (20*J^2 - 40*J + 20)*t^3
      + (-15*J + 15)*t^2 + 6*t + 1;

  when 4 :

    a := (144*J^6 - 632*J^5 + 1081*J^4 - 884*J^3 + 326*J^2 - 28*J - 
      7)*t^8 + (-192*J^5 + 720*J^4 - 960*J^3 + 480*J^2 - 48)*t^7 +
      (112*J^4 - 196*J^3 - 84*J^2 + 308*J - 140)*t^6 + (224*J^3 - 
      672*J^2 + 672*J - 224)*t^5 + (168*J^3 - 546*J^2 + 588*J - 
      210)*t^4 + (-112*J^2 + 224*J - 112)*t^3 + (28*J - 28)*t^2 + 1; 
    b := (1728*J^9 - 10512*J^8 + 26900*J^7 - 37415*J^6 + 30534*J^5 - 
      14833*J^4 + 4312*J^3 - 825*J^2 + 110*J + 1)*t^12 + 
      (-3456*J^8 + 18816*J^7 - 43224*J^6 + 54828*J^5 - 42420*J^4 +
      20856*J^3 - 6336*J^2 + 924*J + 12)*t^11 + (3168*J^7 - 
      19536*J^6 + 50094*J^5 - 68310*J^4 + 52140*J^3 - 20988*J^2 + 
      3366*J + 66)*t^10 + (-9152*J^6 + 44352*J^5 - 85668*J^4 + 
      82192*J^3 - 38808*J^2 + 6864*J + 220)*t^9 + (-4752*J^6 + 
      29304*J^5 - 69201*J^4 + 78804*J^3 - 42966*J^2 + 8316*J + 
      495)*t^8 + (6336*J^5 - 26928*J^4 + 41976*J^3 - 27720*J^2 + 
      5544*J + 792)*t^7 + (-3696*J^4 + 10164*J^3 - 8316*J^2 + 
      924*J + 924)*t^6 + (792*J^2 - 1584*J + 792)*t^5 + (-396*J^3 
      + 1287*J^2 - 1386*J + 495)*t^4 + (264*J^2 - 484*J + 220)*t^3
      + (-66*J + 66)*t^2 + 12*t + 1;

  when 5 :

  a := (429981696*J^16 - 4526751744*J^15 + 21609725184*J^14 - 
    61731738368*J^13 + 117231113696*J^12 - 155400930192*J^11 + 
    146825586029*J^10 - 99124721618*J^9 + 47248922601*J^8 - 
    15563731512*J^7 + 3517900266*J^6 - 600859116*J^5 + 
    96821834*J^4 - 11503208*J^3 + 188841*J^2 - 4370*J - 19)*t^20
    + (-716636160*J^15 + 7110789120*J^14 - 31786306560*J^13 + 
    84431439360*J^12 - 148046356800*J^11 + 180213217560*J^10 - 
    156325857840*J^9 + 98254824120*J^8 - 45668560320*J^7 + 
    16274401200*J^6 - 4597547040*J^5 + 968940720*J^4 - 
    114802560*J^3 + 2523960*J^2 - 68400*J - 360)*t^19 + 
    (567336960*J^14 - 5285952000*J^13 + 22445608960*J^12 - 
    57764274240*J^11 + 101176878720*J^10 - 128341470370*J^9 + 
    121949838450*J^8 - 87649832520*J^7 + 46825091880*J^6 - 
    17659159740*J^5 + 4241049660*J^4 - 520292200*J^3 + 
    15679560*J^2 - 499890*J - 3230)*t^18 + (-283668480*J^13 + 
    3537976320*J^12 - 19115592960*J^11 + 59639127360*J^10 - 
    119900640000*J^9 + 163200922560*J^8 - 153036080640*J^7 + 
    98152686720*J^6 - 41581509120*J^5 + 10749744000*J^4 - 
    1420750080*J^3 + 60064320*J^2 - 2261760*J - 18240)*t^17 + 
    (1447891200*J^12 - 13200652800*J^11 + 53580592800*J^10 - 
    127175846400*J^9 + 194779597725*J^8 - 200091853800*J^7 + 
    138510974700*J^6 - 62975367000*J^5 + 17596695150*J^4 - 
    2623584600*J^3 + 158705100*J^2 - 7079400*J - 72675)*t^16 + 
    (680804352*J^12 - 7109047296*J^11 + 32532338688*J^10 - 
    86047480896*J^9 + 145635676704*J^8 - 164468183040*J^7 + 
    124992233856*J^6 - 62703407232*J^5 + 19701487296*J^4 - 
    3504443904*J^3 + 306453888*J^2 - 16215360*J - 217056)*t^15 +
    (-851005440*J^11 + 7220413440*J^10 - 26951387520*J^9 + 
    58004130240*J^8 - 79193143800*J^7 + 71030602440*J^6 - 
    41863734360*J^5 + 15750351720*J^4 - 3564536040*J^3 + 
    446823000*J^2 - 28009800*J - 503880)*t^14 + (496419840*J^10 
    - 3690754560*J^9 + 11979028800*J^8 - 22210410240*J^7 + 
    25835281920*J^6 - 19481414400*J^5 + 9497093760*J^4 - 
    2888121600*J^3 + 500797440*J^2 - 36990720*J - 930240)*t^13 +
    (-179262720*J^9 + 1256854560*J^8 - 3822255840*J^7 + 
    6576915930*J^6 - 6990875580*J^5 + 4686876870*J^4 - 
    1923902760*J^3 + 434366790*J^2 - 37331580*J - 1385670)*t^12 
    + (85363200*J^8 - 550513600*J^7 + 1508774800*J^6 - 
    2272202400*J^5 + 2021942000*J^4 - 1053603200*J^3 + 
    290175600*J^2 - 28256800*J - 1679600)*t^11 + (10243584*J^8 -
    96602688*J^7 + 352834560*J^6 - 671995116*J^5 + 731500380*J^4
    - 455288184*J^3 + 146131128*J^2 - 15160860*J - 1662804)*t^10
    + (-8536320*J^7 + 54142400*J^6 - 141323520*J^5 + 
    193252800*J^4 - 143852800*J^3 + 52403520*J^2 - 4742400*J - 
    1343680)*t^9 + (3201120*J^6 - 16124160*J^5 + 31603650*J^4 - 
    29195400*J^3 + 11189100*J^2 + 207480*J - 881790)*t^8 + 
    (-711360*J^5 + 2380320*J^4 - 2407680*J^3 + 54720*J^2 + 
    1149120*J - 465120)*t^7 + (-91200*J^4 + 467400*J^3 - 
    855000*J^2 + 672600*J - 193800)*t^6 + (-32832*J^4 + 
    160512*J^3 - 284544*J^2 + 218880*J - 62016)*t^5 + (13680*J^3
    - 41895*J^2 + 42750*J - 14535)*t^4 + (-2280*J^2 + 4560*J - 
    2280)*t^3 + (190*J - 190)*t^2 + 1;

  b := (8916100448256*J^24 - 136342036021248*J^23 + 
    975670391439360*J^22 - 4336644230369280*J^21 + 
    13398528846385152*J^20 - 30506275618506496*J^19 + 
    52934493019074688*J^18 - 71374747273861680*J^17 + 
    75516444221102880*J^16 - 62800290811934817*J^15 + 
    40768321843732527*J^14 - 20273027024648841*J^13 + 
    7404256814051175*J^12 - 1782815758078005*J^11 + 
    169267957804379*J^10 + 56929905429235*J^9 - 
    27594970589565*J^8 + 5425241915325*J^7 - 542716092435*J^6 + 
    27764445285*J^5 - 1684179147*J^4 + 18386841*J^3 + 63945*J^2 
    + 2465*J + 1)*t^30 + (-22290251120640*J^23 + 
    327362264432640*J^22 - 2243330782248960*J^21 + 
    9512585227100160*J^20 - 27898441737600000*J^19 + 
    59882631121251840*J^18 - 97001706525791040*J^17 + 
    120327606784294560*J^16 - 114448382279635860*J^15 + 
    82230414473049750*J^14 - 42651213866974080*J^13 + 
    13860850919372730*J^12 - 873768511846620*J^11 - 
    1835259757706730*J^10 + 1128685169685000*J^9 - 
    355476352886550*J^8 + 66486494834100*J^7 - 7193487646350*J^6
    + 470820092400*J^5 - 30096497250*J^4 + 374373180*J^3 + 
    1401570*J^2 + 66120*J + 30)*t^29 + (26934053437440*J^22 - 
    379258905231360*J^21 + 2463868538449920*J^20 - 
    9754453180569600*J^19 + 26137886281708800*J^18 - 
    49613977007792640*J^17 + 67319840321922960*J^16 - 
    62928769861118520*J^15 + 33890783598148275*J^14 + 
    2387260886626350*J^13 - 24406837918402095*J^12 + 
    25886638613293380*J^11 - 16254866653167285*J^10 + 
    6866852687249250*J^9 - 1976062015536975*J^8 + 
    376341175749600*J^7 - 45688806136575*J^6 + 3758659290450*J^5
    - 254101097925*J^4 + 3617903700*J^3 + 14418945*J^2 + 
    853470*J + 435)*t^28 + (-20948708229120*J^21 + 
    231656759377920*J^20 - 1050922725580800*J^19 + 
    2109577985920000*J^18 + 933326438630400*J^17 - 
    17919719726575680*J^16 + 56244819844426080*J^15 - 
    106386371766312300*J^14 + 141011469300513600*J^13 - 
    137576002368171900*J^12 + 100472307023612160*J^11 - 
    54917931775072860*J^10 + 22185679250644000*J^9 - 
    6470190843734700*J^8 + 1321700050965600*J^7 - 
    185888912607300*J^6 + 18765147484800*J^5 - 1347135030900*J^4
    + 22060833600*J^3 + 91845900*J^2 + 7057440*J + 4060)*t^27 + 
    (-52184728535040*J^20 + 783309972787200*J^19 - 
    5443421933260800*J^18 + 23231063553696000*J^17 - 
    68082366173644800*J^16 + 145095905258025840*J^15 - 
    232222256074738800*J^14 + 284115581392734675*J^13 - 
    267856868132342775*J^12 + 194662754528117850*J^11 - 
    108364119457123530*J^10 + 45637603938824625*J^9 - 
    14295916472084325*J^8 + 3275220088248300*J^7 - 
    545012466648300*J^6 + 65638463480325*J^5 - 5027378483025*J^4
    + 95178465450*J^3 + 400465350*J^2 + 41988375*J + 27405)*t^26    
    + (-32320864124928*J^20 + 538400505692160*J^19 - 
    4093018397982720*J^18 + 18912976662067200*J^17 - 
    59570249328852480*J^16 + 135750691986427008*J^15 - 
    231612638844810240*J^14 + 301794187224123720*J^13 - 
    303474473302317750*J^12 + 236345767468340760*J^11 - 
    142305083952958476*J^10 + 65858066006349600*J^9 - 
    23231103944837130*J^8 + 6184828442386800*J^7 - 
    1223090882781480*J^6 + 170782623917640*J^5 - 
    14031631849770*J^4 + 308806085880*J^3 + 1233616500*J^2 + 
    191365200*J + 142506)*t^25 + (67335133593600*J^19 - 
    945400083148800*J^18 + 6133510356480000*J^17 - 
    24391443933100800*J^16 + 66497559554304000*J^15 - 
    131684596701166800*J^14 + 195817420773018000*J^13 - 
    222976315206493425*J^12 + 196572476626022700*J^11 - 
    134827353284000850*J^10 + 71946737803284300*J^9 - 
    29705307613314975*J^8 + 9351467446224600*J^7 - 
    2168974008031500*J^6 + 342472863256200*J^5 - 
    30374845782975*J^4 + 781838523900*J^3 + 2585191950*J^2 + 
    694547100*J + 593775)*t^24 + (-67335133593600*J^18 + 
    857782578585600*J^17 - 5065781629593600*J^16 + 
    18401810067993600*J^15 - 46003421408083200*J^14 + 
    83898736512228000*J^13 - 115431892681624200*J^12 + 
    122153876823412800*J^11 - 100362698019500400*J^10 + 
    64055359751882400*J^9 - 31460952408514200*J^8 + 
    11620258380777600*J^7 - 3086270461648800*J^6 + 
    541135224396000*J^5 - 52194832774200*J^4 + 1582400404800*J^3
    + 2773386000*J^2 + 2060229600*J + 2035800)*t^23 + 
    (43019668684800*J^17 - 537148363161600*J^16 + 
    3111741627667200*J^15 - 11083892789606400*J^14 + 
    27109563996794400*J^13 - 48128944049802000*J^12 + 
    63882157032508275*J^11 - 64274573026350225*J^10 + 
    49112173882883925*J^9 - 28227287034572175*J^8 + 
    11922285748492350*J^7 - 3542426132402250*J^6 + 
    682995334703850*J^5 - 72270843663150*J^4 + 2603416607775*J^3
    - 3560829525*J^2 + 5086191825*J + 5852925)*t^22 + 
    (-36745967001600*J^16 + 443262749414400*J^15 - 
    2456817247526400*J^14 + 8276687757340800*J^13 - 
    18886286547261600*J^12 + 30777404110949700*J^11 - 
    36773715114022650*J^10 + 32523150163957800*J^9 - 
    21201155303711850*J^8 + 9995910958150200*J^7 - 
    3277751965154100*J^6 + 693946146873600*J^5 - 
    81400697808900*J^4 + 3526136787300*J^3 - 25782584850*J^2 + 
    10587291000*J + 14307150)*t^21 + (-4301966868480*J^16 + 
    78705081354240*J^15 - 585514440034560*J^14 + 
    2477086172444160*J^13 - 6806662627298400*J^12 + 
    12978487155697560*J^11 - 17770789032969465*J^10 + 
    17735739221031570*J^9 - 12899452112094645*J^8 + 
    6732717081402600*J^7 - 2434346368361730*J^6 + 
    569404112696460*J^5 - 74996005283010*J^4 + 3977678565360*J^3
    - 72737000325*J^2 + 18756673650*J + 30045015)*t^20 + 
    (7169944780800*J^15 - 92416770892800*J^14 + 
    533424620620800*J^13 - 1828895859734400*J^12 + 
    4155484342219200*J^11 - 6595554095583300*J^10 + 
    7494861016929600*J^9 - 6133799161097100*J^8 + 
    3576690180964800*J^7 - 1441071438431400*J^6 + 
    377131021675200*J^5 - 56699719291800*J^4 + 3791072188800*J^3
    - 143677602900*J^2 + 28468627200*J + 54627300)*t^19 + 
    (-5676206284800*J^14 + 62090075923200*J^13 - 
    306195171475200*J^12 + 898160278111200*J^11 - 
    1739838161138400*J^10 + 2334224048709975*J^9 - 
    2209415435189775*J^8 + 1471690781451900*J^7 - 
    673642392669900*J^6 + 200963375572050*J^5 - 
    35297864212050*J^4 + 3121029134700*J^3 - 221624156700*J^2 + 
    37179730575*J + 86493225)*t^18 + (2838103142400*J^13 - 
    27553251340800*J^12 + 119848540723200*J^11 - 
    307382337561600*J^10 + 514370198170800*J^9 - 
    586226946489750*J^8 + 460340576060400*J^7 - 
    245841856698600*J^6 + 85822645878000*J^5 - 
    18276140103300*J^4 + 2296514883600*J^3 - 278048154600*J^2 + 
    41881730400*J + 119759850)*t^17 + (-1005161529600*J^12 + 
    8785856332800*J^11 - 34059153866400*J^10 + 
    76843668232800*J^9 - 111219116417325*J^8 + 
    107234680545000*J^7 - 69249696144300*J^6 + 
    29467287480600*J^5 - 8124183161550*J^4 + 1574055034200*J^3 -
    289100277900*J^2 + 40718349000*J + 145422675)*t^16 + 
    (268043074560*J^11 - 2117043912960*J^10 + 7322167414080*J^9 
    - 14528772275760*J^8 + 18223206249600*J^7 - 
    15075561533760*J^6 + 8443356848640*J^5 - 3349607726880*J^4 +
    1031221272960*J^3 - 251290382400*J^2 + 34125854400*J + 
    155117520)*t^15 + (-55842307200*J^10 + 397876438800*J^9 - 
    1247144860800*J^8 + 2280082121325*J^7 - 2735255094075*J^6 + 
    2309748347025*J^5 - 1428196091175*J^4 + 637096739175*J^3 - 
    183087147825*J^2 + 24576432075*J + 145422675)*t^14 + 
    (9307051200*J^9 - 74182672800*J^8 + 265490478900*J^7 - 
    556181851950*J^6 + 743332280400*J^5 - 646001739450*J^4 + 
    354591807300*J^3 - 111599071650*J^2 + 15123958200*J + 
    119759850)*t^13 + (-7436516400*J^8 + 53021297400*J^7 - 
    161874446775*J^6 + 274244353650*J^5 - 278294327625*J^4 + 
    168922218900*J^3 - 56552312025*J^2 + 7883239650*J + 
    86493225)*t^12 + (-1642420800*J^8 + 13778085600*J^7 - 
    47960968500*J^6 + 90563659200*J^5 - 100667308500*J^4 + 
    65994580800*J^3 - 23553971100*J^2 + 3433716000*J + 
    54627300)*t^11 + (-207463680*J^8 + 2384391600*J^7 - 
    10183489200*J^6 + 22382755785*J^5 - 28092689325*J^4 + 
    20377243530*J^3 - 7914055050*J^2 + 1223261325*J + 
    30045015)*t^10 + (172886400*J^7 - 1255027200*J^6 + 
    3691845000*J^5 - 5660528850*J^4 + 4773185400*J^3 - 
    2080039500*J^2 + 343371600*J + 14307150)*t^9 + 
    (-64832400*J^6 + 352976400*J^5 - 757728675*J^4 + 
    797798700*J^3 - 404302050*J^2 + 70235100*J + 5852925)*t^8 + 
    (14407200*J^5 - 57472200*J^4 + 83937600*J^3 - 51051600*J^2 +
    8143200*J + 2035800)*t^7 + (-1357200*J^4 + 3477825*J^3 - 
    2290275*J^2 - 424125*J + 593775)*t^6 + (75168*J^4 - 
    272484*J^3 + 461970*J^2 - 407160*J + 142506)*t^5 + 
    (-31320*J^3 + 90045*J^2 - 86130*J + 27405)*t^4 + (5220*J^2 -
    9280*J + 4060)*t^3 + (-435*J + 435)*t^2 + 30*t + 1;

  end case;
  return a,b;
end intrinsic;



