freeze;

import "getaltgens.m" : GetAltGens;
import "elttoperm.m" : EltToPerm;
import "permasword.m" : PermAsWord;

/* 
 * This function takes the s and t which generate Alt(n), an element
 * x which is odd, and the group, and makes two nice generators for
 * Sym(n)
 */ 

function GetSymGens(group, s, t, x, x_perm, n);
  inv_perm:= Sym(n)!(1, 2);
  even_perm:= x_perm*inv_perm; /*This now lies in alt_n*/
  prog:= PermAsWord(even_perm, n);
  /* billu - change this to an Evaluate (below)
  phi:= hom<Parent(prog) ->  sub<group|s, t> | s, t>;
  y:= phi(prog);
  */
  y := Evaluate(prog, [s,t]);
  t_1:= x^-1*y;                /*This elt should map to (1, 2)*/
  if IsOdd(n) then
     z:= s;
  else
     z:= t_1*s;
  end if;
  s_1:= z*t;                  /*This elt should map to (1, 2, .. , n)*/
  return s_1, t_1;
end function;

/*********************************************************************/

/* 
 * This is the main function. Takes a group "group" and a number n,
 * looks to see if "group" is isomorphic to Alt(n) or Sym(n).
 *
 * Returns 4 things:
 * - a boolean to say if the group is isomorphic to either Alt(n) or
 * Sym(n).
 * - a boolean is_sym to say if the group is isomorphic to Sym(n).
 * - a function BBtoPerm which takes an element elt from an overgroup
 * of "group", and returns true and a permutation corresponding to elt
 * if elt \in "group", and false otherwise.
 * - a function PermtoBB which takes an element from the image of "group"
 * in Sym(n) and returns its preimage in group.
 *
 * The function will wrongly decide that a group is not alternating or
 * symmetric with probability less than 1 in e^5.
 */ 
  

function IsAlternatingOrSymmetric(group, n)
  if n lt 11 then
    "Degree of group must be at least 11"; 
    return "fail",_,_,_; 
  end if;
  bool, s, t, wordfors, wordfort:= GetAltGens(group, n);
  if not bool then      /*This happens when you can't find
                                generators for alt_n inside the group*/
      return false, _,_, _;
  end if;

  is_sym:= false;
  image:= Alt(n);
  t_1:= Identity(group);
  for x in Generators(group) do
      bool, imgx:= EltToPerm(x, n, s, t);
      if not bool then     /*This happens when you cannot
                                    construct the image of one 
                                    of the generators of group. Therefore
                                    group is definitely not alt or sym*/
           return false, _, _,_;
      end if;
      if not IsEven(imgx) then
          is_sym:= true;
          image:= Sym(n);
          odd_perm:= x;
          odd_img:= imgx;
      end if;
  end for;
  if is_sym then
      s_1, t_1:= GetSymGens(group, s, t, odd_perm, odd_img, n);
  end if;

  BBtoPerm:= function(x)
      bool, img:= EltToPerm(x, n, s, t);
      if not bool then
          return false, _;
      elif IsOdd(img) and not is_sym then
          return false, _;
      else
          return bool, img;
      end if;
  end function;

  PermToBB:= function(x)
      if not Degree(Parent(x)) eq n then
  //"Permutation has wrong degree to be in image of BBgroup";
          return false;
      elif IsEven(x) then
          prog:= PermAsWord(x, n);
	  /* replace by an Evaluate - billu 
          phi:= hom<Parent(prog) -> sub<group|s,t>|s,t>;
          return phi(prog);
	  */
	  return Evaluate(prog, [s,t]);
      elif is_sym then
          prog:= PermAsWord((Sym(n)!(1, 2))*x, n);
	  /* replace by an Evaluate - billu 
          phi:= hom<Parent(prog) -> sub<group|s, t>|s,t>;
          return t_1*phi(prog);
	  */
	  return t_1*Evaluate(prog, [s,t]);
      else
          //"Permutation is odd, but image of BBgroup is Alt(", n, ")";
          return false, _, _, _;
      end if;
  end function;
        
  //BBtoPermInv:= map<group -> image | x:->BBtoPerm(x), y:->PermToBB(y)>; 

  return true, is_sym, BBtoPerm, PermToBB;
end function;



intrinsic RecogniseAlternatingOrSymmetric(G::Grp, d::RngIntElt) -> BoolElt, BoolElt, UserProgram, UserProgram
{Constructive recognition of the group G, which will succeed with 
probability > 1-e^5 if G is isomorphic to either Alt(d) or Sym(d)}

    return IsAlternatingOrSymmetric(G, d);

end intrinsic;

