//
// Benchmark of Suzuki stabiliser calculation

ClearVerbose();

// Low and high values of m used in the test, where the field size is
// q = 2^(2 * m + 1)
start := 87;
stop  := 87;

// Number of executions for each field size
nTests := 100;

times := [];

function testSzCopy(G)

    G := DerivedGroupMonteCarlo(G);
    G`RandomProcess := RandomProcessWithWords(G : WordGroup := WordGroup(G));
    P := SuzukiFindOvoidPoints(G : CheckInput := false,
	Randomiser := G`RandomProcess);
    
    SetAssertions(false);
    testTime := Cputime();
    _, logTime := SuzukiStabiliser(G, P : CheckInput := false,
	Randomiser := G`RandomProcess);
    testTime := Cputime(testTime);
    SetAssertions(true);

    delete G;
    return testTime, logTime;
end function;

reals := RealField();
F := GF(2, 2 * StringToInteger(m) + 1 : Optimize := false);
G := Sz(F);

printf "Testing field: %o\n", F;

fieldTime := Cputime();
for i in [1 .. 1000000] do
    _ := Random(F) * Random(F);
end for;
fieldTime := Cputime(fieldTime);

// Initial run to initialise discrete log
_ := testSzCopy(G);

testTime := 0;
logTime  := 0;
for i in [1 .. nTests] do
    G ^:= Random(Generic(G));
    
    curTestTime, curLogTime := testSzCopy(G);
    testTime +:= curTestTime;
    logTime  +:= curLogTime;
end for;

printf "%o %o %o %o\n", char, m, reals ! (testTime / (nTests * fieldTime)),
    reals ! (logTime / (nTests * fieldTime));
