freeze;

// the main ingredients of ClassicalRewrite are the functions
// ClearFirstRow, ClearFirstColumn, SmallerMatrix, SmallerMatrixToSLP,
// DiagonalElementToSLP. These are implemented separately for the 
// different types of classical groups in the black-box and in the 
// char p context.

import "sl-natural.m": SLWordInGen;
import "sp-natural.m": SymplecticWordInGen;
import "su-natural.m": UnitaryWordInGen;
import "orth_plus.m": OmegaPlusWordInGen;
import "orth_minus.m": OmegaMinusWordInGen;
import "orth_circle.m": OmegaCircleWordInGen;
 
import "sl.m": ClearFirstRowSLBB,
               ClearFirstColumnSLBB,
               SmallerMatrixSL,
               SmallerMatrixToSLPSL,
               DiagonalElementToSLPSLBB;


import "spsu.m":ClearFirstRowSpSUBB,
                SmallerMatrixSpSU,
                SmallerMatrixToSLPSp,
                SmallerMatrixToSLPSU,   
                DiagonalElementToSLPSpBB,
                DiagonalElementToSLPSUBB;

import "su34.m":ClearFirstRowSUBBDim4,
                ClearFirstRowSUBBDim3,
                DiagonalElementToSLPSUBBDim4;

DiagonalElementToSLPSUBBDim3 := DiagonalElementToSLPSUBB;


ClearFirstRowSpBB := ClearFirstRowSpSUBB;
ClearFirstRowSUBB := ClearFirstRowSpSUBB;
SmallerMatrixSp := SmallerMatrixSpSU;
SmallerMatrixSU := SmallerMatrixSpSU;

import "charp.m":ClearFirstRowCharP,
                 ClearFirstColumnCharP,
                 DiagonalElementToSLPCharP;

import "charp_orth.m": ClearFirstRowOmegaPlusCharP,
                       SmallerMatrixOmegaPlus,
                       SmallerMatrixToSLPOmegaPlus,
                       DiagonalElementToSLPOmegaPlusCharP,
                       ClearFirstRowOmegaMinusCharP,
                       SmallerMatrixOmegaMinus,
                       SmallerMatrixToSLPOmegaMinus,
                       DiagonalElementToSLPOmegaMinusCharP,
                       ClearFirstRowOmegaCharP,
                       SmallerMatrixOmega,
                       SmallerMatrixToSLPOmega,
                       DiagonalElementToSLPOmegaCharP;

ClearFirstRowSLCharP := ClearFirstRowCharP;
ClearFirstRowSpCharP := ClearFirstRowCharP;
ClearFirstRowSUCharP := ClearFirstRowCharP;
ClearFirstColumnSLCharP := ClearFirstColumnCharP;
DiagonalElementToSLPSLCharP := DiagonalElementToSLPCharP;
DiagonalElementToSLPSpCharP := DiagonalElementToSLPCharP;
DiagonalElementToSLPSUCharP := DiagonalElementToSLPCharP;

// we define a record format to store the invariant of classical groups
bbclassicalattributesrf := recformat< Type : MonStgElt, 
                                   Dimension : RngIntElt,
                                   Field : FldFin >;

// the attribute StabilizerOfE1 holds the stabilizer of e1
AddAttribute( GrpMat, "StabilizerOfE1" );

// the attribute PSubgroup holds the normal p-subgroup of a maximal parabolic
// that is oposite to StabilizerOfE1
AddAttribute( Grp, "PSubgroup" );
AddAttribute( Grp, "Xin" );

AddAttribute( Grp, "IsBlackBoxClassicalGroup" );
AddAttribute( Grp, "AttributesOfBlackBoxClassicalGroup" );
AddAttribute( Grp, "StandardGenerators" );

B := SLPGroup( 9 );

intrinsic IsBlackBoxClassicalGroup( G::Grp ) -> BoolElt 
  { Returns <true> if G is known to be a black-box classical group. }
                      
       return assigned G`IsBlackBoxClassicalGroup and 
       G`IsBlackBoxClassicalGroup;
end intrinsic;

intrinsic IsBlackBoxClassicalGroupWithParameters( G, gens, type, dim, q ) 
  -> BoolElt
{ Returns <true> if G is known to be a black-box classical groups
			with generating set <gens>, type <type>, dimension
			<dim> over the field of <q> elements. }

return assigned G`IsBlackBoxClassicalGroup and 
       G`IsBlackBoxClassicalGroup and
       BBStandardGenerators( G ) eq gens and 
       BBType( G ) eq type and 
       BBDimension( G ) eq dim and 
       BBField( G ) eq GF( q );

end intrinsic;


intrinsic BBStandardGenerators( G::Grp ) -> SeqEnum
  { Returns the standard generators of the black-box classical group G.}
  
  return G`StandardGenerators;
end intrinsic;

intrinsic BBDimension( G::Grp ) -> RngIntElt 
  { Returns the dimension of the black-box classical group G.}
  
  return G`AttributesOfBlackBoxClassicalGroup`Dimension;
end intrinsic;

intrinsic BBField( G::Grp ) -> FldFin
  { Returns the field of the black-box classical group G.}
  
  return G`AttributesOfBlackBoxClassicalGroup`Field;
end intrinsic;


intrinsic BBType( G::Grp ) ->  MonStgElt
  { Returns the field of the black-box classical group G.}
  
  return G`AttributesOfBlackBoxClassicalGroup`Type;
end intrinsic;


BlackBoxClassicalGroup := procedure( ~G, gens, type, dim, q  ) 
  
  if IsBlackBoxClassicalGroupWithParameters( G, gens, type, dim, q ) then
    return;
  end if; 

  if type eq "SU" then
      qq := q^2; 
  else
      qq := q;
  end if;
  
  if type eq "Sp" and dim eq 2 then
      
      // we use that Sp(2,q) is isom to SL(2,q)
      
      newgens := [ gens[1], gens[1], gens[3], gens[4] ];
      gens := newgens;
      type := "SL";
  
  elif type eq "Sp" then
        
        /* the output of ClassicalStandardGenerators contains
          a different root element to the one we usually use
          we append T_{f_1,e_2} as it is more convenient for
          our procedures. We also append the identity that will play
          the role of the new "Y" element.   */
          
        gens[7] := One( G );
        gens[8] := ((gens[6]^gens[2]^2)^gens[1])^-1;
        gens[9] :=  One( G );
        
    elif type eq "SU" and dim eq 2 then
        
        // we use that Sp(2,q) is isom to SL(2,q)
          
        newgens := [ gens[1],gens[1],gens[3],gens[4]];
        gens := newgens;
        type := "SL";
        qq := q;
        
    elif type eq "SU" and IsEven( dim ) then
        
        /* the output of ClassicalStandardGenerators contains
          a different root element to the one we usually use
          we append T_{f_1,e_2} as it is more convenient for
          our procedures. We also append the identity that will play
          the role of the new "Y" element.   */
        
        gens[8] := gens[6]^gens[2]^2;
        gens[9] :=  gens[7]^gens[2]^2;
        
    elif type eq "SU" and IsOdd( dim )  then
        
        /* the output of ClassicalStandardGenerators contains
          a different root element to the one we usually use
          we append T_{f_1,w}(-1/2) as it is more convenient for
          our procedures. We also append the identity that will play
          the role of the new "Y" element.   */
        
        gens[8] := gens[6]^gens[2];
        gens[9] :=  gens[7]^gens[2];
        
    elif type eq "Omega-" then
        
        if #gens eq 4 then
          Append( ~gens, gens[4] );
        end if;
        
    elif type eq "Omega" and dim eq 3 and IsEven( q ) then
        
      newgens := [ gens[1], gens[1], gens[2], gens[3] ];
      gens := newgens;
      type := "SL";
      dim := 2;
        
    end if;

    if assigned G`PSubgroup then delete G`PSubgroup; end if;
    // if assigned G`StabilizerOfE1 then delete G`StabilizerOfE1; end if;
    
    G`StandardGenerators := gens;
    
    G`AttributesOfBlackBoxClassicalGroup := rec< bbclassicalattributesrf | 
                                                 Type := type, 
                                                 Dimension := dim, 
                                                 Field := GF( qq ) >;
    
    G`IsBlackBoxClassicalGroup := true;
end procedure;
                                
Evaluate2 := function( slp, list )
  
    for i in [1..#Generators( Parent( slp ))-#list] do
            Append( ~list, list[1]^0 );
    end for;
    
    return Evaluate( slp, list );
end function;
    
    
    
// the following is a wrapper function for the functions of Elliot's 
// in the natural case      
    
ClassicalRewriteNatural := function( G, g )
    
    dim := BBDimension( G );
    type := BBType( G );
    q := #BBField( G );
        
    /* elif type eq "SU" and dim eq 2 then
        
        a := Round((q+1)/2); s := 1-a; 
        conjmat := GL(2,q^2)![PrimitiveElement( GF( q^2 )),0,0,
                     PrimitiveElement( GF( q^2 ))^s];
        g := (GL(2,q^2)!g)^conjmat; 
        G := sub< GL( dim, q^2 ) | [ (GL(2,q^2)!x)^conjmat : 
                     x in Generators( G )]>;
        type := "SL"; */
        
    
    if type eq "SL" then
        
        conjmat := One( GL( dim, q ));
        
    elif type eq "SU" and IsOdd( dim ) then
        
        list := [1];
        
        for i in [3..dim by 2] do
            Append( ~list, i );
        end for;
        
        for i in [dim-1..2 by -2] do
            Append( ~list, i );
        end for;
        
        conjmat := GL( dim, q )!PermutationMatrix( GF( q ), 
                           Sym( dim )!list );
        
    elif type in { "Sp", "Omega+" } or ( type eq "SU" and IsEven( dim )) then
        
        list := [];
        
        for i in [1..dim-1 by 2] do
            Append( ~list, i );
        end for;
        
        for i in [dim..2 by -2] do
            Append( ~list, i );
        end for;
        
        conjmat := GL( dim, q )!PermutationMatrix( GF( q ), 
                           Sym( dim )!list );
          
    end if;
    
    if type eq "SL" then
       return SLWordInGen( G, g );
    elif type eq "Sp" then
        return SymplecticWordInGen( G, g^(conjmat^-1) );
    elif type eq "SU" then
        return UnitaryWordInGen( G, g^(conjmat^-1));
    elif type eq "Omega+" then
        return OmegaPlusWordInGen( G, g^(conjmat^-1) );
    elif type eq "Omega-" then
       return OmegaMinusWordInGen( G, g );
   elif type eq "Omega" then
        return OmegaCircleWordInGen( G, g );
    end if;
end function;
    

intrinsic ClassicalRewrite( G::Grp, gens::SeqEnum, 
        type::MonStgElt, dim::RngIntElt, q::RngIntElt, g::GrpElt : 
        Method := "choose" ) -> BoolElt, GrpElt
                                
  { Write the element <g> as an SLP in the standard generators of <G>, 
    which must be isomorphic to the classical group of type <type> 
    and dimension <dim> over the field of <q> elements. The string <type>
    must be one of "SL", "Sp", "SU", "Omega", "Omega-" or "Omega+". The parameter 
    <gens> must be a generating set for <G> that satisfies 
    ClassicalStandardPresentation( type, dim, q ). If <type> is one of 
    "Omega", "Omega-" or "Omega+" then <G> must be an absolutely
      irreducible matrix group over a field with the same characteristic as 
      the field GF(q).  <Method> can be either "CharP" or "BB". 
    }
  
  
  // first we need to decorate G with some attributes that describe
  // it as a classical group                        
  
  BlackBoxClassicalGroup( ~G, gens, type, dim, q );

  // we will use Elliots SXWordInGen in the case of the standard 
  // representation with the ClassicalStandardGenerators. Since Elliot's
  // code does not use ClassicalStandardGenerators for Sp and SU 
  // we construct the matrix that transforms ClassicalStandardGenerators
  // to his generators
    
    
    // truetype and type can be different when G is isomorphic 
    // to another classical group
    truetype := type;
    type := BBType( G );
    q2 := case< type | "SU": q^2, default: q >;
    
    // choosing method
    // in the case of the natural representation with the 
    // ClassicalStandardGenerators we call Elliot's code.
      
      if Method eq "choose" then 
                
        if Category( G ) eq GrpMat and 
#CoefficientRing( G ) eq #GF( q2 )  and Dimension( G ) eq dim and 
		     gens eq ClassicalStandardGenerators( type, dim, q ) then
		     

	   return ClassicalRewriteNatural( G, g );
        
        elif Category( G ) eq GrpMat and 
          Characteristic( CoefficientRing( G )) eq
         PrimeBasis( q )[1] and IsAbsolutelyIrreducible( G ) then
            
            Method := "CharP";
            
        else
            
            Method := "BB";
            
        end if;
        
    end if;
    
    gens := BBStandardGenerators( G );

    /* 
      The following function executes the rewriting procedure. 
      The procedure has 5 main steps. Let g be in element in the black-box
      group G isomorphic to SX(n,q). Then in the first step we 
      find elements xl and xr such that xl*g*xr has e_1-row [1,0,...,0]. 
      Then we replace g by  xl*g*xr and we find such elements
      such that the e_1 row and the e_1-column of xl*g*xr are [1,...,0]. 
      We replace g by  xl*g*xr and this element
      is considered as an element of the black-box group isomorphic to 
      SL(n-1,q) or SX(n-2,q) modulo the (e_1,e_1) entry. 
      The third step recovers
      this element as an explicit element of SX(n-[1,2],q) 
      and in the fourth step we write this element as an SLP in the standard
      generators of SX(n-[1,2],q) using the implementation for the 
      natural representation.
        
      After doing these reductions, we obtain an 
      element which corresponds to a
      diagonal matrix and we write such an element as an SLP in the last 
      step. 
    */

   if Method eq "BB" and type eq "SU" and dim eq 4 then 
       str1 := "Dim4";
   elif Method eq "BB" and type eq "SU" and dim eq 3 then
       str1 := "Dim3";
   else
       str1 := "";
   end if;
   
   if type eq "Omega+" then
       type := "OmegaPlus";
   elif type eq "Omega-" then
       type := "OmegaMinus";
   end if;
   
   ClearFirstRow := eval "ClearFirstRow" cat type cat Method cat str1;
      
   if type eq "SL" then
       ClearFirstColumn := eval "ClearFirstColumn" cat "SL" cat Method;
   end if;
      
   SmallerMatrix := eval "SmallerMatrix" cat type; 
   SmallerMatrixToSLP := eval "SmallerMatrixToSLP" cat type;
   DiagonalElementToSLP := eval "DiagonalElementToSLP" cat type 
                           cat Method cat str1; 
   
   if type eq "OmegaPlus" then
       S := gens[1]*gens[4]; progS := B.1*B.4;
   elif type eq "OmegaMinus" and dim eq 4 and IsOdd( q ) then
       S := gens[1]; progS := B.1;
   elif type eq "OmegaMinus" and IsOdd( q ) then
       S := gens[1]^gens[5]; progS := B.1^B.5;
   elif type eq "OmegaMinus" and IsEven( q ) then
       S := (gens[1]*gens[2])^gens[5]; progS := (B.1*B.2)^B.5;
   elif type eq "Omega" and dim eq 3 then
       S := gens[1]; progS := B.1;
   elif type eq "Omega" then
       S := gens[1]^gens[4]; progS := B.1^B.4;
   else     
       S := gens[1]; progS := B.1;
   end if;
   
   pr1l, pr1r, xl, xr := ClearFirstRow( G, g : FirstCall := true ); 
   if Category( pr1l ) eq BoolElt then return false, false; end if;
   g := xl*g*xr;
   
   if type eq "SL" then
   
       pr2l, pr2r, el1, el2 := ClearFirstColumn( G, g );
       if Category( pr2l ) eq BoolElt then return false, false; end if;
       g := el1*g*el2; 
       
   else
       
       g := g^S;
       pr2l, pr2r, el1, el2 := ClearFirstRow( G, g );
       pr2l := pr2l*progS^-1; pr2r := progS*pr2r;
       if Category( pr2l ) eq BoolElt then return false, false; end if;
       g := el1*g*el2; 
       
   end if;
   
   mat := SmallerMatrix( G, g : Method := Method );
   if Category( mat ) eq BoolElt then return false, false; end if;
   
   pr3, pr4, el := SmallerMatrixToSLP( G, g, mat );
   if Category( pr3 ) eq BoolElt then return false, false; end if;
   g := el*g*Evaluate2( pr3, gens )^-1;

   pr5, el := DiagonalElementToSLP( G, g );
   if Category( pr5 ) eq BoolElt then return false, false; end if;
   
   assert Evaluate2( pr5, gens ) eq el;
   
   g := g*el;
   
   p := pr1l^-1*pr2l^-1*pr4^-1*pr5^-1*pr3*pr2r^-1*pr1r^-1;
   
   if truetype eq "SL" then
       
       BB := SLPGroup( 4 );
       p := Evaluate2( p, [ BB.1, BB.2, BB.3, BB.4 ]);
       
   elif truetype eq "Sp" and dim eq 2 then
       
       BB := SLPGroup( 6 );
       p := Evaluate2( p, [ BB.1, BB.1, BB.3, BB.4 ]);
       
   elif truetype eq "Sp" then
       
       BB := SLPGroup( 6 );
       p := Evaluate( p, [BB.1,BB.2,BB.3,BB.4,BB.5,BB.6,
                    One( BB ),((BB.6^BB.2^2)^BB.1)^-1,One(BB)] );
       
   elif truetype eq "SU" and dim eq 2 then
       
       BB := SLPGroup( 7 );
       p := Evaluate2( p, [BB.1,BB.1,BB.3,BB.4] );
       
   elif truetype eq "SU" and IsEven( dim ) then
       
       BB := SLPGroup( 7 );
       p := Evaluate( p, [BB.1,BB.2,BB.3,BB.4,BB.5,BB.6, 
                    BB.7, (BB.6^BB.2^2),BB.7^BB.2^2 ] );
       
   elif truetype eq "SU" and IsOdd( dim ) then
       
       BB := SLPGroup( 7 );
       p := Evaluate( p, [BB.1,BB.2,BB.3,BB.4,BB.5,BB.6,
                    BB.7,(BB.6^BB.2),BB.7^BB.2 ] );
       
   elif truetype eq "Omega+" then
       BB := SLPGroup( 8 );
       p := Evaluate2( p, [BB.1,BB.2,BB.3,BB.4,BB.5,BB.6,
                    BB.7,BB.8]);
   elif truetype eq "Omega-" and dim eq 4 then
       BB := SLPGroup( 3 );
       p := Evaluate2( p, [BB.1,BB.2,BB.3,One(BB),One(BB)] );
   elif truetype eq "Omega-" and dim eq 6 then
       BB := SLPGroup( 4 );
       p := Evaluate2( p, [BB.1,BB.2,BB.3,BB.4,BB.4] );
   elif truetype eq "Omega-" then
       BB := SLPGroup( 5 );
       p := Evaluate2( p, [BB.1,BB.2,BB.3,BB.4,BB.5] );
   elif truetype eq "Omega" and dim eq 3 and IsEven( q ) then
       BB := SLPGroup( 3 );
       p := Evaluate2( p, [BB.1,BB.1,BB.2,BB.3] );
   elif truetype eq "Omega" and dim eq 3 then
       BB := SLPGroup( 3 );
       p := Evaluate2( p, [BB.1,BB.2,BB.3,One(BB)] );
   elif truetype eq "Omega" then
       BB := SLPGroup( 4 );
       p := Evaluate2( p, [BB.1,BB.2,BB.3,BB.4] );
   end if;
   
   return g eq One( G ), p;
end intrinsic;

intrinsic ClassicalRewrite( G::Grp, gens::SeqEnum, 
        type::MonStgElt, dim::RngIntElt, F::FldFin, g::GrpElt : 
        Method := "choose" ) -> GrpElt
                                
  { Executes ClassicalRewrite( G, gens, type, dim, #F, g )}
                                
return ClassicalRewrite( G, gens, type, dim, #F, g : 
                   Method := Method );
end intrinsic;

Xin := function( G )
    
    if assigned G`Xin then
        return G`Xin;
    end if; 
    
    gens := BBStandardGenerators( G );
    q := #BBField( G );
    S := gens[1]; X := gens[2]; T := gens[3]; D := gens[4];
    dim := BBDimension( G );  
    
    xin := [ (T^(S,X))^(-1) ]; xinslp := [ (B.3^(B.1,B.2))^-1 ];
    p := X^-2*S*X; progp := B.2^-2*B.1*B.2;
    
    for i in [1..dim-3] do
        Append( ~xin, (xin[i]^p)^(-1) );
        Append( ~xinslp, (xinslp[i]^progp)^(-1) );
    end for;
    
    G`Xin := < xin, xinslp >;
    
    return < xin, xinslp >;
end function;
 
