freeze;

// these are he

import "MatrixPGroup.m":MatrixPGroupWordInGen;
import "bbclassical.m":B, Evaluate2;
import "spsu.m":X_, Tf1aw;
import "unipotent.m":UnipotentStabiliser;

// G is one of the classical groups. The following function returns the
// subgroup of G which is the stabilizer of the subspace <e_1>.

StabilizerOfE1 := function( G )
    
    if assigned G`StabilizerOfE1 then
        return G`StabilizerOfE1;
    end if;
    
    type := BBType( G );
    gens := BBStandardGenerators( G );
    dim := BBDimension( G );
    q := #BBField( G );
    S := gens[1]; V := gens[2]; T := gens[3]; D := gens[4]; 
    
    H := GL( Degree( G ), BaseRing( G ));
    
    if type eq "SL" and dim gt 2 then  
        
        gens := {T^V^-1, S^-1*V, S^V^-1, D^V^-1, T^S  };

    elif type eq "SL" and dim eq 2 then
        
        gens := { D, T^S };
                
    elif type eq "Sp" and dim eq 4 then
        
        U := gens[5]; X := gens[8]; Y := gens[9];
        gens := { S^V, T^V, D^V, X^V, X^S };
        
    elif type eq "SU" and dim eq 3 and q eq 4 then
        
        gens := { gens[3]^gens[1], gens[6]^gens[1], gens[4]^gens[1], gens[7] };
        
    elif type eq "SU" and dim eq 3 then
        
        X := gens[6]; Y := gens[7];
        gens := { X^S, D, Y };
        
    elif type eq "SU" and dim eq 5  then
        
        U := gens[5]; X := gens[8]; Y := gens[9];
        gens := { S^V, V*U, T^V, D^V, X^V, Y^V, X^S };
        
    elif type eq "SU" and dim eq 4 then
        
        U := gens[5]; X := gens[8]; Y := gens[9]; 
        gens := { S^V, V*U, T^V, D^V, Y^V, X^S };

    elif type in { "Sp", "SU" }  then
        
        U := gens[5]; X := gens[8]; Y := gens[9];
        gens := { S^V, V*U, T^V, D^V, U^V, X^V, Y^V, X^S };
        
    elif type eq "Omega+" then
        
        S := gens[1]; T1 := gens[2]; D1 := gens[3]; U := gens[4];
        T2 := gens[5]; D2 := gens[6]; V := gens[8];
        
        gens := { S^V, T1^V,D1, U^V, T2^V, D2, V*U, T1^S };
        
    elif type eq "Omega-" then
        
        S := gens[1]; T := gens[2]; D := gens[3]; U := gens[4]; V := gens[5];
        gens := { S, T, D, U^V, V*U, T^(S*V) };
        
    elif type eq "Omega" then
        
        S := gens[1]; T := gens[2]; D := gens[3]; V := gens[4];
        U := gens[5]; D1 := gens[6];
        
        gens := { S, T, D, V*U, T^(S*V), D1 };
        
    end if;
    
    St := sub< H | gens >;
    
    G`StabilizerOfE1 := St;
    return St;
        
end function;
    
// In the case of SL we need the transpose of the stabilizer of e1.   
    
StabilizerOfE1Transpose := function( G )    
    
    gens := BBStandardGenerators( G );
    dim := BBDimension( G );
    q := #BBField( G );
    S := gens[1]; X := gens[2]; T := gens[3]; D := gens[4];   
    
    if dim eq 2 then
        return sub< Parent( D ) | D, T >;
    else
        return sub< Parent( D ) | T^X^-1, S^-1*X, S^X^-1, D^X^-1, T >;
    end if;
    
end function;
    

PSubgroup := function( G )
    
    if assigned G`PSubgroup then
        return G`PSubgroup[1],  G`PSubgroup[2];
    end if;   
        
    type := BBType( G );
    gens := BBStandardGenerators( G );
    dim := BBDimension( G );
    q := #BBField( G );
      
    if type eq "SU" and dim eq 3 and q eq 4 then
        P := sub< GL( Dimension( G ), CoefficientRing( G )) |
                  gens[3], gens[4], gens[6] >;
        pgens := [ B.3, B.4, B.6 ];
        G`PSubgroup := < P, pgens >;
        return P, pgens;
        
    end if;

    p := PrimeBasis( q )[1];  
    S := gens[1]; V := gens[2]; T := gens[3]; D := gens[4]; 
        
    if type in { "Sp", "SU" } then
        U := gens[5]; X := gens[8]; Y := gens[9];
    end if;
    
    if type eq "SU" and IsOdd( dim ) then
        X1, pX1 := X_( G );
    end if;
    
    V1 := case< type | "SL":V^-1*S, default:V*U >; 
    pV1 := case< type | "SL":B.2^-1*B.1, default:B.2*B.5  >;
    D1 := case< type | "SL":D^V1, "Sp":D^V1, "SU": Y^V1, default: false >;
    pD1 := case< type | "SL": B.4^pV1, "Sp":B.4^pV1, "SU": B.9^pV1, 
           default: false >;
    
    case type:
      when "SL": T1 := T; pT1 := B.3;
      when "Sp": T1 := X; pT1 := ((B.6^B.2^2)^B.1)^-1;
      when "SU": 
        case IsEven( dim ):
          when true: T1 := X; pT1 := B.6^B.2^2;
          when false: T1 := X1; pT1 := pX1;
        end case;
    end case;
    
    pgens := []; bbgens := [];
    
    
    up := case< type | "SL": dim-1, default:dim-2-( dim mod 2 )>;
    
    for i in [1..up] do
        T2 := T1; pT2 := pT1;
        for j in [1..Round( Log( p, q ))] do
            Append( ~pgens, T2 );
            Append( ~bbgens, pT2 );
            T2 := (T2^(D1)); pT2 := (pT2^pD1);
        end for;
        
        if type in { "Sp", "SU" } and i eq Floor( dim/2-1 ) then
            D1 := D1^(S^V^-1); pD1 := pD1^(B.1^B.2^-1);
            T1 := T1^(S^V^-1); pT1 := pT1^(B.1^B.2^-1);
        end if;
        
        D1 := D1^V1; pD1 := pD1^pV1;
        T1 := (T1^V1)^-1; pT1 := (pT1^pV1)^-1;
    end for;
    
    
    if type in {"Sp", "SU"}  then
        q0 := case< type| "Sp": q, default: Round( Sqrt( q ))>;
        T1 := T; pT1 := B.3;
        for i in [0..Round( Log( p, q0 ))-1] do
            Append( ~bbgens, pT1 );
            Append( ~pgens, T1 );
            T1 := T1^D; pT1 := pT1^B.4;
        end for;
    end if;
    
    if type eq "SU" and IsOdd( dim ) then
        
        z := PrimitiveElement( GF( q ));
        
        for i in [0..Round( Log( p, q ))-1] do
            pr := Tf1aw( G, z^i );
            Append( ~bbgens, pr );
            Append( ~pgens, Evaluate( pr, gens ));
        end for;
    end if;
    
    P := sub< Parent( D ) | pgens >;
    G`PSubgroup := < P, bbgens >;
return P, bbgens;
end function;

PSubgroupTranspose := function( G )
        
    gens := BBStandardGenerators( G );
    dim := BBDimension( G );
    q := #BBField( G );
    p := PrimeBasis( q )[1];  
    S := gens[1]; X := gens[2]; T := gens[3]; D := gens[4]; 
        
    X1 := X^-1*S; pX1 := B.2^-1*B.1;
    D1 := D^X1; pD1 := B.4^pX1;
    
    T1 := T^S; pT1 := B.3^B.1;
    gens := []; bbgens := [];
    for i in [1..dim-1] do
        T2 := T1; pT2 := pT1;
        for j in [1..Round( Log( p, q ))] do
            Append( ~gens, T2 );
            Append( ~bbgens, pT2 );
            T2 := T2^(D1); pT2 := pT2^pD1;
        end for;
        D1 := D1^X1; pD1 := pD1^pX1;
        T1 := T1^X1; pT1 := pT1^pX1;
    end for;
    return sub< Parent( D ) | gens >, bbgens;
end function;
    
   
ClearFirstRowCharP := function( G, g : FirstCall := false )
    
    gens := BBStandardGenerators( G );
    type := BBType( G );
    dim := BBDimension( G );
    V := gens[2]; S := gens[1];
    H := StabilizerOfE1( G );
    P, pgens := PSubgroup( G );
    
    M := GModule( H );
    U := MinimalSubmodule( M );
    
    W := VectorSpace( CoefficientRing( G ), Degree( G ));
    
    U := sub<  W | [ W!(M!x) : x in Basis( U )]>;
    St, C, x, w := UnipotentStabiliser( P, U^g : ComputeBase := false );
    St1, C1, x1, w1 := UnipotentStabiliser( P, U : ComputeBase := false );
    
    c := 0;
    
    if FirstCall and type eq "SL" then
        
        // if the [1,1] entry = 0
        if C ne C1 then
            repeat
                g := g*V;
                St, C, x, w := UnipotentStabiliser( P, U^g : 
                                       ComputeBase := false);
                c := c+1;
                if c eq dim then return false, false, false, false; end if;
            until C eq C1;
        end if;
        
        prog := B.2^c; el := V^c;
        
    elif FirstCall and type in { "Sp", "SU" } then
        
        c := 0;
        CONJ := V;
        
        S1 := One( G ); progS1 := One( B );
        S2 := S; progS2 := B.1;
        for i in [1..Floor( dim/2 )] do
            S1 := S1*S2;
            progS1 := progS1*progS2;
            S2 := S2^V;
            progS2 := progS2^B.2;
        end for;
        
        while C ne C1 do
            c := c+1;
            if c gt dim-1 then
                return false, false, false, false;
            end if;
            if c eq Floor( dim/2 ) then
                CONJ := S1;
            elif c eq Floor( dim/2 )+1 then
                CONJ := V;
            end if;
            g := g*CONJ;
            St, C, x, w := UnipotentStabiliser( P, U^g : ComputeBase := false );
        end while;
        
        if c lt Floor( dim/2 ) then
            prog := B.2^c; el := V^c;
        else
            prog := B.2^-1*progS1*B.2^c; el := V^-1*S1*V^c;
        end if;
        
    else
        
        prog := One( B ); el := One( G );
        
    end if;
        
    p := Evaluate2( w*w1^-1, pgens );
    
    return One( B ), prog*p, One( G ), el*x*x1^-1;
end function;

    
ClearFirstColumnCharP := function( G, g )
    
    H := StabilizerOfE1Transpose( G );
    P, pgens := PSubgroupTranspose( G );
    
    M := GModule( H );
    U := MinimalSubmodule( M );
    
    V := VectorSpace( CoefficientRing( G ), Degree( G ));
    U := sub< V | [ V!(M!x) : x in Basis( U )]>;
    S, C, x, w := UnipotentStabiliser( P, U^g : ComputeBase := false );
    S1, C1, x1, w1 := UnipotentStabiliser( P, U : ComputeBase := false );
    
    p := Evaluate2( w*w1^-1, pgens );
    
    return One( B ), p, One( G ), x*x1^-1;
    
end function;  
    
    
WriteTransvectionAsSLPCharP := function( G, t : GetWE1Entry := true )
    
    dim := BBDimension( G );
    type := BBType( G );
    F := BBField( G );
    P, genspslp := PSubgroup( G );
    p := MatrixPGroupWordInGen( t, P : ComputeBase := false );
    p := Evaluate2( p, genspslp );
    
    q := case< type | "SU": Round( Sqrt( #F )), default: #F >;
    
    stgens := ClassicalStandardGenerators( type, dim, q );
    if type eq "Sp" then
        
        Append( ~stgens, stgens[1]^0 );
        Append( ~stgens, ((stgens[6]^stgens[2]^2)^stgens[1])^-1 );
        Append( ~stgens, stgens[1]^0 );
        
    elif type eq "SU" and IsEven( dim ) then
      
        Append( ~stgens, stgens[6]^stgens[2]^2 );
        Append( ~stgens, stgens[7]^stgens[2]^2 );
        
    elif  type eq "SU" and IsOdd( dim ) then
        
        Append( ~stgens, stgens[6]^stgens[2] );
        Append( ~stgens, stgens[7]^stgens[2] );
        
    end if;    
    
    mat := Evaluate2( p,stgens );
    
    if type eq "SL" then
        vec := [ mat[1,i] : i in [2..dim]];
    elif IsEven( dim ) then
        vec := [];
        
        for i in [3..dim-1 by 2] do
            Append( ~vec, mat[1,i] );
        end for;
        
        for i in [dim..4 by -2] do
            Append( ~vec, mat[1,i] );
        end for;
        
    elif IsOdd( dim ) then
        
        vec := [ mat[1,dim] ]; 
        
        for i in [3..dim-2 by 2] do
            Append( ~vec, mat[1,i] );
        end for;
        
        for i in [dim-1..4 by -2] do
            Append( ~vec, mat[1,i] );
        end for;
          
    end if;
    
    return p, vec;
end function;    
    
// the next function deals with elements in diagonal form
    
    
DiagonalElementToSLPCharP := function( G, g )
    
    dim := BBDimension( G );
    type := BBType( G );
    F := BBField( G );
    z := PrimitiveElement( F );
    q := case< type | "SU": Round( Sqrt( #F )), default: #F >;
    P, genspslp := PSubgroup( G );
    gens := BBStandardGenerators( G );
    T := gens[3]; V := gens[2];
    
    T1 := case< type | "SL":gens[3], default: gens[8] >;
    _, vec := WriteTransvectionAsSLPCharP( G, T1^g );
    
    c := vec[1];
    
    iscentral := func< x | { (x,y) : y in gens } eq { One( G ) }>;
    
    if type eq "SL" then
        
        aset := AllRoots( c^(1-dim), dim );
        bset := [ c*x : x in aset ];
        
        p := B.2^-1*B.1;
        pr := B.4^-1;
        
        for i in [1..dim-2] do
            pr := pr*((B.4^-1)^p);
            p := p*B.2^-1*B.1;
        end for;
        
        centralflag := false;
        for i in [1..#aset] do
            prog := pr^Log( z, bset[i] );
            h := Evaluate2( prog, gens );
            g1 := g*h^-1;
            if g1 eq One( G ) then
                return prog^-1, h^-1;
            elif not centralflag and iscentral( g1 ) then
                centralflag := true;
                prcent := prog^-1;
                hcent := h^-1;
            end if;
        end for;
        
        if centralflag then
            return prcent, hcent;
        end if;
        
    elif type eq "Sp" then
        
        p := B.2*B.5;
        pd := B.4^B.5; pr := pd;
        D := gens[4];
        
        for i in [1..(dim-4)/2] do
            pd := pd^(B.2*B.5);
            pr := pr*pd;
        end for;
        
        pr := pr^Round((q-2)/2);
        
        l1 := Log( z, c^-1 );
        m1 := D^l1;
        
        if g*m1^-1 eq One( G ) then
            return B.4^-l1, m1^-1;
        end if;
        
        m2 := D^(Round(l1 + (q-1)/2) mod (q-1))*Evaluate2( pr, gens );
        
        if g*m2^-1 eq One( G ) then
            return pr^-1*B.4^(-Round(l1 + (q-1)/2) mod (q-1)), m2^-1;
        end if;
        
        if { (x,g*m1^-1) : x in gens } eq { One( G )} then
            return B.4^-l1, m1^-1;
        elif { (x,g*m2^-1) : x in gens } eq { One( G )} then
            return pr^-1*B.4^(-Round(l1 + (q-1)/2) mod (q-1)), m2^-1;
        end if;
        
        return false, false;

    elif type eq "SU" and IsEven( dim )  then
    
        T := gens[3];
        p := MatrixPGroupWordInGen( T^g, P : ComputeBase := false );
        p := Evaluate2( p, genspslp );
        stgens := ClassicalStandardGenerators( type, dim, q );
        mat := Evaluate2( p,stgens );
        c1 := case< IsOdd( q ) | true: mat[1,2]^-1*z^Round((q+1)/2), 
              default: mat[1,2]^-1>;
        
        aset := AllRoots( c1*c^(-dim+2), dim );
        bset := [ c*x : x in aset ];
        
        p := B.2*B.5;
        pd := B.9; pr := pd;
        
        for i in [1..(dim-4)/2] do
            pd := pd^(B.2*B.5);
            pr := pr*pd;
        end for;
        
        pr := pr^(-q+1);
        
        centralflag := false;        
        for i in [1..#aset] do
            n := Log( z^(q-1), bset[i] );
            m := Log( z^(q+1), z^Round(n*(q-1)*(dim-2)/2)*aset[i] );
            
            prog := pr^n*B.4^m;
            h := Evaluate( prog, gens );
            g1 := g*h^-1;
            if g1 eq One( G ) then
                return prog^-1, h^-1;
            elif not centralflag and iscentral( g1 ) then
                centralflag := true;
                prcent := prog^-1;
                hcent := h^-1;
            end if;
        end for;
        
        if centralflag then
            return prcent, hcent;
        end if;
        
    elif type eq "SU" and IsOdd( dim ) then
        
        X := gens[8]; V := gens[2];
        
        T := gens[3];
        p := MatrixPGroupWordInGen( T^g, P : ComputeBase := false );
        p := Evaluate2( p, genspslp );
        stgens := ClassicalStandardGenerators( type, dim, q );
        mat := Evaluate2( p,stgens );
        c1 := case< IsOdd( q )| true: mat[1,2]^-1*z^Round((q+1)/2),
              default:  mat[1,2]^-1 >;
        
        _, vec :=  WriteTransvectionAsSLPCharP( G, X^(V*g*V^-1));
        c2 := vec[1];
        
        bset := AllRoots( c^2*c1*c2^-3, dim );
        aset := [ c2*c^-1*x : x in bset ];
        
        pd := B.9; pr := pd;
        
        for i in [1..(dim-3)/2] do
            pd := pd^(B.2);
            pr := pr*pd;
        end for;
        
        pr := pr^(q-1);
        
        centralflag := false;
        for i in [1..#aset] do
            c := Log( z^(q-1), bset[i] ); 
            gamma := bset[i]^(-dim+3)*aset[i]^(q-1); 
            b := Log( z^(q-1), gamma*z^-Round((q-1)^2*(dim-1)/2*c));  
            a := Log( z^(q+1), aset[i]*z^(-b-(q-1)*c));
            
            prog := B.4^a*B.9^b*pr^c;
            h := Evaluate2( prog, gens );
            g1 := g*h^-1;
            if g1 eq One( G ) then
                return prog^-1, h^-1;
            elif not centralflag and iscentral( g1 ) then
                centralflag := true;
                prcent := prog^-1;
                hcent := h^-1;
            end if;
        end for;
        
        if centralflag then
            return prcent, hcent;
        end if;
            
    end if;

    return false, false;
end function;
   
