freeze;

// this file contains the common functions to deal with symplectic and unitary
// groups
  
/* The SLPGroup B represents the generators of the group G. B.6 and B.7 
  stand for the "new" X and Y.*/
  
import "bbclassical.m":B, Evaluate2;
import "charp.m":ClearFirstRowCharP,WriteTransvectionAsSLPCharP, 
  MakeCentralCharP;
import "su34.m": RewriteBBSUDim3, RewriteBBSUDim4;

import "sp-natural.m": SymplecticWordInGen;
import "su-natural.m": UnitaryWordInGen;

X_ := function( G )
    
    dim := BBDimension( G );
    c := Round( (dim-3)/2 );
    q := Floor( Sqrt( #BBField( G )));
            
    gens := BBStandardGenerators( G ); 
    S := gens[1]; U := gens[5]; V := gens[2]; 
    D := gens[4]; 
    X := gens[8]; Y := gens[9]; T := gens[3]; 
        
    X := X^(V^c); Y := Y^(V^c);
    BX := B.8^(B.2^c); BY := B.9^(B.2^c); 
          
    S1 := S*(Y^V)^(Round((q^2+q)/2));
    C := (X^V)^-1*S1*(X^V*U)^-1*U*S1*U*S1*U*(X^V)^-1;        
    
    WS1 := B.1*(BY^B.2)^(Round((q^2+q)/2));
    WB := (BX^B.2)^-1*WS1*(BX^B.2*B.5)^-1*B.5*WS1*B.5*WS1*B.5*
          (BX^B.2)^-1;
    X1 := ((C^U*S1*(X^V)^-1)^(U*S1*U*V^-2))^-1;
    WX1 := ((WB^B.5*WS1*(BX^B.2)^-1)^(B.5*WS1*B.5*B.2^-2))^-1;
    
    
    X1 := X1^(V^2); WX1 := WX1^(B.2^2);
    
    if IsEven( q ) then
        X1 := X1*((T^S)^V)^(Y^V^2)*T;
        X1 := X1^(D^Round(q/2));
        WX1 := WX1*((B.3^B.1)^B.2)^(B.9^B.2)*B.3;
        WX1 := WX1^(B.4^Round(q/2));
    end if;
    
    return X1, WX1;
end function; 
    
Transvection := function( G, a )
    
    if a eq 0 then
        return One( B );
    end if;
    
    dim := BBDimension( G ); F := BBField( G ); 
    gens := BBStandardGenerators( G );
    type := BBType( G );
    S := gens[1]; V := gens[2]; T := gens[3]; 
    D := gens[4]; U := gens[5]; 
    X := gens[8]; Y := gens[9];
    z := PrimitiveElement( F );
    q := #F; 
    
    if BBType( G ) eq "SU" and IsOdd( q ) then
        q0 := Floor( Sqrt( q ));
        ex := Log( z^(q0+1), a*z^-Floor( (q0+1)/2 ));
        return B.3^B.9^-ex;
    elif BBType( G ) eq "SU" and IsEven( q ) then
        q0 := Floor( Sqrt( q ));
        ex := Log( z^(q0+1), a );
        return B.3^B.9^-ex;
    end if;
    
    // G is symplectic
    
    ex := exists( ns ){ x : x in [0..(#F-1)/2] | 
                      IsOdd( Log( z, z^0+z^(2*x) ))};
    assert ex;

    T1 := T; T2 :=T^D^-ns*T; 
    
    WT1 := B.3; WT2 := B.3^B.4^-ns*B.3;
    
    a1 := 1; a2 := 1+z^(2*ns); exa2 := Log( z, a2 );
    
    ex := Log( z, a );
    
    if IsEven( ex ) then
        return WT1^B.4^Floor((-ex/2));
    else
        return WT2^B.4^Floor((-ex+exa2)/2);
    end if;
    
end function;
    
        
IsZeroEntry := function( G, g )
  
    gens := BBStandardGenerators( G );
    S := gens[1]; T := gens[3]; 

    Q := T^S;
    
    return Q^(Q^g) eq Q;
end function;
    
Tf1aw := function( G, a )
    
    if a eq 0 then
        return One( B ), 0;
    end if;
    
    F := BBField( G );
    q := #F; q0 := Floor( Sqrt( q ));
    z := PrimitiveElement( GF( q )); 
    z0 := PrimitiveElement( GF( q0 ));
    
    if IsEven( q ) then
        x1e1f1 := (Trace (z, GF(q0)))^-1*z;
        x2wf1 := z^-(q0-1);
    else
        x1e1f1 := -1/2;
        x2wf1 := -z^-(q0-1);
    end if;

      
    V, f := VectorSpace( GF( q ), GF( q0 ));
    a1 := One( GF( q ));        
    a2 := z^(q0-1);
    
    m := Eltseq( a1@f ); m := m cat Eltseq( a2@f );
    m := GL( 2, q0 )!m;
    
    v := (a@f)^(m^-1); 
    
    if v[2] eq 0 then
        return B.8^B.4^Log( z0^-1, v[1] ), 
               x1e1f1*z0^(-2*Log( z0^-1, v[1] ));
    elif v[1] eq 0 then
        return (B.8^B.9^B.2)^B.4^Log( z0^-1, v[2] ),  
               x1e1f1*z0^(-2*Log( z0^-1, v[2] ));
    else
        return B.8^B.4^Log( z0^-1, v[1] )*(B.8^B.9^B.2)^B.4^Log( z0^-1, v[2] ), 
               x1e1f1*z0^(-2*Log( z0^-1, v[1] ))+
               x1e1f1*z0^(-2*Log( z0^-1, v[2] ))+
               z0^-Log( z0^-1, v[1] )*x2wf1*z0^-Log( z0^-1, v[2] );
    end if;
    
end function;
    
        
WriteTransvectionAsSLPBB := function( G, g : 
                                  GetWE1Entry := false, 
                                  GetLastEntry := false,
                                  OffSet := 0 )
    
    dim := BBDimension( G ); F := BBField( G ); 
    gens := BBStandardGenerators( G );
    type := BBType( G );
    S := gens[1]; V := gens[2]; T := gens[3]; 
    D := gens[4]; U := gens[5]; 
    X := gens[8]; Y := gens[9];
    
    prog := One( B ); auxprog := One( B );
    R0 := X^(S^V)*T^-1;
    progR0 := B.8;
    T0 := T^V;
        
    CONJ := V*U;
    progCONJ := B.2*B.5;
    
    z := PrimitiveElement( F ); list := [];
    alpha := z^Floor((Sqrt( #F )+1)/2);
    q := #F; 
    if BBType( G ) eq "SU" then
       q0 := Floor( Sqrt( q ));
   end if; 
    
    if BBType( G ) eq "Sp" then
        
        DD := D^-1; WDD := B.4^-1;
        X1 := X; WX1 := B.8;
        
    elif BBType( G ) eq "SU" and IsEven( dim ) then
        
        DD := Y^V^-1; WDD := B.9^B.2^-1;
        X1 := X; WX1 := B.8;
        
    elif BBType( G ) eq "SU" and IsOdd( dim ) then
        
        DD := Y^-1; WDD := B.9^-1;
        X1, WX1 := X_( G );
        
    end if;
    
    R0 := X1^(S^V)*T^-1;
    progR0 := WX1;
    
    for i in [1..(dim-2-( dim mod 2 ))] do
        c := 0; C := ( g, T0 );
        if C ne One( G ) then 
            R := R0;
            repeat
                if c gt q-2 then
                    return false, false;
                end if;
                R := R^DD; c := c+1;
            until C eq R;
            prog := prog*progR0^(WDD^(c+OffSet));
            auxprog := auxprog*progR0^(WDD^c);
            Append( ~list, z^(c+OffSet));
        else
            Append( ~list, F!0 );
        end if;
        
        if i eq Floor( (dim-2)/2 ) then
            CONJ := S^(V^-1);
            progCONJ := B.1^(B.2^-1);
        elif i eq Floor( (dim-2)/2+1 ) then
            CONJ := (V*U)^-1;
            progCONJ := (B.2*B.5)^-1;
        end if;
        
        R0 := R0^CONJ;
        progR0 := progR0^progCONJ;
        T0 := T0^CONJ;
        DD := DD^CONJ; WDD := WDD^progCONJ;
        
    end for;
    
    if BBType( G ) eq "SU" and IsOdd( q ) then
        for i in [Floor((dim-2)/2+1)..dim-2-(dim mod 2 )] do
            list[i] := list[i]*alpha;
        end for;
    end if;
    
    if GetLastEntry then
        
        lambda := 0;
        for i in [1..Floor( dim/2 )-1] do
            if BBType( G ) eq "Sp" then
                lambda := lambda + list[i]*list[dim-1-i];
            elif BBType( G ) eq "SU" then
                lambda := lambda - list[i]*list[dim-1-i-(dim mod 2 )]^q0;
            end if;
        end for;
        
    end if;
    
    if GetWE1Entry then
        
        X2 := X^(Y^V);
        x2 := One( G ); 
        a1 := PrimitiveElement( GF( q0 ))^-1;        
        a2 := z^(q0-1);
        
        found := false;
        
        for i in [1..q0] do
            
            x1 := One( G );
            for j in [1..q0] do
                
                h := g*x1*x2;
                if X^h eq X and X2^h eq X2  then
                    if i eq 1 and j eq 1 then
                        aw := 0;
                        found := true;
                        break i;
                    elif i eq 1 then
                        aw := -a1^(j-2);
                        found := true;
                        break i;
                    elif j eq 1 then
                        found := true;
                        aw := -a2*a1^(i-2);
                        break i;
                    else
                        found := true;
                        aw := -a1^(j-2)-a2*a1^(i-2);
                        break i;
                    end if;
                end if;
                
                if j eq 1 then
                    x1 := X;
                else
                    x1 := x1^D;
                end if;
                
            end for;
            if i eq 1 then
                x2 := X2;
            else
                x2 := x2^D;
            end if;
            
        end for;
        
        if not found then return false, false; end if;
        
        list := Insert( list, 1, aw*z^OffSet );
        pr, b1 := Tf1aw( G, aw*z^OffSet );
        prog := prog*pr;
        if GetLastEntry then
            lambda := lambda + b1;
        end if;
        auxprog := auxprog*Tf1aw( G, aw );
        
    end if;
    
    if GetLastEntry then
        
        Append( ~list, lambda );
        prog := prog;
            
    end if;
    
    return prog, list;
end function;
    
ClearTopRightEntryBB := function( G, g  )
    
    gens := BBStandardGenerators( G ); 
    dim := BBDimension( G ); 
    F := BBField( G );
    type := BBType( G );
    
    S := gens[1]; D := gens[4];  
    U := gens[5]; X := gens[8]; 
    Y := gens[9]; V := gens[2];
    q := #F;
      
    if IsZeroEntry( G, g  ) then
        return One( B ), One( G );
    end if;
    
    if type eq "Sp" then
        DD := D; WDD := B.4;
        X1 := X; WX1 := B.8;
    elif type eq "SU" and IsEven( dim ) then
        DD := Y^V; WDD := B.9^B.2;
        X1 := X; WX1 := B.8;
    elif type eq "SU" and IsOdd( dim ) then
        DD := Y^V; WDD := B.9^B.2;
        X1, WX1 := X_( G );
    end if;
    
    k := 1;
    repeat
        if IsZeroEntry( G, g*X1 ) then 
            return WX1^(WDD^(k-1)), X1;
        end if;
        k := k+1; X1 := X1^DD;
    until k eq q;
    
    return false, false;
end function;
    
    
ClearFirstRowSpSUBB := function( G, g : FirstCall := false )
    
    dim := BBDimension( G ); F := BBField( G ); 
    gens := BBStandardGenerators( G );
    type := BBType( G );
    S := gens[1]; V := gens[2]; T := gens[3]; 
    D := gens[4]; X := gens[8]; Y := gens[9];
    q := #F;
    
    if BBType( G ) eq "Sp" then
        X1 := X;  WX1 := B.8;
    elif BBType( G ) eq "SU" and IsEven( dim ) then
        X1 := X;  WX1 := B.8;
    elif BBType( G ) eq "SU" and IsOdd( dim ) then
        X1, WX1 := X_( G );
    end if;
  
  
    // first we make sure that [1,1] entry non-zero
      
    if FirstCall then  
      
        c := 0;
        CONJ := V;
        
        S1 := One( G ); progS1 := One( B );
        S2 := S; progS2 := B.1;
        for i in [1..Floor( dim/2 )] do
            S1 := S1*S2;
            progS1 := progS1*progS2;
            S2 := S2^V;
            progS2 := progS2^B.2;
        end for;
        
        
        while IsZeroEntry( G, g*S ) do
            c := c+1;
            if c gt dim-1 then
                return false, false;
            end if;
            if c eq Floor( dim/2 ) then
                CONJ := S1;
            elif c eq Floor( dim/2 )+1 then
                CONJ := V;
            end if;
            g := g*CONJ;
        end while;
    
        if c lt Floor( dim/2 ) then
            progR := B.2^c;
        else
            progR := B.2^-1*progS1*B.2^c;
        end if;
        
    else 
        
        progR := One( B );
        
    end if;
        
    // next we make sure that the [1,dim-1] entry
    // is nonzero
      
    if IsZeroEntry( G, g*V^-1 ) then
        g := g*X1^(S^V);
        progR := progR*WX1^(B.1^B.2);
    end if;
    
    prog, z := ClearTopRightEntryBB( G, g );
    if Category( prog ) eq BoolElt then return false, false; end if;
    progR := progR*prog;
    g := g*z;
    
    p1 := X1^S;
    p2 := (T^S)^-1;

    Q := (p1^(p2^g)*p1^-1)^S;
    
    c := 0;
    
    if BBType( G ) eq "Sp" then 
        DD := D^-1;
    elif BBType( G ) eq "SU" and IsEven( dim ) then
        DD := Y^V^-1;
    elif BBType( G ) eq "SU" and IsOdd( dim ) then
        DD := Y^-1;
    end if;
    
    Q1 := Q;
    
    while not IsZeroEntry( G, g*Q1*V^-1 ) do
        if c gt q then
            return false, false;
        end if;
        c := c+1;
        Q1 := Q1^DD;
    end while;
    
    prog1, list := WriteTransvectionAsSLPBB( G, Q : 
                           OffSet := c, 
                           GetLastEntry := true, 
                           GetWE1Entry := IsOdd( dim ));
    
    if Category( prog1 ) eq BoolElt then return false, false; end if;
    
    if list[dim-1] eq 0 then
        pt := One( B );
    elif BBType( G ) eq "Sp" and IsEven( q ) then
        pt := B.3^(B.4^Log( PrimitiveElement( F )^-2, list[dim-1] ));
    else
        pt := Transvection( G, -list[dim-1] );
    end if;
    
    return One( B ), progR*prog1*pt, One( G ), Evaluate2( progR*prog1*pt, gens );
end function;
    
    
SmallerMatrixSpSU := function( G, g : Method := "BB" ) 
    
  type := BBType( G );
  dim := BBDimension( G ); 
  F := BBField( G );  
  
  if ( Method eq "BB" and type eq "SU" and dim eq 4 ) 
     or ( type eq "SU" and dim eq 3 ) then
      return Zero( MatrixAlgebra( F, 0 ));
  end if;
  
  gens := BBStandardGenerators( G ); 

  S := gens[1]; V := gens[2]; 
  U := gens[5]; X := gens[8]; 
  q := #BBField( G ); z := PrimitiveElement( GF( q ));
    
  mat := [];
  
  if Method eq "BB" then
      WriteTransvectionAsSLP := WriteTransvectionAsSLPBB;
  else
      WriteTransvectionAsSLP := WriteTransvectionAsSLPCharP;
  end if;
  

  if BBType( G ) eq "SU" and IsOdd( dim ) then
      p, l := WriteTransvectionAsSLP( G, X^g : GetWE1Entry := true );
      mat := mat cat l;
  end if;
  
  if BBType( G ) eq "Sp" then
      X1 := X;  WX1 := B.8;
  elif BBType( G ) eq "SU" and IsEven( dim ) then
      q0 := Floor( Sqrt( q ));
      X1 := X;  WX1 := B.8;
  elif BBType( G ) eq "SU" and IsOdd( dim ) then
      q0 := Floor( Sqrt( q ));
      X1, WX1 := X_( G );
  end if;
  
  CONJ := V*U;
  X2 := X1;
  
  for i in [1..dim-2-(dim mod 2)] do
      p, l := WriteTransvectionAsSLP( G, X2^g : GetWE1Entry := IsOdd( dim ));
      
      if Category( p ) eq BoolElt then return false; end if;
      
      if i gt (dim-2)/2 and BBType( G ) eq "SU" and IsOdd( q ) then
          l := [ x*z^Floor(-(q0+1)/2) : x in l ];
      end if;
      
      mat := mat cat l;
      
      if i eq Floor( (dim-2)/2 ) then
          CONJ := S^(V^-1);
          progCONJ := B.1^(B.2^-1);
      elif i eq Floor( (dim-2)/2+1 ) then
          CONJ := (V*U)^-1;
          progCONJ := (B.2*B.5)^-1;
      end if;
      
      X2 := X2^CONJ;
  end for;
  
  return MatrixAlgebra( GF( q ), dim-2 )!mat;
end function;    
    
Q_ := function( dim, F )
    
    list := [ Floor((dim+1)/2) ];
    list := list cat [1..Floor((dim-1)/2)]; 
    list := list cat [Floor( (dim+1)/2+1 )..dim];
    Q := GL( dim, (#F)^2 )!PermutationMatrix( GF( (#F)^2 ), Sym( dim )!list );
    return Q;
    
end function;


GeneratorOfCenterSpSUOdd := function( G )
    
    dim := BBDimension( G ); field := BBField( G );
    type := BBType( G );
    gens := BBStandardGenerators( G );
    T := gens[3]; D := gens[4]; S := gens[1]; X := gens[8]; V := gens[2];
    Y := gens[9];
    q := #field;
      
    if type eq "SU" then
        q := Round( Sqrt( q ));
    end if;
    
    if type eq "SU" and IsOdd( dim ) then
        D := Y; progD := B.9;
    else
        D := D; progD := B.4;
    end if;
        
    p := V*S; pp := B.2*B.1;
        
    el1 := D; el := el1;
    pel1 := progD; pel := progD;
    
    for i in [1..Floor( dim/2 )-1] do
        el1 := el1^p; pel1 := pel1^pp;
        el := el*el1; pel := pel*pel1;
    end for;
    
    if type eq "Sp" then
        return pel^Round((q-1)/2), el^Round((q-1)/2);
    elif type eq "SU" and IsOdd( dim ) then
        return pel^Round((q^2-1)/GCD(dim,Round((q+1)))), 
               el^Round((q^2-1)/GCD( dim,Round((q+1))));
    end if;
      
end function;
    
    
GeneratorOfCenterSUEven := function( G );
        
    gens := BBStandardGenerators( G ); 
    dim := BBDimension( G ); 
    F := BBField( G );  
    q :=  Round( Sqrt( #F ));
    type := BBType( G );              
                  
    zz := PrimitiveElement( GF( q^2 ))^(q-1);              
                  
                  
    S := gens[1]; V := gens[2]; T := gens[3]; 
    D := gens[4]; U := gens[5]; Y := gens[9];
        
    S1 := S*S^V; V1 := V*U; progS1 := B.1*B.1^B.2; progV1 := B.2*B.5;
    
    Y2 := Y*Y^S1; progY2 := B.9*B.9^(B.1*B.1^B.2); 
    
    Y1 := One( G ); progY1 := One( B );
    
    for i in [1..(dim-2)/2] do
        Y1 := Y1*Y2; Y2 := Y2^V1; 
        progY1 := progY1*progY2; progY2 := progY2^(B.2*B.5);
    end for;
    
    Y1 := Y1^Round((q+1)/GCD( dim, q+1));
    progY1 := progY1^Round((q+1)/GCD( dim, q+1));
    
    if (zz^Round(-dim/2+1))^Round((q+1)/GCD(q+1,dim)) ne 
               zz^Round((q+1)/GCD(q+1,dim)) then
        Y1 := Y1*D^Round((q-1)/2);
        progY1 := progY1*B.4^Round((q-1)/2);
    end if;
    
    return progY1, Y1; 
    
end function;
    
SmallerMatrixToSLPSp := function( G, g, mat )
    
    dim := BBDimension( G );
    F := BBField( G );
    
    form := 0;
    for i in [1..(dim-2)/2] do
        form := form + mat[1,i]*mat[dim-2,dim-1-i]-
                mat[1,dim-1-i]*mat[dim-2,i];
    end for;
    
    if not IsSquare( (F!form)^-1 ) then return false, false, false; end if;
    
    s := Sqrt( (F!form)^-1 );
    mat := s*mat;
    
    a, w := SymplecticWordInGen( Sp( dim-2, F ), GL( dim-2, F )!mat );
    assert a;
    prog := Evaluate2( w, 
                     [B.1^B.2,B.2*B.5,B.3^B.2,B.4^B.2,B.5^B.2,
                      B.6 ]);
    
    return prog, One( B ), One( G );
    
    
end function;
    
SmallerMatrixToSLPSU := function( G, g, mat )  
    
    
    if Dimension( Parent( mat )) eq 0 then
        return One( B ), One( B ), One( G );
    end if;
    
    dim := BBDimension( G );
    q := #BBField( G );
    gens := BBStandardGenerators( G );
    S := gens[1]; Y := gens[9]; U := gens[5];
    
    if IsEven( dim ) then
                
        q0 := Floor( Sqrt( q ));
        form := 0;
        for i in [1..dim-2] do
            form := form+mat[1][i]*mat[dim-2][dim-i-1]^q0;
        end for;
        
        s := Root( form^-1, q0+1 );
        mat := GL( dim-2, q )![ x*s : x in Eltseq( mat )];
        
        z := PrimitiveElement( GF( q ));
        det := Determinant( mat ); 
        s := Log( z^(-q0+1), det^-1 );
        pr := One( MatrixAlgebra( GF( q ), dim-2 ));
        pr[1,1] := z^s;
        pr[dim-2,dim-2] := (z^-q0)^s;
        mat := pr*mat;
        a, w := UnitaryWordInGen( SU( dim-2, q0 ), GL( dim-2,q )!mat );
        assert a;
        
        if IsOdd( q0 ) then
            w := Evaluate( w, [B.1^B.2,B.2*B.5,B.3^B.2,B.4^B.2,
                         B.5^B.2,B.6,
                         B.7] );
            
        elif dim eq 4 then
            e := Round( q0/4 );
            w := Evaluate2( w, [B.1^(B.4^-e*B.2),One(B),B.3^(B.4^-e*B.2),
                                 B.4^B.2,One(B),One(B),One(B)] );
        else
            w := Evaluate2( w, [B.1^B.2,B.2*B.5,B.3^B.2,B.4^B.2,
                         B.5^B.2,B.8^B.2^-2,B.9^B.2^-2] );
        end if;
        
        return w, (B.9^B.1)^-s, (Y^S)^-s;
                
    else
        
        q0 := Floor( Sqrt( q ));
        form := mat[1][1]*mat[1][1]^q0;
        for i in [2..dim-2] do
            form := form+mat[1][i]*mat[1][dim-i]^q0;
        end for;
        
        s := Root( form^-1, q0+1 );
        mat := GL( dim-2, q )![ x*s : x in Eltseq( mat )];
        
        z := PrimitiveElement( GF( q ));
        det := Determinant( mat ); 
        s := Log( z^(-q0+1), det^-1 );
        pr := One( MatrixAlgebra( GF( q ), dim-2 ));
        pr[2,2] := z^s;
        pr[dim-2,dim-2] := (z^-q0)^s;
        
        mat := pr*mat;
        Q := Q_( dim-2, GF( q0 ));
        
        a, w := UnitaryWordInGen( SU( dim-2, q0 ), 
                        GL( dim-2,q )!(Q^-1*mat*Q) );
        assert a;
        
        if IsOdd( q0 ) then
            w := Evaluate2( w, [B.1^B.2,B.2*B.5,B.3^B.2,B.4^B.2,
                         B.5^B.2,B.8^(B.2^-1),B.9^(B.2^-1)] );
        elif dim eq 5 and q0 eq 2 then
            w := Evaluate2( w, [B.1^B.2,B.2*B.5,B.3^B.2,(B.8^B.2^-1)^B.9,
                         B.5^B.2,B.8^B.2^-1,B.9^B.2^-1] );
            
        else
            w := Evaluate2( w, [B.1^B.2,B.2*B.5,B.3^B.2,B.4^B.2,
                         B.5^B.2,B.8^B.2^-1,B.9^B.2^-1] );
        end if;
        
        
        return w, (B.9^B.5)^s, (Y^U)^s;
            
    end if;
end function;
        

DiagonalElementToSLPSpBB := function( G, g )
    
    gens := BBStandardGenerators( G );
    X := gens[8]; D := gens[4];
    q := #BBField( G );
    type := BBType( G );
    
    i := 0;
    while X^g ne X do
        if i gt q-2 then
            return false, false;
        end if;
        i := i+1;
        g := g*D;
    end while; 
    
    prog := B.4^i; el := D^i;
    
    if not (type eq "Sp" and IsEven( q )) and g ne One( G ) then
        slpcent, gencent := GeneratorOfCenterSpSUOdd( G );
        prog := prog*slpcent;
        el := el*gencent;
    end if;
    
    return prog, el;
end function;


DiagonalElementToSLPSUBB := function( G, g )     
    
    gens := BBStandardGenerators( G );
    S := gens[1]; V := gens[2]; T := gens[3]; D := gens[4];
    U := gens[5]; Y := gens[9]; X := gens[6];
    q := #BBField( G );
    dim := BBDimension( G );
    q0 := Round( Sqrt( q ));
    
    
    if IsEven( dim ) then
        i := 0;
        while T^g ne T do
            if i gt q-2 then return false, false; end if;
            g := g*D;
            i := i+1;
        end while; 
        
        prog := B.4^i; 
        
        S1 := S*S^V; V1 := V*U;
        
        Y2 := Y*Y^S1; progY2 := B.9*B.9^(B.1*B.1^B.2); 
        
        Y1 := One( G ); progY1 := One( B );
        
        for i in [1..(dim-2)/2] do
            Y1 := Y1*Y2; Y2 := Y2^V1; 
            progY1 := progY1*progY2; progY2 := progY2^(B.2*B.5);
        end for;
        
        flag := false; i := 0;
        while U^g ne U do
            if i gt q-2 then return false, false; end if;
            g := g*Y1;
            i := i+1;
            if i eq q0+1 then 
                flag := true;
                break;
            end if;
        end while;
        
        prog := prog*progY1^i;
        
        if flag then
            g := g*D^(Round((q0-1)/2));
            prog := prog*B.4^(Round((q0-1)/2));
            flag := false; i := 0;
            while U^g ne U do
                if i gt q-2 then return false, false; end if;
                g := g*Y1;
                i := i+1;
            end while;
            prog := prog*progY1^i;
        end if;
        
        c := 0;
        pcent, cent := GeneratorOfCenterSUEven( G );
    
        while g ne One( G ) do
            g := g*cent;
            c := c+1;
            if c eq GCD( q0+1, dim ) then
                break;
            end if;
        end while;
        
        return prog*pcent^c, Evaluate( prog*pcent^c, gens );
        
    else
        
        Y1 := Y; progY1 := B.9;
        X1 := X_( G );
        X2 := X1^V^-1;
        
        c := 0;
        while X2^g ne X2 do
            if c gt q-1 then return false, false; end if;
            g := g*Y1;
            c := c+1;
        end while;
        
        prog := progY1^c;
        
        Z1 := Y*Y^S; progZ1 := B.9*B.9^B.1;
        
        Z := One( G ); progZ := One( B );
        
        for i in [1..(dim-1)/2] do
            Z := Z*Z1; progZ := progZ*progZ1;
            Z1 := Z1^V; progZ1 := progZ1^B.2;
        end for;
        
        c := 0;
        
        while (X,g) ne One( G ) do
            if c gt q-1 then return false, false; end if;
            g := g*Z;
            c := c+1;
        end while;
        
        prog := prog*progZ^c;
        
        c := 0;
        pcent, cent := GeneratorOfCenterSpSUOdd( G );
        
        while g ne One( G ) do
            g := g*cent;
            c := c+1;
            if c eq GCD( q0+1, dim ) then
                break;
            end if;
        end while;
        
        return prog*pcent^c, Evaluate( prog*pcent^c, gens );
        
    end if;
end function;
    
    
    

