freeze;

/* Code for maximal subgroups of Sym(n)/Alt(n) taken from
 * package/Group/GrpPerm/max/identify.m
 * Main function is AltSymMax at end of file
 */

RFG := recformat<order:RngIntElt, length:RngIntElt,
                      subgroup:GrpPerm, presentation:GrpFP>;

IntransitiveSubgroupsAltSym := function(n,sym)
// return the intransitive maximal subgroups of Alt(n) or Sym(n)
local subs, A, P;
  assert n ge 4;
  subs:=[ ];
  A := Alt(n);
  for d in [1..(n-1) div 2] do
    if (sym) then
      P := DirectProduct(Sym(d),Sym(n-d));
      AssertAttribute(P, "Order", Factorial(d) * Factorial(n-d));
      Append(~subs, rec< RFG | subgroup:=P, order:=Order(P),
                     length:=Factorial(n) div Order(P)> );
    else
      P := d eq 1 select sub < Sym(n) | DirectProduct(Alt(d),Alt(n-d)) >
                    else sub < Sym(n) | DirectProduct(Alt(d),Alt(n-d)),
                              [2,1] cat [3..d] cat [d+2,d+1] cat [d+3..n] >;
      AssertAttribute(P, "Order", Factorial(d) * Factorial(n-d) div 2);
      Append(~subs, rec< RFG | subgroup:=P, order:=Order(P),
                     length:=Factorial(n) div (2*Order(P)) > );

    end if;
  end for;

  return subs;
end function;

ImprimitiveSubgroupsAltSym := function(n,sym)
// return the imprimitive maximal subgroups of Alt(n) or Sym(n) intersected
// with Alt(n).  sym is true according to whether we have Sym(n) or Alt(n). 
local subs, A, P;
  subs:=[];
  A := Alt(n);
  for d in Divisors(n) do if d ne 1 and d ne n then
    if n eq 8 and d eq 2 and not sym then
      continue; // the only case of nonmaximality - lies in AGL(3,2)
    end if;
    P := WreathProduct(Sym(d),Sym(n div d));
    AssertAttribute(P, "Order", Factorial(d)^(n div d) * Factorial(n div d));
    if (sym) then
      Append(~subs, rec< RFG | subgroup:=P, order:=Order(P),
                     length:=Factorial(n) div Order(P)> );
    else
      P := P meet A;
      Append(~subs, rec< RFG | subgroup:=P, order:=Order(P),
                     length:=Factorial(n) div (2*Order(P))> );
    end if;
  end if; end for;

  return subs;
end function;

PrimitiveSubgroupsAltSym := function(n,sym)
// return the primitive maximal subgroups of Alt(n) or Sym(n) intersected
// with Alt(n). sym is true according to whether we have Sym(n) or Alt(n). 

  local maxprim, subs, x, A, P;

  /* The n-th component of the following list consists of a list of 3 lists.
   * (i) those d such that PrimitiveGroup(n,d) is maximal in Sym(n)
   * (ii) those d such that PrimitiveGroup(n,d) is maximal in Alt(n)
   * (iii) those d as in (ii) for which there are two conjugacy classes
   *       of corresponding subgroups in Alt(n). We conjugate
   *      PrimitiveGroup(n,d) by (1,2) to get one in the other class.
   */
maxprim:=[
[[ ], [ ], [ ]], 					  // 1
[[ ], [ ], [ ]], 					  // 2
[[ ], [ ], [ ]], 					  // 3
[[ ], [ ], [ ]], 					  // 4
[[3], [2], [ ]], 					  // 5
[[2], [1], [ ]], 					  // 6
[[4], [5], [5]], 					  // 7
[[5], [3], [3]], 					  // 8
[[7], [6,9], [9]], 					  // 9
[[7], [6], [ ]],	 				  // 10
[[4], [6], [6]],	  				  // 11
[[2], [4], [4]],	 				  // 12
[[6], [5,7], [7]], 					  // 13
[[2], [1], [ ]],	 				  // 14
[[ ], [4], [4]],	 				  // 15
[[ ], [20], [20]],	 				  // 16
[[5], [8], [8]], 			 		  // 17
[[2], [1], [ ]],	 				  // 18
[[6], [5], [ ]],	 				  // 19
[[2], [1], [ ]],	 				  // 20
[[1,3,7], [2,6], [ ]], 					  // 21
[[2], [1], [ ]],	 				  // 22
[[4], [5], [5]],	 				  // 23
[[2], [3], [3]],	 				  // 24
[[22,26], [21,24], [ ]],				  // 25
[[5], [2], [ ]],					  // 26
[[11], [13,10], [13]],					  // 27
[[11], [9,12], [12]],					  // 28
[[ ], [ ], [ ]],					  // 29
[[ ], [ ], [ ]],	 				  // 30
[[ ], [9,10], [9,10]],					  // 31
[[ ], [3], [3]],					  // 32
[[ ], [2], [2]],					  // 33
[[ ], [ ], [ ]],					  // 34
[[ ], [4], [4]],					  // 35
[[19, 8], [18,20], [20]],				  // 36
[[ ], [ ], [ ]],					  // 37
[[ ], [ ], [ ]],					  // 38
[[ ], [ ], [ ]],					  // 39
[[4,6], [2,5], [ ]],					  // 40
[[ ], [ ], [ ]],					  // 41
[[ ], [ ], [ ]],					  // 42
[[ ], [ ], [ ]],					  // 43
[[ ], [ ], [ ]],					  // 44
[[5], [4,7], [7]],					  // 45
[[ ], [ ], [ ]],					  // 46
[[ ], [ ], [ ]],					  // 47
[[ ], [ ], [ ]],					  // 48
[[33,38], [32,37], [ ]],				  // 49
[[ 6 ],[ 2, 7 ],[ 7 ]], //50
[[],[],[]], //51
[[],[ 1 ],[ 1 ]], //52
[[],[],[]], //53
[[],[],[]], //54
[[ 2, 3, 6 ],[ 5 ],[]], //55
[[ 6, 7 ],[ 2, 4 ],[]], //56
[[],[ 1, 3 ],[ 1, 3 ]], //57
[[],[],[]], //58
[[],[],[]], //59
[[ 5 ],[ 4 ],[]], //60
[[],[],[]], //61
[[],[],[]], //62
[[ 4 ],[ 1, 6 ],[ 6 ]], //63
[[],[ 64, 72 ],[ 64, 72 ]], //64
[[ 2 ],[ 1, 5, 7, 11 ],[ 5, 7, 11 ]], //65
[[],[ 5 ],[ 5 ]], //66
[[],[],[]], //67
[[ 3 ],[ 2 ],[]], //68
[[],[],[]], //69
[[],[],[]], //70
[[],[],[]], //71
[[],[],[]], //72
[[],[ 14 ],[ 14 ]], //73
[[],[],[]], //74
[[],[],[]], //75
[[],[],[]], //76
[[],[ 2 ],[ 2 ]], //77
[[ 2, 4 ],[ 1, 3 ],[]], //78
[[],[],[]], //79
[[],[],[]], //80
[[145, 153],[144, 152],[]], //81
[[ 8 ],[ 6 ],[]], //82
[[],[],[]], //83
[[ 3 ],[ 1 ],[]], //84
[[ 4 ],[ 3 ],[]], //85
[[],[],[]], //86
[[],[],[]], //87
[[],[],[]], //88
[[],[],[]], //89
[[],[],[]], //90
[[ 6 ],[ 4, 5, 8 ],[ 4, 8 ]], //91
[[],[],[]], //92
[[],[],[]], //93
[[],[],[]], //94
[[],[],[]], //95
[[],[],[]], //96
[[],[],[]], //97
[[],[],[]], //98
[[],[],[]], //99
[[ 32, 34, 36 ],[ 33, 35 ],[]], //100
[[],[],[]], //101
[[],[ 1 ],[ 1 ]], //102
[[],[],[]], //103
[[],[],[]], //104
[[ 3, 9 ],[ 7, 8 ],[ 7 ]], //105
[[],[],[]], //106
[[],[],[]], //107
[[],[],[]], //108
[[],[],[]], //109
[[],[],[]], //110
[[],[],[]], //111
[[],[ 8 ],[ 8 ]], //112
[[],[],[]], //113
[[],[],[]], //114
[[],[],[]], //115
[[],[],[]], //116
[[],[ 3 ],[ 3 ]], //117
[[],[],[]], //118
[[],[ 2 ],[ 2 ]], //119
[[ 10 ],[ 8, 19, 21 ],[ 19, 21 ]], //120
[[ 48, 54 ],[ 47, 53, 55 ],[ 55 ]], //121
[[ 5 ],[ 3 ],[]], //122
[[],[],[]], //123
[[],[],[]], //124
[[ 33, 43 ],[ 32, 40 ],[]], //125
[[ 12, 13, 17 ],[ 6, 8, 10, 15 ],[ 6 ]], //126
[[],[ 13 ],[ 13 ]], //127
[[],[ 3 ],[ 3 ]], //128
[[],[ 2 ],[ 2 ]], //129
[[ 5 ],[ 2 ],[]], //130
[[],[],[]], //131
[[],[],[]], //132
[[],[ 1 ],[ 1 ]], //133
[[],[],[]], //134
[[],[ 3 ],[ 3 ]], //135
[[ 3, 12 ],[ 10, 11 ],[ 10 ]], //136
[[],[],[]], //137
[[],[],[]], //138
[[],[],[]], //139
[[],[],[]], //140
[[],[],[]], //141
[[],[],[]], //142
[[],[],[]], //143
[[],[ 10, 12, 14],[ 10, 12, 14]], //144
[[],[],[]], //145
[[],[],[]], //146
[[],[],[]], //147
[[],[],[]], //148
[[],[],[]], //149
[[],[],[]], //150
[[],[],[]], //151
[[],[],[]], //152
[[],[ 4 ],[ 4 ]], //153
[[],[],[]], //154
[[],[ 1 ],[ 1 ]], //155
[[ 3, 7 ],[ 1, 6 ],[]], //156
[[],[],[]], //157
[[],[],[]], //158
[[],[],[]], //159
[[],[],[]], //160
[[],[],[]], //161
[[ 5 ],[ 3 ],[]], //162
[[],[],[]], //163
[[],[],[]], //164
[[ 4 ],[ 2, 5 ],[ 5 ]], //165
[[],[],[]], //166
[[],[],[]], //167
[[ 5 ],[ 4 ],[]], //168
[[ 68, 73 ],[ 67, 72 ],[]], //169
[[ 5 ],[ 3 ],[]], //170
[[ 2, 4 ],[ 1, 3 ],[]], //171
[[],[],[]], //172
[[],[],[]], //173
[[],[],[]], //174
[[ 2 ],[ 1, 4 ],[ 4 ]], //175
[[],[ 3, 4 ],[ 3, 4 ]], //176
[[],[],[]], //177
[[],[],[]], //178
[[],[],[]], //179
[[],[],[]], //180
[[],[],[]], //181
[[],[],[]], //182
[[],[ 2 ],[ 2 ]], //183
[[],[],[]], //184
[[],[],[]], //185
[[],[ 1 ],[ 1 ]], //186
[[],[],[]], //187
[[],[],[]], //188
[[],[],[]], //189
[[],[ 4 ],[ 4 ]], //190
[[],[],[]], //191
[[],[],[]], //192
[[],[],[]], //193
[[],[],[]], //194
[[],[],[]], //195
[[ 8 ],[],[]], //196
[[],[],[]], //197
[[],[],[]], //198
[[],[],[]], //199
[[],[],[]], //200
[[],[],[]], //201
[[],[],[]], //202
[[],[ 1 ],[ 1 ]], //203
[[],[],[]], //204
[[],[],[]], //205
[[],[],[]], //206
[[],[],[]], //207
[[],[ 3 ],[ 3 ]], //208
[[],[],[]], //209
[[ 4 ],[ 2, 3 ],[ 2 ]], //210
[[],[],[]], //211
[[],[],[]], //212
[[],[],[]], //213
[[],[],[]], //214
[[],[],[]], //215
[[],[ 20 ],[ 20 ]], //216
[[],[],[]], //217
[[],[],[]], //218
[[],[],[]], //219
[[ 3 ],[ 2 ],[]], //220
[[],[],[]], //221
[[],[],[]], //222
[[],[],[]], //223
[[],[],[]], //224
[[ 10 ],[ 8 ],[]], //225
[[],[],[]], //226
[[],[],[]], //227
[[],[],[]], //228
[[],[],[]], //229
[[],[],[]], //230
[[],[ 4 ],[ 4 ]], //231
[[],[],[]], //232
[[],[],[]], //233
[[],[ 2 ],[ 2 ]], //234
[[],[],[]], //235
[[],[],[]], //236
[[],[],[]], //237
[[],[],[]], //238
[[],[],[]], //239
[[],[],[]], //240
[[],[],[]], //241
[[],[],[]], //242
[[ 34 ],[ 33 ],[]], //243
[[ 4 ],[ 3 ],[]], //244
[[],[],[]], //245
[[],[],[]], //246
[[],[],[]], //247
[[],[ 1 ],[ 1 ]], //248
[[],[],[]], //249
[[],[],[]], //250
[[],[],[]], //251
[[],[],[]], //252
[[ 3, 7 ],[ 1, 2, 4, 6 ],[ 1, 4 ]], //253
[[],[],[]], //254
[[],[ 2 ],[ 2 ]], //255
[[],[ 238, 242 ],[ 238, 242 ]], //256
[[],[ 13 ],[ 13 ]], //257
[[],[],[]], //258
[[],[],[]], //259
[[],[],[]], //260
[[],[],[]], //261
[[],[],[]], //262
[[],[],[]], //263
[[],[],[]], //264
[[],[],[]], //265
[[],[ 1 ],[ 1 ]], //266
[[],[],[]], //267
[[],[],[]], //268
[[],[],[]], //269
[[],[],[]], //270
[[],[],[]], //271
[[],[],[]], //272
[[],[ 6 ],[ 6 ]], //273
[[],[],[]], //274
[[],[ 2 ],[ 2 ]], //275
[[],[ 4, 6 ],[ 4, 6 ]], //276
[[],[],[]], //277
[[],[],[]], //278
[[],[],[]], //279
[[ 11, 12, 14, 22 ],[ 9, 10, 13, 19 ],[]], //280
[[],[],[]], //281
[[],[],[]], //282
[[],[],[]], //283
[[],[],[]], //284
[[],[ 1 ],[ 1 ]], //285
[[ 2 ],[ 1 ],[]], //286
[[],[],[]], //287
[[],[],[]], //288
[[ 80, 95 ],[ 79, 93 ],[]], //289
[[ 5 ],[ 3 ],[]], //290
[[],[],[]], //291
[[],[],[]], //292
[[],[],[]], //293
[[],[],[]], //294
[[],[],[]], //295
[[],[],[]], //296
[[ 2 ],[ 1 ],[]], //297
[[],[],[]], //298
[[],[],[]], //299
[[ 5, 7, 9 ],[ 2, 6, 8 ],[]], //300
[[],[],[]], //301
[[],[],[]], //302
[[],[],[]], //303
[[],[],[]], //304
[[],[],[]], //305
[[],[],[]], //306
[[],[ 13 ],[ 13 ]], //307
[[],[],[]], //308
[[],[],[]], //309
[[],[],[]], //310
[[],[],[]], //311
[[],[],[]], //312
[[],[],[]], //313
[[],[],[]], //314
[[ 2 ],[ 1, 3 ],[ 3 ]], //315
[[],[],[]], //316
[[],[],[]], //317
[[],[],[]], //318
[[],[],[]], //319
[[],[],[]], //320
[[],[],[]], //321
[[],[],[]], //322
[[],[],[]], //323
[[ 8 ],[],[]], //324
[[ 10 ],[7, 9, 12 ],[7, 12 ]], //325
[[],[],[]], //326
[[],[],[]], //327
[[],[],[]], //328
[[],[],[]], //329
[[],[ 2, 4 ],[ 2, 4 ]], //330
[[],[],[]], //331
[[],[],[]], //332
[[],[],[]], //333
[[],[],[]], //334
[[],[],[]], //335
[[],[ 6, 7 ],[ 6, 7 ]], //336
[[],[],[]], //337
[[],[],[]], //338
[[],[],[]], //339
[[],[],[]], //340
[[ 2 ],[ 1 ],[]], //341
[[],[],[]], //342
[[ 76, 88 ],[ 75, 85 ],[]], //343
[[ 6 ],[ 2, 5 ],[ 2 ]], //344
[[],[],[]], //345
[[],[],[]], //346
[[],[],[]], //347
[[],[],[]], //348
[[],[],[]], //349
[[],[],[]], //350
[[ 4, 7, 9 ],[ 3, 6, 8 ],[]], //351
[[],[],[]], //352
[[],[],[]], //353
[[],[],[]], //354
[[],[],[]], //355
[[],[],[]], //356
[[ 5 ],[ 2 ],[]], //357
[[],[],[]], //358
[[],[],[]], //359
[[ 16 ],[ 11 ],[]], //360
[[ 86, 90 ],[ 85, 88 ],[]], //361
[[ 5 ],[ 4 ],[]], //362
[[],[],[]], //363
[[ 5, 9 ],[ 3, 7, 8 ],[ 7 ]], //364
[[],[],[]], //365
[[],[],[]], //366
[[],[],[]], //367
[[],[],[]], //368
[[ 3 ],[ 2 ],[]], //369
[[],[],[]], //370
[[],[],[]], //371
[[],[],[]], //372
[[],[],[]], //373
[[],[],[]], //374
[[],[],[]], //375
[[],[],[]], //376
[[],[],[]], //377
[[],[ 7, 9 ],[ 7, 9 ]], //378
[[],[],[]], //379
[[],[],[]], //380
[[],[ 2 ],[ 2 ]], //381
[[],[],[]], //382
[[],[],[]], //383
[[],[],[]], //384
[[],[],[]], //385
[[],[],[]], //386
[[],[],[]], //387
[[],[],[]], //388
[[],[],[]], //389
[[],[],[]], //390
[[],[],[]], //391
[[],[],[]], //392
[[],[],[]], //393
[[],[],[]], //394
[[],[],[]], //395
[[],[ 2 ],[ 2 ]], //396
[[],[],[]], //397
[[],[],[]], //398
[[],[],[]], //399
[[ 11, 14 ],[ 8, 9, 13 ],[ 8 ]], //400
[[],[],[]], //401
[[],[],[]], //402
[[],[],[]], //403
[[],[],[]], //404
[[],[],[]], //405
[[ 2, 4 ],[ 1, 3 ],[]], //406
[[],[],[]], //407
[[],[],[]], //408
[[],[],[]], //409
[[],[],[]], //410
[[],[],[]], //411
[[],[],[]], //412
[[],[],[]], //413
[[],[],[]], //414
[[],[],[]], //415
[[],[ 5 ],[ 5 ]], //416
[[],[],[]], //417
[[],[],[]], //418
[[],[],[]], //419
[[],[],[]], //420
[[],[],[]], //421
[[],[],[]], //422
[[],[],[]], //423
[[],[],[]], //424
[[ 1 ],[],[]], //425
[[],[],[]], //426
[[],[],[]], //427
[[],[],[]], //428
[[],[],[]], //429
[[],[],[]], //430
[[],[],[]], //431
[[],[],[]], //432
[[],[],[]], //433
[[],[],[]], //434
[[],[ 4 ],[ 4 ]], //435
[[],[],[]], //436
[[],[],[]], //437
[[],[],[]], //438
[[],[],[]], //439
[[],[],[]], //440
[[ 22 ],[ 20 ],[]], //441
[[],[],[]], //442
[[],[],[]], //443
[[],[],[]], //444
[[],[],[]], //445
[[],[],[]], //446
[[],[],[]], //447
[[],[],[]], //448
[[],[],[]], //449
[[],[],[]], //450
[[],[],[]], //451
[[],[],[]], //452
[[],[],[]], //453
[[],[],[]], //454
[[],[ 2 ],[ 2 ]], //455
[[],[ 2 ],[ 2 ]], //456
[[],[],[]], //457
[[],[],[]], //458
[[],[],[]], //459
[[],[],[]], //460
[[],[],[]], //461
[[],[ 6 ],[ 6 ]], //462
[[],[],[]], //463
[[],[],[]], //464
[[ 2, 3, 5 ],[ 1, 4 ],[]], //465
[[],[],[]], //466
[[],[],[]], //467
[[],[],[]], //468
[[],[],[]], //469
[[],[],[]], //470
[[],[],[]], //471
[[],[],[]], //472
[[],[],[]], //473
[[],[],[]], //474
[[],[],[]], //475
[[],[],[]], //476
[[],[],[]], //477
[[],[],[]], //478
[[],[],[]], //479
[[],[],[]], //480
[[],[],[]], //481
[[],[],[]], //482
[[],[],[]], //483
[[ 8 ],[],[]], //484
[[],[],[]], //485
[[],[],[]], //486
[[],[],[]], //487
[[],[],[]], //488
[[],[],[]], //489
[[],[],[]], //490
[[],[],[]], //491
[[],[],[]], //492
[[],[],[]], //493
[[],[],[]], //494
[[ 4 ],[ 2, 8 ],[ 8 ]], //495
[[  ],[ 8, 10 ],[ 8, 10  ]], //496
[[],[],[]], //497
[[],[],[]], //498
[[],[],[]], //499
[[],[],[]], //500
[[],[],[]], //501
[[],[],[]], //502
[[],[],[]], //503
[[],[ 4 ],[ 4 ]], //504
[[],[],[]], //505
[[],[ 1 ],[ 1 ]], //506
[[],[],[]], //507
[[],[],[]], //508
[[],[],[]], //509
[[],[],[]], //510
[[],[ 1 ],[ 1 ]], //511
[[],[ 36, 56 ],[ 36, 56 ]], //512
[[],[ 9, 12 ],[ 9, 12 ]], //513
[[],[],[]], //514
[[],[],[]], //515
[[],[],[]], //516
[[],[],[]], //517
[[],[],[]], //518
[[],[],[]], //519
[[ 7 ],[ 4, 6 ],[ 4 ]], //520
[[],[],[]], //521
[[],[],[]], //522
[[],[],[]], //523
[[],[],[]], //524
[[ 6 ],[ 4, 5 ],[ 5 ]], //525
[[],[],[]], //526
[[],[ 2 ],[ 2 ]], //527
[[ 7 ],[ 5, 6 ],[ 5 ]], //528
[[ 58, 63 ],[ 57, 61 ],[]], //529
[[ 5 ],[ 4 ],[]], //530
[[],[],[]], //531
[[],[],[]], //532
[[],[],[]], //533
[[],[],[]], //534
[[],[],[]], //535
[[],[],[]], //536
[[],[],[]], //537
[[],[],[]], //538
[[],[],[]], //539
[[ 8 ],[ 7 ],[]], //540
[[],[],[]], //541
[[],[],[]], //542
[[],[],[]], //543
[[],[],[]], //544
[[],[],[]], //545
[[],[],[]], //546
[[],[],[]], //547
[[],[],[]], //548
[[],[],[]], //549
[[],[],[]], //550
[[],[],[]], //551
[[],[],[]], //552
[[],[ 1 ],[ 1 ]], //553
[[],[],[]], //554
[[],[],[]], //555
[[],[],[]], //556
[[],[],[]], //557
[[],[],[]], //558
[[],[],[]], //559
[[ 4 ],[ 2, 3 ],[ 2 ]], //560
[[],[ 2 ],[ 2 ]], //561
[[],[],[]], //562
[[],[],[]], //563
[[],[],[]], //564
[[],[],[]], //565
[[],[],[]], //566
[[],[ 5 ],[ 5 ]], //567
[[],[],[]], //568
[[],[],[]], //569
[[],[],[]], //570
[[],[],[]], //571
[[],[],[]], //572
[[],[],[]], //573
[[],[ 1 ],[ 1 ]], //574
[[],[],[]], //575
[[],[ 9 ],[ 9 ]], //576
[[],[],[]], //577
[[],[],[]], //578
[[],[],[]], //579
[[],[],[]], //580
[[],[],[]], //581
[[],[],[]], //582
[[],[],[]], //583
[[],[],[]], //584
[[],[ 4 ],[ 4 ]], //585
[[],[],[]], //586
[[],[],[]], //587
[[],[],[]], //588
[[],[],[]], //589
[[],[],[]], //590
[[],[],[]], //591
[[],[],[]], //592
[[],[],[]], //593
[[],[],[]], //594
[[ 2 ],[ 1 ],[]], //595
[[],[],[]], //596
[[],[],[]], //597
[[],[],[]], //598
[[],[],[]], //599
[[],[],[]], //600
[[],[],[]], //601
[[],[],[]], //602
[[],[],[]], //603
[[],[],[]], //604
[[],[],[]], //605
[[],[],[]], //606
[[],[],[]], //607
[[],[],[]], //608
[[],[],[]], //609
[[],[],[]], //610
[[],[],[]], //611
[[],[],[]], //612
[[],[],[]], //613
[[],[],[]], //614
[[],[],[]], //615
[[],[ 2 ],[ 2 ]], //616
[[],[],[]], //617
[[],[],[]], //618
[[],[],[]], //619
[[],[ 1 ],[ 1 ]], //620
[[],[],[]], //621
[[],[],[]], //622
[[],[],[]], //623
[[],[],[]], //624
[[ 647, 692, 696 ],[ 646, 689, 695 ],[]], //625
[[ 8 ],[ 7 ],[]], //626
[[],[],[]], //627
[[],[],[]], //628
[[],[],[]], //629
[[],[ 2 ],[ 2 ]], //630
[[],[],[]], //631
[[],[],[]], //632
[[],[],[]], //633
[[],[],[]], //634
[[],[],[]], //635
[[],[],[]], //636
[[],[],[]], //637
[[],[],[]], //638
[[],[],[]], //639
[[],[],[]], //640
[[],[],[]], //641
[[],[],[]], //642
[[],[],[]], //643
[[],[],[]], //644
[[],[],[]], //645
[[],[],[]], //646
[[],[],[]], //647
[[],[],[]], //648
[[],[],[]], //649
[[],[],[]], //650
[[],[ 1, 5 ],[ 1, 5 ]], //651
[[],[],[]], //652
[[],[],[]], //653
[[],[],[]], //654
[[],[],[]], //655
[[],[],[]], //656
[[],[ 2 ],[ 2 ]], //657
[[],[],[]], //658
[[],[],[]], //659
[[ 5 ],[ 2 ],[]], //660
[[],[],[]], //661
[[],[],[]], //662
[[],[],[]], //663
[[],[],[]], //664
[[],[],[]], //665
[[ 2, 4 ],[ 1, 3 ],[]], //666
[[],[],[]], //667
[[],[],[]], //668
[[],[],[]], //669
[[],[],[]], //670
[[],[ 2 ],[ 2 ]], //671
[[],[ 6 ],[ 6 ]], //672
[[],[],[]], //673
[[],[],[]], //674
[[],[],[]], //675
[[ 28 ],[],[]], //676
[[],[],[]], //677
[[],[],[]], //678
[[],[],[]], //679
[[ 2 ],[ 1 ],[]], //680
[[],[],[]], //681
[[],[],[]], //682
[[],[],[]], //683
[[],[],[]], //684
[[],[],[]], //685
[[],[],[]], //686
[[],[],[]], //687
[[],[],[]], //688
[[],[],[]], //689
[[],[],[]], //690
[[],[],[]], //691
[[],[],[]], //692
[[ 4 ],[ 3 ],[]], //693
[[],[],[]], //694
[[],[],[]], //695
[[],[],[]], //696
[[],[],[]], //697
[[],[],[]], //698
[[],[],[]], //699
[[],[],[]], //700
[[],[],[]], //701
[[],[],[]], //702
[[],[ 4 ],[ 4 ]], //703
[[],[],[]], //704
[[],[],[]], //705
[[],[],[]], //706
[[],[],[]], //707
[[],[],[]], //708
[[],[],[]], //709
[[],[],[]], //710
[[],[],[]], //711
[[],[],[]], //712
[[],[],[]], //713
[[],[],[]], //714
[[ 2 ],[ 1 ],[]], //715
[[],[],[]], //716
[[],[],[]], //717
[[],[],[]], //718
[[],[],[]], //719
[[],[],[]], //720
[[],[],[]], //721
[[],[],[]], //722
[[],[],[]], //723
[[],[],[]], //724
[[],[],[]], //725
[[],[],[]], //726
[[],[],[]], //727
[[],[],[]], //728
[[ 471, 495, 499 ],[ 470, 492, 498 ],[]], //729
[[ 3, 13 ],[ 2, 12 ],[]], //730
[[],[],[]], //731
[[],[],[]], //732
[[],[],[]], //733
[[],[],[]], //734
[[],[],[]], //735
[[],[],[]], //736
[[],[],[]], //737
[[],[],[]], //738
[[],[],[]], //739
[[],[],[]], //740
[[ 2 ],[ 1 ],[]], //741
[[],[],[]], //742
[[],[],[]], //743
[[],[],[]], //744
[[],[],[]], //745
[[],[],[]], //746
[[],[],[]], //747
[[],[],[]], //748
[[],[],[]], //749
[[ 1 ],[],[]], //750
[[],[],[]], //751
[[],[],[]], //752
[[],[],[]], //753
[[],[],[]], //754
[[],[],[]], //755
[[],[ 5 ],[ 5 ]], //756
[[],[ 26 ],[ 26 ]], //757
[[],[],[]], //758
[[],[ 1 ],[ 1 ]], //759
[[],[],[]], //760
[[],[],[]], //761
[[],[],[]], //762
[[],[],[]], //763
[[],[],[]], //764
[[ 2 ],[ 1 ],[]], //765
[[],[],[]], //766
[[],[],[]], //767
[[],[],[]], //768
[[],[],[]], //769
[[],[],[]], //770
[[],[],[]], //771
[[],[],[]], //772
[[],[],[]], //773
[[],[],[]], //774
[[ 1 ],[],[]], //775
[[],[],[]], //776
[[],[],[]], //777
[[],[],[]], //778
[[],[],[]], //779
[[],[ 2 ],[ 2 ]], //780
[[],[ 1 ],[ 1 ]], //781
[[],[],[]], //782
[[],[],[]], //783
[[],[ 36 ],[ 36 ]], //784
[[],[],[]], //785
[[],[],[]], //786
[[],[],[]], //787
[[],[],[]], //788
[[],[],[]], //789
[[],[],[]], //790
[[],[],[]], //791
[[],[ 2 ],[ 2 ]], //792
[[],[],[]], //793
[[],[],[]], //794
[[],[],[]], //795
[[],[],[]], //796
[[],[],[]], //797
[[],[],[]], //798
[[],[],[]], //799
[[],[],[]], //800
[[],[],[]], //801
[[],[],[]], //802
[[],[],[]], //803
[[],[],[]], //804
[[],[],[]], //805
[[ 8 ],[ 6 ],[]], //806
[[],[],[]], //807
[[],[],[]], //808
[[],[],[]], //809
[[],[],[]], //810
[[],[],[]], //811
[[],[],[]], //812
[[],[],[]], //813
[[],[],[]], //814
[[],[],[]], //815
[[],[ 2 ],[ 2 ]], //816
[[],[],[]], //817
[[],[],[]], //818
[[ 4 ],[ 3, 6 ],[ 6 ]], //819
[[ 2, 11, 20, 22 ],[ 1, 8, 17, 21 ],[]], //820
[[],[],[]], //821
[[],[],[]], //822
[[],[],[]], //823
[[],[],[]], //824
[[],[],[]], //825
[[],[],[]], //826
[[],[],[]], //827
[[],[],[]], //828
[[],[],[]], //829
[[],[],[]], //830
[[],[],[]], //831
[[],[],[]], //832
[[],[],[]], //833
[[],[],[]], //834
[[],[],[]], //835
[[],[],[]], //836
[[],[],[]], //837
[[],[],[]], //838
[[],[],[]], //839
[[],[ 2, 4 ],[ 2, 4 ]], //840
[[ 110, 114 ],[ 109, 113 ],[]], //841
[[ 5 ],[ 3 ],[]], //842
[[],[],[]], //843
[[],[],[]], //844
[[],[],[]], //845
[[],[],[]], //846
[[],[],[]], //847
[[],[],[]], //848
[[],[],[]], //849
[[],[],[]], //850
[[],[],[]], //851
[[],[],[]], //852
[[],[],[]], //853
[[],[],[]], //854
[[],[],[]], //855
[[],[],[]], //856
[[],[],[]], //857
[[],[],[]], //858
[[],[],[]], //859
[[],[],[]], //860
[[],[ 4 ],[ 4 ]], //861
[[],[],[]], //862
[[],[],[]], //863
[[],[],[]], //864
[[],[],[]], //865
[[],[],[]], //866
[[],[],[]], //867
[[],[],[]], //868
[[],[],[]], //869
[[],[],[]], //870
[[],[ 1 ],[ 1 ]], //871
[[],[],[]], //872
[[],[],[]], //873
[[],[],[]], //874
[[],[],[]], //875
[[],[],[]], //876
[[],[],[]], //877
[[],[],[]], //878
[[],[],[]], //879
[[],[ 1 ],[ 1 ]], //880
[[],[],[]], //881
[[],[],[]], //882
[[],[],[]], //883
[[],[],[]], //884
[[],[],[]], //885
[[],[],[]], //886
[[],[],[]], //887
[[],[],[]], //888
[[],[],[]], //889
[[],[],[]], //890
[[],[ 4 ],[ 4 ]], //891
[[],[],[]], //892
[[],[],[]], //893
[[],[],[]], //894
[[],[],[]], //895
[[],[],[]], //896
[[],[],[]], //897
[[],[],[]], //898
[[],[],[]], //899
[[ 8 ],[],[]], //900
[[],[],[]], //901
[[],[],[]], //902
[[ 2, 4 ],[ 1, 3 ],[]], //903
[[],[],[]], //904
[[],[],[]], //905
[[],[],[]], //906
[[],[],[]], //907
[[],[],[]], //908
[[],[],[]], //909
[[ 3 ],[ 1 ],[]], //910
[[],[],[]], //911
[[],[],[]], //912
[[],[],[]], //913
[[],[],[]], //914
[[],[],[]], //915
[[],[],[]], //916
[[],[],[]], //917
[[],[],[]], //918
[[],[],[]], //919
[[],[],[]], //920
[[],[],[]], //921
[[],[],[]], //922
[[],[],[]], //923
[[],[],[]], //924
[[],[],[]], //925
[[],[],[]], //926
[[],[],[]], //927
[[],[],[]], //928
[[],[],[]], //929
[[],[],[]], //930
[[],[],[]], //931
[[],[],[]], //932
[[],[],[]], //933
[[],[],[]], //934
[[],[],[]], //935
[[],[],[]], //936
[[],[],[]], //937
[[],[],[]], //938
[[],[],[]], //939
[[],[],[]], //940
[[],[],[]], //941
[[],[],[]], //942
[[],[],[]], //943
[[],[],[]], //944
[[],[ 2 ],[ 2 ]], //945
[[],[ 4 ],[ 4 ]], //946
[[],[],[]], //947
[[],[],[]], //948
[[],[],[]], //949
[[],[],[]], //950
[[],[],[]], //951
[[],[],[]], //952
[[],[],[]], //953
[[],[],[]], //954
[[],[],[]], //955
[[],[],[]], //956
[[],[],[]], //957
[[],[],[]], //958
[[],[],[]], //959
[[],[ 5, 8 ],[ 5, 8 ]], //960
[[ 126, 132 ],[ 125, 131 ],[]], //961
[[ 5 ],[ 4 ],[]], //962
[[],[],[]], //963
[[],[],[]], //964
[[],[],[]], //965
[[],[],[]], //966
[[],[],[]], //967
[[],[],[]], //968
[[],[ 2 ],[ 2 ]], //969
[[],[],[]], //970
[[],[],[]], //971
[[],[],[]], //972
[[],[],[]], //973
[[],[],[]], //974
[[],[],[]], //975
[[],[],[]], //976
[[],[],[]], //977
[[],[],[]], //978
[[],[],[]], //979
[[],[ 2 ],[ 2 ]], //980
[[],[],[]], //981
[[],[],[]], //982
[[],[],[]], //983
[[],[],[]], //984
[[],[],[]], //985
[[],[],[]], //986
[[],[],[]], //987
[[],[],[]], //988
[[],[],[]], //989
[[ 2 ],[ 1 ],[]], //990
[[],[],[]], //991
[[],[],[]], //992
[[],[ 2 ],[ 2 ]], //993
[[],[],[]], //994
[[],[],[]], //995
[[],[],[]], //996
[[],[],[]], //997
[[],[],[]], //998
[[],[],[]], //999
[[],[ 105 ],[ 105 ]], //1000
[[],[ 2 ],[ 2 ]], //1001
[[],[],[]], //1002
[[],[],[]], //1003
[[],[],[]], //1004
[[],[],[]], //1005
[[],[],[]], //1006
[[],[],[]], //1007
[[],[ 2, 7 ],[ 2, 7 ]], //1008
[[],[],[]], //1009
[[],[],[]], //1010
[[],[],[]], //1011
[[],[],[]], //1012
[[],[],[]], //1013
[[],[],[]], //1014
[[],[ 1 ],[ 1 ]], //1015
[[],[],[]], //1016
[[],[],[]], //1017
[[],[],[]], //1018
[[],[],[]], //1019
[[],[],[]], //1020
[[],[],[]], //1021
[[],[],[]], //1022
[[],[2],[2]], //1023
[[],[ 105, 113 ],[ 105, 113 ]], //1024
[[],[ 2, 6 ],[ 2, 6 ]], //1025
[[],[],[]], //1026
[[],[],[]], //1027
[[],[],[]], //1028
[[],[],[]], //1029
[[],[],[]], //1030
[[],[],[]], //1031
[[],[],[]], //1032
[[],[],[]], //1033
[[],[],[]], //1034
[[],[ 2 ],[ 2 ]], //1035
[[],[],[]], //1036
[[],[],[]], //1037
[[],[],[]], //1038
[[],[],[]], //1039
[[3 ],[1,5 ],[ 5 ]], //1040
[[],[],[]], //1041
[[],[],[]], //1042
[[],[],[]], //1043
[[],[],[]], //1044
[[],[ 1 ],[ 1 ]], //1045
[[],[],[]], //1046
[[],[],[]], //1047
[[],[],[]], //1048
[[],[],[]], //1049
[[],[],[]], //1050
[[],[],[]], //1051
[[],[],[]], //1052
[[],[],[]], //1053
[[],[],[]], //1054
[[],[],[]], //1055
[[],[],[]], //1056
[[],[ 2 ],[ 2 ]], //1057
[[],[],[]], //1058
[[],[],[]], //1059
[[],[],[]], //1060
[[],[],[]], //1061
[[],[],[]], //1062
[[],[],[]], //1063
[[],[],[]], //1064
[[],[],[]], //1065
[[ 5 ],[ 4 ],[]], //1066
[[],[],[]], //1067
[[],[],[]], //1068
[[],[],[]], //1069
[[],[],[]], //1070
[[],[ 2 ],[ 2 ]], //1071
[[],[],[]], //1072
[[],[],[]], //1073
[[],[],[]], //1074
[[],[],[]], //1075
[[],[],[]], //1076
[[],[],[]], //1077
[[],[],[]], //1078
[[],[],[]], //1079
[[3, 9 ],[2, 8],[ ]], //1080
[[ 2, 4 ],[ 1, 3 ],[]], //1081
[[],[],[]], //1082
[[],[],[]], //1083
[[],[],[]], //1084
[[ 1 ],[],[]], //1085
[[],[],[]], //1086
[[],[],[]], //1087
[[],[],[]], //1088
[[ 8 ],[ 7 ],[]], //1089
[[],[],[]], //1090
[[],[],[]], //1091
[[ 5 ],[ 4 ],[]], //1092
[[],[25 ],[25 ]], //1093
[[],[],[]], //1094
[[],[],[]], //1095
[[],[],[]], //1096
[[],[],[]], //1097
[[],[],[]], //1098
[[],[],[]], //1099
[[ 4 ],[ 2, 3 ],[ 3 ]], //1100
[[],[],[]], //1101
[[],[],[]], //1102
[[],[],[]], //1103
[[],[],[]], //1104
[[ 2 ],[ 1, 5 ],[ 5 ]], //1105
[[],[],[]], //1106
[[],[ 2 ],[ 2 ]], //1107
[[],[],[]], //1108
[[],[],[]], //1109
[[],[],[]], //1110
[[],[],[]], //1111
[[],[],[]], //1112
[[],[],[]], //1113
[[],[],[]], //1114
[[],[],[]], //1115
[[],[],[]], //1116
[[],[],[]], //1117
[[],[],[]], //1118
[[],[],[]], //1119
[[  ],[ 5, 14 ],[ 14 ]], //1120
[[],[],[]], //1121
[[],[],[]], //1122
[[],[],[]], //1123
[[],[],[]], //1124
[[],[],[]], //1125
[[],[],[]], //1126
[[],[],[]], //1127
[[],[ 4 ],[ 4 ]], //1128
[[],[],[]], //1129
[[],[],[]], //1130
[[],[],[]], //1131
[[],[],[]], //1132
[[],[],[]], //1133
[[],[],[]], //1134
[[],[],[]], //1135
[[],[],[]], //1136
[[],[],[]], //1137
[[],[],[]], //1138
[[],[],[]], //1139
[[ 2 ],[ 1 ],[]], //1140
[[],[],[]], //1141
[[],[],[]], //1142
[[],[],[]], //1143
[[],[],[]], //1144
[[],[],[]], //1145
[[],[],[]], //1146
[[],[],[]], //1147
[[],[],[]], //1148
[[],[],[]], //1149
[[],[],[]], //1150
[[],[],[]], //1151
[[],[],[]], //1152
[[],[],[]], //1153
[[],[],[]], //1154
[[],[],[]], //1155
[[ 4 ],[],[]], //1156
[[],[],[]], //1157
[[],[],[]], //1158
[[],[],[]], //1159
[[],[],[]], //1160
[[],[],[]], //1161
[[],[],[]], //1162
[[],[],[]], //1163
[[],[],[]], //1164
[[],[],[]], //1165
[[],[],[]], //1166
[[],[],[]], //1167
[[],[],[]], //1168
[[],[],[]], //1169
[[],[],[]], //1170
[[],[],[]], //1171
[[],[],[]], //1172
[[],[],[]], //1173
[[],[],[]], //1174
[[],[],[]], //1175
[[ 5, 9 ],[ 3, 7, 8 ],[ 7 ]], //1176
[[],[],[]], //1177
[[],[],[]], //1178
[[],[],[]], //1179
[[],[],[]], //1180
[[],[],[]], //1181
[[],[],[]], //1182
[[],[],[]], //1183
[[],[],[]], //1184
[[],[],[]], //1185
[[],[],[]], //1186
[[],[],[]], //1187
[[],[],[]], //1188
[[],[],[]], //1189
[[],[],[]], //1190
[[],[],[]], //1191
[[],[],[]], //1192
[[],[],[]], //1193
[[],[],[]], //1194
[[],[],[]], //1195
[[],[],[]], //1196
[[],[],[]], //1197
[[],[],[]], //1198
[[],[],[]], //1199
[[],[],[]], //1200
[[],[],[]], //1201
[[],[],[]], //1202
[[],[],[]], //1203
[[],[],[]], //1204
[[],[],[]], //1205
[[],[],[]], //1206
[[],[],[]], //1207
[[],[],[]], //1208
[[],[],[]], //1209
[[],[ 1 ],[ 1 ]], //1210
[[],[],[]], //1211
[[],[],[]], //1212
[[],[],[]], //1213
[[],[],[]], //1214
[[],[],[]], //1215
[[],[],[]], //1216
[[],[],[]], //1217
[[],[],[]], //1218
[[],[],[]], //1219
[[],[],[]], //1220
[[],[],[]], //1221
[[],[],[]], //1222
[[],[],[]], //1223
[[],[],[]], //1224
[[ 12, 19 ],[ 10, 18, 21 ],[ 21 ]], //1225
[[],[],[]], //1226
[[],[],[]], //1227
[[],[],[]], //1228
[[],[],[]], //1229
[[],[],[]], //1230
[[],[],[]], //1231
[[],[],[]], //1232
[[],[],[]], //1233
[[],[],[]], //1234
[[],[],[]], //1235
[[],[],[]], //1236
[[],[],[]], //1237
[[],[],[]], //1238
[[],[],[]], //1239
[[],[],[]], //1240
[[],[],[]], //1241
[[],[],[]], //1242
[[],[],[]], //1243
[[],[],[]], //1244
[[],[],[]], //1245
[[],[],[]], //1246
[[],[],[]], //1247
[[],[],[]], //1248
[[],[],[]], //1249
[[],[],[]], //1250
[[],[],[]], //1251
[[],[],[]], //1252
[[],[],[]], //1253
[[],[],[]], //1254
[[],[],[]], //1255
[[],[],[]], //1256
[[],[],[]], //1257
[[],[],[]], //1258
[[],[],[]], //1259
[[],[],[]], //1260
[[],[],[]], //1261
[[],[],[]], //1262
[[],[],[]], //1263
[[],[],[]], //1264
[[],[],[]], //1265
[[],[],[]], //1266
[[],[],[]], //1267
[[],[],[]], //1268
[[],[],[]], //1269
[[],[],[]], //1270
[[],[],[]], //1271
[[],[],[]], //1272
[[],[],[]], //1273
[[],[],[]], //1274
[[ 2 ],[ 1 ],[]], //1275
[[],[],[]], //1276
[[],[],[]], //1277
[[],[],[]], //1278
[[],[],[]], //1279
[[],[],[]], //1280
[[],[],[]], //1281
[[],[],[]], //1282
[[],[],[]], //1283
[[],[],[]], //1284
[[],[],[]], //1285
[[],[],[]], //1286
[[],[ 2 ],[ 2 ]], //1287
[[],[ 2 ],[ 2 ]], //1288
[[],[],[]], //1289
[[],[],[]], //1290
[[],[],[]], //1291
[[],[],[]], //1292
[[],[],[]], //1293
[[],[],[]], //1294
[[],[],[]], //1295
[[],[ 127, 131, 133 ],[ 127, 131, 133 ]], //1296
[[],[],[]], //1297
[[],[],[]], //1298
[[],[],[]], //1299
[[],[],[]], //1300
[[],[],[]], //1301
[[],[],[]], //1302
[[],[],[]], //1303
[[],[],[]], //1304
[[],[],[]], //1305
[[],[],[]], //1306
[[],[],[]], //1307
[[],[],[]], //1308
[[],[],[]], //1309
[[],[],[]], //1310
[[],[],[]], //1311
[[],[],[]], //1312
[[],[],[]], //1313
[[],[],[]], //1314
[[],[],[]], //1315
[[],[],[]], //1316
[[],[],[]], //1317
[[],[],[]], //1318
[[],[],[]], //1319
[[],[ 2 ],[ 2 ]], //1320
[[],[],[]], //1321
[[],[],[]], //1322
[[],[],[]], //1323
[[],[],[]], //1324
[[],[],[]], //1325
[[],[ 2 ],[ 2 ]], //1326
[[],[],[]], //1327
[[],[],[]], //1328
[[],[],[]], //1329
[[ 2 ],[ 1 ],[]], //1330
[[ 76, 90 ],[ 75, 89 ],[]], //1331
[[ 8 ],[ 4, 7 ],[ 4 ]], //1332
[[],[],[]], //1333
[[],[],[]], //1334
[[],[],[]], //1335
[[],[],[]], //1336
[[],[],[]], //1337
[[],[],[]], //1338
[[],[],[]], //1339
[[],[],[]], //1340
[[],[],[]], //1341
[[],[],[]], //1342
[[],[],[]], //1343
[[],[],[]], //1344
[[],[],[]], //1345
[[],[],[]], //1346
[[],[],[]], //1347
[[],[],[]], //1348
[[],[],[]], //1349
[[],[],[]], //1350
[[],[],[]], //1351
[[],[],[]], //1352
[[],[],[]], //1353
[[],[],[]], //1354
[[],[],[]], //1355
[[],[],[]], //1356
[[],[],[]], //1357
[[],[],[]], //1358
[[],[],[]], //1359
[[],[ 3 ],[ 3 ]], //1360
[[],[],[]], //1361
[[],[],[]], //1362
[[],[],[]], //1363
[[],[],[]], //1364
[[ 4, 12 ],[ 2, 6, 11 ],[ 6 ]], //1365
[[],[],[]], //1366
[[],[],[]], //1367
[[],[],[]], //1368
[[ 136, 140 ],[ 135, 138 ],[]], //1369
[[ 5 ],[ 4 ],[]], //1370
[[],[],[]], //1371
[[],[],[]], //1372
[[],[],[]], //1373
[[],[],[]], //1374
[[],[],[]], //1375
[[],[],[]], //1376
[[],[],[]], //1377
[[ 2, 4 ],[ 1, 3 ],[]], //1378
[[],[],[]], //1379
[[],[],[]], //1380
[[],[],[]], //1381
[[],[],[]], //1382
[[],[],[]], //1383
[[],[],[]], //1384
[[],[],[]], //1385
[[],[],[]], //1386
[[],[],[]], //1387
[[],[],[]], //1388
[[],[],[]], //1389
[[],[],[]], //1390
[[],[],[]], //1391
[[],[],[]], //1392
[[],[],[]], //1393
[[],[],[]], //1394
[[],[ 2 ],[ 2 ]], //1395
[[],[],[]], //1396
[[],[],[]], //1397
[[],[],[]], //1398
[[],[],[]], //1399
[[],[],[]], //1400
[[],[],[]], //1401
[[],[],[]], //1402
[[],[],[]], //1403
[[],[],[]], //1404
[[],[],[]], //1405
[[],[],[]], //1406
[[],[ 2 ],[ 2 ]], //1407
[[],[ 4 ],[ 4 ]], //1408
[[],[],[]], //1409
[[],[],[]], //1410
[[],[],[]], //1411
[[],[],[]], //1412
[[],[],[]], //1413
[[],[],[]], //1414
[[],[],[]], //1415
[[],[],[]], //1416
[[],[],[]], //1417
[[],[],[]], //1418
[[],[],[]], //1419
[[],[],[]], //1420
[[],[],[]], //1421
[[],[],[]], //1422
[[],[],[]], //1423
[[],[],[]], //1424
[[],[],[]], //1425
[[],[],[]], //1426
[[],[],[]], //1427
[[],[],[]], //1428
[[],[],[]], //1429
[[],[],[]], //1430
[[],[ 4 ],[ 4 ]], //1431
[[],[],[]], //1432
[[],[],[]], //1433
[[],[],[]], //1434
[[],[ 1 ],[ 1 ]], //1435
[[],[],[]], //1436
[[],[],[]], //1437
[[],[],[]], //1438
[[],[],[]], //1439
[[],[],[]], //1440
[[],[],[]], //1441
[[],[],[]], //1442
[[],[],[]], //1443
[[ 8 ],[],[]], //1444
[[],[],[]], //1445
[[],[],[]], //1446
[[],[],[]], //1447
[[],[],[]], //1448
[[],[],[]], //1449
[[],[],[]], //1450
[[],[],[]], //1451
[[],[],[]], //1452
[[],[],[]], //1453
[[],[],[]], //1454
[[],[],[]], //1455
[[],[ 2, 3 ],[ 2, 3 ]], //1456
[[],[],[]], //1457
[[],[],[]], //1458
[[],[],[]], //1459
[[],[],[]], //1460
[[],[],[]], //1461
[[],[],[]], //1462
[[],[ 1 ],[ 1 ]], //1463
[[ 3, 6 ],[ 1, 5 ],[]], //1464
[[],[],[]], //1465
[[],[],[]], //1466
[[],[],[]], //1467
[[],[],[]], //1468
[[],[],[]], //1469
[[],[],[]], //1470
[[],[],[]], //1471
[[],[],[]], //1472
[[],[],[]], //1473
[[],[],[]], //1474
[[],[],[]], //1475
[[],[],[]], //1476
[[],[],[]], //1477
[[],[],[]], //1478
[[],[],[]], //1479
[[],[],[]], //1480
[[],[],[]], //1481
[[],[],[]], //1482
[[],[],[]], //1483
[[],[],[]], //1484
[[ 2 ],[ 1 ],[]], //1485
[[],[],[]], //1486
[[],[],[]], //1487
[[],[],[]], //1488
[[],[],[]], //1489
[[],[],[]], //1490
[[],[],[]], //1491
[[],[],[]], //1492
[[],[],[]], //1493
[[],[],[]], //1494
[[],[],[]], //1495
[[],[],[]], //1496
[[],[],[]], //1497
[[],[],[]], //1498
[[],[],[]], //1499
[[],[],[]], //1500
[[],[],[]], //1501
[[],[],[]], //1502
[[],[],[]], //1503
[[],[],[]], //1504
[[],[],[]], //1505
[[],[],[]], //1506
[[],[],[]], //1507
[[],[],[]], //1508
[[],[],[]], //1509
[[],[],[]], //1510
[[],[],[]], //1511
[[],[],[]], //1512
[[],[],[]], //1513
[[],[],[]], //1514
[[],[],[]], //1515
[[],[],[]], //1516
[[],[],[]], //1517
[[],[],[]], //1518
[[],[],[]], //1519
[[],[],[]], //1520
[[ 4 ],[ 2 ],[]], //1521
[[],[],[]], //1522
[[],[],[]], //1523
[[],[],[]], //1524
[[],[],[]], //1525
[[],[],[]], //1526
[[],[],[]], //1527
[[],[],[]], //1528
[[],[],[]], //1529
[[],[],[]], //1530
[[],[],[]], //1531
[[],[],[]], //1532
[[],[],[]], //1533
[[],[],[]], //1534
[[],[],[]], //1535
[[],[],[]], //1536
[[],[],[]], //1537
[[],[],[]], //1538
[[],[],[]], //1539
[[],[ 1, 3, 5 ],[ 1, 3, 5 ]], //1540
[[],[],[]], //1541
[[],[],[]], //1542
[[],[],[]], //1543
[[],[],[]], //1544
[[],[],[]], //1545
[[],[],[]], //1546
[[],[],[]], //1547
[[],[],[]], //1548
[[],[],[]], //1549
[[ 2, 7 ],[ 1, 5 ],[ ]], //1550
[[],[],[]], //1551
[[],[],[]], //1552
[[],[],[]], //1553
[[],[],[]], //1554
[[],[],[]], //1555
[[],[],[]], //1556
[[],[],[]], //1557
[[],[],[]], //1558
[[],[],[]], //1559
[[],[],[]], //1560
[[],[],[]], //1561
[[],[],[]], //1562
[[],[],[]], //1563
[[],[],[]], //1564
[[],[],[]], //1565
[[],[],[]], //1566
[[],[],[]], //1567
[[],[],[]], //1568
[[],[],[]], //1569
[[],[],[]], //1570
[[],[],[]], //1571
[[],[],[]], //1572
[[],[],[]], //1573
[[],[],[]], //1574
[[],[ 4 ],[ 4 ]], //1575
[[],[],[]], //1576
[[],[],[]], //1577
[[],[],[]], //1578
[[],[],[]], //1579
[[],[],[]], //1580
[[],[],[]], //1581
[[],[],[]], //1582
[[],[],[]], //1583
[[],[ 1 ],[ 1 ]], //1584
[[],[],[]], //1585
[[],[],[]], //1586
[[],[],[]], //1587
[[],[],[]], //1588
[[],[],[]], //1589
[[],[],[]], //1590
[[],[],[]], //1591
[[],[],[]], //1592
[[],[],[]], //1593
[[],[],[]], //1594
[[],[],[]], //1595
[[ 4 ],[ 1, 2, 3 ],[ 1, 2 ]], //1596
[[],[],[]], //1597
[[],[],[]], //1598
[[],[],[]], //1599
[[],[ 16, 19, 20 ],[ 16, 19, 20 ]], //1600
[[],[],[]], //1601
[[],[],[]], //1602
[[],[],[]], //1603
[[],[],[]], //1604
[[],[],[]], //1605
[[],[],[]], //1606
[[],[],[]], //1607
[[],[],[]], //1608
[[],[],[]], //1609
[[],[],[]], //1610
[[],[],[]], //1611
[[],[],[]], //1612
[[],[],[]], //1613
[[],[],[]], //1614
[[],[],[]], //1615
[[],[],[]], //1616
[[],[],[]], //1617
[[],[],[]], //1618
[[],[],[]], //1619
[[],[],[]], //1620
[[],[],[]], //1621
[[],[],[]], //1622
[[],[],[]], //1623
[[],[],[]], //1624
[[],[],[]], //1625
[[],[],[]], //1626
[[],[],[]], //1627
[[],[],[]], //1628
[[],[],[]], //1629
[[],[],[]], //1630
[[],[],[]], //1631
[[],[ 2 ],[ 2 ]], //1632
[[],[],[]], //1633
[[],[],[]], //1634
[[],[],[]], //1635
[[],[],[]], //1636
[[],[],[]], //1637
[[],[],[]], //1638
[[],[],[]], //1639
[[],[],[]], //1640
[[],[],[]], //1641
[[],[],[]], //1642
[[],[],[]], //1643
[[],[],[]], //1644
[[],[],[]], //1645
[[],[],[]], //1646
[[],[],[]], //1647
[[],[],[]], //1648
[[],[],[]], //1649
[[],[],[]], //1650
[[],[],[]], //1651
[[],[],[]], //1652
[[],[ 2 ],[ 2 ]], //1653
[[],[],[]], //1654
[[],[],[]], //1655
[[],[],[]], //1656
[[],[],[]], //1657
[[],[],[]], //1658
[[],[],[]], //1659
[[],[],[]], //1660
[[],[],[]], //1661
[[],[],[]], //1662
[[],[],[]], //1663
[[],[],[]], //1664
[[],[],[]], //1665
[[],[],[]], //1666
[[],[],[]], //1667
[[],[],[]], //1668
[[],[],[]], //1669
[[],[],[]], //1670
[[],[],[]], //1671
[[],[],[]], //1672
[[],[],[]], //1673
[[],[],[]], //1674
[[],[],[]], //1675
[[],[],[]], //1676
[[],[],[]], //1677
[[],[],[]], //1678
[[],[],[]], //1679
[[],[],[]], //1680
[[ 188, 192 ],[ 187, 191 ],[]], //1681
[[ 5 ],[ 4 ],[]], //1682
[[],[],[]], //1683
[[],[],[]], //1684
[[],[],[]], //1685
[[],[],[]], //1686
[[],[],[]], //1687
[[],[],[]], //1688
[[],[],[]], //1689
[[],[],[]], //1690
[[],[],[]], //1691
[[],[],[]], //1692
[[],[],[]], //1693
[[],[],[]], //1694
[[],[],[]], //1695
[[],[],[]], //1696
[[],[],[]], //1697
[[],[],[]], //1698
[[],[],[]], //1699
[[],[],[]], //1700
[[],[],[]], //1701
[[],[],[]], //1702
[[],[],[]], //1703
[[],[],[]], //1704
[[],[],[]], //1705
[[],[],[]], //1706
[[],[],[]], //1707
[[],[],[]], //1708
[[],[],[]], //1709
[[],[],[]], //1710
[[ 3, 5 ],[ 1, 2, 4 ],[ 1 ]], //1711
[[],[],[]], //1712
[[],[],[]], //1713
[[],[],[]], //1714
[[],[],[]], //1715
[[],[ 4 ],[ 4 ]], //1716
[[],[],[]], //1717
[[],[],[]], //1718
[[],[],[]], //1719
[[],[],[]], //1720
[[],[],[]], //1721
[[],[],[]], //1722
[[],[ 17 ],[ 17 ]], //1723
[[],[],[]], //1724
[[],[],[]], //1725
[[],[],[]], //1726
[[],[],[]], //1727
[[],[ 24 ],[ 24 ]], //1728
[[],[],[]], //1729
[[],[],[]], //1730
[[],[],[]], //1731
[[],[],[]], //1732
[[],[],[]], //1733
[[],[],[]], //1734
[[],[],[]], //1735
[[],[],[]], //1736
[[],[],[]], //1737
[[],[],[]], //1738
[[],[],[]], //1739
[[],[],[]], //1740
[[],[],[]], //1741
[[],[],[]], //1742
[[],[],[]], //1743
[[],[],[]], //1744
[[],[],[]], //1745
[[],[],[]], //1746
[[],[],[]], //1747
[[],[],[]], //1748
[[],[],[]], //1749
[[ 3 ],[ 1, 4 ],[ 4 ]], //1750
[[],[],[]], //1751
[[],[],[]], //1752
[[],[],[]], //1753
[[],[],[]], //1754
[[],[ 2 ],[ 2 ]], //1755
[[],[],[]], //1756
[[],[],[]], //1757
[[],[],[]], //1758
[[],[],[]], //1759
[[],[],[]], //1760
[[],[],[]], //1761
[[],[],[]], //1762
[[],[],[]], //1763
[[ 8 ],[],[]], //1764
[[],[],[]], //1765
[[],[],[]], //1766
[[],[],[]], //1767
[[],[],[]], //1768
[[],[],[]], //1769
[[],[ 4 ],[ 4 ]], //1770
[[],[ 2, 4 ],[ 2, 4 ]], //1771
[[],[],[]], //1772
[[],[],[]], //1773
[[],[],[]], //1774
[[],[],[]], //1775
[[],[],[]], //1776
[[],[],[]], //1777
[[],[],[]], //1778
[[],[],[]], //1779
[[],[],[]], //1780
[[],[],[]], //1781
[[ 2 ],[ 1 ],[]], //1782
[[],[],[]], //1783
[[],[],[]], //1784
[[],[ 3 ],[ 3 ]], //1785
[[],[],[]], //1786
[[],[],[]], //1787
[[],[],[]], //1788
[[],[],[]], //1789
[[],[],[]], //1790
[[],[],[]], //1791
[[],[],[]], //1792
[[],[],[]], //1793
[[],[],[]], //1794
[[],[],[]], //1795
[[],[],[]], //1796
[[],[],[]], //1797
[[],[],[]], //1798
[[],[],[]], //1799
[[ 2 ],[ 1 ],[]], //1800
[[],[],[]], //1801
[[],[],[]], //1802
[[],[],[]], //1803
[[],[],[]], //1804
[[],[],[]], //1805
[[],[],[]], //1806
[[],[],[]], //1807
[[],[],[]], //1808
[[],[],[]], //1809
[[],[],[]], //1810
[[],[],[]], //1811
[[],[],[]], //1812
[[],[],[]], //1813
[[],[],[]], //1814
[[],[],[]], //1815
[[],[],[]], //1816
[[],[],[]], //1817
[[],[],[]], //1818
[[],[],[]], //1819
[[],[ 2 ],[ 2 ]], //1820
[[],[],[]], //1821
[[],[],[]], //1822
[[],[],[]], //1823
[[],[],[]], //1824
[[],[],[]], //1825
[[],[],[]], //1826
[[],[],[]], //1827
[[],[],[]], //1828
[[],[],[]], //1829
[[ 2, 4 ],[ 1, 3 ],[]], //1830
[[],[],[]], //1831
[[],[],[]], //1832
[[],[],[]], //1833
[[],[],[]], //1834
[[],[],[]], //1835
[[],[],[]], //1836
[[],[],[]], //1837
[[],[],[]], //1838
[[],[],[]], //1839
[[],[],[]], //1840
[[],[],[]], //1841
[[],[],[]], //1842
[[],[],[]], //1843
[[],[],[]], //1844
[[],[],[]], //1845
[[],[],[]], //1846
[[],[],[]], //1847
[[],[],[]], //1848
[[ 126, 130 ],[ 125, 128 ],[]], //1849
[[ 5 ],[ 3 ],[]], //1850
[[],[],[]], //1851
[[],[],[]], //1852
[[],[],[]], //1853
[[],[],[]], //1854
[[],[],[]], //1855
[[],[],[]], //1856
[[],[],[]], //1857
[[],[],[]], //1858
[[],[],[]], //1859
[[],[],[]], //1860
[[],[],[]], //1861
[[],[],[]], //1862
[[],[],[]], //1863
[[],[],[]], //1864
[[],[],[]], //1865
[[],[],[]], //1866
[[],[],[]], //1867
[[],[],[]], //1868
[[],[],[]], //1869
[[],[],[]], //1870
[[],[],[]], //1871
[[],[],[]], //1872
[[],[],[]], //1873
[[],[],[]], //1874
[[],[],[]], //1875
[[],[],[]], //1876
[[],[],[]], //1877
[[],[],[]], //1878
[[],[],[]], //1879
[[],[],[]], //1880
[[],[],[]], //1881
[[],[],[]], //1882
[[],[],[]], //1883
[[],[],[]], //1884
[[],[],[]], //1885
[[],[],[]], //1886
[[],[],[]], //1887
[[],[],[]], //1888
[[],[],[]], //1889
[[],[],[]], //1890
[[],[ 1, 5 ],[ 1, 5 ]], //1891
[[],[],[]], //1892
[[],[ 2 ],[ 2 ]], //1893
[[],[],[]], //1894
[[],[],[]], //1895
[[],[],[]], //1896
[[],[],[]], //1897
[[],[],[]], //1898
[[],[],[]], //1899
[[],[],[]], //1900
[[],[],[]], //1901
[[],[],[]], //1902
[[],[],[]], //1903
[[],[],[]], //1904
[[],[],[]], //1905
[[],[],[]], //1906
[[],[],[]], //1907
[[],[],[]], //1908
[[],[],[]], //1909
[[],[],[]], //1910
[[],[],[]], //1911
[[],[],[]], //1912
[[],[],[]], //1913
[[],[],[]], //1914
[[],[],[]], //1915
[[],[],[]], //1916
[[],[],[]], //1917
[[],[],[]], //1918
[[],[],[]], //1919
[[],[],[]], //1920
[[],[],[]], //1921
[[],[],[]], //1922
[[],[],[]], //1923
[[],[],[]], //1924
[[],[],[]], //1925
[[],[],[]], //1926
[[],[],[]], //1927
[[],[],[]], //1928
[[],[],[]], //1929
[[],[],[]], //1930
[[],[],[]], //1931
[[],[],[]], //1932
[[],[],[]], //1933
[[],[],[]], //1934
[[],[],[]], //1935
[[],[ 8 ],[ 8 ]], //1936
[[],[],[]], //1937
[[],[],[]], //1938
[[],[],[]], //1939
[[],[],[]], //1940
[[],[],[]], //1941
[[],[],[]], //1942
[[],[],[]], //1943
[[],[],[]], //1944
[[],[],[]], //1945
[[],[],[]], //1946
[[],[],[]], //1947
[[],[],[]], //1948
[[],[],[]], //1949
[[],[],[]], //1950
[[],[],[]], //1951
[[],[],[]], //1952
[[ 1, 3 ],[ 2 ],[]], //1953
[[],[],[]], //1954
[[],[],[]], //1955
[[],[],[]], //1956
[[],[],[]], //1957
[[],[],[]], //1958
[[],[],[]], //1959
[[],[],[]], //1960
[[],[],[]], //1961
[[],[],[]], //1962
[[],[],[]], //1963
[[],[],[]], //1964
[[],[],[]], //1965
[[],[],[]], //1966
[[],[],[]], //1967
[[],[],[]], //1968
[[],[],[]], //1969
[[],[],[]], //1970
[[],[],[]], //1971
[[],[],[]], //1972
[[],[],[]], //1973
[[],[],[]], //1974
[[],[],[]], //1975
[[],[],[]], //1976
[[],[],[]], //1977
[[],[],[]], //1978
[[],[],[]], //1979
[[],[],[]], //1980
[[],[],[]], //1981
[[],[],[]], //1982
[[],[],[]], //1983
[[],[],[]], //1984
[[],[],[]], //1985
[[],[],[]], //1986
[[],[],[]], //1987
[[],[],[]], //1988
[[],[],[]], //1989
[[],[],[]], //1990
[[],[],[]], //1991
[[],[],[]], //1992
[[],[],[]], //1993
[[],[],[]], //1994
[[],[],[]], //1995
[[],[],[]], //1996
[[],[],[]], //1997
[[],[],[]], //1998
[[],[],[]], //1999
[[],[],[]], //2000
[[],[],[]], //2001
[[ 2 ],[ 1 ],[]], //2002
[[],[],[]], //2003
[[],[],[]], //2004
[[],[],[]], //2005
[[],[],[]], //2006
[[],[],[]], //2007
[[],[],[]], //2008
[[],[],[]], //2009
[[],[],[]], //2010
[[],[],[]], //2011
[[],[],[]], //2012
[[],[],[]], //2013
[[],[],[]], //2014
[[],[ 2 ],[ 2 ]], //2015
[[],[ 16, 18 ],[ 16, 18 ]], //2016
[[],[],[]], //2017
[[],[],[]], //2018
[[],[],[]], //2019
[[],[],[]], //2020
[[],[],[]], //2021
[[],[],[]], //2022
[[],[],[]], //2023
[[ 3 ],[ 2 ],[]], //2024
[[ 32, 33 ],[ 30, 34 ],[ 34 ]], //2025
[[],[],[]], //2026
[[],[],[]], //2027
[[],[],[]], //2028
[[],[],[]], //2029
[[],[],[]], //2030
[[],[],[]], //2031
[[],[],[]], //2032
[[],[],[]], //2033
[[],[],[]], //2034
[[],[],[]], //2035
[[],[],[]], //2036
[[],[],[]], //2037
[[],[],[]], //2038
[[],[],[]], //2039
[[],[],[]], //2040
[[],[],[]], //2041
[[],[],[]], //2042
[[],[],[]], //2043
[[],[],[]], //2044
[[],[],[]], //2045
[[],[],[]], //2046
[[],[ 1 ],[ 1 ]], //2047
[[],[ 12 ],[ 12 ]], //2048
[[],[ 2 ],[ 2 ]], //2049
[[],[],[]], //2050
[[],[],[]], //2051
[[],[],[]], //2052
[[],[],[]], //2053
[[],[],[]], //2054
[[],[],[]], //2055
[[],[],[]], //2056
[[],[],[]], //2057
[[ 2 ],[ 1 ],[]], //2058
[[],[],[]], //2059
[[],[],[]], //2060
[[],[],[]], //2061
[[],[],[]], //2062
[[],[],[]], //2063
[[],[],[]], //2064
[[],[],[]], //2065
[[],[],[]], //2066
[[],[],[]], //2067
[[],[],[]], //2068
[[],[],[]], //2069
[[],[],[]], //2070
[[],[],[]], //2071
[[],[],[]], //2072
[[],[],[]], //2073
[[],[],[]], //2074
[[],[],[]], //2075
[[],[],[]], //2076
[[],[],[]], //2077
[[],[],[]], //2078
[[],[ 2 ],[ 2 ]], //2079
[[ 17 ],[ 15, 16 ],[ 15 ]], //2080
[[],[],[]], //2081
[[],[],[]], //2082
[[],[],[]], //2083
[[],[],[]], //2084
[[],[],[]], //2085
[[],[],[]], //2086
[[],[],[]], //2087
[[],[],[]], //2088
[[],[],[]], //2089
[[],[],[]], //2090
[[],[],[]], //2091
[[],[],[]], //2092
[[],[],[]], //2093
[[],[],[]], //2094
[[],[],[]], //2095
[[],[],[]], //2096
[[],[],[]], //2097
[[],[],[]], //2098
[[],[],[]], //2099
[[],[],[]], //2100
[[],[],[]], //2101
[[],[],[]], //2102
[[],[],[]], //2103
[[],[],[]], //2104
[[],[],[]], //2105
[[ 5 ],[ 4 ],[]], //2106
[[ 2 ],[ 1 ],[]], //2107
[[],[],[]], //2108
[[ 2 ],[ 1 ],[]], //2109
[[],[],[]], //2110
[[],[],[]], //2111
[[],[],[]], //2112
[[],[],[]], //2113
[[],[],[]], //2114
[[],[],[]], //2115
[[ 4 ],[],[]], //2116
[[],[],[]], //2117
[[],[],[]], //2118
[[],[],[]], //2119
[[],[],[]], //2120
[[],[],[]], //2121
[[],[],[]], //2122
[[],[],[]], //2123
[[],[],[]], //2124
[[],[],[]], //2125
[[],[],[]], //2126
[[],[],[]], //2127
[[],[],[]], //2128
[[],[],[]], //2129
[[],[],[]], //2130
[[],[],[]], //2131
[[],[],[]], //2132
[[],[],[]], //2133
[[],[],[]], //2134
[[],[],[]], //2135
[[],[],[]], //2136
[[],[],[]], //2137
[[],[],[]], //2138
[[],[],[]], //2139
[[],[],[]], //2140
[[],[],[]], //2141
[[],[],[]], //2142
[[],[],[]], //2143
[[],[],[]], //2144
[[],[ 2 ],[ 2 ]], //2145
[[],[],[]], //2146
[[],[],[]], //2147
[[],[],[]], //2148
[[],[],[]], //2149
[[],[],[]], //2150
[[],[],[]], //2151
[[],[],[]], //2152
[[],[],[]], //2153
[[],[],[]], //2154
[[],[],[]], //2155
[[],[],[]], //2156
[[],[],[]], //2157
[[],[],[]], //2158
[[],[],[]], //2159
[[],[],[]], //2160
[[],[],[]], //2161
[[],[ 1 ],[ 1 ]], //2162
[[],[],[]], //2163
[[],[],[]], //2164
[[],[],[]], //2165
[[],[],[]], //2166
[[],[],[]], //2167
[[],[],[]], //2168
[[],[],[]], //2169
[[],[],[]], //2170
[[],[],[]], //2171
[[],[],[]], //2172
[[],[],[]], //2173
[[],[],[]], //2174
[[],[],[]], //2175
[[],[],[]], //2176
[[],[],[]], //2177
[[],[],[]], //2178
[[],[],[]], //2179
[[],[],[]], //2180
[[],[],[]], //2181
[[],[],[]], //2182
[[],[],[]], //2183
[[],[],[]], //2184
[[],[],[]], //2185
[[],[],[]], //2186
[[ 71 ],[ 70 ],[]], //2187
[[ 4 ],[ 3 ],[]], //2188
[[],[],[]], //2189
[[],[],[]], //2190
[[],[],[]], //2191
[[],[],[]], //2192
[[],[],[]], //2193
[[],[],[]], //2194
[[],[],[]], //2195
[[],[],[]], //2196
[[ 158, 170 ],[ 157, 169 ],[]], //2197
[[ 6 ],[ 2, 5 ],[ 2 ]], //2198
[[],[],[]], //2199
[[],[],[]], //2200
[[],[],[]], //2201
[[],[],[]], //2202
[[],[],[]], //2203
[[],[],[]], //2204
[[],[],[]], //2205
[[],[],[]], //2206
[[],[],[]], //2207
[[],[],[]], //2208
[[ 70, 74 ],[ 69, 73 ],[]], //2209
[[ 5 ],[ 3 ],[]], //2210
[[ 2, 4 ],[ 1, 3 ],[]], //2211
[[],[],[]], //2212
[[],[],[]], //2213
[[],[],[]], //2214
[[],[],[]], //2215
[[],[],[]], //2216
[[],[],[]], //2217
[[],[],[]], //2218
[[],[],[]], //2219
[[],[],[]], //2220
[[],[],[]], //2221
[[],[],[]], //2222
[[],[],[]], //2223
[[],[],[]], //2224
[[],[],[]], //2225
[[],[],[]], //2226
[[],[],[]], //2227
[[],[],[]], //2228
[[],[],[]], //2229
[[],[],[]], //2230
[[],[],[]], //2231
[[],[],[]], //2232
[[],[],[]], //2233
[[],[],[]], //2234
[[],[],[]], //2235
[[],[],[]], //2236
[[],[],[]], //2237
[[],[],[]], //2238
[[],[],[]], //2239
[[],[],[]], //2240
[[],[],[]], //2241
[[],[],[]], //2242
[[],[],[]], //2243
[[],[],[]], //2244
[[],[],[]], //2245
[[],[],[]], //2246
[[],[],[]], //2247
[[],[],[]], //2248
[[],[],[]], //2249
[[],[],[]], //2250
[[],[],[]], //2251
[[],[],[]], //2252
[[],[],[]], //2253
[[],[],[]], //2254
[[],[],[]], //2255
[[],[],[]], //2256
[[],[ 1 ],[ 1 ]], //2257
[[],[],[]], //2258
[[],[],[]], //2259
[[],[],[]], //2260
[[],[],[]], //2261
[[],[],[]], //2262
[[],[],[]], //2263
[[],[],[]], //2264
[[],[],[]], //2265
[[],[],[]], //2266
[[],[],[]], //2267
[[],[],[]], //2268
[[],[],[]], //2269
[[],[],[]], //2270
[[],[],[]], //2271
[[],[],[]], //2272
[[],[],[]], //2273
[[],[],[]], //2274
[[],[],[]], //2275
[[],[],[]], //2276
[[],[],[]], //2277
[[],[ 4 ],[ 4 ]], //2278
[[],[],[]], //2279
[[],[],[]], //2280
[[],[],[]], //2281
[[],[],[]], //2282
[[],[],[]], //2283
[[],[],[]], //2284
[[],[],[]], //2285
[[],[],[]], //2286
[[],[],[]], //2287
[[],[],[]], //2288
[[],[],[]], //2289
[[],[],[]], //2290
[[],[],[]], //2291
[[],[],[]], //2292
[[],[],[]], //2293
[[],[],[]], //2294
[[],[ 2 ],[ 2 ]], //2295
[[],[],[]], //2296
[[],[],[]], //2297
[[],[],[]], //2298
[[],[],[]], //2299
[[ 3 ],[ 1, 2 ],[ 1 ]], //2300
[[],[],[]], //2301
[[],[],[]], //2302
[[],[],[]], //2303
[[],[ 8, 10 ],[ 8, 10 ]], //2304
[[],[],[]], //2305
[[],[],[]], //2306
[[],[],[]], //2307
[[],[],[]], //2308
[[],[],[]], //2309
[[],[],[]], //2310
[[],[],[]], //2311
[[],[],[]], //2312
[[],[],[]], //2313
[[],[],[]], //2314
[[],[],[]], //2315
[[],[],[]], //2316
[[],[],[]], //2317
[[],[],[]], //2318
[[],[],[]], //2319
[[],[],[]], //2320
[[],[],[]], //2321
[[],[],[]], //2322
[[],[],[]], //2323
[[],[],[]], //2324
[[],[],[]], //2325
[[],[],[]], //2326
[[],[],[]], //2327
[[],[],[]], //2328
[[],[],[]], //2329
[[],[],[]], //2330
[[],[],[]], //2331
[[],[],[]], //2332
[[],[],[]], //2333
[[],[],[]], //2334
[[],[],[]], //2335
[[],[],[]], //2336
[[],[],[]], //2337
[[],[],[]], //2338
[[],[],[]], //2339
[[],[],[]], //2340
[[],[],[]], //2341
[[],[],[]], //2342
[[],[],[]], //2343
[[],[],[]], //2344
[[],[],[]], //2345
[[ 2 ],[ 1 ],[]], //2346
[[],[],[]], //2347
[[],[],[]], //2348
[[],[],[]], //2349
[[],[],[]], //2350
[[],[],[]], //2351
[[],[],[]], //2352
[[],[],[]], //2353
[[],[],[]], //2354
[[],[],[]], //2355
[[],[],[]], //2356
[[],[],[]], //2357
[[],[],[]], //2358
[[],[],[]], //2359
[[],[],[]], //2360
[[],[],[]], //2361
[[],[],[]], //2362
[[],[],[]], //2363
[[],[],[]], //2364
[[],[],[]], //2365
[[],[],[]], //2366
[[],[],[]], //2367
[[],[],[]], //2368
[[],[],[]], //2369
[[],[],[]], //2370
[[],[],[]], //2371
[[],[],[]], //2372
[[],[],[]], //2373
[[],[],[]], //2374
[[],[],[]], //2375
[[],[],[]], //2376
[[],[],[]], //2377
[[],[],[]], //2378
[[],[],[]], //2379
[[ 4, 6, 9 ],[ 2, 5, 8 ],[]], //2380
[[],[],[]], //2381
[[],[],[]], //2382
[[],[],[]], //2383
[[],[],[]], //2384
[[],[],[]], //2385
[[],[],[]], //2386
[[],[],[]], //2387
[[],[],[]], //2388
[[],[],[]], //2389
[[],[],[]], //2390
[[],[],[]], //2391
[[],[],[]], //2392
[[],[],[]], //2393
[[],[],[]], //2394
[[],[],[]], //2395
[[],[],[]], //2396
[[],[],[]], //2397
[[],[],[]], //2398
[[],[],[]], //2399
[[],[],[]], //2400
[[1179,1229,1233],[1178,1226,1232],[]], //2401
[[ 8 ],[ 6 ],[]], //2402
[[],[],[]], //2403
[[],[],[]], //2404
[[],[],[]], //2405
[[],[],[]], //2406
[[],[],[]], //2407
[[],[],[]], //2408
[[],[],[]], //2409
[[],[],[]], //2410
[[],[],[]], //2411
[[],[],[]], //2412
[[],[],[]], //2413
[[],[],[]], //2414
[[],[ 2 ],[ 2 ]], //2415
[[],[],[]], //2416
[[],[],[]], //2417
[[],[],[]], //2418
[[],[],[]], //2419
[[],[],[]], //2420
[[],[],[]], //2421
[[],[],[]], //2422
[[],[],[]], //2423
[[],[],[]], //2424
[[],[],[]], //2425
[[],[],[]], //2426
[[],[],[]], //2427
[[],[],[]], //2428
[[],[],[]], //2429
[[],[],[]], //2430
[[],[],[]], //2431
[[],[],[]], //2432
[[],[],[]], //2433
[[],[],[]], //2434
[[],[],[]], //2435
[[],[],[]], //2436
[[],[],[]], //2437
[[],[],[]], //2438
[[],[],[]], //2439
[[],[ 2 ],[ 2 ]], //2440
[[],[],[]], //2441
[[],[],[]], //2442
[[],[],[]], //2443
[[],[],[]], //2444
[[],[],[]], //2445
[[],[],[]], //2446
[[],[],[]], //2447
[[ 5 ],[ 2 ],[]], //2448
[[],[],[]], //2449
[[],[],[]], //2450
[[ 4 ],[ 3 ],[]], //2451
[[],[],[]], //2452
[[],[],[]], //2453
[[],[],[]], //2454
[[],[],[]], //2455
[[],[],[]], //2456
[[],[ 2 ],[ 2 ]], //2457
[[],[],[]], //2458
[[],[],[]], //2459
[[],[],[]], //2460
[[],[],[]], //2461
[[],[],[]], //2462
[[],[],[]], //2463
[[],[],[]], //2464
[[],[ 2 ],[ 2 ]], //2465
[[],[],[]], //2466
[[],[],[]], //2467
[[],[],[]], //2468
[[],[],[]], //2469
[[],[],[]], //2470
[[],[],[]], //2471
[[],[],[]], //2472
[[],[],[]], //2473
[[],[],[]], //2474
[[],[],[]], //2475
[[],[],[]], //2476
[[],[],[]], //2477
[[],[],[]], //2478
[[],[],[]], //2479
[[],[],[]], //2480
[[],[],[]], //2481
[[],[],[]], //2482
[[],[],[]], //2483
[[],[],[]], //2484
[[ 2, 4 ],[ 1, 3 ],[]], //2485
[[],[],[]], //2486
[[],[],[]], //2487
[[],[],[]], //2488
[[],[],[]], //2489
[[],[],[]], //2490
[[],[],[]], //2491
[[],[],[]], //2492
[[],[],[]], //2493
[[],[],[]], //2494
[[],[],[]], //2495
[[],[],[]], //2496
[[],[],[]], //2497
[[],[],[]], //2498
[[],[],[]]  //2499
];
if sym then
    subs := [PrimitiveGroup(n,d) : d in maxprim[n][1]];
else 
    x:=Sym(n)!(1,2);
    subs := [ ];
    for d in maxprim[n][2] do
        P:=PrimitiveGroup(n,d);
        Append(~subs,P);
        if d in maxprim[n][3] then
            Append(~subs,P^x);
        end if;
    end for;
end if;

/***************************************************
 *Finally we deal with the generic cases. AGL(1, p) is
 *always Sym(p) maximal if p gt 24, with its normal subgroup
 *of index 2 being alt maximal.
 *PGL(2, p) is always Sym(p+1) maximal for p gt 23 (so 
 *degree gt 24), with PSL(2, p) being
 *alt(p+1) maximal.
 ***************************************************/
   
if n gt 23 and IsPrime(n) then
    if sym then
       Append(~subs, AGL(1, n));
    else
       Append(~subs, PrimitiveGroup(n, #Divisors(n-1) -1));
                 /*This is n:((n-1)/2)*/
    end if;
end if;
if n gt 24 and IsPrime(n-1) then
    if sym then
       Append(~subs, PGL(2, n-1));
    else
       Append(~subs, PSL(2, n-1));
    end if;
end if;

  subs := sym select [ rec<RFG | subgroup := s, order := Order(s),
                      length := Factorial(n) div Order(s) > : s in subs ] 
                else [ rec<RFG | subgroup := s, order := Order(s),
                      length := Factorial(n) div (2*Order(s)) > : s in subs ]; 
return subs;
end function;

AltSymMax := function(n, sym)
   //maximal subgroups of Sym(n) (sym=true) or Alt(n) (sym=false) 
   assert n ge 4;
   subs := IntransitiveSubgroupsAltSym(n,sym) cat
           ImprimitiveSubgroupsAltSym(n,sym) cat
           PrimitiveSubgroupsAltSym(n,sym);
  if sym then Append(~subs, rec<RFG |
              subgroup := Alt(n), order := Factorial(n) div 2, length:=1 > );
  end if;
  return subs;
end function;

intrinsic MaximalSubgroupsAltSym(n::RngIntElt, sym::BoolElt) -> SeqEnum
{Maximal subgroups of the standard groups Alt(n) and Sym(n). Set sym true
for Sym, false for Alt}
    require n ge 4:"degree must be at least 4";
    return AltSymMax(n, sym);
end intrinsic;

intrinsic MaximalSubgroupsAlt(n::RngIntElt) -> SeqEnum
{Maximal subgroups of the standard groups Alt(n). }
    require n ge 4:"degree must be at least 4";
    return AltSymMax(n, false);
end intrinsic;

intrinsic MaximalSubgroupsSym(n::RngIntElt) -> SeqEnum
{Maximal subgroups of the standard groups Sym(n). }
    require n ge 4:"degree must be at least 4";
    return AltSymMax(n, true);
end intrinsic;

intrinsic MaximalSubgroupsAltSym(G::GrpPerm) -> SeqEnum
{Maximal subgroups of G, where G is assumed to be isomorphic to either
Alt(n) or Sym(n), n ge 4.}

    n := Degree(G);
    require n ge 4:"degree must be at least 4";
    if IsAltsym(G) then
	maxsubs := AltSymMax(n, not IsEven(G));
	S, f := StandardGroup(G);
	if S cmpeq G then
	    return maxsubs;
	end if;
	for i := 1 to #maxsubs do
	    maxsubs[i]`subgroup := sub<G| (maxsubs[i]`subgroup)@@f>;
	    (maxsubs[i]`subgroup)`Order := maxsubs[i]`order;
	end for;
	return maxsubs;
    end if;

    error "Unable to compute maximal subgroups";

end intrinsic;
