freeze;

Z := IntegerRing();
Q := RationalField();

intrinsic Lattice(S::MonStgElt, n::RngIntElt) -> Lat
{The standard lattice given by S and n}
    if S eq "A" then
	requirege n, 1;
	B := RMatrixSpace(Z, n, n + 1) ! 0;
	for i := 1 to n do
	    B[i][i] := -1;
	    B[i][i + 1] := 1;
	end for;
	return LatticeWithBasis(B);
    elif S eq "B" then
	requirege n, 1;
	return StandardLattice(n);
    elif S eq "C" then
	return Lattice("D", n);
    elif S eq "D" then
	requirege n, 3;
	B := RMatrixSpace(Z, n, n) ! 0;
	B[1][1] := -1;
	B[1][2] := -1;
	for i := 2 to n do
	    B[i][i - 1] := -1;
	    B[i][i] := 1;
	end for;
	return LatticeWithBasis(B);
    elif S eq "E" then
	require n in [6, 7, 8]: "Argument 2 must be 6, 7, or 8";

	B := RMatrixSpace(Q, n, 8) ! 0;
	if n eq 8 then
	    B[1][1] := 2;
	    s := 1;
	elif n eq 7 then
	    s := 0;
	else
	    s := -1;
	end if;
	for i := Max(1 - s, 1) to 8 - 2 do
	    B[i + s][i] := -1;
	    B[i + s][i + 1] := 1;
	end for;
	for j := 1 to 8 do
	    B[n][j] := n eq 8 or j le 4 select 1/2 else -1/2;
	end for;
	return LatticeWithBasis(B);
    elif S eq "F" then
	require n in [4]: "Argument 2 must be 4";
	return Lattice("D", n);
    elif S eq "G" then
	require n in [2]: "Argument 2 must be 2";
	return Lattice("A", n);
    elif S eq "Kappa" then
        require n in [1 .. 13]: "Argument 2 should be in the range [1 .. 13]";
	return KappaLattice(n);
    elif S eq "Lambda" then
        require n in [1 .. 31]: "Argument 2 should be in the range [1 .. 31]";
	return LaminatedLattice(n);
    else
	require false:
	    "Argument 1 should be one of " cat
		"\"A\", \"B\", \"C\", \"D\", \"E\", \"F\", \"G\", " cat
		"\"Kappa\", or \"Lambda\"";
    end if;
end intrinsic;


intrinsic BKZ(L::Lat,n::RngIntElt) -> Lat, Mtrx
{} // need to pass VARARGS
 d:=Dimension(L); C:=Cholesky(L); // need to think more about precision
 M:=Matrix(d,d,[Round(x) : x in Eltseq(BasisMatrix(C)*10^100)]);
 B:=BKZ(M,n); T:=B*ChangeRing(M,Q)^(-1); // BKZ does not return T
 TZ:=ChangeRing(T,Z);
 BM:=ChangeRing(BasisMatrix(L),Q); IPM:=ChangeRing(InnerProductMatrix(L),Q);
 if IsGLattice(L) then S:=[T*g*Inverse(T) : g in Generators(Group(L))];
  G:=sub<GL(d,Integers())|S>; return LatticeWithBasis(G,T*BM,IPM),TZ; end if;
 return LatticeWithBasis(T*BM,IPM),TZ; end intrinsic;

/*
> L:=LLL(LATTICES[1] : Deep,Delta:=0.9999);
> n:=36;
> d:=Dimension(L); C:=Cholesky(L);
> M:=Matrix(d,d,[Round(x) : x in Eltseq(BasisMatrix(C)*10^100)]);
> B:=BKZ(M,n); T:=ChangeRing(B*Inverse(ChangeRing(M,Rationals())),Integers());
> GM:=T*GramMatrix(L)*Transpose(T);
> Nagens(L);
> ag:=ActionGenerator(L,1);
 */
