
/* L-function of Hecke characters */

freeze;

declare type GrossenChar;

Z:=Integers();

function cyclo_comp(raw,P) C<i>:=ComplexField(P);
 if Type(Parent(raw)) eq RngInt then return 0; end if;
 m:=Modulus(Parent(raw)); r:=Integers()!raw;
  // zr:=r*(zm div m); return zetam^zr; probably Exp is just as fast
 return Exp(2*Pi(C)*i*(r/m)); end function;

function cyclo_mult(a,b)
 m1:=Modulus(Parent(a)); r1:=Z!a; m2:=Modulus(Parent(b)); r2:=Z!b;
 M:=LCM(m1,m2); r1:=r1*(M div m1); r2:=r2*(M div m2); r:=r1+r2;
 return Integers(M)!r;
 g:=Gcd(M,r); return Integers(M div g)!(r div g); end function;
/* probably want to use second reduction */

////////////////////////////////////////////////////////////////////////

declare attributes GrossenChar: hecke,evalf,type,wt,dirich,zetas,clreps,roots;

intrinsic PrintNamed
 (gr::GrossenChar,level::MonStgElt,name::MonStgElt) {}
 if name ne "$" then
  printf "Grossencharacter %o of type %o for Hecke character %o in "*
         "Group of Hecke characters of modulus of norm %o over %o",
          name,gr`type,gr`hecke,Norm(Modulus(gr`hecke`Parent)),
          gr`hecke`Parent`NF;
 else printf "Grossencharacter of type %o for Hecke character %o in "*
             "Group of Hecke characters of modulus of norm %o",
             gr`type,gr`hecke,Norm(Modulus(gr`hecke`Parent)),
             gr`hecke`Parent`NF; end if; end intrinsic;

intrinsic IsCoercible (gr::GrossenChar,X::.) -> BoolElt,. {}
 return false,"Illegal coercion"; end intrinsic;

intrinsic 'in' (x::.,gr::GrossenChar) -> BoolElt {}
 return false; end intrinsic;

intrinsic Conductor(gr::GrossenChar) -> RngOrdIdl, SeqEnum
{ Given a Hecke Grossencharacter,
  return its conductor as an ideal and a sequence of real places }
 return Conductor(DirichletRestriction(gr`hecke)/gr`dirich); end intrinsic;

intrinsic Modulus(gr::GrossenChar) -> RngOrdIdl, SeqEnum
{ Given a Hecke Grossencharacter,
  return its modulus as an ideal and a sequence of real places }
 return gr`hecke`Parent`Modulus,gr`hecke`Parent`ooReal; end intrinsic;

intrinsic IsPrimitive(gr::GrossenChar) -> BoolElt
{ Whether or not a Grossencharacter is primitive }
 C,ooC:=Conductor(gr);
 return C eq gr`hecke`Parent`Modulus and ooC eq gr`hecke`Parent`ooReal;
 end intrinsic;

function IsCMField(K)
 d:=Degree(K); r,c:=Signature(K); if r ne 0 then return false; end if;
 if d eq 2 and r eq 0 then return true; end if;
 P:=[F[1] : F in Subfields(K) | 2*Degree(F[1]) eq d and IsTotallyReal(F[1])];
 return #P gt 0; end function;

////////////////////////////////////////////////////////////////////////

declare attributes FldNum: extension_field,Hip,grossen_array,ext_roots;

function cyclo_get_power(z,m)
 if Type(Parent(z)) eq FldRe then z:=ComplexField(Parent(z))!z; end if;
 e:=Round(Real(m*Imaginary(Log(z))/(2*Pi(Parent(z)))));
 if e lt 0 then e:=e+m; end if; return e; end function;

function cyclo_str(z,m)
 return Sprintf("zeta_%o^%o",m,cyclo_get_power(z,m)); end function;

////////////////////////////////////////////////////////////////////////

intrinsic '@'(e::RngIntElt,gr::GrossenChar : Precision:=0) -> RngElt {}
 K:=gr`hecke`Parent`NF;
 return '@'(e*IntegerRing(K),gr : Precision:=Precision); end intrinsic;

intrinsic '@'(e::FldRatElt,gr::GrossenChar : Precision:=0) -> RngElt {}
 K:=gr`hecke`Parent`NF;
 return '@'(e*IntegerRing(K),gr : Precision:=Precision); end intrinsic;

intrinsic '@'(e::RngOrdElt,gr::GrossenChar : Precision:=0) -> RngElt {}
 K:=gr`hecke`Parent`NF;
 return '@'(K!e,gr : Precision:=Precision); end intrinsic;

intrinsic '@'(e::FldNumElt,gr::GrossenChar : Precision:=0) -> RngElt {}
 H:=gr`hecke`Parent; K:=Parent(e);
 require K eq H`NF: "Element must be compatible with the character";
 return '@'(e*IntegerRing(K),gr : Precision:=Precision); end intrinsic;

intrinsic '@'(J::RngOrdFracIdl,gr::GrossenChar : Precision:=0) -> FldComElt {}
 require gr`hecke`Parent`NF eq NumberField(Order(J)):
 "Grossencharacter and ideal must be over the same number field";
 prec:= Precision gt 0 select Precision else GetPrecision(0.5);
 psi:=gr`hecke; v:='@'(J,psi : Raw);
 if Type(Parent(v)) eq RngInt then return ComplexField(prec)!0.0; end if;
 e,z:=gr`evalf(J); r:=ComplexField(prec)!1.0; Hip:=NumberField(Order(J))`Hip;
 for i in [1..#Hip] do
  ip:=Hip[i]; t:=gr`type[i]; ev:=Evaluate(e,ip : Precision:=prec);
  r*:=ev^t[1]*ComplexConjugate(ev)^t[2]; end for;
 return r*cyclo_comp(cyclo_mult(v,z),prec); end intrinsic;

intrinsic RawEval(J::RngOrdFracIdl,gr::GrossenChar) ->
 FldNumElt,FldCycElt,FldCycElt {}
 require gr`hecke`Parent`NF eq NumberField(Order(J)):
 "Grossencharacter and ideal must be over the same number field";
 v:=gr`hecke(J); if v eq 0 then return 0,0,0; end if;
 e,z:=gr`evalf(J); r:=Z!z; m:=Modulus(Parent(z));
 return e,CyclotomicField(m).1^r,v; end intrinsic;

/************************************************************************/

function AbelianGroupExtension(Q,E,R) // 0 -> E -> A -> Q -> 0
// extend Q by E, with relations R, each <q,m,e> sending m*q to e
 ngQ:=Ngens(Q); ngE:=Ngens(E);
 F:=FreeAbelianGroup(ngQ+ngE);
 REL:=[Order(E.i)*F.(ngQ+i) : i in [1..ngE]];
 fQtoF:=func<q | F ! (Eltseq(q) cat [0 : i in [1..ngE]])>;
 fEtoF:=func<e | F ! ([0 : i in [1..ngQ]] cat Eltseq(e))>;
 FtoQE:=function(e) T:=Eltseq(e);
   return Q![T[i] : i in [1..ngQ]],E![T[i+ngQ] : i in [1..ngE]]; end function;
 S:=[fQtoF(r[1])*r[2]-fEtoF(r[3]) : r in R]; A,qmap:=quo<F|REL cat S>;
 EtoA:=hom<E->A|[qmap(fEtoF(E.i)) : i in [1..ngE]]>;
 AtoQ:=hom<A->Q|[FtoQE((A.i) @@ qmap) : i in [1..Ngens(A)]]>;
 return A,EtoA,AtoQ; end function;

function Principalise(x) b,e:=IsPrincipal(x); assert b; return e; end function;

////////////////////////////////////////////////////////////////////////

procedure grossenideal_catalogue(GI,K) n:=Norm(GI[1]);
 if not n in Keys(K`grossen_array) then K`grossen_array[n]:=<GI>;
 else Append(~K`grossen_array[n],GI); end if; end procedure;

function grossenideal_exists(I,K)
 if not assigned K`grossen_array then
  K`grossen_array:=AssociativeArray(Integers()); return false,_; end if;
 A:=K`grossen_array;
 n:=Norm(I); if not n in Keys(A) then return false,_; end if;
 for gi in A[n] do if gi[1] eq I then return true,gi; end if; end for;
 return false,_; end function;
 
function initialise_grossenideal(I) K:=NumberField(Order(I)); POW:=[]; CLR:=[];
 b,GI:=grossenideal_exists(I,K); if b then return GI; end if; oo:=[]; ROO:=[];
 CL,clmp:=ClassGroup(K); RR,rrm:=RayResidueRing(I,oo); U,um:=UnitGroup(K);
 TOR,tm:=TorsionUnitGroup(K); cyc:=tm(TOR.1);
 H:=K`extension_field; _<y>:=PolynomialRing(K);
 for i in [1..Ngens(CL)] do c:=clmp(CL.i); m:=MakeCoprime(c,I);
  c:=c*m; CLR cat:=[c]; o:=Order(CL.i);
  e:=(K!(K`ext_roots[i]^o))*m^o; r:=K`ext_roots[i]*m;
  ROO cat:=[r]; POW cat:=[<CL.i,o,e@@rrm>]; end for;
 A,RRtoA,AtoCL:=AbelianGroupExtension(CL,RR,POW); ZE:=[];
 for j in [1..#POW] do o:=POW[j][2];
  mult:=hom<A->A|[o*A.i : i in [1..Ngens(A)]]>; DS,m1,m2:=DirectSum(A,CL);
  h:=hom<A->DS|[m1(mult(A.i))+m2(AtoCL(A.i)) : i in [1..Ngens(A)]]>;
  b,ret:=HasPreimage(m1(RRtoA(POW[j][3]))+m2(POW[j][1]),h); assert b;
  ZE cat:=[ret]; end for;
 DU,dual_map:=Dual(A : UseSameGroup); GI:=<I,CLR,ROO,ZE,0,0,DU,dual_map,RRtoA>;
 grossenideal_catalogue(GI,K); return GI; end function;

procedure ensure_field(K) if assigned(K`extension_field) then return; end if;
 CL,mp:=ClassGroup(K); _<y>:=PolynomialRing(K); // already comp: HeckeCharacter
 if #CL eq 1 then K`extension_field:=K; K`ext_roots:=[];
                  K`Hip:=InfinitePlaces(K); return; end if;
 cgp:=[w where _,w:=IsPrincipal(mp(CL.i)^Order(CL.i)) : i in [1..Ngens(CL)]];
 H:=ext<K|[y^Order(CL.i)-cgp[i] : i in [1..Ngens(CL)]]>;
 K`ext_roots:=[Root(H!cgp[i],Order(CL.i)) : i in [1..Ngens(CL)]];
 ip:=[[ooH : ooH in InfinitePlaces(H) | Extends(ooH,ooK)]
           : ooK in InfinitePlaces(K)];
 K`extension_field:=H; K`Hip:=[lifts[1] : lifts in ip]; end procedure;

function simple_pullback(e,h)
 o:=Order(e); if o eq 1 then return Domain(h).0; end if;
 H:=Parent(e); HS:=sub<H|e>; S:=HS@@h; G:=Domain(h); A:=sub<G|G.0>;
 for F in Factorization(o) do p:=F[1]; v:=F[2]; W:=Sylow(S,p);
  WS:=sub<W|[W.i : i in [1..Ngens(W)] |  sub<H|h(W.i)> eq (o div (p^v))*HS]>;
  A:=A+sub<G|pPrimaryComponent(WS,p).1>; end for;
 hm:=hom<A->H|[h(G!A.i) : i in [1..Ngens(A)]]>; return G!(e@@hm); end function;

////////////////////////////////////////////////////////////////////////

function get_eval(ZETAS,ROOTS,CLREPS,CHI)
 K:=CHI`Parent`NF; CL,clmp:=ClassGroup(K); ng:=Ngens(CL); H:=K`extension_field;
 function princ_eval(e) b,alpha:=IsPrincipal(e); assert b;
  return K!alpha,-'@'(K!alpha,CHI : Raw); end function;
 function feval(e) im:=-(e@@clmp); expo:=Eltseq(im); zd:=Integers(1)!0;
  mult:=&*[Codomain(clmp) | CLREPS[i]^expo[i] : i in [1..ng]];
  ev,z:=princ_eval(e*mult); dv:=&*[H | ROOTS[i]^expo[i] : i in [1..ng]];
  if ng ne 0 then zd:=&+[ZETAS[i]*expo[i] : i in [1..ng]]; end if;
  return (H!ev)/dv,cyclo_mult(z,zd); end function; return feval; end function;

procedure force_clreps_and_roots(gr,CLREPS,ROOTS,oldchi) ng:=#CLREPS;
 K:=gr`hecke`Parent`NF; CL,clmp:=ClassGroup(K); H:=K`extension_field;
 for i in [1..ng] do // e1*e2 is change in roots, see get_eval above
  e1:=Principalise(CLREPS[i]/gr`clreps[i]); gr`clreps[i]:=CLREPS[i];
  e2:=ROOTS[i]/(gr`roots[i]*e1); gr`roots[i]:=ROOTS[i];
  gr`zetas[i]:=cyclo_mult(gr`zetas[i],'@'(K!(e1*e2),oldchi : Raw)); end for;
 gr`evalf:=get_eval(gr`zetas,gr`roots,gr`clreps,gr`dirich); end procedure;

function make_grossencharacter(chi,psi,I,wt,T)
 K:=NumberField(Order(I)); CL,clmp:=ClassGroup(K); ng:=Ngens(CL);
 GI:=initialise_grossenideal(I); H:=K`extension_field; ID:=Codomain(clmp);
 gr:=New(GrossenChar); gr`wt:=wt; gr`type:=T;
 gr`clreps:=GI[2]; gr`roots:=GI[3]; ZE:=GI[4];
 Z:=Integers(); DU:=GI[7]; pair:=GI[8]; RRtoA:=GI[9];
 M:=Modulus(Codomain(pair)); CYC:=CyclotomicField(M);
 RR:=Domain(RRtoA); A:=Codomain(RRtoA); AB:=AbelianGroup([M]);
 DS,ic:=DirectSum([AB : i in [1..Ngens(RR)]]);
 h:=hom<DU->DS|
    [&+[ic[j](Domain(ic[j])![Z!pair(DU.i,RRtoA(RR.j))]) : j in [1..Ngens(RR)]]
      : i in [1..Ngens(DU)]]>;
 SD,sp:=Dual(RR : UseSameGroup); m:=Modulus(Codomain(sp)); o:=M div m;
 sd:=SD!Eltseq(chi);
 tar:=&+[ic[j](Domain(ic[j])![o*(Z!sp(sd,RR.j))]) : j in [1..Ngens(RR)]];
 ch:=simple_pullback(tar,h); gr`zetas:=[* pair(ze,ch) : ze in ZE *];
 gr`dirich:=chi; gr`hecke:=psi;
 gr`evalf:=get_eval(gr`zetas,gr`roots,gr`clreps,chi); return gr; end function;

////////////////////////////////////////////////////////////////////////

// Evaluate works to KANT prec here !?
function check_1modI_is_OK(I,T) oo:=[]; K:=NumberField(Order(I));
 U,um:=UnitGroup(K); RR,rm:=RayResidueRing(I,oo); ip:=InfinitePlaces(K);
 A:=[ K!um(u) : u in Generators(U) ]; AA:=[ u^Order(u @@ rm) :  u in A ];
 for i in [1..#AA] do ev:=[Evaluate(AA[i],pl) : pl in ip];
  pr:=&*[ev[i]^T[i][1]*ComplexConjugate(ev[i])^T[i][2] : i in [1..#ev]];
  b:=Abs(pr-1) lt 10^(-Precision(pr)/2); if b then continue; end if;
  return false,
    Sprintf("\noo-type should be trivial on all"*
	    " totally positive units that are 1 mod I\n"*
	    "Fails for %o which gives %o\n",K!A[i],ComplexField(10)!pr),K!A[i];
 end for; return true,_,_; end function;

function check_chi_is_OK(chi,T) K:=NumberField(Order(chi`Parent`Modulus));
 U,um:=UnitGroup(K); ip:=InfinitePlaces(K); A:=[K!um(u) : u in Generators(U)];
 MOD:=CyclotomicOrder(chi`Parent`TargetRing);
 for i in [1..#A] do ev:=[Evaluate(A[i],pl) : pl in ip];
  prod:=&*[ev[i]^T[i][1]*ComplexConjugate(ev[i])^T[i][2] : i in [1..#ev]];
  pr:=Precision(prod); b:=Abs(prod-ComplexField(pr)!chi(A[i])) lt 10^(-pr/2);
  if b then continue; end if;
  return false,
    Sprintf("\noo-type must be compatible with Dirichlet character on units\n"*
	    "Fails for %o which gives %o rather than %o\n",K!A[i],
	    cyclo_str(prod,MOD),cyclo_str(ComplexField()!chi(A[i]),MOD));
 end for; return true,_; end function;

function get_chi(I,T) K:=NumberField(Order(I)); oo:=[]; IM:=[];
 U,um:=UnitGroup(K); ip:=InfinitePlaces(K); A:=[K!um(u) : u in Generators(U)];
 DG:=DirichletGroup(I,oo); MOD:=CyclotomicOrder(DG`TargetRing);
 rm:=DG`ResidueMap; RR:=Domain(rm);
 for i in [1..#A] do ev:=[Evaluate(A[i],pl) : pl in ip];
  prod:=&*[ev[i]^T[i][1]*ComplexConjugate(ev[i])^T[i][2] : i in [1..#ev]];
  k:=cyclo_get_power(prod,MOD); r:=A[i]@@rm; IM cat:=[<r,k>]; end for;
 DS,ic:=DirectSum([AbelianGroup([MOD]) : i in [1..#A]]); Z:=Integers();
 h:=hom<DG`AbGrp->DS|
        [&+[ic[i](Domain(ic[i])![Z!DG`Pairing(IM[i][1],RR.j)]) : i in [1..#A]]
	  : j in [1..Ngens(RR)]]>;
 t:=&+[ic[i](Domain(ic[i])!IM[i][2]) : i in [1..#A]];
 b,e:=HasPreimage(t,h); if not b then return false,_; end if;
 e:=simple_pullback(t,h); assert h(e) eq t;
 return true,DG!Eltseq(e); end function;

////////////////////////////////////////////////////////////////////////

intrinsic Grossencharacter(psi::GrpHeckeElt,T::SeqEnum[SeqEnum[RngIntElt]])
 -> GrossenChar
{Given a Hecke character along with a compatible infinity-type,
 try to find a suitable Dirichlet character on the units,
 and then return the induced Hecke Grossencharacter}
 if psi`Parent`issubgroup then
  return Grossencharacter(psi`Parent`supergroup!psi,T); end if;
 K:=psi`Parent`NF; I:=psi`Parent`Modulus; oo:=psi`Parent`ooReal;
 require IsCMField(K): "Field must be of CM-type";
 require Degree(K) eq 2*#T: "Degree must equal number of oo-type components";
 require #Set([&+t : t in T]) eq 1: "All oo-type parts must have same weight";
 b,S:=check_1modI_is_OK(I,T); require b: S;
 b,chi:=get_chi(I,T); require b: "No suitable Dirichlet character exists";
 ensure_field(K);
 return make_grossencharacter(chi,psi,I,&+T[1],T); end intrinsic;

intrinsic Grossencharacter
 (psi::GrpHeckeElt,chi::GrpDrchNFElt,T::SeqEnum[SeqEnum[RngIntElt]])
 -> GrossenChar
{Given a Hecke character along with a compatible infinity-type and
 Dirichlet character, return the induced Hecke Grossencharacter}
 if psi`Parent`issubgroup then
  return Grossencharacter(psi`Parent`supergroup!psi,chi,T); end if;
 if chi`Parent`issubgroup then
  return Grossencharacter(psi,chi`Parent`supergroup!chi,T); end if;
 K:=psi`Parent`NF; I:=psi`Parent`Modulus; oo:=psi`Parent`ooReal;
 require K eq chi`Parent`NF: "psi and chi must be defined over same field";
 require I eq chi`Parent`Modulus and oo eq chi`Parent`ooReal:
 "psi and chi must have same modulus";
 require IsCMField(K): "Field must be of CM-type";
 require Degree(K) eq 2*#T: "Degree must equal number of oo-type components";
 require #Set([&+t : t in T]) eq 1: "All oo-type parts must have same weight";
 if Degree(K) gt 2 then "WARNING: Not tested beyond quadratic fields"; end if;
 b,S:=check_1modI_is_OK(I,T); require b: S;
 b,S:=check_chi_is_OK(chi,T); require b: S;
 ensure_field(K);
 return make_grossencharacter(chi,psi,I,&+T[1],T); end intrinsic;

intrinsic GrossenCheck(I::RngOrdIdl,T::SeqEnum[SeqEnum[RngIntElt]]) ->
 BoolElt, FldNumElt
{Given an ideal in a CM field and an oo-type, determine whether
 the oo-type trivialises all units that are 1 mod I}
 K:=NumberField(Order(I)); require IsCMField(K): "Field must be of CM-type";
 require Degree(K) eq 2*#T: "Degree must equal number of oo-type components";
 require #Set([&+t : t in T]) eq 1: "All oo-type parts must have same weight";
 b,_,e:=check_1modI_is_OK(I,T);
 if b then return true,_; else return false,e; end if; end intrinsic;

function GrossenCopy(gr) cp:=New(GrossenChar);
 cp`wt:=gr`wt; cp`type:=gr`type; cp`hecke:=gr`hecke; cp`dirich:=gr`dirich;
 cp`zetas:=gr`zetas; cp`roots:=gr`roots; cp`clreps:=gr`clreps;
 cp`evalf:=gr`evalf; return cp; end function;

intrinsic Extend(gr::GrossenChar,I::RngOrdIdl) -> GrossenChar
{Extend a Grossencharacter to the given ideal}
 require gr`hecke`Parent`NF eq NumberField(Order(I)):
  "Grossencharacter and ideal must be over the same number field";
 require I subset Modulus(gr):
  "The modulus of the new group must be divisible by that of the old";
 // if I eq Modulus(gr) then return gr; end if;
 GR:=GrossenCopy(gr); GR`dirich:=Extend(gr`dirich,DirichletGroup(I));
 GR`hecke:=Extend(gr`hecke,HeckeCharacterGroup(I));
 GI:=initialise_grossenideal(I); // handles zetas correct from force_clreps ?!
 force_clreps_and_roots(GR,GI[2],GI[3],gr`dirich); return GR; end intrinsic;

intrinsic '*'(gr::GrossenChar,psi::GrpHeckeElt) -> GrossenChar {}
 require gr`hecke`Parent`NF eq psi`Parent`NF:
 "Grossencharacter and Hecke character must be defined over the same field";
 if psi`Parent`Modulus eq gr`hecke`Parent`Modulus and
    psi`Parent`ooReal eq gr`hecke`Parent`ooReal then
  GR:=GrossenCopy(gr); GR`hecke*:=psi; return GR; end if;
 nh:=gr`hecke*psi; GR:=Extend(gr,Modulus(nh)); GR`hecke:=nh;
 return GR; end intrinsic;

intrinsic '*'(psi::GrpHeckeElt,gr::GrossenChar) -> GrossenChar {}
 require gr`hecke`Parent`NF eq psi`Parent`NF:
 "Grossencharacter and Hecke character must be defined over the same field";
 return '*'(gr,psi); end intrinsic;

intrinsic '*'(gr::GrossenChar,chi::GrpDrchNFElt) -> GrossenChar {}
 require gr`hecke`Parent`NF eq chi`Parent`NF:
 "Grossencharacter and Dirichlet character must be defined over same field";
 return gr*HeckeLift(chi); end intrinsic;

intrinsic '*'(chi::GrpDrchNFElt,gr::GrossenChar) -> GrossenChar {}
 require gr`hecke`Parent`NF eq chi`Parent`NF:
 "Grossencharacter and Dirichlet character must be defined over same field";
 return gr*HeckeLift(chi); end intrinsic;

intrinsic '/'(psi::GrpHeckeElt,gr::GrossenChar) -> GrossenChar {}
 require gr`hecke`Parent`NF eq psi`Parent`NF:
 "Grossencharacter and Hecke character must be defined over the same field";
 return psi*(gr^(-1)); end intrinsic;

intrinsic '/'(gr::GrossenChar,psi::GrpHeckeElt) -> GrossenChar {}
 require gr`hecke`Parent`NF eq psi`Parent`NF:
 "Grossencharacter and Hecke character must be defined over the same field";
 return gr*(psi^(-1)); end intrinsic;

intrinsic '/'(chi::GrpDrchNFElt,gr::GrossenChar) -> GrossenChar {}
 require gr`hecke`Parent`NF eq chi`Parent`NF:
 "Grossencharacter and Dirichlet character must be defined over same field";
 return chi*(gr^(-1)); end intrinsic;

intrinsic '/'(gr::GrossenChar,chi::GrpDrchNFElt) -> GrossenChar {}
 require gr`hecke`Parent`NF eq chi`Parent`NF:
 "Grossencharacter and Dirichlet character must be defined over same field";
 return gr*(chi^(-1)); end intrinsic;

////////////////////////////////////////////////////////////////////////

intrinsic '^'(gr::GrossenChar,n::RngIntElt) -> GrossenChar {}
 GR:=New(GrossenChar);
 GR`dirich:=gr`dirich^n; GR`hecke:=gr`hecke^n; GR`wt:=n*gr`wt;
 GR`zetas:=[* x*n : x in gr`zetas *]; GR`roots:=gr`roots; GR`clreps:=gr`clreps;
 GR`type:=[[n*gr`type[i][1],n*gr`type[i][2]] : i in [1..#gr`type]];
 GR`evalf:=get_eval(GR`zetas,GR`roots,GR`clreps,GR`dirich);
 return GR; end intrinsic;

intrinsic '*'(g1::GrossenChar,g2::GrossenChar) -> GrossenChar {}
 require g1`hecke`Parent`NF eq g2`hecke`Parent`NF:
  "Grossencharacters must be in same field";
 g3:=New(GrossenChar); g3`wt:=g1`wt+g2`wt;
 g3`dirich:=g1`dirich*g2`dirich; g3`hecke:=g1`hecke*g2`hecke;
 TYPE:=[[g1`type[i][1]+g2`type[i][1],
	 g1`type[i][2]+g2`type[i][2]] : i in [1..#g1`type]];
 g3`type:=TYPE; I:=Modulus(g3`hecke); GI:=initialise_grossenideal(I);
 G1:=GrossenCopy(g1); force_clreps_and_roots(G1,GI[2],GI[3],g1`dirich);
 G2:=GrossenCopy(g2); force_clreps_and_roots(G2,GI[2],GI[3],g2`dirich);
 assert G1`roots eq G2`roots; assert G1`roots eq GI[3];
 ZETAS:=[* cyclo_mult(G1`zetas[i],G2`zetas[i]) : i in [1..#G1`zetas] *];
 CLREPS:=GI[2]; ROOTS:=GI[3];
 g3`zetas:=ZETAS; g3`roots:=ROOTS; g3`clreps:=CLREPS;
 g3`evalf:=get_eval(ZETAS,ROOTS,CLREPS,g3`dirich); return g3; end intrinsic;

intrinsic '/'(g1::GrossenChar,g2::GrossenChar) -> GrossenChar {}
 require g1`hecke`Parent`NF eq g2`hecke`Parent`NF:
 "Grossencharacters must be in same field"; return g1*(g2^(-1)); end intrinsic;

intrinsic '/'(e::RngIntElt,gr::GrossenChar) -> GrossenChar {}
 require e eq 1: "Numerator must be 1"; return gr^(-1); end intrinsic;

////////////////////////////////////////////////////////////////////////

intrinsic Restrict(gr::GrossenChar,I::RngOrdIdl) -> GrossenChar
{Restrict a Grossencharacter to the given ideal}
 require gr`hecke`Parent`NF eq NumberField(Order(I)):
  "Grossencharacter and ideal must be over the same number field";
 require Modulus(gr) subset I:
 "The modulus of the grossencharacter must contain the ideal";
 require I subset Conductor(gr):
 "The ideal must be a subset of the conductor of the grossencharacter";
 GR:=GrossenCopy(gr); dr:=DirichletRestriction(gr`hecke);
 GR`dirich:=Restrict(gr`dirich/DirichletRestriction(gr`hecke),I);
 GR`hecke:=Restrict(gr`hecke/gr`hecke,I);
 GR`zetas:=[* cyclo_mult(gr`zetas[i],-'@'(gr`clreps[i],gr`hecke : Raw))
			: i in [1..#gr`clreps] *];
 GR`evalf:=get_eval(GR`zetas,GR`roots,GR`clreps,GR`dirich);
 return GR; end intrinsic;

intrinsic AssociatedPrimitiveGrossencharacter(gr::GrossenChar) -> GrossenChar
{Returned the associated primitive Grossencharacter of a given one}
 if IsPrimitive(gr) then return gr; end if;
 return Restrict(gr,Conductor(gr)); end intrinsic;

intrinsic Weight(psi::GrossenChar) -> RngIntElt
{Given a Hecke Grossencharacter, returns its weight}
 return psi`wt; end intrinsic;

intrinsic CentralCharacter(psi::GrossenChar) -> GrpDrchNFElt
{Given a Hecke Grossencharacter, compute its central Dirichlet
 character down to Q (normalised to have weight 0).}
 n:=Norm(Modulus(psi)); Q:=NumberField(Polynomial([-1,1]) : DoLinearExtension);
 G:=DirichletGroup(n*IntegerRing(Q),[1]); // include oo
 mI:=G`ResidueMap; RI:=Domain(mI); reps:=[mI(RI.i) : i in [1..Ngens(RI)]];
 MO:=MaximalOrder(NumberField(Order(Modulus(psi))));
 X:=<<r,(Integers()!Rationals()!((Q!e)/(Q!r)))*
        cyclo_mult('@'(r*MO,psi`hecke : Raw),z)>
        where e,z:=psi`evalf(r*MO) : r in reps>;
 ch:=DirichletCharacter(G,X); ch:=AssociatedPrimitiveCharacter(ch);
 return sub<Parent(ch)|[ch]>!ch; end intrinsic;

intrinsic GrossenTwist(Y::GrossenChar,D::List) -> Grossenchar
{Given a Grossencharacter, find its twist by Hilbert characters that match
 the given data (if possible)}
 S:=HilbertCharacterSubgroup(Parent(Y`hecke));
 X:=[s : s in Elements(S) | &and[Round(Real((s*Y)(x[1]))) eq x[2] : x in D]];
 require #X le 1: "Data do not uniquely specify a Grossencharacter";
 require #X ge 1: "Data are incompatible"; return X[1]*Y; end intrinsic;
