
freeze;

intrinsic Hom(M::ModDed, N::ModDed) -> ModDed, Map
{The Hom-Module of homomorphisms between M and N};

    require BaseRing(M) eq BaseRing(N) : 
		"Orders of domain and codomain are different";

    PM := PseudoBasis(M);
    PN := PseudoBasis(N);

    S := [PN[i][1]/PM[j][1] : i in [1 .. #PN], j in [1 .. #PM]];

    H := Module(S);

    m := map<H -> PowerStructure(Map) | 
	    x :-> hom<M -> N | 
	    Matrix(FieldOfFractions(BaseRing(N)), Ngens(M), Ngens(N), Eltseq(x))
	    : ModuleBasis := true>
	 >;

    return H, m;

end intrinsic;

//-------------
//
// Miscellaneous algorithms for pseudomatrices.
//
//-------------

/*intrinsic Transpose(P::PMat) -> PMat
  {Returns the transpose of the pseudomatrix.}

  M := Matrix(P);
  I := CoefficientIdeals(P);
  return PseudoMatrix(I, Transpose(M));
end intrinsic;*/

intrinsic Basis(P::PMat) -> SeqEnum
  {Returns the basis of the pseudomatrix.}

  M := Matrix(P);
  return [Eltseq(M[i]) : i in [1..Nrows(M)]];
end intrinsic;

intrinsic Module(P::PMat) -> ModDed
  {Returns the module defined by the rows of the pseudomatrix.}
  return Module([<c[i], m[i]> : i in [1..#c]]) where c := CoefficientIdeals(P) where m := Matrix(P);
end intrinsic;

intrinsic PseudoMatrix(M::ModDed:Generators := false) -> PMat
  {Returns the pseudomatrix used as "basis" for the module.}
  if Generators then
    l := PseudoGenerators(M);
  else
    l := PseudoBasis(M);
  end if;
  return PseudoMatrix([x[1] :x in l], Matrix([x[2] : x in l]));
end intrinsic;

intrinsic '*'(M::PMat, A::RngOrdFracIdl) -> PMat
  {}
  C := CoefficientIdeals(M);
  return PseudoMatrix([x*A : x in C], Matrix(M));
end intrinsic;

intrinsic '*'(A::RngOrdFracIdl, M::PMat) -> PMat
  {}
  C := CoefficientIdeals(M);
  return PseudoMatrix([x*A : x in C], Matrix(M));
end intrinsic;

intrinsic Order(O::RngOrd, M::PMat:Check := true) -> RngOrd
  {The order where the basis is given via the transformation in M.}
  n := Degree(O);
  c := CoefficientIdeals(M);
  if #c eq 0 then
    c := [1*CoefficientRing(O) : i in [1..n]];
  else
    assert #c eq n;
  end if;
  m := Matrix(M);
  assert Nrows(m) eq n;
  M := Module([<c[i], m[i]> : i in [1..n]]);
  return Order(O, M:Check := Check);
end intrinsic;

intrinsic PseudoMatrix(M::Mtrx[RngOrd]) -> PMat
  {The pseudomatrix with trivial ideals and matrix M.}
  C := CoefficientRing(M);
  C := MaximalOrder(C);
  return PseudoMatrix([1*C : i in [1..Nrows(M)]], M);
end intrinsic;

//from Markus
intrinsic Generators(M::ModDed : Minimal:= false) -> []
    {A (minimal) sequence of generators of the module M over its base ring}
    if Minimal then
	MM:= SteinitzForm(M);
	Gens:= [];
	P:= PseudoGenerators(MM); 
	n:= #P;
	for i in [1..n] do
	    p:= P[i];
	    ok, x:= IsPrincipal(p[1]);
	    assert i eq n or ok;
	    if ok then 
		Append(~Gens, p[2] * x); 
	    else 
		Gens cat:= [ g*p[2] : g in Generators(p[1]) ]; 
	    end if;
	end for;
    else
	P:= PseudoBasis(M);
	Gens:= [ g*p[2] : g in Generators(p[1]), p in P ];
    end if;
    return Gens;
end intrinsic;

