freeze;
//

intrinsic IsEven(G::GrpPerm) -> BoolElt
{Tests if G leq  A_n}

  if forall(x){x : x in Generators(G) |IsEven(x)} then
     return true;
  end if;
  return false;
end intrinsic;

intrinsic WreathProduct(B::GSet :  SubAlt:=false, NormalAlt:=false) ->GrpPerm, GrpPerm, GrpPerm
{Compute wreath product S_l wr S_m corresponding to given block system}

  l:=#Representative(B);
  m:=#B;

  if NormalAlt then
     //"Warning: be careful when applying this (sorting of roots!!!)";
     H1:=AlternatingGroup(l);
  else
     H1:=SymmetricGroup(l);
  end if;

  if SubAlt then
     H2:=AlternatingGroup(m);
  else
     H2:=SymmetricGroup(m);
  end if;

  W,ll,f1:=WreathProduct(H1, H2);
  kern:=sub<W|[Image(x):x in ll]>;
  comp:=Image(f1);
  L:=[];
  for a in B do
    for b in a do
      Append(~L,b);
    end for;
  end for;
  g:=SymmetricGroup(m*l) ! L;
  return W^g, kern^g, comp^g;
end intrinsic;

intrinsic InternalWreathProductSm(B::GSet) ->GrpPerm
{invariant under Sm}

  if assigned B`WreathSm then
     return B`WreathSm;
  end if;

  l:=#Representative(B);
  m:=#B;
  W,ker,H:=WreathProduct(B);
  A:=ker meet AlternatingGroup(m*l);

  U:=sub<W|H,A>;

  B`WreathSm:=U;
  return U;
end intrinsic;

intrinsic InternalWreathProductD(B::GSet) ->GrpPerm
{invariant under D}

  if assigned B`WreathD then
     return B`WreathD;
  end if;

  W:=WreathProduct(B:SubAlt:=true);
  B`WreathD:=W;
  return W;
end intrinsic;

intrinsic InternalWreathProductS1(B::GSet) ->GrpPerm
{invariant under S1}

  if assigned B`WreathS1 then
     return B`WreathS1;
  end if;

  W:=WreathProduct(B:NormalAlt:=true);
  B`WreathS1:=W;
  return W;
end intrinsic;

intrinsic InternalSubgroupSum(G::Grp, H1::Grp, H2::Grp) -> Grp
{H1 and H2 are subgroups of index 2 in G. Computes third subgroup of index 2}

  assert H1 ne H2;
  HH:=H1 meet H2;
  cnt := 0;
  repeat
    h1 := Random(G);
    cnt +:= 1;
  until h1 notin H1 and h1 notin H2;
  vprint GaloisGroup, 3: "used ", cnt , " random elements";

  H3 := sub<G|HH, h1>;
  assert #G/#H3 eq 2;
  assert H3 ne H1 and H3 ne H2;
  return H3;

  Q,f:=quo<G|HH>;
  t:=Subgroups(Q:OrderEqual:=2);
  t:=[x`subgroup:x in t];
  t:=[x:x in t| x ne f(H1) and x ne f(H2)];
  assert #t eq 1;
  return t[1] @@f;

end intrinsic;

intrinsic InternalWreathProductDSm(B::GSet) ->GrpPerm
{invariant under DSm}

  if assigned B`WreathDSm then
     return B`WreathDSm;
  end if;


  W:=WreathProduct(B);
  W1:=InternalWreathProductD(B);
  W2:=InternalWreathProductSm(B);
  // W/(W1 meet W2) has index 4. Compute third intermediate group

  B`WreathDSm:=InternalSubgroupSum(W,W1,W2);
  return B`WreathDSm;
end intrinsic;

intrinsic InternalWreathProductS2(B::GSet) ->GrpPerm
{invariant under S2}

  if false and assigned B`WreathS2 then
     return B`WreathS2;
  end if;

  W:=WreathProduct(B);
  W1:=WreathProduct(B:NormalAlt:=true);
  L:=[SetToSequence(x):x in B];
  g:=&*[W ! (x[1],x[2]):x in L]; //is odd for all blocks! 
  U:=sub<W|W1,g>;
  B`WreathS2:=U;

  return U;
end intrinsic;

