freeze;

declare attributes GaloisData : Scale, SubfieldLattice, KnownSubgroup,
                                Bound, Time, Maps, MinPrec;
declare attributes FldNum  : GaloisData;
declare attributes GrpPerm : Blocks, Sub2;
declare attributes GSet: WreathS1, WreathS2, WreathD,  WreathD,
                              WreathSm, WreathDSm, RelG, SubG;

import "GalInvar.m" : SLPRGg;

//currently not really used. It determines if the function check 
//computes the multivariate polynomial for invariance checks or just tries to
//evaluate at random points...
cheat := true;
//max_cost is the cut-off point for the old generic invariant computation.
//If more than max_cost monomials have to be considered, 
//we use the new method.
max_cost := 250000;
//max_cost := 0000;

// DEFAULT_MAX_PREC is the cut off for precision: if we need more
// than DEFAULT_MAX_PREC, then we assume the decent fails...
DEFAULT_MAX_PREC := Infinity();

//triggers the use of factorisation (over Z_p) for the computation of roots.
UseFactor :=  not true;
//decides if adaptive invariants or classical invariants are used
Classical := false;
//decides how expensive the inner invariant in BlockQuotients is allowed to be
MaxInnerInvar := Infinity();
//cut-off for ShortCosets
MaxShortCosets := 50;

// for colorized output...(changes the background color)
if true then
  red := "[41m";
  green := "[42m";
  RED := "[43m";
  blue := "[44m";
  purple := "[45m";
  BLUE := "[46m";
  grey := "[47m";
  normal := "[49m";
else
  red := "";
  green := "";
  RED := "";
  blue := "";
  purple := "";
  BLUE := "";
  grey := "";
  normal := "";
end if;

color_array := [red, green, RED, blue, purple, BLUE, grey, normal];

function _MaxSub_(X:IsTransitive := false)
  vprint GaloisGroup, 3: "Calling maximal subgroups on", X:Magma;
  return MaximalSubgroups(X: IsTransitive := IsTransitive);
end function;

function shallow_copy(G)
  return sub<Sym(Degree(G))|GeneratorsSequence(G)>;
end function;

AddAttribute(RngInt, "UseCache");
AddAttribute(RngInt, "CacheLevel");
AddAttribute(RngInt, "Invar");
AddAttribute(RngInt, "Group");

AddAttribute(RngInt, "Error");
intrinsic PutInZ(A) -> RngIntElt
  {}
  Z := Integers();
  if assigned Z`Error then
    Append(~Z`Error, A);
  else
    Z`Error := [*A*];
  end if;
  return #Z`Error;
end intrinsic;

_Cache:= Random(2^30);
_Z := Integers();
__Cache := ideal<_Z|_Cache>;

function _GetGaloisCacheLevel()
  return ideal<_Z|_Cache>`CacheLevel;
end function;

procedure _SetGaloisCacheLevel(l)
  Cache := ideal<_Z|_Cache>;
  Cache`CacheLevel := l;
  if l gt 0 then
    Cache`UseCache := true;
  else
    Cache`UseCache := false;
  end if;
end procedure;

function UseCache() 
  Cache := ideal<_Z|_Cache>;
  return Cache`UseCache;
end function;

_SetGaloisCacheLevel(0);

inv_fmt := recformat<G:GrpPerm, H:GrpPerm,
                     I:UserProgram,
                     c:RngIntElt>;  
ci_fmt := recformat<Hashes, invs>;

function CacheHasInvar(G,H)
  Cache := ideal<_Z|_Cache>;
  if not assigned Cache`Invar then
    return false, _;
  else
    ci := Cache`Invar;
  end if;
  
  if <G,H> in ci`Hashes then
    p := Position(ci`Hashes, <G, H>);
    assert ci`invs[p]`G eq G and ci`invs[p]`H eq H;
    ci`invs[p]`c +:=1;
    Cache`Invar := ci;
    return true, ci`invs[p]`I;
  end if;
  return false, _;
end function;

procedure CachePutInvar(G, H, I)
  Cache := ideal<_Z|_Cache>;
  if assigned Cache`Invar then
    ci := Cache`Invar;
  else
    ci := rec<ci_fmt|Hashes := {@@}, invs := []>;
  end if;
  if <G, H> in ci`Hashes then
    p := Position(ci`Hashes, <G, H>);
    assert ci`invs[p]`G eq G and ci`invs[p]`H eq H;
    ci`invs[p]`I := I;
  else
    Include(~ci`Hashes, <G, H>);
    Append(~ci`invs, rec<inv_fmt| G := G, H := H, I := I, c := 0>);
  end if;
  Cache`Invar := ci;
  return;
end procedure;

forward GaloisRootAddPermutation;
procedure CacheNormalise(~G, S)
  if Degree(G) le 23 then
    H := TransitiveGroup(a,b) where b,a := TransitiveGroupIdentification(G);
    fl, p := MyIsConjugate(Sym(Degree(G)), H, G);
    assert H^p eq G;
    GaloisRootAddPermutation(S, p);
    G := H;
  end if;
end procedure;


grp_fmt := recformat<G:GrpPerm, S,
                     c:RngIntElt>;  
gi_fmt := recformat<Hashes, grps>;


function CacheHasGroup(G)
  Cache := ideal<_Z|_Cache>;
  if not assigned Cache`Group then
    return false, _;
  else
    ci := Cache`Group;
  end if;
  
  if G in ci`Hashes then
    p := Position(ci`Hashes, G);
    assert ci`grps[p]`G eq G;
    ci`grps[p]`c +:=1;
    Cache`Group := ci;
    return true, ci`grps[p]`S;
  end if;
  return false, _;
end function;

procedure CachePutGroup(G, S)
  Cache := ideal<_Z|_Cache>;
  if assigned Cache`Group then
    ci := Cache`Group;
  else
    ci := rec<gi_fmt|Hashes := {@@}, grps := []>;
  end if;
  if G in ci`Hashes then
    error "should not happen";
  else
    Include(~ci`Hashes, G);
    Append(~ci`grps, rec<grp_fmt| G := G, S := S, c := 0>);
  end if;
  Cache`Group := ci;
  "added group";
  return;
end procedure;

function TestAllPartitions(G)
  local   L, par, par1, f, Im, ker;

  L:=AllPartitions(G);
  L:=[MinimalPartition(G:Block:=x): x in L];
  for par in L do
    f, Im, ker:= BlocksAction(G, par);
    par1:=Random(Set(par));
    sub:=OrbitImage(Stabilizer(G,par1),par1);
    if Degree(Im) le 23 and Degree(sub) le 23 then
      vprint GaloisGroup, 2: "Subfield: ", TransitiveGroupIdentification(Im),
                             "rel : ", TransitiveGroupIdentification(sub);
    end if;
  end for;
  return true;
end function;

intrinsic WeakValuation(x::RngSerLaurElt[FldPad]) -> RngIntElt
  {}
  a,b,c := Eltseq(x);
  i := 1;
  for i := 1 to #a do
    if not IsWeaklyZero(a[i]) then
      return b+i-1, Minimum([Valuation(a[x]) : x in [1..i]] cat [0]);;
    end if;
  end for;
  return b+#a, Minimum([Valuation(a[x]) : x in [1..#a]]);
end intrinsic;


intrinsic WeakValuation(x::RngSerPowElt[FldPad]) -> RngIntElt
  {}
  a,b,c := Eltseq(x);
  i := 1;
  for i := 1 to #a do
    if not IsWeaklyZero(a[i]) then
      return b+i-1, Minimum([Valuation(a[x]) : x in [1..i]] cat [0]);;
    end if;
  end for;
  return b+#a, Minimum([Valuation(a[x]) : x in [1..#a]]);
end intrinsic;

intrinsic WeakValuation(x::FldReElt) -> RngIntElt
  {}
  if IsWeaklyZero(x) then
    return Infinity();
  end if;
  return -Log(Abs(x));
end intrinsic;
intrinsic WeakValuation(x::FldComElt) -> RngIntElt
  {}
  if IsWeaklyZero(x) then
    return Infinity();
  end if;
  return -Log(Abs(x));
end intrinsic;
 
intrinsic WeakValuation(x::RngElt) -> RngIntElt
  {}
  if IsWeaklyZero(x) then
    return Infinity();
  end if;
  return Valuation(x);
end intrinsic;
    

intrinsic IsInt(x::RngElt, B::RngElt, S::GaloisData:EasyNonInt := false) -> BoolElt, RngIntElt
  {}
  if Type(x) eq RngIntElt then
    error "the int here implies that block-quotient found a wrong invariant";
    return false, _;
  end if;

  if assigned S`IsInt then
    return S`IsInt(x, B, S:EasyNonInt := EasyNonInt);
  elif S`Type eq "Complex" then
     f := Round(Re(x));
    return (Re(x) - f)^2 + Im(x)^2 lt 10^Round(-S`Prec*0.8), f;
  elif S`Type eq "p-Adic" then
    if Precision(Parent(x)) cmpne Infinity() then
      error "wrong ring";
    end if;
    f, y := IsCoercible(S`Base, x);
    if f then
      if EasyNonInt then
        return true;
      end if;
      x := y;
      f := Integers()!x;
      p := Precision(x);
      p := Characteristic(ResidueClassField(S`Base))^p;
//      f := Integers(p)!f;
//      return RationalReconstruction(f);
      if f gt p/2 then
        f := f-p;
      end if;
      fl :=  Abs(f) le B;
      if fl then
        vprint GaloisGroup, 4:  "INT: ", x, "(", f, ")";
      else
        vprint GaloisGroup, 4: "no int - size error";
      end if;
      return fl, f;
    else
      vprint GaloisGroup, 4: "no int - ring error";
      return false, _;
    end if;
  else
    error "not implemented yet";
  end if;
end intrinsic;

function InternalHenselLift(f, x, k, R)
  fx := Evaluate(f, x);
  fs := Derivative(f);
  fss := Derivative(fs);
  assert Abs(fx/(Evaluate(fs, x))^2*Evaluate(fss, x)) lt 1/2;
  //according to
  // http://numbers.computation.free.fr/Constants/Algorithms/newton.html
  // there is a thm that guarantees convergence under this assumption
  if Abs(fx) ge 1/2 then
    p := Precision(x);
    p := 2*Ceiling(Log(Abs(fx)));
   elif IsWeaklyZero(fx)  then
    p := Precision(x);
  else
    p := Ceiling(-Log(Abs(fx))/Log(10))+1;
  end if;
  kw := Ceiling(k*1.5);
  while p lt kw do
    p := Minimum(2*p, kw);
    C := R(Ceiling(p*1.5 + 5));
    delta :=
        Evaluate(Polynomial(C, f), C!x)/Evaluate(Polynomial(C, Derivative(f)), C!x);
    //Ceiling(-Log(Abs(delta)));
    // it would be nice if we could assume that delta is getting smaller...    
    x := C!x - delta;
    if Abs(delta) lt 1 then
      p := 2*Ceiling(-Log(Abs(delta))/Log(10)) + Maximum(1, Ceiling(Log(1+Abs(x))))
           + Ceiling(Log(10+Abs(fx)))
           + Ilog2(Maximum([Abs(x) : x in Eltseq(f)]))
           + Degree(f);  
    end if;
  end while;
  return R(k)!x;
end function;

intrinsic HenselLift(f::RngUPolElt, x::FldReElt, k::RngIntElt) -> FldComElt
  {Lifts the real root x of f to a precision of k}
  return InternalHenselLift(f, x, k, RealField);
end intrinsic;

intrinsic HenselLift(f::RngUPolElt, x::FldComElt, k::RngIntElt) -> FldComElt
  {Lifts the complex root x of f to a precision of k}
  return InternalHenselLift(f, x, k, ComplexField);
end intrinsic;
    

procedure GaloisRootAddPermutation(S, p)
  if assigned S`Permutation then
    S`Permutation := p*S`Permutation;
  else
    S`Permutation := p;
  end if;
end procedure;

intrinsic GaloisRoot(f::RngUPolElt, n::RngIntElt, S::GaloisData:Prec := 20, Bound := false) -> RngElt
  {Returns the n-th root of f in the ring defined somehow by S with a precision of Prec}
  assert S`f eq f;
  return GaloisRoot(n, S:Prec := Prec, Bound := Bound);
end intrinsic;

function PrecisionFromBound(Bound, S)
    if assigned S`GetPrec then
      Prec := S`GetPrec(Bound, S);
    elif S`Type eq "Complex" then
      Prec := Floor((Log(Bound)+20)/Log(10)*1.5);
    elif S`Type eq "p-Adic" then
      if Bound eq 0 then 
        Bound := 1;
      end if;
      Prec := Floor(Log(Bound)/Log(Characteristic(ResidueClassField(S`Base)))+5);
    else
      error "unsupported ring type...";
    end if;
    vprint GaloisGroup, 3: "GaloisRoot: using a prec of ", Prec, 
                           "as computed from the bound ", Bound;
    return Prec;
end function;

intrinsic GaloisRoot(n::RngIntElt, S::GaloisData:Prec := 20, Bound := false) -> RngElt
  {Returns the n-th root of f in the ring defined somehow by S with a precision of Prec}

  if assigned S`Permutation then
    n := n^(S`Permutation^1);
  end if;
  if Bound cmpne false then
    Prec := PrecisionFromBound(Bound, S);
  end if;
  if assigned S`MinPrec then
    Prec := Maximum(S`MinPrec, Prec);
  end if;
  if assigned S`Roots then
    if Precision(S`Roots[n]) ge Prec then
      r := S`Roots[n];
      if assigned S`Scale and S`Scale ne 1 then
        if assigned S`BaseMap then
          r *:= S`BaseMap(S`Scale, Prec);
        else
          r *:= S`Scale;
        end if;
      end if;
      r := ChangePrecision(r, Prec);
      return r;
    end if; 
    if not assigned S`f or assigned S`GetRoots then
      assert assigned S`GetRoots;
      old := S`Roots;
      S`GetRoots(S:Prec := Prec);
      assert #S`Roots eq #old;
      r := S`Roots[n];
      assert Precision(r) ge Prec;
      if assigned S`Scale then
        if assigned S`BaseMap then
          r *:= S`BaseMap(S`Scale, Prec);
        else
          r *:= S`Scale;
        end if;
      end if;
      assert Precision(r) ge Prec;
      r := ChangePrecision(r, Prec);
      assert Precision(r) ge Prec;
      return r;
    end if;
    f := S`f;
    if S`Type eq "p-Adic" or S`Type in {"p-Adic (FldFun over Q)", "p-Adic (FldFun over Fq)"} then
      if assigned S`BaseMap then
        g := Polynomial([S`BaseMap(x, Prec+Prec) : x in Eltseq(f)]);
	R := S`Ring;
        g := Polynomial(R, g);
      else
        R := ChangePrecision(S`Ring, Prec+Prec); // too crude - but Prec itself
              // may be too low if f is not squarefree mod p
              // which can happen in the reducible polynomial case (coming
              // from SolveByRadicals: f := x^4 - x^3 - 7*x^2 + 2*x + 9, 
              //                   Prime := 7
              // f*CyclotomicPolynomial(3) is mod 7 not squarefree.
        g := Polynomial(R, f);
      end if;
      vprint GaloisGroup, 4: "Lifting roots in ", R, 
                     " to precision ", Prec, " starting with ", S`Roots;
      if assigned S`Time then
        rt := Cputime();
      end if;
      if UseFactor and (not assigned S`Scale or S`Scale eq 1) then
        C := ResidueClassField(BaseRing(S`Ring));
        vprint GaloisGroup, 4: "Factorising over the ResidueClassField";
        vtime GaloisGroup, 4: lR := Factorisation(Polynomial(C, f));
        if exists{x : x in lR | x[2] ne 1} then
          vprint GaloisGroup, 3: "Polynomial not square-free over the residue class field";
        end if;
        RR := Integers(BaseRing(R));
        vprint GaloisGroup, 4: "Lifting the factorisation";
        if #lR gt 1 and forall{x : x in lR | x[2] eq 1} then
          vtime GaloisGroup, 4:
          lf := HenselLift(Polynomial(RR, f), [Polynomial(RR, x[1]) : x in lR]);
        else
          lf:=[Polynomial(RR,f)];
          lR := [<Polynomial(C, f), 1>];
        end if;//trivial case, only one factor
        lf := [ Polynomial(R, x) : x in lf];
        C := ResidueClassField(S`Ring);
        lR := [Polynomial(C, x[1]) : x in lR];
        vprint GaloisGroup, 4: "Lifting roots";
        vtime GaloisGroup, 4: S`Roots := [S`Ring |
            HenselLift(lf[[n : n in [1..#lR]|Evaluate(lR[n], C!x) eq 0][1]],
                                   R!x, Prec) : x in S`Roots];
	assert #S`Roots eq #Set(S`Roots);
      else
        vprint GaloisGroup, 4: "Lifting roots...";
        if S`Type eq "p-Adic" then
          S`Roots := ChangeUniverse(S`Roots, R);
        end if;
//"This is how we get the roots";
        vtime GaloisGroup, 4: 
          S`Roots := [S`Ring| HenselLift(g, x, Prec) : x in S`Roots];
        assert forall{x : x in S`Roots | Precision(x) ge Prec};
	if Type(Prec) eq RngIntElt then
	    assert &and [Valuation(Evaluate(g, r)) ge Prec : r in S`Roots];
	end if;
	// reducible poly with multiple roots will not satisfy this
	// assert #Set(S`Roots) le #S`Roots;
      end if;
      if assigned S`Time then
        S`Time`Lift +:= Cputime(rt);
      end if;
      if Type(Prec) eq Tup then
        S`Ring`DefaultPrecision := Maximum(Prec[1], S`Ring`DefaultPrecision);
      else
        S`Ring`DefaultPrecision := Maximum(Prec, S`Ring`DefaultPrecision);
      end if;
    else
      if assigned S`BaseMap then
        g := Polynomial([S`BaseMap(x, Prec) : x in Eltseq(f)]);
	S`Roots := [ HenselLift(g, x, Prec) : x in S`Roots];
      else
	  S`Roots := [ HenselLift(f, x, Prec) : x in S`Roots];
      end if;
      assert forall{x : x in S`Roots | Precision(x) ge Prec};
      assert #Set(S`Roots) eq #S`Roots;
    end if;
    r := S`Roots[n];
    if assigned S`Scale then
      if assigned S`BaseMap then
        r *:= S`BaseMap(S`Scale, Prec);
      else
        r *:= S`Scale;
      end if;
    end if;
    r := ChangePrecision(r, Prec);
    return r;
  end if;
  if not assigned S`f then
    S`GetRoots(S:Prec := Prec);
    r := S`Roots[n];
    if assigned S`Scale then
      if assigned S`BaseMap then
        r *:= S`BaseMap(S`Scale, Prec);
      else
        r *:= S`Scale;
      end if;
    end if;
    r := ChangePrecision(r, Prec);
    return r;
  end if;
  f := S`f;
  if S`Type eq "Complex" then
    pPrec := Maximum(40, Prec);
    S`Roots := [x[1] : x in Roots(Polynomial(ComplexField(pPrec), f))];
    S`Ring := ComplexField(pPrec);
    r := ChangePrecision(S`Roots[n], Prec);
    if assigned S`Scale then
      r *:= S`Scale;
    end if;
    S`Prec := pPrec;
    return r;
  elif S`Type eq "p-Adic" then
    if assigned S`Time then
      rt := Cputime();
    end if;
    f := Polynomial(Integers(), f*Lcm([Denominator(x) : x in Eltseq(f)]));
    if not assigned S`Ring then
        vprint GaloisGroup, 3: "Finding splitting field";
	vtime GaloisGroup, 3: R := SplittingField(f, S`Base);
	S`Ring := R;
    else
	R := S`Ring;
    end if;
    vprint GaloisGroup, 3: "Computing initial roots";
    vtime GaloisGroup, 3: S`Roots := [ x[1] : x in Roots(Polynomial(R, f))];
    if assigned S`Time then
      S`Time`Lift +:= Cputime(rt);
    end if;
    S`Ring`DefaultPrecision := Maximum(Prec, S`Ring`DefaultPrecision);
    r := ChangePrecision(S`Roots[n], Prec);
    if assigned S`Scale then
      if assigned S`BaseMap then
        r *:= S`BaseMap(S`Scale, Prec);
      else
        r *:= S`Scale;
      end if;
    end if;
    return r;
  else
    error "not implemented yet";
  end if;
end intrinsic;

function My_np(x)
  return NextPrime(x);
end function;

intr_subfields := Subfields;

function PrimeChoose(K, S:Subfields := true, Current := false, 
                          Prime := false, SmallDeg := false,
                          NextPrime := My_np)
//Choose a prime which does not divide a discriminant 
// of any subfield polynomial...
//
//Additionally inertia small when ShortCosets is not used...
//
//Otherwise do not choose the identity (still looking for good heuristics)
//
//returns CycleStructures of group elements...

  if GetVerbose("Subfields") eq 0 then
    SetVerbose("Subfields", 0); 
    // Problem is that SetVerbose(Subfields) and SetVerbose(GaloisGroup)
    // have a non-empty intersection, but at least for now, I'm not interested
    // in the printing from c.
  end if;
  vprint GaloisGroup, 2: "PrimeChoose started....";
  if Subfields then
    vprint GaloisGroup, 2: "computing Subfields ...";
    if S cmpne false and assigned S`Subfields then
      sub := S`Subfields;
    else
      // _, Subfields := IsIntrinsic("Subfields"); 
      Subfields := intr_subfields;
      vtime GaloisGroup, 2: sub:=Subfields(K:Current := Current);
      if S cmpne false then
        S`Subfields := sub;
      end if;
    end if;
    d:=Lcm([Integers() | d*Denominator(d)^2
      where d := Discriminant(DefiningPolynomial(x[1])): x in sub]);
  else
    d := 1;
    sub := [];
  end if;
  f:=DefiningPolynomial(K);
  f := Lcm([Denominator(x) : x in Eltseq(f)])*f;
  f := Polynomial(Integers(), f);
  d:=d*LeadingCoefficient(f);  

  if Prime cmpeq false then
    p:=NextPrime(Degree(K));//Small primes may lead to problems
  else
    p := Prime;
  end if;
  cands:=[];
  shape:=[];
  cycleshape:={};
  if SmallDeg then
    MaxCand := Maximum(20*Degree(K), 20);
  else
    MaxCand := Maximum(5*Degree(K), 20);
  end if;
  b := 8;
  no_b := 0;
  while #cands lt MaxCand do
    if GetVerbose("GaloisGroup") ge 2 and #cands mod 50 eq 0 then
      "So far, ", #cands, "primes have been considered.";
      if #shape eq 0 then
        "No useable prime so far, prime: ", p;
      else
        "Best degree is", Minimum(shape), "prime: ", p;
      end if;
    end if;
    if d mod p ne 0 or Prime cmpne false then
       no_b +:= 1;
       if no_b gt 100 then
         b +:= 1;
         no_b := 0;
       end if;
       L := Polynomial(GF(p), f);
       if not IsSquarefree(L) then
         p:=NextPrime(p);
         continue;
       end if;
       L:=Factorization(L);
       assert forall(x) {x : x in L | x[2] eq 1};
       L:=[Degree(x[1]):x in L];
       if Lcm(L) gt Degree(f) div b or Prime cmpne false or SmallDeg then
         Append(~cands,p);
         Append(~shape,Lcm(L));// Ignore cycle structure for a moment
       end if;
       Include(~cycleshape, L);
    end if;
    if Prime cmpne false then
      break;
    end if;
    if SmallDeg and #shape ge 1 and Minimum(shape) eq 1 then
      break;
    end if;
    p:=NextPrime(p);
  end while;
  if #sub gt 1 or SmallDeg then
     m,pos:=Minimum(shape);
     p:=cands[pos];
  else //primitive case
     m,pos:=Maximum(shape); //not the optimal strategy...
     p:=cands[pos];
  end if;
  L:=Factorization(PolynomialRing(GF(p)) ! f);
  vprint GaloisGroup, 1: "Choose p= ",p," of type :",[Degree(x[1]):x in L];
  return p, cycleshape;
end function;

function IdentifyBlockSystem(K, L, S : UseGalois:=false, ShortOK := false) 
//intrinsic IdentifyBlockSystem(K::FldNum, L::Tup, S::GaloisData 
//     : UseGalois:=false, ShortOK := false) 
//                                   ->GrpPermElt,GrpPerm
//{Identifies the block system of a subfield acting on roots mod p}
//tup is the output of an entry of Subfields
// (almost): tup[2] is the image of the primitive element of the
// subfield in the LARGE field.

//   if S`Type eq "Complex" then error "Not implemented yet";end if;

   prec := 1;
   prec_cnt := 0;
   if S`Type eq "Complex" then
     prec_cnt_max := 20;
     R := GetDefaultRealField();
     if assigned S`MinPrec then
       prec := S`MinPrec;
     else
       prec := S`Prec;
     end if;
   else
     prec_cnt_max := 10;
   end if;
   if S`Type eq "p-Adic (FldFun over Q)" then
     prec := <prec, 2*prec>;
   end if;
     function val(x)
       if S`Type eq "p-Adic (FldFun over Q)" then
         v, vv := WeakValuation(x);
         v := <v, vv>;
       elif S`Type eq "Complex" then
         if IsWeaklyZero(x) then return Infinity(); end if;
         v := -Log(Abs(x))/Log(10);
       else
         v := Valuation(x);
       end if;  
       return v;
     end function;
     function cmp(a,b)
       P := Parent(a);
       if Type(P) cmpeq FldRe then
         return a-b;
       elif Type(P) eq FldCom then
         if Re(a) eq Re(b) then
           return Im(a)- Im(b);
         else
           return Re(a) - Re(b);
         end if;
       elif Type(P) eq FldFin then
	return Hash(a) - Hash(b);
       elif BaseRing(P) cmpeq P then
         return Rationals()!a - Rationals()!b;
       else
	 cv := Valuation(a) - Valuation(b);
	 if cv ne 0 then
	    return cv;
	 end if;
         a := Eltseq(a);
         b := Eltseq(b);
	 cv := #a - #b;
	 if cv ne 0 then
	    return cv;
	 end if;
         i := 1;
         while i lt #a and i lt #b and cmp(a[i], b[i]) eq 0 do
           i +:= 1;
         end while;
         if i gt #a or i gt #b then
           return 0;
         else 
           return cmp(a[i], b[i]);
         end if;
       end if;
     end function;

   repeat
     if S`Type eq "Complex" then
       SetDefaultRealField(RealField(prec));
     end if;
     r := GaloisRoot(1, S:Prec := prec);  //re-computes all roots to Prec 
                                          //at least prec.
     r := [ChangePrecision(x, prec) : x in S`Roots];
     if assigned S`Permutation then
       r := PermuteSequence(r, S`Permutation);
     end if;
     /*CF: Note, we cannot use GaloisRoot here as we need to map
      * field elements, thus polynomials in a proper root of S`f, not
      * in some multiple.
      * In all (most) other parts, we need to have integral numbers,
      * so the scaling is correct, necessary and important.
      */ 
     R := Universe(r);
     F := FieldOfFractions(R);
     r := [F!ChangePrecision(x,prec): x in r];
     // I guess prec 1 is enough for
     // unramified, for ramified, it has to be larger...
     elt:=L[2];
     l:=ElementToSequence(elt);
     //p:=Characteristic(ResidueClassField(S`Base));  
     if Type(Universe(l)) eq FldFun then
	 d := Lcm([Denominator(x, EquationOrderFinite(Universe(l))) : x in l]);
     else
	 d := Lcm([Denominator(x) : x in l]);
     end if;
     den := d;
     if assigned S`BaseMap then
       d := S`BaseMap(d, prec);
     else
       d := R!d;
     end if;
     v := val(d);
     if v ge prec then
       prec := prec+prec;
       prec_cnt +:= 1;
       vprint GaloisGroup, 2: "DenVal error: Increasing precision to ", prec;
       if prec_cnt ge prec_cnt_max then
S`f;
         error "too much(0)";
       end if;
       continue;
     end if;
     if assigned S`BaseMap then
       l := [ S`BaseMap(x*den, prec) : x in l];
       l := [x/d : x in l];
     end if;
     h:= Polynomial(l);
     
     Z := Integers();
     rr := [Evaluate(h, x) : x in r];
     if exists{x : x in rr | val(x) ge prec} then
       prec := prec + prec;
       prec_cnt +:= 1;
       vprint GaloisGroup, 2: "Eval error(1): Increasing precision to ", prec;
       if prec_cnt ge prec_cnt_max then
S`f;
         error "too much(1)";
       end if;
       continue;
     end if;

     if ISA(Type(K), FldNum) or Characteristic(K) gt 0 then
       new_pr := Ceiling(Minimum([Precision(x)-val(x) : x in rr]));
       if new_pr lt 1 then
         prec := prec+prec;
         prec_cnt +:=1;
         vprint GaloisGroup, 2: "Eval error(4): Increasing precision to ", prec;
         if prec_cnt ge prec_cnt_max then
S`f;
           error "too much(4)";
         end if;
         continue;
       end if;
       //TODO: fix precision things!!
     else
       new_pr := <prec[1], prec[1]>;
     end if;

     rr := [ ChangePrecision(y, new_pr) : y in rr];
     if S`Type eq "Complex" then
       eps := 10^-(new_pr div 4 + 1);
       for i in [1..#rr] do
         for j in [i+1..#rr] do
           if Abs(rr[i]-rr[j]) lt eps then
             rr[j] := rr[i];
           end if;
         end for;
       end for;  
     end if;
     if #Set(rr) ne Degree(L[1]) then
       prec := prec+prec;
       prec_cnt +:= 1;
       vprint GaloisGroup, 2: "Eval error(2): Increasing precision to ", prec;
       if prec_cnt ge prec_cnt_max then
#Set(rr);
Degree(L[1]);
prec, prec_cnt, prec_cnt_max;
S`f;
         error "too much(2)";
       end if;
       continue;
     end if;
     g := DefiningPolynomial(L[1]);
     if assigned S`BaseMap then
       g := Polynomial([S`BaseMap(x, prec) : x in Eltseq(g)]);
     end if;
     gs := Derivative(g);
     assert S`Type eq "Complex" or forall{x : x in Set(rr) | IsWeaklyZero(Evaluate(g, x))};
     if exists{x : x in Set(rr) | val(Evaluate(g,x)) le (v+v where v := val(Evaluate(gs, x)))} then
       prec := prec+prec;
       prec_cnt +:= 1;
       vprint GaloisGroup, 2: "Eval error(2.5): Increasing precision to ", prec;
       if prec_cnt ge prec_cnt_max then
S`f;
         error "too much(2.5)";
       end if;
       continue;
     end if;
     rr_srt, perm := Sort(rr, cmp);
     if UseGalois and Degree(L[1]) gt 3 then 
        SS:=InternalGaloisDataCreate(S`Prime);
        for i in GetAttributes(MakeType("GaloisData")) do
          if assigned S``i then
            SS``i := S``i;
          end if;
        end for;
        SS`Type:=S`Type;
        /*SS`Base:=S`Base;
        SS`Ring:=S`Ring;
        SS`Subfields := S`Subfields;
        SS`SubfieldLattice := S`SubfieldLattice;
        */
        delete SS`KnownSubgroup;
        delete SS`CycleTypes;
        delete SS`Scale;
        if ISA(Type(K), FldFun) then
          delete SS`Subfields;
          delete SS`SubfieldLattice;
          SS`f := DefiningPolynomial(L[1]);
          delete SS`max_comp;
          delete SS`GetRoots;
	  if Characteristic(K) eq 0 then
	      _ := InternalGalQt_ComputeRoots(SS`f, SS`Prime, <5, 5>:S := SS, pAdic := false);
	  else 
	      _ := InternalGalFqt_setup(Polynomial(FieldOfFractions(CoefficientRing(SS`f)), SS`f), SS`Prime, 5:S := SS);
	  end if;
          delete SS`GetRoots;
	  assert #Set(SS`Roots) eq #SS`Roots;
        end if;
        m:=Degree(L[1]);
        n:=Degree(K);
        d:= n div m;
        l:=[1+i*d: i in [0..m-1]];
        q := [ R!Evaluate(h, x) : x in S`Roots];
        q := PermuteSequence(q, perm);
//q;
        if SS`Type eq "Complex" then
          SS`Roots:=[ q[i]: i in l]; 
        else
          SS`Roots:=[ SS`Ring| q[i]: i in l]; 
        end if;
/*
m, n, d, l, q, h;
SS`Roots;
SS`f;
S`Roots;
S`f;
*/
	assert #Set(SS`Roots) eq #SS`Roots;
        SS`Prec:=Minimum([Precision(x) : x in SS`Roots]);
        vprint GaloisGroup, 2: "Setting root precision to ", SS`Prec;
        ff := DefiningPolynomial(L[1]);
        if assigned S`BaseMap then
          SS`BaseMap := S`BaseMap;
          ff := Polynomial([S`BaseMap(x, prec) : x in Eltseq(ff)]);
//SS`Prec;
//prec;
//SS`Roots;
//[Valuation(Evaluate(ff, r)) : r in SS`Roots];
        end if;

        if exists{x : x in SS`Roots | WeakValuation(Evaluate(ff, x)) lt 0} then
          assert SS`Type eq "Complex";
          prec := prec+prec;
          prec_cnt +:= 1;
          vprint GaloisGroup, 2: "Newton error: Increasing precision to ", prec;
          continue;
        end if;

        ffs := Derivative(ff);
        if exists(x){ x : x in SS`Roots | 
            WeakValuation(Evaluate(ff, x)) le 2*WeakValuation(Evaluate(ffs, x))} then
          prec := prec+prec;
          prec_cnt +:= 1;
          vprint GaloisGroup, 2: "Newton error: Increasing precision to ", prec;
          continue;
        end if;
        vprint GaloisGroup, 1: RED, "Recurse...", normal;
        IndentPush();
        H,ss, SS:=GaloisGroup(L[1]:Ring:=SS, ShortOK := ShortOK);
        assert ISA(Type(K), FldFun) or forall{x : x in [1..#l] | 
          val(q[l[x]] - SS`Roots[x]) ge 0.25*Minimum(Precision(q[l[x]]), SS`Prec)};
        if assigned SS`Permutation then
          H := H^SS`Permutation^1;
        end if;
        vprint GaloisGroup, 1: "done, group size is ", #H;
        IndentPop();
        delete SS`Subfields;
        delete SS`SubfieldLattice;
        break;
     else
       if Characteristic(K) ne 2 and IsSquare(Discriminant(DefiningPolynomial(L[1]))) then
         H:=AlternatingGroup(Degree(L[1]));
       else
         H:=SymmetricGroup(Degree(L[1]));
       end if;
       break;
     end if;
   until false;

   return perm, H;
end function;

function GaloisStartingGroup(K, S: Current := false, ShortOK := false, Subfields := true) 
//intrinsic GaloisStartingGroup(K::FldNum, S::GaloisData: 
//               Subfields := true, ShortOK := false) 
//               -> GrpPerm
//{Computes subgroup G of S_n such that Gal(f) \leq G}
//Now it also returns rejected subgroups, i.e. 
//normal subgroups which do not contain Gal(f)
//returns know automorphisms...

  if not Subfields then
    return Sym(Degree(K)),[],[];
  end if;

  // _, Subfields := IsIntrinsic("Subfields");
  Subfields := intr_subfields;

  if assigned S`Subfields then
    L := S`Subfields;
    if assigned S`SubfieldLattice then
      LL := S`SubfieldLattice;
    else
      LL := InternalGaloisSubfieldLattice(K:Current := Current);
      S`SubfieldLattice := LL;
    end if;
    T := InternalGaloisSubfieldLatticeTop(LL);
    T := InternalGaloisSubfieldLatticeGetField(LL, T);
    KL := InternalGaloisSubfieldLatticeFindField(LL, K);
    L := [x : x in L | InternalGaloisSubfieldLatticeSubset(LL, InternalGaloisSubfieldLatticeFindField(LL, x[1]), KL)];
    L := [<x[1], K!T!x[1].1> : x in L];
  else
    if ISA(Type(K), FldFun) then
      // Hope to remove this when I am sure of the new subfields algo for char 0
      if Characteristic(K) eq 0 then
	  LL, L := InternalGaloisSubfieldLattice(K, S:Current := Current);
      else
	  LL, L := GeneratingSubfieldsLattice(K, S);
      end if;
    else
      LL, L := InternalGaloisSubfieldLattice(K:Current := Current);
    end if;
    S`Subfields := L;
    S`SubfieldLattice := LL;
    L := [<x[1], K!x[2](x[1].1)> : x in L];
    T := K;
  end if;

  n:=Degree(K);
  df:=Discriminant(DefiningPolynomial(K));
  is_square_df:= Characteristic(K) ne 2 and IsSquare(df);
  if is_square_df then
     G:=AlternatingGroup(n);
  else
     G:=SymmetricGroup(n);
  end if;
  //remove trivial subfields:
  L:=[sub: sub in L |Degree(sub[1]) lt n and Degree(sub[1]) gt 1];
  if #L eq 0 then //primitive extension
     return G, [], [];
  end if;

  aut:=[];
  ML := InternalGaloisSubfieldLatticeMaximalSubfields(LL, 
                             InternalGaloisSubfieldLatticeFindField(LL, K));
  for s in ML do
    sub := <Domain(L), K!L(Domain(L).1)> where L := InternalGaloisSubfieldLatticeGetFieldMap(LL, s);
    if sub[1] eq K then
      continue;
    end if;
    m:=Degree(sub[1]);
    d:=n div m;
    perm,H:=IdentifyBlockSystem(K,sub,S:UseGalois:=true, ShortOK := ShortOK);
    W:=WreathProduct(SymmetricGroup(d), H);
    W:=W^perm; //Gal(f) is subgroup of W !!
    vprint GaloisGroup, 2: "Computed H ",H," and block ",MinimalPartition(W);
    G:=G meet W;
    if false and d eq 2 then //automorphism found
      B:=MinimalPartition(W);//blocks of size 2
      B:=[SetToSequence(x):x in B];
      L:=[x+1: x in [1..n]];//Create (1,2)(3,4)...(n-1,n)
      for x in [1..n] do
        if x mod 2 eq 0 then
           L[x]:=x-1;
        end if;
      end for;
      new:=(Sym(n) ! L)^(perm);
      assert new in G;
      ChangeUniverse(~aut,G);
      Append(~aut,(G ! L)^perm);
    end if;
    assert IsTransitive(G);
    if #G eq Degree(K) then
      vprint GaloisGroup, 1: green, "not all subfields neccessary", normal;
      break;
    end if;
  end for;
//"Have gone through maximal subfields";

/*Try to improve intersection by using invariants Sm and DSm*/
/*New loop since I want to check all subfields*/

  /*
  Deleted at end of function - this should be good enough
  delete S`SubfieldLattice;
  delete S`Subfields;
  */

  reject:=[];

    L := S`Subfields;
  if InternalGaloisSubfieldLatticeGetField(LL, InternalGaloisSubfieldLatticeTop(LL)) ne K then
    L := [x : x in L | Degree(x[1]) le Degree(K)];
    KL := InternalGaloisSubfieldLatticeFindField(LL, K);
    L := [x : x in L | InternalGaloisSubfieldLatticeSubset(LL, InternalGaloisSubfieldLatticeFindField(LL, x[1]), KL)];
    //Subfields(K:Current := Current);
  end if;
  if is_square_df then 
//In this case we cannot improve the starting group, but we can add some groups
//in reject, if we know that the disc of the subfield is not a square...
    for s in L do
      sub := <s[1], K!s[2](s[1].1)>;
      if sub[1] eq K then
        continue;
      end if;
      m:=Degree(sub[1]);
      d:=n div m;
      dg:=Discriminant(DefiningPolynomial(sub[1]));
      if not IsSquare(dg) then
        perm,H:=IdentifyBlockSystem(K,sub,S:UseGalois:=false);
        W:=WreathProduct(SymmetricGroup(d), Alt(m));
        W:=W^perm; //Gal(f) is subgroup of W !!
        Append(~reject,W); //Is it a better idea to intersect WW with G before inserting?      
      end if;
    end for;
    if UseCache() then
      CacheNormalise(~G, S);
    end if;
    delete S`Subfields;
    delete S`SubfieldLattice;
    return G, reject, aut;
  end if;


  vprint GaloisGroup, 2: "Computed G of size", FactoredOrder(G),
                         "and try to improve";
  vtime GaloisGroup, 2:  
  for s in L do
    sub := <s[1], K!s[2](s[1].1)>;
    if sub[1] eq K then
      continue;
    end if;
    m:=Degree(sub[1]);
    d:=n div m;
    //if d le 3 then continue; end if;
    dg:=Discriminant(DefiningPolynomial(sub[1]));
    disc:=dg*df;
    perm,H:=IdentifyBlockSystem(K,sub,S:UseGalois:=false);
//assume here that H=A_n or S_n
    W:=WreathProduct(SymmetricGroup(d), H);
    B:=MinimalPartition(W);
    if not IsEven(H) and Characteristic(K) ne 2 and IsSquare(disc) then //we can improve!
      if IsOdd(d) then
        WW:=InternalWreathProductSm(B);
      else
        WW:=InternalWreathProductDSm(B);
      end if;
      WW:=WW^perm; //Gal(f) is subgroup of W !!      
      G:=G meet WW;
      assert IsTransitive(G);
      //dg cannot be a square because df is not and disc is...
      WW:=InternalWreathProductD(B);
      WW:=WW^perm;
      Append(~reject,WW); //Is it a better idea to intersect WW with G before inserting?
      if #G eq Degree(K) then
        vprint GaloisGroup, 1: green, "not all subfields neccessary", normal;
        break;
      end if;
    else  //found one to reject
      if IsOdd(H) then //we can add three subgroups to reject
         Append(~reject,InternalWreathProductD(B)^perm);
         Append(~reject,InternalWreathProductDSm(B)^perm);
         Append(~reject,InternalWreathProductSm(B)^perm);
      else//here we can only add one (the other one is now the same...)
          //D was used to descent...
        if IsOdd(d) then
           WW:=InternalWreathProductSm(B);
        else
           WW:=InternalWreathProductDSm(B);
        end if;
        WW:=WW^perm;
        Append(~reject,WW);
      end if;      
    end if;
  end for;
  vprint GaloisGroup, 2: "improved to ",FactoredOrder(G), " also have ",#reject," rejected subgroups";
  if UseCache() then
    CacheNormalise(~G, S);
  end if;
  delete S`Subfields;
  delete S`SubfieldLattice;
  return G, reject, aut;
end function;    

intrinsic InternalBound(S::GaloisData, tschirni::RngUPolElt, inv::RngSLPolElt, 
                idx::RngIntElt: B := false, LowTest := false) -> .
  {}
  //RED, "\nInternslBound called\n\n", normal;
  if S`Type eq "Complex" then
    nop := NumberOfOperations(inv, "+") + NumberOfOperations(inv, "-") +
           Degree(tschirni);
    nop := Minimum(1000, nop);       
  else
    nop := 0;
  end if;

//"assigned", assigned S`Bound;
  if Type(S`max_comp) eq RngIntElt then
    if LowTest then
      return idx * 10^nop;
    else
      return 10^nop * Bound(inv, Evaluate(tschirni, S`max_comp))^idx;
    end if;
  end if;
  if assigned S`Bound then
    //"calling provate bound function";
    x := 10^nop * S`Bound(tschirni, inv, idx: LowTest := LowTest);
    //"returning", x, Type(x), Parent(x);
    return x;
  end if;
  error "case not implemented (yet)";
end intrinsic;

function invar_coeff_ring(S)
    if Characteristic(S`Base) eq 0 then
	return Integers();
    elif ISA(Type(S`Prime), RngElt) then
	return Parent(S`Prime);
    else
	return Parent(Minimum(S`Prime));
    end if;
end function;

function get_tschirni_R(R, coeff_bound, degree_bound, all_tschirni)
    if Characteristic(R) eq 0 then
	repeat
	  tschirni := Polynomial([Random(coeff_bound) : x in [1..Random(degree_bound)]] cat [1]);
	until Degree(tschirni) ge 1 and tschirni notin all_tschirni;  
    else
	Fq := CoefficientRing(R);
	repeat
	  tschirni := Polynomial([Polynomial([Random(Fq) : y in [1 .. Random(1, coeff_bound)]]) : x in [1..Random(degree_bound)]] cat [1]);
	until Degree(tschirni) ge 1 and tschirni notin all_tschirni;  
    end if;
    return tschirni;
end function;

function get_tschirni(S, coeff_bound, degree_bound, all_tschirni)
    return get_tschirni_R(invar_coeff_ring(S), coeff_bound, degree_bound, all_tschirni);
end function;

function get_all_tschirni_universe(S)
    if Characteristic(S`Base) eq 0 then
	return PolynomialRing(Integers());
    elif ISA(Type(S`Prime), RngElt) then
	return PolynomialRing(Parent(S`Prime));
    else
	return PolynomialRing(Parent(Minimum(S`Prime)));
    end if;
end function;

function apply_tschirni(S, tschirni, X)
    if Characteristic(S`Base) eq 0 then
	return [Evaluate(tschirni, x) : x in X];
    else
	T := Polynomial([S`BaseMap(c, AbsolutePrecision(X[1])) : c in Coefficients(tschirni)]);
	return [Evaluate(T, x) : x in X];
    end if;
end function;

function eval_invar(S, i, r)
    if Characteristic(S`Base) eq 0 then
	return Evaluate(i, r);
    else
	ii := 1;
    end if;
end function;

intrinsic Stauduhar(G::GrpPerm, H::GrpPerm, 
                    S::GaloisData, B::RngIntElt:
                    AlwaysTransform:= false, Coset := false, 
                    PreCompInvar := false, GalRel := false,
                    Tschirni := false, UseInvarEnv := false,
                    AllTschirni := false, ProofEffort := 10)
                            -> RngIntElt, GrpElt, BoolElt, RngSLPolElt
  {return 1 if the Galois group is contained in H, -1 is it is probably in H and 5.10 was conclusive, -2 if if it probable in H, and 0 otherwise}

    L := H;
    x := ext<Integers()|>.1;
    max_comp := B;
    n := Degree(G);

    vprint GaloisGroup, 2: "Stauduhar: Start with subgroup of index ",
                            Index(G,L);
    if IsTransitive(L) then                        
      vprint GaloisGroup, 2: "with block structure ", TestAllPartitions(L);
    else
      vprint GaloisGroup, 2: "(intransitive case)";
    end if;  
    if Degree(H) le 23 and GetVerbose("GaloisGroup") ge 2 
       and IsTransitive(G) then
       //TransitiveGroupIdent is expensive...
       idH:=TransitiveGroupIdentification(H);
       idG:=TransitiveGroupIdentification(G);
       vprintf GaloisGroup, 2: "and number <%o%oT%o, %oT%o%o>\n", 
                      green, n, idG, n, idH, normal;
    end if;
    vprint GaloisGroup, 5: "Generators of G ", Generators(G);
    if PreCompInvar cmpeq false then
      i := GaloisGroupInvariant(invar_coeff_ring(S), G, L:GalRel := GalRel); 
    else
	assert assigned PreCompInvar`GalInvType;
      i := PreCompInvar;
    end if;
    /*
    CAN CAUSE COERCION ISSUES IN CHECK WHICH WE NEED TO SORT OUT
    assert forall{x : x in Generators(L) | IsInvariant(SLPolynomialRing(CoefficientRing(get_all_tschirni_universe(S)), Rank(Parent(i)) : Global)!i, x)};
    assert exists{x : x in Generators(G) | not IsInvariant(SLPolynomialRing(CoefficientRing(get_all_tschirni_universe(S)), Rank(Parent(i)) : Global)!i, x)};
    */
    R := CoefficientRing(get_all_tschirni_universe(S));
    assert forall{x : x in Generators(L) | IsInvariant(SLPRGg(R, i), x)};
    assert exists{x : x in Generators(G) | not IsInvariant(SLPRGg(R, i), x)};
    assert assigned i`GalInvType;
    vprint GaloisGroup,2: "Using invar of type: ", i`GalInvType;
    vprint GaloisGroup,2: "Should have cost ", NumberOfOperations(i, "*");
    if AlwaysTransform then
      tschirni := get_tschirni(S, 3, Degree(G), {x});
    elif Tschirni cmpne false and UseInvarEnv then
      tschirni := Tschirni;
    else
      tschirni := x;
    end if;
    if AllTschirni cmpne false then
      all_tschirni := AllTschirni;
    else
      all_tschirni := {get_all_tschirni_universe(S) | x, tschirni};
    end if;

    if Coset cmpeq false and 
       (S`Type notin {"p-Adic (FldFun over Q)"}//, "p-Adic (FldFun over Fq)"} 
	   or assigned S`KnownSubgroup) and
	   // should count number of multiplications in each powering and add
	   // for exponents > 2 there will be at least 2 multiplications
	   // NumberOfOperations will count the multiplications in the powerings
	   ((#G div #L)*(NumberOfOperations(i, "*")) gt 10000) then
      vprint GaloisGroup, 1: "use short cosets";
      if not assigned S`KnownSubgroup then
        // we need to compute U = Gal(S`Ring) as a subgroup of H<G
        // and then to implement 5.12 on page 115
        // Lets start with U:
        R, mR := ResidueClassField(S`Ring);
        r := [ GaloisRoot(i, S: Prec := 1)@mR : i in [1..n]];
        rt := [ x^q : x in r] where q := #ResidueClassField(S`Base);
        p := [ Position(rt, x) : x in r];
      else
        p := S`KnownSubgroup.1; //TODO: think about better groups!
                                //      and better use of them.
        vprintf GaloisGroup, 2: "Using a subgroup of size %m:\n%o\n", #S`KnownSubgroup, S`KnownSubgroup, "Maximal";                        
      end if;
      vprint GaloisGroup, 2: "Using a permutation", p;

      if IsNormal(G, H) and assigned S`KnownSubgroup and not S`KnownSubgroup subset H then
          vprint GaloisGroup, 1: "(normal) subgroup is not possible";
          return 0, 1, true, _;
      end if;

      if (#G div #L) lt MaxShortCosets then //simply take all cosets
        cos := [x : x in Transversal(G, L)];
      elif IsNormal(G,H) then
//JK: For normal subgroups all cosets produce integers or none. Therefore there is an easy algorithm
//to produce ShortCosets, simply take one. In order to use 5.10, add a second one
         vprint GaloisGroup, 2: "normal subgroup";
         gen:={x: x in Generators(G) | not x in H};
         assert #gen gt 0;
         cos := [G.0, Random(gen)];      
      else  
        cos := ShortCosets(G!p, H, G);
      end if;
      if cos cmpeq false then
S`f, S`Prime, G, H;
        error "Too many short cosets, wrong prime ....";
      end if;
      if #cos lt (#G div #L) then 
//and not IsNormal(G, H) then //JK: Do not understand the not IsNormal part
        UseShort := true;
      else
        //OK, we tried to use ShortCosets, but it didn't help...
        UseShort := false;
      end if;
      //to give it a chance, try to add some elementes of G\H...
      //(to help 5.10)??
      vprint GaloisGroup, 2: "Short: Have to consider ", 
                             #cos, " cosets initially";
      if #cos eq 0 then
        vprint GaloisGroup, 1: "no cosets remaining, group not possible";
        return 0, 1, true, _;
      end if;
      if #cos lt 10 and not AlwaysTransform and UseShort then
        for i in [1..10] do
          g := Random(G);
          if forall(x){x : x in cos | x*g^-1 notin H} then
            Append(~cos, g);
          end if;
        end for;
        vprint GaloisGroup, 2: "supplemented to  ", #cos;
      end if;
    elif Coset cmpeq false then
      cos:=RightTransversal(G,L);
      cos:=[g: g in cos];
      UseShort := false;
      vprint GaloisGroup, 2: "Long: Have to consider ", #cos, 
                             " cosets initially";
    else                         
      cos := [ g : g in Coset];
      vprint GaloisGroup, 2: "Long: Have to consider ", #cos, 
                             " cosets initially";
      UseShort := false;
    end if;
    success := false;
    all_pf_flag := true;

    // When we evaluate at the roots in the splitting field it is easier 
    // for compatibility to have the invariant over the integers if possible
    is_c, eval_i := IsCoercible(SLPolynomialRing(Integers(), Rank(Parent(i)) : Global), i);

    /*
    Keep index_limit not too large since it is used as an upper bound
    to get BB_lift. Increasing index_limit increases BB_lift and so precision
    unnecessarily when the result is not recorded as being proven
    */
    index_limit := 45;
    if S`Type eq "p-Adic (FldFun over Q)" then
	index_limit := index_limit div 3;
    end if;
    mult_limit := ProofEffort*10000;
    repeat
      pf_flag := true;
      if Index(G, H) gt index_limit then
        pf_flag := false;
        BB_lift := InternalBound(S, tschirni, i, index_limit: B := B);
      else
        BB_lift := InternalBound(S, tschirni, i, Index(G, H): B := B);
      end if;
      BB_test := InternalBound(S, tschirni, i, 1: B:= B);
//TotalDegreeAbstract(i), S`max_comp, Index(G, H);
	Prec := PrecisionFromBound(BB_lift, S);
	if Type(Prec) eq RngIntElt then
	    mult_cost := NumberOfOperations(i, "*")*Prec*Log(Prec);
	else
	    mult_cost := NumberOfOperations(i, "*")*Prec[1]*Prec[2]*Log(Prec[1])*Log(Prec[2]);
	end if;

      if Index(G, H) gt index_limit and (S`Type eq "p-Adic (FldFun over Q)" or
      			Index(G, H)*mult_cost gt index_limit*mult_limit) then
        pf_flag := false;
        vprintf GaloisGroup, 2: 
          "Large index (%o) subgroup, using a bound of %o instead\n",
            Index(G, H), index_limit div 3;
	if S`Type ne "p-Adic (FldFun over Q)" then
	    BB_lift := InternalBound(S, tschirni, i, index_limit div 3:B := B);
        end if;
      else
      end if;
      vprintf GaloisGroup, 2: "Log2 of Bound: %o (%o, %o)\n", 
                                                Ilog2(BB_lift), B, tschirni;
      vprintf GaloisGroup, 2: "Log2 of Bound: %o (%o, %o)\n", 
                                                Ilog2(BB_test), B, tschirni;
      all_pf_flag := all_pf_flag and pf_flag;
      inv := i;
      if Ilog2(BB_lift) gt 300 and not false then
        vprint GaloisGroup, 2: 
          "Numbers are largish, try to check for duplicates first...";
        repeat
          b :=  InternalBound(S, tschirni, inv, #cos: B := B, LowTest);
          r := func<|[ GaloisRoot(i, S: Bound := b) : i in [1..n]]>();
          r := func<|apply_tschirni(S, tschirni, r)>();
          ir := {};
          dbl := 0;
          n_cos := #cos;
          for g in cos do
	    if not is_c then
		assert assigned S`BaseMap;
if not IsExport() then
"\n\n\nne Evaluating using base map 1";
end if;
		x := Evaluate(i, PermuteSequence(r, g), map<CoefficientRing(Parent(i)) -> Universe(r) | y :-> S`BaseMap(y, PrecisionFromBound(b, S))>);
	    else
		x := Evaluate(eval_i, PermuteSequence(r, g));
	    end if;
            if x in ir then dbl +:= 1; 
              if dbl gt 5 then
                  n_cos := Infinity();
                  break;
                end if;
              end if;
            Include(~ir, x);
//"IsInt 1 : ", b;
            if not IsInt(x, b, S:EasyNonInt) then
              Exclude(~cos, g);
              additional := g;
            end if;
          end for;
          if #cos eq 0 then
            vprint GaloisGroup, 2: "removed all cosets";
            return 0, _, _, _;
          else
            vprint GaloisGroup, 2: "only", #cos, "are remaining after low prec";
//CF            if #cos eq 2 then return i, tschirni, cos; end if;
          end if;

          // TODO: if evaluation is in a "too large" ring, then remove the
          // coset as it cannot correspond to s.th. useful!
          // eg: in the absolute case if we're not landing in Z_p we're out.
          if #ir lt n_cos then
            vprint GaloisGroup, 2: "Tschirni - short cut";
	    tschirni := get_tschirni(S, 3, Minimum(Degree(G), 2+#all_tschirni), all_tschirni);
            // I guess, one should also check that the new element is actually
            // primitive.
            Include(~all_tschirni, tschirni);
	    if not IsExport() then 
		if #all_tschirni ge 3*Degree(G) then
		    S`f;
		  error "way too many transforms in shortcut!";
		end if;
	    end if;
          else
            break;
          end if;
        until false;
        if Index(G, H) gt index_limit then
          pf_flag := false;
          BB_lift := InternalBound(S, tschirni, i, index_limit: B := B);
        else
          BB_lift := InternalBound(S, tschirni, i, Index(G, H): B := B);
        end if;
        BB_test := InternalBound(S, tschirni, i, 1: B:= B);
	Prec := PrecisionFromBound(BB_lift, S);
	if Type(Prec) eq RngIntElt then
	    mult_cost := NumberOfOperations(i, "*")*Prec*Log(Prec);
	else 
	    mult_cost := NumberOfOperations(i, "*")*Prec[1]*Prec[2]*Log(Prec[1])*Log(Prec[2]);
	end if;
//"INDEX TEST", Index(G, H), mult_cost;
        if Index(G, H) gt index_limit and (S`Type eq "p-Adic (FldFun over Q)" or
			Index(G, H)*mult_cost gt index_limit*mult_limit) then
          pf_flag := false;
          vprintf GaloisGroup, 2: 
            "Large index (%o) subgroup, using a bound of %o instead\n",
              Index(G, H), index_limit;
	  if S`Type ne "p-Adic (FldFun over Q)" then
	      BB_lift := InternalBound(S, tschirni, i, index_limit div 3: B:= B);
	  end if;
        end if;
        vprintf GaloisGroup, 2: "Log2 of Bound now: %o (%o, %o)\n", 
                                                  Ilog2(BB_lift), B, tschirni;
        if assigned additional then
          Include(~cos, additional);
        end if;
        if UseInvarEnv and Tschirni cmpne tschirni then
          vprint GaloisGroup, 1: "Tschirni has changed, returning invariant";
          InternalGaloisInvarEnvChangeInvar(S, inv, tschirni, cos, all_tschirni);
          return 2, _, _, _;
        end if;
      end if;

      if not IsExport() then
	  max_prec := 2*50000;
      else
	  max_prec := DEFAULT_MAX_PREC;
      end if;

      if Ilog2(BB_lift) gt max_prec then 
S`max_comp;
tschirni;
TotalDegreeAbstract(i), NumberOfOperations(i, "*"), NumberOfOperations(i, "+"), NumberOfOperations(i, "^"), NumberOfTerms(i), i`GalInvType;
BB_lift, BB_test, Prec, Index(G, H);
      error "Precision too large?";
        vprintf GaloisGroup, 1:
          "Precision (%o) too large, descent failed.\n", Ilog2(BB_lift);
        vprint GaloisGroup, 1: "Maximal precision is ", max_prec;
        return 0, _, _, _; 
      end if;

      r := func<|[ GaloisRoot(i, S:Bound := BB_lift): i in [1..n]]>();
      r := func<|apply_tschirni(S, tschirni, r)>();

      //Check invariant for each g in cos, i.e. permute roots accordingly
      foundlist:=[];foundg:=[];
      no_cos := 0;
      for g in cos do
        no_cos +:= 1;
        if no_cos mod 100 eq 0 then
          vprint GaloisGroup, 2: "Done ", no_cos, " cosets so far";
        end if;
	if not is_c then
	    assert assigned S`BaseMap;
if not IsExport() then
"\n\n\nne Evaluating using base map 2";
end if;
	    z := Evaluate(i, PermuteSequence(r, g), map<CoefficientRing(Parent(i)) -> Universe(r) | y :-> S`BaseMap(y, PrecisionFromBound(BB_lift, S))>);
	else
	    z := Evaluate(eval_i, PermuteSequence(r, g));
	end if;
//"IsInt 2 : ", BB_test, BB_lift;
        isInt, Int := IsInt(z, BB_test, S);
        if isInt then
          Append(~foundlist, Int);
          Append(~foundg,g);
        end if;
      end for;
//Seqset(foundlist);
//#foundg;
//#foundlist;
      if #foundlist gt 0 then  //Check, if there are single zeros!
        multi:=SequenceToMultiset(foundlist);
        vprint GaloisGroup, 2: "found elements with multiplicity:", 
                                [ Multiplicity(multi, x) : x in Set(multi)];
        mm:=[x:x in Set(multi) |Multiplicity(multi,x) eq 1];
        if #mm gt 0 then
           success := true;
           vprint GaloisGroup, 1: BLUE, "DESCENT", normal;
//"UseShort", UseShort;
           if UseShort and not IsNormal(G, H) then 	
// FOR JK : This is what I really wanted to avoid
// As above comments on normal subgroups all cosets give integers or none.
// The Galois group either lies in H or not (when H is normal as all conjugates 
// are the same) so we want to return a definite answer here and stop looking
// as that can be really expensive.
             if #foundlist lt #cos or #foundlist gt 1 then
	       assert assigned i`GalInvType;
               return -1, [foundg[Position(foundlist, g)] : g in mm], 
                 pf_flag, i;
             else 
	       assert assigned i`GalInvType;
               return -2, [foundg[Position(foundlist, g)] : g in mm], 
                 pf_flag, i;
             end if;
           end if;
	   assert assigned i`GalInvType;
           return 1, [foundg[Position(foundlist, g)] : g in mm], pf_flag, i;
        else
          cos := foundg; // according to KG, it's sufficient to consider
                         // the cosets that produced integers at this point.
	  tschirni := get_tschirni(S, 3, Minimum(Degree(G), 2+#all_tschirni), all_tschirni);
          // I guess, one should also check that the new element is actually
          // primitive.
          Include(~all_tschirni, tschirni);
	  if not IsExport() then 
	      if #all_tschirni ge 100 then
		#all_tschirni;
S`f;
		error "way too many transforms!";
	      end if;
	  end if;
          vprint GaloisGroup, 1: 
            "Tschirni must be applied: using", tschirni;
          vprint GaloisGroup, 1: 
            "Number of cosets now ", #cos;
          if UseInvarEnv then
            vprint GaloisGroup, 1: "Tschirni has changed, returning invariant";
            InternalGaloisInvarEnvChangeInvar(S, inv, tschirni, cos, all_tschirni);
            return 2, _, _, _;
          end if;
          success := false;
        end if;
      else
        success := true; // no root, next subgroup
      end if;
    until success;
    vprint GaloisGroup, 1: "no descent";
    return 0, _, _, _;
end intrinsic;

function CycStrucToSequence(L)
  LL:=[];
  for a in L do
    for b in [1..a[2]] do
      Append(~LL,a[1]);
    end for;
  end for;
  return Sort(LL);
end function;

function TestShapes(G, shapes)
  if Degree(G) gt 30 then
    return true;
  end if;

  c:=ConjugacyClasses(G);
  c:={CycleStructure(x[3]):x in c};
  c:={CycStrucToSequence(x):x in c};
  return shapes subset c;  
end function;

function GenericStauduhar(G, S, max_comp, 
                             MaxSubGp, Sieve:
              ShortOK := false, Subfields := true, GalRel := false, 
              SubfieldsComplete := false,
              Reject:=[],Subgrp:=sub<G|>, ProofEffort := 10);

  vprintf GaloisGroup, 1: "\n%oStart Generic Stauduhar Algo%o\n", RED, normal;

  rejectedNormalSubGroups := Reject;
  repeat
    vprint GaloisGroup, 2: "Doing group of size", FactoredOrder(G);
    vprintf GaloisGroup, 2: "Known subgroup f size %m: \n%o\n", #Subgrp, Subgrp, "Maximal";
    vprint GaloisGroup, 2: "Rejected subgroups", Reject;

    if IsTransitive(G) then
      vprint GaloisGroup, 3: "with block structure",TestAllPartitions(G);
    else
      vprint GaloisGroup, 3: "(intransitive case)";
    end if;
    if UseCache() then
      r_a, r_b := GetSeed();
      SetSeed(1,0);
    end if;
    if _GetGaloisCacheLevel() gt 1 then
      fl, LU := CacheHasGroup(G);
      if not fl then
        LU := MaxSubGp(G);
        CachePutGroup(G, LU);
      end if;
    else
      LU := MaxSubGp(G);
    end if;
    if UseCache() then
      SetSeed(r_a, r_b);
    end if;
    LU:=[x:x in LU];
    assert G notin LU;
    vprint GaloisGroup, 1: "Have to consider ", #LU, 
                           "subgroups (classes of them)";
    InternalGaloisInvarEnvInit(S, G, LU); // maybe add intransitive groups!!
    InternalGaloisInvarEnvSieve(S, Sieve);
    vprint GaloisGroup, 1: "Reduce to", InternalGaloisInvarEnvNumber(S);

    InternalGaloisInvarEnvSieve(S, 
      func<x|forall{t : t in rejectedNormalSubGroups | not x subset t}>);
    vprint GaloisGroup, 1: "Further reduce to (using normal subgroups)", 
      InternalGaloisInvarEnvNumber(S);
    InternalGaloisInvarEnvSieve(S, 
      func<x|not IsNormal(G, x) or Subgrp subset x>);
    vprint GaloisGroup, 1: "Further reduce to (using known subgroups)", 
      InternalGaloisInvarEnvNumber(S);
    if InternalGaloisInvarEnvNumber(S) eq 0 then
      InternalGaloisInvarEnvClear(S);
      return G, [GaloisRoot(i, S:Bound := 1) : i in [1..Degree(G)]], S;
    end if;
    no_LU :=0;
    found := false;
    possible := [];
    Res := {};
    repeat
      f, L, I, T := InternalHasGaloisInvarEnvNext(S);
      if not f then break; end if;
if not IsExport() and Characteristic(invar_coeff_ring(S)) ne 0 then 
"ne Using ", I`GalInvType;
end if;
      A := T[3];
      C := T[2];
      T := T[1];
      if #C eq 0 then C := false; end if;
      if #A eq 0 then A := false; end if;
      vprint GaloisGroup, 1: "doing Stauduhar for group", Position(LU, L);
      res, g, pf_flag, invar := 
        Stauduhar(G, L, S, max_comp: GalRel := GalRel, PreCompInvar := I, 
                                                       Tschirni := T,
                                                       Coset := C,  
                                                       AllTschirni := A,  
                                                       UseInvarEnv, 
						       ProofEffort := ProofEffort);
      vprint GaloisGroup, 2: red, "Stauduhar returns ", res, normal;
      if res eq 2 then continue; end if; // invar was changed.
      if res eq 1 or (ShortOK and res eq -1) then
        assbl:=false;
        if assigned G`Blocks and Subfields and SubfieldsComplete then
           assbl:=true;
           Bl:=G`Blocks;
        end if;
        if #g gt 1 then 
          vprint GaloisGroup, 2: 
                   "multiple descents possible, try to intersect...";
          if UseCache() then
            GaloisRootAddPermutation(S, g[1]);
            g := [x*g[1]^-1 : x in g];
          end if;
          GG := &meet [ L^x : x in g];
	  if A cmpeq false then A := {}; end if;
	  if C cmpeq false then C := []; end if;
	  if not Sieve(GG) then
	      T := get_tschirni(S, 3, Minimum(Degree(G), 2+#A), A);
	      Include(~A, T);
	      if not IsExport() then 
		  if #A ge 40 then
    S`f;
		    error "way too many transforms!";
		  end if;
	      end if;
		InternalGaloisInvarEnvChangeInvar(S, I, T, C, A);
	    continue;
	  end if;
	  G := GG;
          if _GetGaloisCacheLevel() gt 1 then
            CacheNormalise(~G, S);
          end if;
          Append(~S`DescentChain, <shallow_copy(G), pf_flag>);
          vprint GaloisGroup, 2: "Size of G ", FactoredOrder(G), 
                                 " without intersect: ", FactoredOrder(L);
          if #L ne #G then
            vprint GaloisGroup, 1: 
              RED, "Intersection reduced to ", FactoredOrder(G), 
                   " from ", FactoredOrder(L), normal;
          end if;

        else
          if UseCache() then
            GaloisRootAddPermutation(S, g[1]);
            G := L;
          else
            G := L^g[1];
          end if;
          Append(~S`DescentChain, <shallow_copy(G), pf_flag>);
        end if;
        found := true;
        if assbl then
          G`Blocks:=Bl;
          assert #Bl eq #AllPartitions(G);
        end if;
	break;
      elif res eq 0 then
        found := false;
        if IsNormal(G, L) then
          Append(~rejectedNormalSubGroups, L);
          vprint GaloisGroup, 1: purple, "added wrong subgroup", normal;
        end if;
      elif res eq -1 or res eq -2 then // should correspond to short cosets
	assert assigned invar`GalInvType;
        Append(~possible, <L,g, pf_flag, invar>);
        Include(~Res, res);
      else
        error "Error: this case should not be possible...";
      end if;
    until false; 

    if GetVerbose("GaloisGroup") ge 1 then
      "OK, we've tried enough (all) (maximal) subgroups, and here's what we've got:";
      if found then
        "A descent was definitely found";
      else
        "No descent is (so far) identified";
        if #possible gt 0 then
          "But there are", #possible, " possiblilies (through use of short cosets)";
          if -1 in Res then
            "and we KNOW there is a descent possible";
          end if;
        end if;
      end if;
    end if;
    if not found and #possible eq 1 and -1 in Res then
      found := true;
      if assigned G`Blocks and Subfields and SubfieldsComplete then      
         Bl:=G`Blocks;
         G := possible[1][1];
         if UseCache() then
           GaloisRootAddPermutation(possible[1][2][1]);
         else
           G := &meet [ G^x : x in possible[1][2]];
         end if;
       G`Blocks:=Bl;
       assert #Bl eq #AllPartitions(G);
      else
         G:=possible[1][1];
         if UseCache() then
           GaloisRootAddPermutation(possible[1][2][1]);
         else
           G := &meet [ G^x : x in possible[1][2]];
         end if;
      end if;
      Append(~S`DescentChain, <shallow_copy(G), possible[1][3]>);
    end if;

    if not found and #possible gt 1 then
      vprint GaloisGroup, 1:  
        "Sorry, short cosets weren't conclusive trying some transformations\n";
      vprint GaloisGroup, 1: "still ", #possible, "subgroups possible";  
      IndentPush();
      for i in [1..#possible] do
        no_t := 0;
        s_g := Set(possible[i][2]);
        vprint GaloisGroup, 2: BLUE, 
          "starting with ", #s_g, " possibilities in this group", normal;
        repeat
          IndentPush();
	  assert assigned possible[i][4]`GalInvType;
          res, g, pf_flag := Stauduhar(G, possible[i][1], S, max_comp:
                                    AlwaysTransform, Coset := s_g, 
                                    PreCompInvar := possible[i][4], 
				    ProofEffort := ProofEffort);
          IndentPop();                                                
	  if assigned g then
	      s_g := s_g meet Set(g);
	      no_t +:=1;
	  end if;
        until no_t gt 3 or res eq 0 or #s_g eq 0;
        if res eq 0 or #s_g eq 0 then
          Undefine(~possible, i);
          vprint GaloisGroup, 2: BLUE, 
            "group no longer possible\n", normal;
        else
          possible[i][2] := [x : x in s_g];
          vprint GaloisGroup, 2: BLUE, 
            "left  with ", #s_g, " possibilities", normal;
        end if;
      end for;
      IndentPop();
      possible := [x : x in possible];
      
      if #possible gt 1 then
        vprint GaloisGroup, 2: 
                       "multiple descents possible, try to intersect...";
        L := &meet [ &meet [ x[1]^g : g in x[2]] : x in possible];
        possible := [<L, [L.0], false>];
        vprint GaloisGroup, 1: 
          RED, "Intersect all possibilities, size left:", #L, normal;
        assert IsTransitive(G) and #G gt Degree(G);
      end if;

      if #possible eq 1 then
        found := true;
        if assigned G`Blocks and Subfields and SubfieldsComplete then      
           Bl:=G`Blocks;
           G := possible[1][1];
           if UseCache() then
             GaloisRootAddPermutation(possible[1][2]);
           else
//why?	     Exclude(~possible[1][2], G.0);
	     //assert #possible[1][2] eq 1;
             G := G^possible[1][2][1];
           end if;
           G`Blocks:=Bl;
           assert #Bl eq #AllPartitions(G);
        else
           G:=possible[1][1];
           if UseCache() then
             GaloisRootAddPermutation(possible[1][2]);
           else
//why?	     Exclude(~possible[1][2], G.0);
	     //assert #possible[1][2] eq 1;
             G := G^possible[1][2][1];
           end if;
        end if;
        Append(~S`DescentChain, <shallow_copy(G), possible[1][3]>);
      else
        error "Sorry, short cosets weren't conclusive, don't know what to do";
      end if;
    end if;
    if not found then
      //It might be the case that tschirhausen roots are stored in r...
      r := [ GaloisRoot(i, S:Bound := 1) : i in [1..Degree(G)]];
      InternalGaloisInvarEnvClear(S);
      return G, r, S;
    end if;
  until IsCyclic(G);
  //It might be the case that tschirhausen roots are stored in r...
  r := [ GaloisRoot(i, S:Bound := 1) : i in [1..Degree(G)]];
  InternalGaloisInvarEnvClear(S);
  return G, r, S;
end function;


intrinsic EnterStauduhar(G::GrpPerm, K::FldNum:Subfields:=false) -> GrpPerm, [], GaloisData
{Enter Galois Group algorithm}
  S:=K`GaloisData;
  if Subfields then
time    G2:=GaloisStartingGroup(K,S:Subfields:=true);
    GG:= G meet G2;
  else
    GG:=G;
  end if;nummm:=#GG;
GG;  
  d:=Discriminant(K);
  isEven:=IsSquare(d);
  return GenericStauduhar(GG, S, S`max_comp, 
      func<X|[ x`subgroup : x in MaximalSubgroups(X:IsTransitive)]>,
      func<X|IsEven(X) eq isEven and TestShapes(X, S`CycleTypes)>:
        ShortOK := true, Subfields := Subfields);
end intrinsic;


function GetGaloisDataOrGroup(K: 
  Ring := false, Type := "p-Adic", Prime := false, Prec := false, 
  ShortOK := false, Subfields := true, Current := false, 
  SOnly := false, Time := false, NextPrime := My_np)

  if ISA(ExtendedType(K), FldCyc) then
    Cyclo := true;
    Subfields := false;
  else
    Cyclo := false;
  end if;
  UserPrime := Prime;
  if Ring cmpne false then
    UserPrime := Ring`Prime;
  end if;

  if Prec cmpeq false then
    Prec := 20;
    MinPrec := false;
  else
    MinPrec := Prec;
  end if;
  f := DefiningPolynomial(K);
  shapes:={};
  if Ring cmpeq false then
    if assigned K`GaloisData and 
      (Prime cmpeq false or Prime cmpeq K`GaloisData`Prime) then
      S := K`GaloisData;
      if assigned S`Permutation then
        delete S`Permutation;
      end if;
      if assigned S`Scale then
        delete S`Scale;
      end if;
      if S`Type ne "Complex" then
        Prime := S`Prime;
        UserPrime := S`Prime;
      end if;
      if assigned S`Subfields then
        delete S`Subfields;
      end if;
      if assigned S`SubfieldLattice then
        delete S`SubfieldLattice;
      end if;
    else
      if Type cmpeq "p-Adic" then
        if Prime cmpeq false then
           Prime,shapes:=PrimeChoose(K, false: Current := Current,
                      Subfields := Subfields, NextPrime := NextPrime);
        end if;
        S := InternalGaloisDataCreate(Prime);
        S`Type := "p-Adic";
        S`Base := pAdicRing(Prime);
        S`Prec := Prec;
      elif Type cmpeq "Complex" then
        S := InternalGaloisDataCreate(4);
        S`Type := "Complex";
        S`Prec := Prec;
        S`Base := GetDefaultRealField();
        if MinPrec cmpne false then
          S`MinPrec := MinPrec;
        end if;
      else
        error "Wrong type, only p-Adic and Complex are supported yet";
      end if;
      K`GaloisData := S;
    end if;
  else
    S := Ring;
  end if;
  if #shapes eq 0 then
     _,shapes:=PrimeChoose(K, S:Subfields := Subfields, Prime := Prime, Current := Current);
  end if;
  S`CycleTypes := shapes;
  S`f := f;
  if Time then
    S`Time := rec<recformat<Lift>|Lift := 0.0>;
  end if;
  n := Degree(K);
  // Sn/ An test:
  // if there is a cycle of length n/2 < p < n-2 then we are Sn or An
  // (p prime)
  poss_p := { p : p in [n div 2+1 .. n-3] | IsPrime(p)};
  d := Discriminant(K);
  isEven := IsSquare(d);
  if #(poss_p meet Set( &cat [x : x in shapes])) ne 0 then
    vprint GaloisGroup, 1: "Sn or An found from cycle shapes";

    prime := PrimeChoose(K, S:Subfields := Subfields, SmallDeg,
                                           NextPrime := NextPrime);
    if S`Type ne "Complex" and prime cmpne Prime and UserPrime cmpeq false then
      vprint GaloisGroup, 1: "Changing prime to ", prime;
      Prime := prime;
      S`Base := pAdicRing(Prime);
      S`Prime := Prime;
      delete S`Roots;
      delete S`Ring;
    end if;

    if isEven then
      vprint GaloisGroup, 1: "An found";
      // Since An is a normal subgroup in Sn (the only index 2 subgroup)
      // the ordering of the roots does not matter (I think)
      r := [ GaloisRoot(i, S:Bound := 1) : i in [1..Degree(f)]];
      S`max_comp := Round(Abs(Eltseq(f)[1])+1);
      S`DescentChain := [<Alt(n), true>];
      G := Alt(n); 
    else
      vprint GaloisGroup, 1: "Sn found";
      r := [ GaloisRoot(i, S:Bound := 1) : i in [1..Degree(f)]];
      S`max_comp := Round(Abs(Eltseq(f)[1])+1);
      S`DescentChain := [<Sym(n), true>];
      G := Sym(n); 
    end if;
      if SOnly then
	return S;
      end if;
      return G, r, S;
  end if;

  if not SOnly and not Cyclo then
    vprint GaloisGroup, 1: "Compute starting group: ";
    vprint GaloisGroup, 2: "==========================";
    G,reject,aut:=GaloisStartingGroup(K,S:ShortOK := ShortOK,
                                          Current := Current);
    if S`Type ne "Complex"  and UserPrime cmpeq false and #FactoredOrder(G) eq 1 then
      vprint GaloisGroup, 1: "p-Group of order ", FactoredOrder(G), 
                             "found, looking for different prime...";
      prime := PrimeChoose(K, S:Subfields := Subfields, SmallDeg,
                                             NextPrime := NextPrime);
      if prime ne Prime then
        vprint GaloisGroup, 1: "Changing prime to ", prime;
        Prime := prime;
        S`Base := pAdicRing(Prime);
        S`Prime := Prime;
        delete S`Roots;
        delete S`Ring;
        G,reject,aut := GaloisStartingGroup(K, S: ShortOK := ShortOK);
      end if;
    end if;
    S`DescentChain := [<shallow_copy(G), true>];
    assert IsTransitive(G);
    vprint GaloisGroup, 2: "Start with group of size", FactoredOrder(G);
    if Degree(G) le 23 and GetVerbose("GaloisGroup") ge 2 then
       num:=TransitiveGroupIdentification(G);
       vprint GaloisGroup, 2: "and number ", num;
    end if;
    vprint GaloisGroup, 2: "==========================";
  else
    S`DescentChain := [];
  end if;

  r := [ GaloisRoot(i, S:Bound := 1) : i in [1..n]];
  //cyclo test...
  if Cyclo then
    R, mR := UnitGroup(Integers(CyclotomicOrder(K)));
    G := Sym(n);
    g := [];
    C, mC := ResidueClassField(Parent(r[1]));
    rt := [ GaloisRoot(i, S:Prec := 1)@mC : i in [1..n]];
    for i in Generators(R) do
      p := Integers()!(i@mR);
      Append(~g, [Position(rt, x^p) : x in rt]);
    end for;
    S`max_comp := 1;
    G := sub<G|g>;
    S`DescentChain := [<G, true>];
    S`KnownSubgroup := G;
    if SOnly then
      return S;
    end if;
    return G, [ GaloisRoot(i, S:Prec := 1) : i in [1..n]], S;
  end if;

  vprint GaloisGroup, 3: "Computed roots, do complex";
  if Maximum([Abs(x) : x in Eltseq(f)]) lt 2^400 then
    vtime GaloisGroup, 3: comp_roots:=Roots(f,ComplexField());
    if (not IsMonic(f)) or LCM([Denominator(x) : x in Eltseq(f)]) ne 1 then
      S`Scale := LCM([Denominator(x) : 
             x in Eltseq(Polynomial(Rationals(), f)/LeadingCoefficient(f))]);
      comp_roots := [ <x[1]*S`Scale, x[2]> : x in comp_roots];
    end if;
    vprint GaloisGroup, 3: "done\n";
    max_comp:=Ceiling(Maximum([Abs(x[1]):x in comp_roots]));
//f;
//comp_roots;
//max_comp;
    if max_comp gt 2^BitPrecision(RealField()) then
	//increase precision and try again
	comp_roots := Roots(f, ComplexField(Ceiling(Log(10, max_comp))));
	if assigned S`Scale then
	    comp_roots := [ <x[1]*S`Scale, x[2]> : x in comp_roots];
	end if;
	max_comp:=Ceiling(Maximum([Abs(x[1]):x in comp_roots]));
	assert max_comp le 2^BitPrecision(RealField(Ceiling(Log(10, max_comp))));
//max_comp;
    end if;
  else
    c := Eltseq(f);_n := #c;
    max_comp := Maximum([Abs(c[_n-i]/c[_n])^(1/i) : i in [1.._n-2]]);
    max_comp := 2*Maximum(max_comp, Abs(c[1]/2/c[_n])^(1/(_n-1)));
    max_comp := Ceiling(max_comp); //Fujiwara bound
    if (not IsMonic(f)) or LCM([Denominator(x) : x in Eltseq(f)]) ne 1 then
      S`Scale := LCM([Denominator(x) : 
             x in Eltseq(Polynomial(Rationals(), f)/LeadingCoefficient(f))]);
      max_comp *:= S`Scale;
    end if;
  end if;
  S`max_comp := max_comp;
//Compute known subgroup
  pr := 2;
  repeat
    r := [ GaloisRoot(i, S: Prec := pr) : i in [1..n]];
    if S`Type cmpeq "Complex" then
      rt := [ ComplexConjugate(x) : x in r];
      p := [ t where _, t := Minimum([Abs(rt[y]-r[x]) : y in [1..#rt]]) : x in [1..#r]];
    else
      rt := [ ChangePrecision(GaloisImage(x, 1), pr) : x in r];
      p := [ Position(rt, x) : x in r];
    end if;
    pr *:= 2;
    if pr gt 20 then
      error "cannot get Frobenius";
    end if;
  until #Set(p) eq n;  
  if #Set(p) eq n then
    S`KnownSubgroup := sub<Sym(n)|p>; // TODO: use automorphisms
    vprint GaloisGroup, 2: "Frobenius permutation: ", S`KnownSubgroup.1;
  else
    vprint GaloisGroup, 2: "prime mildly ramified, cannot compute Frobenius";
    //TODO: fix that!!!
  end if;

  if SOnly then
    return S;
  end if;

  if #Set(p) eq n then
    vprint GaloisGroup, 2: "Automorphisms: ",aut;
    if aut eq [] then
      knownsub := sub<G|p>;
    else
      knownsub:=sub<G|p, aut>;
    end if;
  elif aut eq [] then
    knownsub := sub<G|>;
  else
    knownsub:=sub<G|aut>;
  end if;

  vprint GaloisGroup, 2: "known subgroup ", knownsub,
                         " of order ", FactoredOrder(knownsub);

  if Subfields then
    return GenericStauduhar(G, S, max_comp, 
      func<X|[ x`subgroup : x in _MaxSub_(X:IsTransitive)]>,
      func<X|IsEven(X) eq isEven and 
             (Current or AllPartitions(G) eq AllPartitions(X)) and 
             TestShapes(X, shapes)>:
        ShortOK := ShortOK, Subfields := Subfields, 
        SubfieldsComplete := not Current,  
        Reject:=reject,Subgrp:=knownsub);
  else
    return GenericStauduhar(G, S, max_comp, 
      func<X|[ x`subgroup : x in _MaxSub_(X:IsTransitive)]>,
      func<X|IsEven(X) eq isEven and TestShapes(X, shapes)>:
        ShortOK := ShortOK, Subfields := Subfields, Reject:=reject,Subgrp:=knownsub);
  end if;    
end function;

intrinsic GaloisGroup(K::FldNum[FldRat]: 
  Ring := false, Type := "p-Adic", Prime := false, Prec := false, 
  ShortOK := false, Subfields := true, Current := false, Old := false,
  Time := false, NextPrime := My_np)
                           -> Grp, [], GaloisData
  {Computes the galois group of (the normal closure) of the field.}

  require IsSimple(K): "The field must be a simple extension of the rationals";

  if Old then
    if Type eq "p-Adic" then
      Al := "pAdic";
    else
      Al := "Real";
    end if;
    return InternalGaloisGroup(K: Al := Al);
  end if;

  return GetGaloisDataOrGroup(K : Ring := Ring, Type := Type, Prime := Prime, Prec := Prec, ShortOK := ShortOK, Subfields := Subfields, Current := Current, Time := Time, NextPrime := NextPrime, SOnly := false);

end intrinsic;

intrinsic GaloisData(K::FldNum[FldRat]: 
  Prime := false, NextPrime := My_np, Subfields := true)
                           -> GaloisData
  {Computes the data structure of the galois group of (the normal closure) of the field.}

  require IsSimple(K): "The field must be a simple extension";
  return GetGaloisDataOrGroup(K : Prime := Prime, NextPrime := NextPrime, SOnly := true, Subfields := Subfields);

end intrinsic;

intrinsic GaloisGroup(f::RngUPolElt[FldRat]: Ring := false, 
          Type := "p-Adic", Prime := false, Prec := false, 
          ShortOK := false, Old := false, Time := false,
          NextPrime := My_np)
            -> Grp, [], GaloisData
  {Computes the galois group of f}
  d := Lcm([Denominator(x) : x in Eltseq(f)]);
  return GaloisGroup(Polynomial(Integers(), f*d) : Type := Type,
    Ring := Ring, Prime := Prime, ShortOK := ShortOK, NextPrime := NextPrime,
      Type := Type, Prec := Prec);

end intrinsic;

intrinsic GaloisGroup(f::RngUPolElt[FldAlg]: Ring := false, 
          Prime := false, ShortOK := false, NextPrime := My_np)
            -> Grp, [], GaloisData
  {"}
  d := Lcm([Denominator(x) : x in Eltseq(f)]);
  return GaloisGroup(Polynomial(Integers(CoefficientRing(f)), f*d) :
    Ring := Ring, Prime := Prime, ShortOK := ShortOK, NextPrime := NextPrime);

end intrinsic;

forward galois_group, galois_data_copy_for_factor;

intrinsic GaloisGroup(f::RngUPolElt[RngInt]: Ring := false, Type := "p-Adic", 
          Prime := false, ShortOK := false, NextPrime := My_np, Prec := 40)
            -> Grp, [], GaloisData
  {Compute the Galois group of the splitting field of the monic and squarefree
    polynomial.}

  require Degree(f) gt 0: "Polynomial must not be constant";
    return galois_group(f, Ring, Prime, ShortOK, NextPrime:Type := Type, Prec := Prec);
end intrinsic;

intrinsic GaloisGroup(f::RngUPolElt[RngOrd[RngInt]]: Ring := false, 
          Prime := false, ShortOK := false, NextPrime := My_np)
            -> Grp, [], GaloisData
  {Compute the Galois group of the splitting field of the monic and squarefree
    polynomial.}

  require Degree(f) gt 0: "Polynomial must not be constant";
    return galois_group(f, Ring, Prime, ShortOK, NextPrime);
end intrinsic;

intrinsic GaloisGroup(f::RngUPolElt[RngOrd]: Ring := false, 
          Prime := false, ShortOK := false, NextPrime := My_np)
            -> Grp, [], GaloisData
  {Compute the Galois group of the splitting field of the monic and squarefree
    polynomial.}

  require Degree(f) gt 0: "Polynomial must not be constant";
  A := AbsoluteOrder(CoefficientRing(f));
  f := Polynomial(A, f);
  G, R, S := galois_group(f, Ring, Prime, ShortOK, NextPrime);

  return G, R, S;

end intrinsic;

function get_prime(f, llf, lf, ff, next_prime)

      p := Degree(f);
      lp := [];
      np := Maximum(20, Degree(f));
      k := CoefficientRing(f);

    if CoefficientRing(f) cmpeq Integers() then
    fs := &* [x[1] : x in llf | Degree(x[1]) gt 0];
      d := LeadingCoefficient(f);
      d *:= Discriminant(fs);
      while #lp lt 1 and np gt 0 do
        repeat
          p := next_prime(p);
        until d mod p ne 0;
        np -:= 1;
        c :=  Factorization(Polynomial(GF(p), ff));
        if exists{x : x in c | x[1] eq Parent(x[1]).1} then
         //we want the linear factors to not vanish
         //Seems like a good idea....
          continue;
        end if;
        c := [Integers()| Degree(x[1]) : x in c];
        c := Sort(c);
        if c notin [x[2] : x in lp] or lp eq [] then
          Append(~lp, <p, c, Lcm(c)>);
        end if;
      end while;
      vprint GaloisGroup, 2: "Found some possible primes: ", lp;
    else 
      f := Polynomial(FieldOfFractions(k), f);
      while #lp lt Maximum(1, Degree(ff) div 5) and np gt 0 do
        repeat
          p := next_prime(p);
        until Valuation(Discriminant(EquationOrder(k)), p) eq 0;
	lc := Lcm([Denominator(x) : x in Eltseq(f/LeadingCoefficient(f))]);
	repeat
	    p := next_prime(p);
	until Valuation(lc, p) eq 0 and Valuation(Discriminant(EquationOrder(k)), p) eq 0;
	P := Decomposition(MaximalOrder(k), p);
	if forall(x){x : x in P | Degree(x[1]) gt Maximum(Degree(k) div 3, 1)} then
	    continue;
	end if;
	for j in [1 .. #P] do
	    F, mF := ResidueClassField(P[j][1]);
	    fact := Factorization(Polynomial([mF(x) : x in Eltseq(ff)]));
	    if exists(x){x : x in fact | x[2] gt 1} then
		for LP in lp do
		    if Minimum(LP[1]) eq p then
			Exclude(~lp, LP);
		    end if;
		end for;
		break;
	    end if;
	    c := [Integers() | Degree(x[1]) : x in fact];
	    Append(~lp, <P[j][1], c, Lcm(c)>);
	end for;
      end while;
   end if;
  l := Maximum([Degree(x[1]) : x in lf] cat [1]) div 6;
  u := Maximum([Degree(x[1]) : x in lf] cat [1]) div 2;
  if not exists(x){x : x in lp | x[3] in [l..u]} then
    _, x := Minimum([x[3] : x in lp]);
    x := lp[x];
  end if;
  Prime := x[1];
  vprint GaloisGroup, 1: "Using Prime", Prime, " of shape ", x[2];
  return Prime;
end function;

forward galois_group_for_all;

function galois_group(f, ring, prime, ShortOK, next_prime:Type := "p-Adic", Prec := 40)

    if IsIrreducible(Polynomial(FieldOfFractions(CoefficientRing(f)), f)) then  
	K:=NumberField(Polynomial(FieldOfFractions(CoefficientRing(f)), f):Check := false, DoLinearExtension);
	// Prec only can be set in the absolute case
	if CoefficientField(K) eq Rationals() then
	    return GaloisGroup(K: Ring := ring, Prime := prime, Type := Type,
			    ShortOK := ShortOK, NextPrime := next_prime,
                            Prec := Prec
			    );
	else
	    return GaloisGroup(K: Ring := ring, Prime := prime, Type := Type,
			    ShortOK := ShortOK, NextPrime := next_prime);
	end if;
    end if;
    if Type ne "p-Adic" then
      error "For reducible polynomials, the Type cannot be set";
    end if;  


    if CoefficientRing(f) cmpne Integers() and not IsMonic(f) then
	lllf, lc := Factorization(Polynomial(FieldOfFractions(CoefficientRing(f)), f));
	llf := [];
	Px := PolynomialRing(CoefficientRing(f));
	for lf in lllf do
	    if IsCoercible(Px, lf[1]) then
		Append(~llf, <Px!lf[1], lf[2]>);
	    else
		Append(~llf, <Px!(lc*lf[1]), lf[2]>);
	    end if;
	end for;
	assert #llf eq #lllf;
    else
	llf, lc := Factorization(f);
    end if;
    lf := [car<Parent(f), Integers()>|x : x in llf | Degree(x[1]) gt 1];  // we'll add the linears later
          // also, since the multiplicity is irrelevant for the group, we
          // deal with this also later..

    ff := &* [Parent(f) | x[1] : x in lf];

    vprintf GaloisGroup, 1: "\n%oInatransitive case!%o\n", green, normal;
    vprint GaloisGroup, 1: 
                  green, "computing Galois groups of factors...", normal;
    IndentPush();

    if prime cmpeq false then  
	prime := get_prime(f, llf, lf, ff, next_prime);
    end if;  

    return galois_group_for_all(f, prime, ShortOK, lf, llf, ff);
end function;

function galois_group_for_all(f, prime, ShortOK, lf, llf, ff);

    lG := [];
    if #lf gt 0 then
	G, _, S := GaloisGroup(lf[1][1] : Prime := prime, ShortOK := ShortOK);
	if assigned S`Permutation then    
	    Append(~lG, <G^S`Permutation, S>);
	else
	    Append(~lG, <G, S>);
	end if;

	Sall := InternalGaloisGroupProductSetup(S, ff, lf);

	for i in lf[2 .. #lf] do
	  G, _, S := GaloisGroup(i[1]:Prime := prime, Ring := galois_data_copy_for_factor(Sall, i[1]),
	      ShortOK := ShortOK);
	  f2 := Polynomial(S`Ring, [S`BaseMap(x, 5) : x in Eltseq(f)]);
//"f2 = ", f2;
//"S`Roots = ", S`Roots;
	  assert &and[Valuation(Evaluate(f2, x)) gt 0 : x in S`Roots];
	  //S`Roots;
	  if assigned S`Permutation then    
	    Append(~lG, <G^S`Permutation, S>);
	  else
	    Append(~lG, <G, S>);
	  end if;
	end for;
    end if;

    IndentPop();

    lin_p := 0;
    if lG ne [] then
      G, R, S := InternalGaloisGroupProduct(Sall, lG, ff, lf:ShortOK := ShortOK);
      //assert not assigned S`GetRoots;
    else
      if exists(i){i : i in [1..#llf] | Degree(llf[i][1]) gt 0} then
        G, R, S := GaloisGroup(llf[i][1]:Prime := prime);
        lin_p := i;
      else
        error "Polynomial must not be constant";
      end if;
    end if;

    if assigned S`BaseMap then
	f1 := Polynomial([S`Ring!S`BaseMap(x, 1) : x in Eltseq(f)]);
    else
	f1 := Polynomial(S`Ring, f);
    end if;
//[Evaluate(f1, x) : x in S`Roots];
    assert &and[Valuation(Evaluate(f1, x)) gt 0 : x in S`Roots];
    /*
    Handle multiple roots and linear factors
    */
    F, mF := ResidueClassField(S`Ring);
    r := [mF(x) : x in S`Roots];
    i := 0;
    pr := 1;
    nG := [G];
    lm := [*hom<G -> G | [G.i : i in [1..Ngens(G)]]>*];
    for j in [1..#llf] do
      if Degree(llf[j][1]) eq 0 then
        continue;
      end if;
      if Degree(llf[j][1]) gt 1 then
        i +:= 1;
	pr +:= Degree(llf[j][1]);
        assert llf[j][1] eq lf[i][1];
      end if;
      if Degree(llf[j][1]) eq 1 then
        if lG eq [] and j eq lin_p then
          s := 2;
        else
          s := 1;
        end if;
        for k in [s..llf[j][2]] do
	  if assigned S`BaseMap then
	      rfb, mrfb := ResidueClassField(S`Base);
	      Append(~r, Roots(Polynomial(F, Polynomial(rfb, [mrfb(S`BaseMap(c, 1)) : c in Coefficients(llf[j][1])])), F)[1][1]);
	  else
	      Append(~r, Roots(Polynomial(F, Polynomial(ResidueClassField(S`Base), Coefficients(llf[j][1]))), F)[1][1]);
	  end if;
          Append(~nG, Sym(1));
          Append(~lm, hom<G-> Sym(1)|[Sym(1).0 : l in [1..Ngens(G)]]>);
	  // Fix max_comp : but HOW? This depends on the coeff ring!
	  // Get a GaloisData for each?
	  _, _, sl := GaloisGroup(llf[j][1] : Prime := prime, Ring := galois_data_copy_for_factor(S, llf[j][1]));
	  S`max_comp := Maximum(S`max_comp, sl`max_comp);
        end for;
      elif llf[j][2] ge 2 then
	pr -:= Degree(llf[j][1]);
        for k in [2..llf[j][2]] do
          r cat:= r[pr..pr+Degree(llf[j][1])-1];
          Append(~nG, lG[i][1]);
          Append(~lm, S`Maps[2][i]);
        end for;
	pr +:= Degree(llf[j][1])*llf[j][2];
      end if;
//f1 := Polynomial([S`Ring!S`BaseMap(x, 1) : x in Eltseq(f)]);
//[Valuation(Evaluate(f1, x)) : x in r];
//assert &and[Valuation(Evaluate(f1, x)) gt 0 : x in r];
    end for;
    nG, inc := DirectProduct(nG);
    h := hom<G -> nG | [&* [ (G.i)@lm[x]@inc[x] : x in [1..#lm] ] : i in [1..Ngens(G)]]>;
    S`Roots := ChangeUniverse(r, S`Ring);
//S`Roots;
    r := [ChangePrecision(x, 1) : x in S`Roots];
    if assigned S`BaseMap then
	f1 := Polynomial([S`Ring!S`BaseMap(x, 1) : x in Eltseq(f)]);
    else
	f1 := Polynomial(S`Ring, f);
    end if;
//[Valuation(Evaluate(f1, x)) : x in S`Roots];
    assert &and[Valuation(Evaluate(f1, x)) gt 0 : x in S`Roots];
    // Need to make sure S`Roots satisfy the hensel condition
    if assigned S`BaseMap then
    g := Polynomial([S`BaseMap(c, 5) : c in Coefficients(f)]);
    dg := Derivative(g);
    pr := 1;
//r;
    //assert &and[Valuation(Evaluate(g, x)) gt 2*Valuation(Evaluate(dg, x)) : x in r]; 
    while false and &or[Valuation(Evaluate(g, x)) le 2*Valuation(Evaluate(dg, x)) : x in r] do
	pr +:= 1;
//pr;
	r := [ChangePrecision(x, pr) : x in S`Roots];
//r;
	if pr gt AbsolutePrecision(LeadingCoefficient(g)) then
	    g := Polynomial([S`BaseMap(c, pr + 5) : c in Coefficients(f)]);
	    dg := Derivative(g);
//g;
	end if;
//[<Valuation(Evaluate(g, x)), 2*Valuation(Evaluate(dg, x))> : x in r];
    end while;
    end if;
    S`Roots := r;
//S`Roots;
    // need to overwrite GetRoots to handle the linears and the multiplicities
    if assigned S`GetRoots and Degree(f) gt Degree(ff) then
	old_get_roots := S`GetRoots;
	procedure GetRoots(SS : Prec := 20)
	    assert #SS`Roots eq Degree(f);
	    if Precision(Representative(SS`Roots)) ge Prec then
		return;
	    end if;

	    //removes roots from linear factors but lifts the rest
	    old_get_roots(SS : Prec := Prec);
	    assert #SS`Roots eq Degree(ff);
	    assert #SS`Roots lt Degree(f);

	    //replace roots from linear factors 
	    pr := 1;
	    for j in [1 .. #llf] do
		if Degree(llf[j][1]) eq 1 then
		    /*
		    Think this is wrong and final assertion is wrong
		    if lG ne [] or j ne lin_p then
			Append(~SS`Roots, Roots(Polynomial(S`Ring, [S`BaseMap(c, Prec) : c in Coefficients(llf[j][1])]))[1][1]);
		    end if;
		    */
		    if lG eq [] and j eq lin_p then
		      s := 2;
		    else
		      s := 1;
		    end if;
		    for k in [s .. llf[j][2]] do
			Append(~SS`Roots, Roots(Polynomial(S`Ring, [S`BaseMap(c, Prec) : c in Coefficients(llf[j][1])]))[1][1]);
		    end for;
		elif llf[j][2] ge 2 then
		    for k in [2 .. llf[j][2]] do
			SS`Roots cat:= SS`Roots[pr .. pr + Degree(llf[j][1]) - 1];
		    end for;
		    pr +:= Degree(llf[j][1])*llf[j][2];
		else
		    pr +:= Degree(llf[j][1]);
		end if;
	    end for;
	    //assert #SS`Roots eq Degree(SS`f);
	end procedure;
	S`GetRoots := GetRoots;
    end if;
    S`f := &* [ x[1] : x in llf];
    assert #Image(h) eq #G;
    G := Image(h);
    S`DescentChain := [<shallow_copy(G), true>];
    return Image(h), S`Roots, S;
end function;

intrinsic InternalGaloisGroupProductSetup(S1::GaloisData, f::RngUPolElt, lf::[]) -> GaloisData
{}
    Prime := S1`Prime;

    S := InternalGaloisDataCreate(Prime);
    S`Base := S1`Base;
    assert not assigned S`BaseMap;
    if not assigned S1`BaseMap then
    // Was using ExtendedReals here as the second factor in the CartesianProduct
    // but it caused problems in the function field cases and I can't find an 
    // example which actually needs it ... yet.
	base_map := map<CartesianProduct(FieldOfFractions(CoefficientRing(f)), Integers()) -> S1`Base | t :-> S1`Base!ChangePrecision(S1`Base, t[2])!t[1]>;
    else 
	if S1`Type eq "p-Adic (FldFun over Q)" then
	    base_map := map<CartesianProduct(FieldOfFractions(CoefficientRing(f)), CartesianProduct(Integers(), Integers())) -> S1`Ring | t :-> S1`BaseMap(t[1], t[2])>;
	else
	    base_map := map<CartesianProduct(FieldOfFractions(CoefficientRing(f)), Integers()) -> S1`Ring | t :-> S1`BaseMap(t[1], t[2])>;
	end if;
    end if;

    S`BaseMap := base_map;
    S`Ring := SplittingField(f, base_map);
    S`Type := S1`Type;
    if assigned S1`Prec then
	S`Prec := S1`Prec;
    end if;
    //S`Scale := LCM([LeadingCoefficient(v[1]) : v in lf]);
    if assigned S1`IsInt then
	function IsInt(x, B, SS : EasyNonInt := false)
	    pr := AbsolutePrecision(x);
	    if pr eq 0 then
	      return true, 0;
	    end if;
	    f, c := IsCoercible(SS`Base, x);
	    if not f then
	      vprint GaloisGroup, 1: "no int - ring error";
	      return false, _;
	    end if;
	    return S1`IsInt(x, B, S1 : EasyNonInt := EasyNonInt);
	end function;
	S`IsInt := IsInt;
    end if;

    if assigned S1`max_comp then
	S`max_comp := S1`max_comp;
    end if;

    // Need this to store the prime
    if assigned S1`S then
	S`S := S1`S;
    end if;

    // This is fine since KQ (and hence the mapping is based on the prime
    // which is common among the factors)
    if assigned S1`mKQ then
	S`mKQ := S1`mKQ;
    end if;
    return S;

end intrinsic;

function galois_data_copy_for_factor(S, f)
    SC := InternalGaloisDataCreate(S`Prime);
    SC`f := f;
    SC`Base := S`Base;
    // this will get overwritten so don't bother confusing self by setting here
    if not false and assigned S`BaseMap then
	SC`BaseMap := S`BaseMap;
    end if;
    if ISA(Type(S`Ring), RngSer) or BaseRing(BaseRing(S`Ring)) eq PrimeRing(S`Ring) then
	SC`Ring := S`Ring;
    else
	SC`Ring := UnramifiedExtension(BaseRing(BaseRing(S`Ring)), InertiaDegree(S`Ring)*InertiaDegree(BaseRing(S`Ring)));
    end if;
    SC`Type := S`Type;
    if assigned S`Prec then
	SC`Prec := S`Prec;
    end if;
    //SC`Scale := S`Scale;

    // NOT SURE ABOUT THIS, ARE THEY REALLY THE SAME? NO, BUT WHAT DO I DO? I
    // NEED TO SET IT TO SOMETHING!
    // RECOMPUTE HERE AS IN QtComputeRoots
    if assigned S`S then
    end if;

    if assigned S`mKQ then
	SC`mKQ := S`mKQ;
	ff := Polynomial([S`mKQ(x) : x in Eltseq(f)]);
	assert IsSquarefree(ff);
	KK := NumberField(ff : Check := false);
	SC`S := GaloisData(KK: Prime := S`S`Prime);
    end if;

    return SC;
end function;

intrinsic InternalGaloisGroupProduct(S::GaloisData, lG::[],f::RngUPolElt, lf::[] : ShortOK := false, Time := false) -> GrpPerm, [], GaloisData
  {}

    require forall{x : x in lG | x[2]`Prime eq lG[1][2]`Prime} :
      "Primes must agree";
    Prime := lG[1][2]`Prime;
    F, mF := ResidueClassField(S`Ring);

    /*
    Split into factors with disjoint splitting fields to all others
    and factors with possibly non-disjoint splitting fields to some others
    */
    /*
    If the group orders = degree of splitting fields are pairwise coprime then 
    there can be no overlapping of splitting fields.
    */
//[Order(x[1]) : x in lG];
    maybe_coprime := {x : x in [1..#lG], y in [1..#lG] | x ne y and GCD(Order(lG[x][1]), Order(lG[y][1])) ne 1}; 
//maybe_coprime;

    if #maybe_coprime gt 0 then

	if Type(S`Prime) eq RngIntElt then
	    d := [i in maybe_coprime select Discriminant(lf[i][1]) else 1 : i in [1 .. #lf]];
//d;
	    non_coprime := {x : x in maybe_coprime, y in maybe_coprime | x ne y and GCD(d[x], d[y]) ne 1};
	elif Type(S`Prime) eq RngUPolElt and Characteristic(F) gt 0 then
	    /* 
	    A PREVIOUSLY TROUBLESOME EXAMPLE
	    k := GF(17);
	    K<a> := FunctionField(k);
	    _<y> := PolynomialRing(K);
	    f := y^2 + 3;
	    f *:= Polynomial(K, [a, 0, 0, 0, 0, 1]);
	    G, R, S := GaloisGroup(f);

	    something to do with the exact constant fields of extensions
	    by each factor - could be expensive!!!
	    Otherwise if the GCD of the discriminants is 1 then we need to avoid 
	    the quick check
	    */

	    /*
	    What we need here is to know whether the intersection of the splitting
	    fields is a constant field (unramified) extension 
	    But how do we determine what the constant field extensions in 
	    the splitting fields are?
	    And we have no hope except over a rational function field
	    Only need to bother with this if the discriminants are coprime
	    If the discriminants are not coprime then cannot do anything here
	    */

	    /*
	    I THINK WE STILL NEED TO CHECK THE DISCRIMINANTS ARE PAIRWISE 
	    COPRIME HERE
	    THOSE THAT ARE NOT PAIRWISE COPRIME COULD INTERSECT IN A
	    RAMIFIED EXTENSION WHICH WE DON'T CHECK FOR BELOW
	    WE NEED TO CONTINUE HERE ONLY FOR THOSE FACTORS WHOSE 
	    DISCRIMINANTS ARE COPRIME
	    */
	    d := [i in maybe_coprime select Discriminant(lf[i][1]) else 1 : i in [1 .. #lf]];
//d;
	    disc_non_coprime := {x : x in maybe_coprime, y in maybe_coprime | x ne y and GCD(d[x], d[y]) ne 1};
	    //maybe_coprime diff:= disc_non_coprime;
//"Maybe coprime after disc check : ", maybe_coprime;

	    /*
	    Check whether there is a possibility of an unramified
	    equivalently constant field extension
	    IN THE INTERSECTION OF SPLITTING FIELDS (but first in the intersection
	    of stem fields)
	    */
	    de := [i in maybe_coprime select DimensionOfExactConstantField(FunctionField(lf[i][1])) else 1: i in [1 .. #lf]];
//de;
	    /*
	    really_non_coprime should be those whose stem fields possibly contain
	    an unramified/constant field extension
	    Remove these from non_coprime for the moment (put back in later)
	    and check whether the rest may have an unramified extension in the
	    splitting field
	    */
	    decfst_non_coprime := {x : x in maybe_coprime, y in maybe_coprime | x ne y and GCD(de[x], de[y]) ne 1};
	    non_coprime := disc_non_coprime join decfst_non_coprime;
	    maybe_coprime diff:= non_coprime;
//"Maybe coprime after stem check : ", maybe_coprime;

	    if #maybe_coprime gt 0 then
		/*
		No constant field extension in intersection of stem fields, what 
		about intersection of splitting fields?
		*/
//[CyclicSubgroups(x[1]) : x in lG];
		/* 
		constant field extension will be cyclic in char p
		but its Galois group will be a normal subgroup of
		the whole Galois group (of the factor) whose quotient
		in the whole Galois group is cyclic with the same order
		*/
		cs := [x in maybe_coprime join non_coprime select CyclicSubgroups(lG[x][1]) else [] : x in [1 .. #lG]];
		maybe_coprime := {x : x in maybe_coprime | #cs[x] ne 0};
//"After cyclic subgroup check : ", maybe_coprime;

		Fq := CoefficientRing(S`Prime);
		still_non_coprime := {};
		non_coprime_tested := {};
		for x in maybe_coprime do
		    r1 := LCM([Integers() | r`order : r in cs[x]]);
		    Fqr1 := ext<Fq | r1>;
		    Include(~non_coprime_tested, x);
//(maybe_coprime diff non_coprime_tested) join non_coprime;
		    for y in (maybe_coprime diff non_coprime_tested) join non_coprime do
			r2 := LCM([Integers() | r`order : r in cs[y]]);
			Fqr2 := ext<Fq | r2>;
			if GCD(r1, r2) ne 1 then
			    still_non_coprime join:= {x, y};
			end if;
			if Fqr1 meet Fqr2 ne Fq then
		    	    still_non_coprime join:= {x, y};
			end if;
		    end for;
		end for;
//still_non_coprime;
		maybe_coprime := still_non_coprime diff non_coprime;

		// Maybe only do this if there is a small proportion of factors 
		// in non_coprime since the normal subgroups could be expensive
		ns := [x in maybe_coprime join non_coprime select NormalSubgroups(lG[x][1]) else [] : x in [1 .. #lG]];
		still_non_coprime := {};
		non_coprime_tested := {};
		for x in maybe_coprime do
		    cqx := [r : r in ns[x] | IsCyclic(lG[x][1]/r`subgroup)];
		    cqxo := [r`order : r in cqx];
		    fix_fieldsx := [GaloisSubgroup(lG[x][2], r`subgroup) : r in cqx];
		    ecf_dim_x := LCM([DimensionOfExactConstantField(FunctionField(f)) : f in fix_fieldsx]);
		    Include(~non_coprime_tested, x);
//(maybe_coprime diff non_coprime_tested) join non_coprime;
		    for y in (maybe_coprime diff non_coprime_tested) join non_coprime do
			cqy := [r : r in ns[y] | IsCyclic(lG[y][1]/r`subgroup)];
			cqyo := [r`order : r in cqy];
			fix_fieldsy := [GaloisSubgroup(lG[y][2], r`subgroup) : r in cqy];
			ecf_dim_y := LCM([DimensionOfExactConstantField(FunctionField(f)) : f in fix_fieldsy]);
//cqxo, cqyo;
//ecf_dim_x, ecf_dim_y;
			if GCD(ecf_dim_x, ecf_dim_y) ne 1 and GCD(LCM(cqxo), LCM(cqyo)) ne 1 then 
			    still_non_coprime join:= {x, y};
			end if;
		    end for;
		end for;
		maybe_coprime := still_non_coprime;
	    end if;
	    non_coprime join:= maybe_coprime;
	else
	    non_coprime := maybe_coprime;
	end if;	// Coefficient ring
    else
	non_coprime := maybe_coprime;
    end if; //#maybe_coprime gt 0
if not IsExport() then
"ne Final non_coprime : ", non_coprime;
end if;

    S`Prec := 1;
    if #non_coprime gt 0 then
	Snc := galois_data_copy_for_factor(S, &*[lf[x][1] : x in non_coprime]);
	if assigned S`IsInt then
	    Snc`IsInt := S`IsInt;
	end if;
    end if;
    S`f := f;
    vprint GaloisGroup, 1: green, "computing starting group", normal;

    if assigned lG[1][2]`GetRoots then
      procedure GetRoots(SS:Prec := 20)
        if Precision(Representative(SS`Roots)) ge Prec then
	  return;
        end if;
	// This needs to be done here for the reducible case because of the 
	// different splitting field setup to the irreducible case
        SetPrecision(SS`Ring, Prec);
	  r := [];
	  start_roots := 1;
	  //assert SS`Scale eq 1;
	  for i in [1..#lG] do
	    /*
	    What block in SS`Roots are being lifted? make this r1
	    F1 := ResidueClassField(lG[i][2]`Ring);
	    [SS`Ring|F!F1!x : x in lG[i][2]`Roots];
	    */
	    r1 := SS`Roots[start_roots .. start_roots + Degree(lf[i][1]) - 1];
	    start_roots +:= Degree(lf[i][1]);
	    f1 := Polynomial([S`Ring!S`BaseMap(x, Prec) : x in Eltseq(lf[i][1])]);
	    //f1 := Polynomial([lG[i][2]`Ring!lG[i][2]`BaseMap(x, Prec) : x in Eltseq(lf[i][1])]);
//[Valuation(Evaluate(f1, x)) : x in SS`Roots];
	    r1 := [HenselLift(f1, x, Prec) : x in r1];
	    r cat:= r1;
	  end for;
	    SS`Roots := [SS`Ring|ChangePrecision(SS`Ring!x, Prec) : x in r];
	    f2 := Polynomial([S`Ring!S`BaseMap(x, Prec) : x in Eltseq(f)]);
	    assert &and[Valuation(Evaluate(f2, x)) gt 0: x in SS`Roots];
      end procedure;
      S`GetRoots := GetRoots;
      if #non_coprime gt 0 then
	  procedure GetRoots(SS:Prec := 20)
	    if Precision(Representative(SS`Roots)) ge Prec then
	      return;
	    end if;
	    // This needs to be done here for the reducible case because of the 
	    // different splitting field setup to the irreducible case
	    SetPrecision(SS`Ring, Prec);
	      r := [];
	      start_roots := 1;
	      //assert SS`Scale eq 1;
	      for i in [1..#lG] do
		if i in non_coprime then
		    /*
		    What block in SS`Roots are being lifted? make this r1
		    F1 := ResidueClassField(lG[i][2]`Ring);
		    [SS`Ring|F!F1!x : x in lG[i][2]`Roots];
		    */
		    r1 := SS`Roots[start_roots .. start_roots + Degree(lf[i][1]) - 1];
		    start_roots +:= Degree(lf[i][1]);
		    f1 := Polynomial([S`Ring!S`BaseMap(x, Prec) : x in Eltseq(lf[i][1])]);
		    //f1 := Polynomial([lG[i][2]`Ring!lG[i][2]`BaseMap(x, Prec) : x in Eltseq(lf[i][1])]);
    //[Valuation(Evaluate(f1, x)) : x in SS`Roots];
		    r1 := [HenselLift(f1, x, Prec) : x in r1];
		    r cat:= r1;
		else
		    start_roots +:= Degree(lf[i][1]);
		end if;
	      end for;
		SS`Roots := [SS`Ring|ChangePrecision(SS`Ring!x, Prec) : x in r];
		f2 := Polynomial([S`Ring!S`BaseMap(x, Prec) : x in Eltseq(f)]);
		assert &and[Valuation(Evaluate(f2, x)) gt 0: x in SS`Roots];
	  end procedure;
	  Snc`GetRoots := GetRoots;
      end if;
    end if;  
    if assigned lG[1][2]`GetPrec then
	//S`GetPrec := func<B, S|Maximum([i[2]`GetPrec(B, i[2]) : i in lG])>;
	function GP(B, S)
	    precs := {i[2]`GetPrec(B, i[2]) : i in lG};
	    assert #precs eq 1;
	    return Rep(precs);
	end function;
	S`GetPrec := GP;
	if #non_coprime gt 0 then
	function GP(B, S)
	    precs := {lG[i][2]`GetPrec(B, lG[i][2]) : i in non_coprime};
	    assert #precs eq 1;
	    return Rep(precs);
	end function;
	Snc`GetPrec := GP;
	end if;
    end if;
    if assigned lG[1][2]`Bound then
	S`Bound := func<t, inv, idx:LowTest := false|Maximum([i[2]`Bound(t, inv, idx : LowTest := LowTest) : i in lG])>;
	if #non_coprime gt 0 then
	Snc`Bound := func<t, inv, idx:LowTest := false|Maximum([lG[i][2]`Bound(t, inv, idx : LowTest := LowTest) : i in non_coprime])>;
	end if;
    end if;
    scales := [x[2]`Scale : x in lG | assigned x[2]`Scale];
    S`Scale := #scales eq 0 select 1 else LCM(scales);
    if #non_coprime gt 0 then 
    scales := [lG[x][2]`Scale : x in non_coprime | assigned lG[x][2]`Scale];
    Snc`Scale := #scales eq 0 select 1 else LCM(scales);
    end if;
      
    pr := 1;
    r := [];
    rnc := [];
    for i in [1 .. #lG] do
      F1, mF1 := ResidueClassField(lG[i][2]`Ring);
      Embed(F1, F);
//"Product root stuff";
//ResidueClassField(CoefficientRing(lG[i][2]`Ring));
      r1 := [F|mF1(x) : x in lG[i][2]`Roots];
      //check finite field coercion
      assert &and[Evaluate(MinimalPolynomial(mF1(x)), F!mF1(x))  eq 0 : x in lG[i][2]`Roots];
      //check coercion back into p-adic
      assert &and[Valuation(Evaluate(Polynomial(S`Ring, MinimalPolynomial(mF1(x))), S`Ring!F!mF1(x))) gt 0 : x in lG[i][2]`Roots];
      if assigned lG[i][2]`BaseMap then
	  f1 := Polynomial([lG[i][2]`Ring!lG[i][2]`BaseMap(x, pr) : x in Eltseq(lf[i][1])]);
	  assert &and[Valuation(Evaluate(f1, x)) gt 0: x in lG[i][2]`Roots];
	  f1 := Polynomial(F, Polynomial(F1, [mF1(x) : x in Eltseq(f1)]));
	  assert &and[Evaluate(f1, mF1(x)) eq 0: x in lG[i][2]`Roots];
	  assert &and[Evaluate(f1, mF1(x)) eq 0: x in r1];
//"lG[i][2]`Roots ok", i;
      end if;
      f2 := Polynomial([S`Ring!S`BaseMap(x, pr) : x in Eltseq(f)]);
      if assigned S`Roots then
	  assert &and[Valuation(Evaluate(f2, x)) gt 0: x in S`Roots];
	  assert &and[Valuation(Evaluate(f2, x)) gt 0: x in lG[i][2]`Roots];
      end if;
      assert &and[Valuation(Evaluate(f2, S`Ring!x)) gt 0: x in r1];
      assert &and[Valuation(Evaluate(f2, S`Ring!x)) gt 0: x in ChangeUniverse(r1, F1)];
      f2 := Polynomial([S`Ring!S`BaseMap(x, pr) : x in Eltseq(lf[i][1])]);
      //Can't coerce from lG[i][2]`Ring to S`Ring
      //assert &and[Valuation(Evaluate(f2, S`Ring!x)) gt 0: x in lG[i][2]`Roots];
      assert &and[Valuation(Evaluate(f2, S`Ring!x)) gt 0: x in r1];
      if ISA(Type(Domain(mF1)), RngSer) then
	  f2 := Polynomial(F, [mF1(c) : c in Eltseq(f2)]);
	  assert &and[Evaluate(f2, x) eq 0: x in r1];
      else
	  assert &and[Evaluate(Polynomial(F, f2), x) eq 0: x in r1];
	  f2 := Polynomial(F, f2);
      end if;
      if assigned lG[i][2]`BaseMap then
	  assert f1 eq f2;
      end if;
//"r1 roots ok", i;
      r cat:= r1;
      if i in non_coprime then
	rnc cat:= r1;
      end if;
    end for;
    while not assigned S`GetRoots and #Set(r) ne Degree(f) do
      pr +:= 2;
      r := [];
      rnc := [];
      for i in [1 .. #lG] do
	F1 := ResidueClassField(lG[i][2]`Ring);
	r1 := [S`Ring|F!F1!x : x in lG[i][2]`Roots];
	f1 := Polynomial(S`Ring, lG[i][2]`f);
	r1 := [HenselLift(f1, x, pr) : x in r1];
	r cat:= r1;
	if i in non_coprime then
	    rnc cat:= r1;
	end if;
      end for;
    end while;
    S`Roots := [S`Ring|ChangePrecision(S`Ring!x, pr) : x in r];
    if #non_coprime gt 0 then
    Snc`Roots := [Snc`Ring|ChangePrecision(Snc`Ring!x, pr) : x in rnc];
    end if;
    f1 := Polynomial([S`Ring!S`BaseMap(x, pr) : x in Eltseq(f)]);
    assert &and[Valuation(Evaluate(f1, x)) gt 0 : x in S`Roots];

    G, incG, projG := DirectProduct([i[1] : i in lG]);
    if (CoefficientRing(f) cmpeq Integers()) then
	max_comp := Ceiling(Abs(LeadingCoefficient(f))*
        Maximum([ x[2]`max_comp/Abs(LeadingCoefficient(x[2]`f)) : x in lG ]));
	if #non_coprime gt 0 then
	max_compnc := Ceiling(Abs(LeadingCoefficient(f))*
        Maximum([ lG[x][2]`max_comp/Abs(LeadingCoefficient(lG[x][2]`f)) : x in non_coprime ]));
	end if;
    elif Type(S`Scale) eq RngIntElt then
	max_comp := Maximum([x[2]`max_comp/x[2]`Scale : x in lG]);
	max_comp := Integers()!max_comp*S`Scale;
	if #non_coprime gt 0 then
	max_compnc := Maximum([lG[x][2]`max_comp/lG[x][2]`Scale : x in non_coprime]);
	max_compnc := Integers()!max_compnc*Snc`Scale;
	end if;
    elif Type(CoefficientRing(f)) eq RngFunOrd then
	max_comp := 2^30;
	max_compnc := 2^30;
//S`Scale;
	for ip in InfinitePlaces(FunctionField(CoefficientRing(f))) do
	    max_comp := Minimum(max_comp, Minimum([-x[2]`max_comp + Valuation(x[2]`Scale, ip): x in lG]) + Valuation(S`Scale, ip));
	    if #non_coprime gt 0 then
	    max_compnc := Minimum(max_compnc, Minimum([-lG[x][2]`max_comp + Valuation(lG[x][2]`Scale, ip): x in non_coprime]) + Valuation(Snc`Scale, ip));
	    end if;
//Valuation(S`Scale, ip);
//[Valuation(x[2]`Scale, ip) : x in lG];
	end for;
	max_comp := -max_comp;
	max_compnc := -max_compnc;
//"Product max_comp :", max_comp;
    else
	max_comp := Minimum([-(x[2]`max_comp - Degree(x[2]`Scale)) : x in lG]);
	max_comp := -max_comp;
	max_comp +:= Degree(S`Scale);
	assert max_comp ge 0;
	if #non_coprime gt 0 then
	max_compnc := Minimum([-(lG[x][2]`max_comp - Degree(lG[x][2]`Scale)) : x in non_coprime]);
	max_compnc := -max_compnc;
	max_compnc +:= Degree(Snc`Scale);
	end if;
    end if;

//RESET the max_comp of the factors?
for i in [1 .. #lG] do
if assigned lG[i][2]`max_comp then
lG[i][2]`max_comp := Maximum(lG[i][2]`max_comp, max_comp);
end if;
end for;

    S`DescentChain := [<shallow_copy(G), true>];
    if #non_coprime gt 0 then
    Snc`DescentChain := [<shallow_copy(G), true>];
    end if;
    if assigned S`max_comp then
	S`max_comp := Maximum(S`max_comp, max_comp);
    else
	S`max_comp := max_comp;
    end if;
    if #non_coprime gt 0 then 
    if assigned Snc`max_comp then
	Snc`max_comp := Maximum(Snc`max_comp, max_compnc);
    else
	Snc`max_comp := max_compnc;
    end if;
    end if;
    S`Maps := <incG, projG>;

    vprint GaloisGroup, 1: green, "done, and now the descents...", normal;
    if #non_coprime eq 0 then
	return G, S`Roots, S;
    end if;
    coprime := {1 .. #lG} diff non_coprime;
//#coprime gt 0;
    if not false and #coprime gt 0 then
//coprime;
	coprime := Sort(Setseq(coprime));
	non_coprime := Sort(Setseq(non_coprime));
	Gc, incGc, projGc := DirectProduct([lG[i][1] : i in coprime]);
	Gnc, incGnc, projGnc := DirectProduct([lG[i][1] : i in non_coprime]);
	m := map<Gnc -> G | x :-> &*[ incG[non_coprime[i]](projGnc[i](x)): i in [1..#non_coprime]]>;
	hmnc := hom<Gnc -> G | [m(Gnc.i) : i in [1..Ngens(Gnc)]]>;

	function Sieve(X)
	  if assigned S`Permutation then
	    p := S`Permutation;
	    X := X^p^-1;
	  end if;
	  return forall{z : z in [1..#non_coprime] | projG[non_coprime[z]](hmnc(X)) eq lG[non_coprime[z]][1]};
	end function;

	_gnc, _rnc, _snc := GenericStauduhar(Gnc, Snc, Type(max_compnc) eq RngIntElt select max_compnc else 1, 
		  func<G|[x`subgroup : x in _MaxSub_(G)]>,
		  Sieve:
		  ShortOK := ShortOK, Subfields := false);

	/*
	map Gc and gnc back to subgroups of G and take direct product there
	Already done the non-coprime descent so no more descent necessary
	*/
	//hmnc := hom<_gnc -> G | [m(_gnc.i) : i in [1..Ngens(_gnc)]]>;
	_gnc := hmnc(_gnc);

	m := map<Gc -> G | x :-> &*[ incG[coprime[i]](projGc[i](x)): i in [1..#coprime]]>;
	hmc := hom<Gc -> G | [m(x) : x in Generators(Gc)]>;
	Gc := hmc(Gc);

if not IsExport() then
"ne Done the map back", Gc, _gnc;
end if;
	GG, incGG, projGG := DirectProduct(Gc, _gnc);
	m := map<GG -> G | x :-> &*[ G!(projGG[1](x)), G!(projGG[2](x))]>;
	hm := hom<GG -> G | [m(x) : x in Generators(GG)]>;
	GG := hm(GG);
	/*
	Insert Snc`Roots into S`Roots
	*/
	root_cnt := 1;
	root_cnt_nc := 1;
	for i in [1 .. #lG] do 
	    if i in non_coprime then
		for j in [1 .. Degree(lf[i][1])] do
		    S`Roots[root_cnt] := _rnc[root_cnt_nc];
		    root_cnt +:= 1;
		    root_cnt_nc +:= 1;
		end for;
	    else
		root_cnt +:= Degree(lf[i][1]);
	    end if;
	end for;
	return GG, S`Roots, S;
    end if;
    function Sieve(X)
      if assigned S`Permutation then
        p := S`Permutation;
        X := X^p^-1;
      end if;
      return forall{z : z in [1..#lG] | projG[z](X) eq lG[z][1]};
    end function;
    return GenericStauduhar(G, S, Type(max_comp) eq RngIntElt select max_comp else 1, 
                  func<G|[x`subgroup : x in _MaxSub_(G)]>,
                  Sieve:
                  ShortOK := ShortOK, Subfields := false);
end intrinsic;

    

