freeze;
//

intrinsic ShortCosets(p::GrpPCElt, H::GrpPC, G::GrpPC:Limit := 10^6) -> []
  {Finds representatives for all right cosets of G modulo H that contain p.}

  assert H subset G;
  vprint GaloisGroup, 2: "ShortCosets (PC group): start";
  if Index(G, H) le 1000 then
    vprint GaloisGroup, 2: 
       "Small index case in ShortCosets, using the direct approach";
    c := RightTransversal(G, H);
    R := [ x : x in c | p^x^-1 in H];
  else
    if #H gt 2^40 then
      vprint GaloisGroup, 2: "ShortCosets: computing core";
      vtime GaloisGroup, 2: U := Core(G, H);
      if #U gt 1 then
        GG, mGG := quo<G|U>;
        HH := mGG(H);
        p := p@mGG;
      else
        GG := G;
        HH := H;
      end if;
    else
      GG := G;
      HH := H;
    end if;
    C := ConjugacyClasses(HH);
    R := [];
    CGp := Centralizer(GG, p);
    for c in C do
      f, g := IsConjugate(GG, c[3], p);
      if f then
        assert p eq c[3]^g;
        CHg := Centralizer(HH^g, p);
        if Index(CGp, CHg) gt Limit-#R then
          vprint GaloisGroup, 2: "2:ShortCosets: returning false";
          return false;
        end if;
        A := RightTransversal(CGp, CHg);
        R cat:= [g*x : x in A];
        if #R gt Limit then
          vprint GaloisGroup, 2: "1:ShortCosets: returning false";
          return false;
        end if;
      end if;
    end for;
  end if;
  vprint GaloisGroup, 2: "ShortCosets: returning a coset of size ", #R, 
                         " instead of ", #G div #H;
  
  if IsIdentical(GG, G) then
    return R;
  else
    return [x@@mGG : x in R];
  end if;
end intrinsic;


intrinsic ShortCosets(p::GrpPermElt, H::GrpPerm, G::GrpPerm:Limit := 10^6) -> []
  {Finds representatives for all right cosets of G modulo H that contain p.}

  assert H subset G;
  vprint GaloisGroup, 2: "ShortCosets: start";
  if Index(G, H) le 1000 then
    vprint GaloisGroup, 2: 
       "Small index case in ShortCosets, using the direct approach";
    c := RightTransversal(G, H);
    R := [ x : x in c | p^x^-1 in H];
  elif Degree(G) gt 40 and IsSolvable(G) then
    vprint GaloisGroup, 2: "Transfer over to PC-Groups...";
    vtime GaloisGroup, 2: GG, mGG := PCGroup(G);
    vtime GaloisGroup, 2: HH := mGG(H);
    pp := mGG(p);
    IndentPush();
    R := ShortCosets(pp, HH, GG:Limit := Limit);
    IndentPop();
    if R cmpne false then
      R := [x@@mGG : x in R];
      assert forall{x : x in R | p^x^-1 in H};
    end if;
  else
    c := CycleStructure(p);
    C := ConjugacyClasses(H);
    C := [ x : x in C | CycleStructure(x[3]) eq c];
    vprint GaloisGroup, 2: "Removed some (hopefully) remaining: ", #C;
    R := [];
    CGp := Centralizer(G, p);
    for c in C do
      f, g := IsConjugate(G, c[3], p);
      if f then
        assert p eq c[3]^g;
        CHg := Centralizer(H^g, p);
        if Index(CGp, CHg) gt Limit-#R then
          vprint GaloisGroup, 2: "2:ShortCosets: returning false, Index is", Index(CGp, CHg);
          return false;
        end if;
        A := RightTransversal(CGp, CHg);
        R cat:= [g*x : x in A];
        if #R gt Limit then
          vprint GaloisGroup, 2: "1:ShortCosets: returning false, too many cosets found", #G;
          return false;
        end if;
      end if;
    end for;
  end if;
  vprint GaloisGroup, 2: "ShortCosets: returning a coset of size ", #R, 
                         " instead of ", #G div #H;
  
  return R;
end intrinsic;

declare attributes GrpPerm : OptimalClass;

intrinsic InternalGetChain(G::GrpPerm, H::GrpPerm:MaxLen := Infinity(), 
                          SubGroup := false, Small := false, MaxLimit := Infinity(), StartLimit := 100) -> []
  {Computes a chain of maximal subgroups of G down to H}

  require H subset G : "H must be a subgoup of G";

  C := [G];
  if #H ne 1 then
    if IsSymmetric(G) then
      if assigned H`OptimalClass then
        p, h := Explode(H`OptimalClass);
        h := H!h;
      else
        c := Classes(H);
        ct := [ CycleStructure(x[3]) : x in c ];
        cs := Set(ct);
        p := Infinity();
        for a in cs do
          l := [c[x] :x in [1..#ct] | ct[x] eq a];
          pp := &+ [#Centralizer(G, x[3])*x[2]/#H : x in l];
          if pp lt p then
            p := pp;
            h := l[1][3];
          end if;
        end for;
        H`OptimalClass := <p, Eltseq(h)>;
      end if;
      StartLimit := Maximum(StartLimit, p);
      if p gt MaxLimit then
        return -1, _;
      end if;
    else
      h := [ Order(H.i) : i in [1..Ngens(H)]];
      _, h := Maximum(h);
      h := H.h;
    end if;
  else
    h := H.0;
  end if;
  while #G gt #H and MaxLen ge #C do
    assert H subset G;
    if SubGroup cmpeq false then
      M := MaximalSubgroups(G:IsTransitive := IsTransitive(H));
      M := [x`subgroup : x in M];
    else
      M := [SubGroup];
    end if;
    S := [];
    for m in M do
      if #m lt #H then
        continue;
      end if;
      if IsNormal(G, m) then
        vprint Invariant, 2: "IsNormal";
        c := [G.1];
      elif Index(G, m) lt 100 then
        vprint Invariant, 2: "Transversal";
        c := RightTransversal(G, m);
      else
        vprint Invariant, 2: "ShortCoset";
        c := ShortCosets(h, m, G:Limit := StartLimit);
        if c cmpeq false or #c gt StartLimit then
          vprint Invariant, 1: "Too many short cosets (", 
                        c cmpeq false select "way to many" else #c, 
                                                     "), trying again";
          hh := Order(h);
          li := StartLimit;
          attempt := 0;
          repeat
            repeat
              x := Random(H);
            until Order(x) ge hh;
            cc := ShortCosets(x, m, G:Limit := li);
            attempt +:= 1;
            if attempt mod 6 eq 0 then
              hh *:= 0.99;
              // printf "li was %o", li;
              if li eq MaxLimit then
                return -1, -1;
              end if;
              li := Round(li*1.4);
              li := Minimum(MaxLimit, li);
            end if;
          until cc cmpne false;
          if c cmpeq false or #cc lt #c then
            c := cc;
            h := x;
          end if;
        end if;
      end if;
      vprint Invariant, 1: "using ", #c, " cosets";
      for i in c do
        if H subset m^i then
          Append(~S, <i, m>);
        end if;
      end for;
    end for;
    if #S gt 1 then
      vprint Invariant, 1: 
        "I'm confused - multiple decents possible, choosing the largest";
    end if;
    if #S eq 0 then
      vprint Invariant, 1: "must be done - no decent possible...";
      if SubGroup cmpne false then
        return false, _;
      end if;
      assert C[#C] eq H;
      return C;
    end if;
    if Small then
      _, i := Maximum([Order(x[2]) : x in S]);
    else
      _, i := Minimum([Order(x[2]) : x in S]);
    end if;
    G := S[i][2]^S[i][1];
    if SubGroup cmpne false then
      if G eq H then
        return true, S[1][1];
      else
        return false, _;
      end if;
    end if;
    Append(~C, G);
  end while;
  if SubGroup cmpne false then
    return false, _;
  end if;

  return C;
end intrinsic;
    
intrinsic MyIsMaximal(G::GrpPerm, H::GrpPerm) -> BoolElt
  {}
 
  require H subset G: "H must e a subgroup of G";
  if #H eq #G then return false; end if;
  for i in [1..10] do
    S := sub<G|H, Random(G)>;
    if S ne G and S ne H then
      return false;
    end if;
  end for;
  C := InternalGetChain(G, H:MaxLen := 2);  
  if #C lt 2 or #C[2] ne #H then
    return false;
  else
    return true;
  end if;
end intrinsic;

intrinsic MyIsConjugate(G::GrpPerm, H1::GrpPerm, H2::GrpPerm) -> BoolElt, GrpPermElt
  {}

  function ret_false()
    return false;
  end function;
  function ret_true()
    return true;
  end function;
  
  if #H1 ne #H2 then
    return false, _;
  end if;
  if H1 eq H2 then
    return true, G.0;
  end if;
  if IsAbelian(H1) then
    if not IsAbelian(H2) then 
      return false, _;
    end if;
    if AbelianInvariants(H1) ne AbelianInvariants(H2) then
      return false, _;
    else
      return IsConjugate(G, H1, H2);
    end if;
  end if;
  a,b := InternalGetChain(G, H2:SubGroup := H1, MaxLimit := 10^6);

  if a cmpeq -1 then
    return IsConjugate(G, H1, H2);
  end if;
  if a then
    return ret_true(), b;
  else 
    return ret_false(), _;
  end if;
  return InternalGetChain(G, H2:SubGroup := H1);
end intrinsic;

intrinsic MyIsConjugateSubgroup(G::GrpPerm, H1::GrpPerm, H2::GrpPerm) -> BoolElt, GrpPermElt
  {}

  if Degree(H1) ne Degree(H2) then
    "Intersting case";
    if Degree(H1) mod Degree(H2) ne 0 then
      return false; // degrees don't match
    end if;
    A := AllPartitions(H1);
    A := [a : a in A | #a eq Degree(H1) div Degree(H2)];
    for a in A do
      O := Orbit(H1, a);
      h1 := StandardGroup(ActionImage(H1, O));
      if MyIsConjugateSubgroup(Generic(h1), h1, H2) then
        return true;
      end if;
    end for;
    return false;
  end if;
  if assigned H2`OptimalClass then
    c := H2!H2`OptimalClass[2];
  else
    c2 := Classes(H2);
    ct := [ CycleStructure(x[3]) : x in c2 ];
     cs := Set(ct);
     p := Infinity();
     for a in cs do
       l := [c2[x] :x in [1..#ct] | ct[x] eq a];
       pp := &+ [#Centralizer(G, x[3])*x[2]/#H2 : x in l];
       if pp lt p then
         p := pp;
         h := l[1][3];
       end if;
     end for;
     H2`OptimalClass := <p, Eltseq(h)>;
     c := h;
  end if;
  c := ShortCosets(c, H1, G);
  if c cmpeq false then
    return IsConjugateSubgroup(G, H1, H2);
  end if;
  for i in c do
    if H2 subset H1^i then
      return true, i;
    end if;
  end for;
  return false, _;
end intrinsic;

intrinsic MyIsConjugateQuotient(G::GrpPerm, H1::GrpPerm, H2::GrpPerm) -> BoolElt, GrpPermElt
  {Essentially cheks if H2 could be the Galois group of a subfield of H1}

    if not IsTransitive(H1) then
      O := Orbits(H1);
      np := Set(CartesianProduct(O));
      gs:= GSet(H1, np, map<car<np, H1> -> np| x:-><x[1][i]^x[2] : i in [1..#x[1]]>>);
      gs := Orbit(H1, gs[1]);
      mp := Action(H1, gs);
      H1 := mp(H1);
      H1 := StandardGroup(H1);
      G := Generic(H1);
    end if;
    if not IsTransitive(H2) then
      O := Orbits(H2);
      np := Set(CartesianProduct(O));
      gs:= GSet(H2, np, map<car<np, H2> -> np| x:-><x[1][i]^x[2] : i in [1..#x[1]]>>);
      gs := [Orbit(H2, x) : x in gs];
      _, p := Minimum([#x :x in gs]);
      gs := gs[p];
      mp := Action(H2, gs);
      H2 := mp(H2);
      H2 := StandardGroup(H2);
    end if;

    if Degree(H1) mod Degree(H2) ne 0 then
      return false; // degrees don't match
    end if;
    
    A := AllPartitions(H1);
    Include(~A, Support(H1));
    Include(~A, {1});
    A := [a : a in A | #a eq Degree(H1) div  Degree(H2)];
    for a in A do
      S := Stabiliser(H1, a);
      N := CosetImage(H1, S);
      assert Degree(N) eq Degree(H2);
      if MyIsConjugateSubgroup(Generic(N), N, H2) then
        return true;
      end if;
    end for;
    return false;
end intrinsic;

